﻿--[[
Copyright (c) 2015, Bernhard Voigt <tritium@noisywires.net>
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the copyright owner nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
]]--
--[[
LibQuestForGlory provides information about reputation rewards of quests.
It takes possible bonus modifiers into account and keeps track of changes in the bonus.
Fires 'BonusChanged' event.
]]--
local MAJOR_VERSION = "LibQuestForGlory-1.0"
local MINOR_VERSION = 2

if not LibStub then error(MAJOR_VERSION .. " requires LibStub") end

local QFG, oldLib = LibStub:NewLibrary(MAJOR_VERSION, MINOR_VERSION)

if not QFG then
	return
end

if oldLib then
	oldLib = {}
	for k, v in pairs(QFG) do
		QFG[k] = nil
		oldLib[k] = v
	end
end

QFG.callbacks = QFG.callbacks or LibStub("CallbackHandler-1.0"):New(QFG)

-- local references
local GetOwnedBuildingInfo   = _G.C_Garrison.GetOwnedBuildingInfo
local GetPlots               = _G.C_Garrison.GetPlots
local GetQuestWorldMapAreaID = _G.GetQuestWorldMapAreaID
local UnitBuff               = _G.UnitBuff

local LE_FOLLOWER_TYPE_GARRISON_6_0 = _G.LE_FOLLOWER_TYPE_GARRISON_6_0

local isHuman = false

do
	local _, raceEn = UnitRace("player")
	
	isHuman = raceEn == "Human"
end

local playerHasLvl3TradingPost = nil

-- extra effort to check for trading post (data only available when inside garrison)
local watchFrame = CreateFrame("Frame")

watchFrame:RegisterEvent("GARRISON_UPDATE")
watchFrame:RegisterEvent("GARRISON_BUILDING_ACTIVATED")
watchFrame:RegisterEvent("GARRISON_BUILDING_UPDATE")
watchFrame:RegisterEvent("GARRISON_BUILDING_REMOVED")
watchFrame:RegisterEvent("PLAYER_AURAS_CHANGED")

-- TODO: WoW celebration package
local bonusBuffs = {
    ["Spirit of Sharing"]               = { bonus = 0.1   },
    ["Grim Visage"]                     = { bonus = 0.1   },
    ["Unburdened"]                      = { bonus = 0.1   },
    ["Banner of Cooperation"]           = { bonus = 0.05  },
    ["Standard of Unity"]               = { bonus = 0.1   },
    ["Battle Standard of Coordination"] = { bonus = 0.15  },
    ["Nazgrel's Fervor"]                = { bonus = 0.1,  faction = 947 }, -- Thrallmar
    ["Trollbane's Command"]             = { bonus = 0.1,  faction = 946 }, -- Honor Hold
    ["A'dal's Song of Battle"]          = { bonus = 0.1,  faction = 935 }, -- The Sha'tar
    ["WHEE!"]                           = { bonus = 0.1,  excludes = { "Darkmoon Top Hat" } },
    ["Darkmoon Top Hat"]                = { bonus = 0.1,  excludes = { "WHEE!" } },
	["Cenarion Favor "]                 = { bonus = 0.25, faction = 609 }, -- Cenarion Circle
}

local TRADING_POST_LVL3_ID    = 145
local TRADING_POST_LVL3_BONUS = 0.2

local DRAENOR_MIN_MAP_ID =  941
local DRAENOR_MAX_MAP_ID = 1009

local DRAENOR_MAP_ID_EXCEPTIONS = {
	[951] = true, -- Timeless Isle
	[953] = true, -- Orgrimmar Raid
	[955] = true, -- Celestial Challenge
}

local INDEX_OFFSET = 10000

--[[
quest = {
	factionId = rewardAmount,
}
]]
local quests = {}

local mapIds = {}

local unpackedQuests = {}

local function UnpackQuest(questId) 
	if not questId or unpackedQuests[questId] then
		return
	end
	
	unpackedQuests[questId] = {}
	
	local questOffset = questId * INDEX_OFFSET
	
	for key in pairs(quests) do
		local faction = key - questOffset
		
		if faction > 0 and faction < INDEX_OFFSET then
			unpackedQuests[questId][faction] = quests[key]
		end
	end
end

--- Gets the base reputation reward of a quest for a given faction without any bonuses.
--
-- @param questId The id of the quest to query.
-- @param factionId The id of the faction to query.
function QFG:GetBaseReward(questId, factionId)
	UnpackQuest(questId)
	
	return unpackedQuests[questId] and unpackedQuests[questId][factionId] or nil
end

--- Gets the actual reputation reward of a quest for a given faction with all bonuses.
--
-- @param questId The id of the quest to query.
-- @param factionId The id of the faction to query.
function QFG:GetReward(questId, factionId)	
	return self:GetBaseReward(questId, factionId) * (1 + self:GetBonus(questId, factionId))
end

--- Gets the base reputation bonus the player will receive.
--
function QFG:GetBaseBonus()
	local bonus = 0

	-- human racial trait "Diplomacy"
	if isHuman then
		bonus = bonus + 0.1
	end	

	local exclude = {}
	
	-- bonus for all factions 
	for buff, buffInfo in pairs(bonusBuffs) do
		if not buffInfo.faction and not exclude[buff] and UnitBuff("player", buff) then
			bonus = bonus + buffInfo.bonus
			
			if buffInfo.excludes then
				for toExclude in pairs(buffInfo.excludes) do
					exclude[toExclude] = true
				end
			end
		end
	end
	
	return bonus
end

--- Gets the reputation bonus the player will receive for given quest.
--
-- @param questId The id of the quest to query.
function QFG:GetQuestBonus(questId)
	local bonus = 0

	-- quest happening in draenor get +20% for Garrison Trading Outpost level 3
	if self:IsDraenorQuest(questId) and self:PlayerHasTradingPostLvl3() then
		bonus = bonus + TRADING_POST_LVL3_BONUS
	end		
	
	return bonus
end

--- Gets the reputation bonus the player will receive for given faction.
--
-- @param factionId The id of the faction to query.
function QFG:GetFactionBonus(factionId)
	local bonus = 0

	-- bonus for given factions
	for buff, buffInfo in pairs(bonusBuffs) do
		if buffInfo.faction == factionId and UnitBuff("player", buff) then
			bonus = bonus + buffInfo.bonus
		end
	end
	
	return bonus
end

--- Gets the total reputation bonus the player will receive for given quest and faction.
--
-- @param questId The id of the quest to query.
-- @param factionId The id of the faction to query.
function QFG:GetBonus(questId, factionId)
	local bonus = 0

	return QFG:GetBaseBonus() + self:GetQuestBonus(questId) + self:GetFactionBonus(factionId)
end

--- Sets the value if player owns a trading post. 
--  If addon can provide that info (maybe it stored the info in options) 
--  this can help to work around the issue that we cannot find out about garrison buildings outside the garrison.
--
function QFG:SetPlayerHasTradingPostLvl3(hasLvl3Post)
	hasLvl3Post = hasLvl3Post and true or nil
	
	if playerHasLvl3TradingPost == hasLvl3Post then
		return
	end
	
	playerHasLvl3TradingPost = hasLvl3Post
	
	self.callbacks:Fire("BonusChanged")
end

--- Returns true if trading post has been found, false if not and nil if it could not be established (player was never in garrison). 
--
function QFG:PlayerHasTradingPostLvl3()
	return playerHasLvl3TradingPost
end

--- Returns the map id for the quest.
--
-- @param questId The id of the quest to query.
function QFG:GetQuestMapId(questId)
	if not mapIds[questId] then
		mapIds[questId] = GetQuestWorldMapAreaID(questId) or 0
	end
	
	return mapIds[questId]
end

--- Returns true if quest happens in Draenor.
--
-- @param questId The id of the quest to query.
function QFG:IsDraenorQuest(questId)
	local mapId = self:GetQuestMapId(questId)

	return mapId and mapId >= DRAENOR_MIN_MAP_ID and mapId <= DRAENOR_MAX_MAP_ID and not DRAENOR_MAP_ID_EXCEPTIONS[mapId] and true or false
end

--- Tries to determine if player has trading post.
--
function QFG:CheckForLvl3TradingPost()
	local plots = GetPlots(LE_FOLLOWER_TYPE_GARRISON_6_0)
	
	if not plots or not next(plots) then
		return
	end

	local hasLvl3TradingPost = nil
	
	for key, value in pairs(plots) do
		local buildingID = GetOwnedBuildingInfo(value.id)

		if buildingID == TRADING_POST_LVL3_ID then
			hasLvl3TradingPost = true
			
			break
		end		
	end
	
	self:SetPlayerHasTradingPostLvl3(hasLvl3TradingPost)
end

function QFG:CheckBonusBuffs()
	local changed = false
	
	for buff, buffInfo in pairs(bonusBuffs) do
		local active = UnitBuff("player", buff) and true or nil
		
		if buffInfo.active ~= active then
			buffInfo.active = active
			
			changed = true
		end
	end
	
	if changed then
		self.callbacks:Fire("BonusChanged")	
	end
end

watchFrame:SetScript("OnEvent", function(self, event, ...)
	if event == "PLAYER_AURAS_CHANGED" then
		QFG:CheckBonusBuffs()
	else
		QFG:CheckForLvl3TradingPost()
	end
end)

-- iterators:
-- IterateFactionIds
do --Do-end block for iterators
	local emptyTbl = {}
	local tablestack = setmetatable({}, {__mode = 'k'})

	local function KeyOnlyIter(t, prestate)
		if not t then 
			return nil 
		end

		if t.iterator then
			local key = t.iterator(t.t, prestate)

			if key then
				return key
			end				
		end
		
		tablestack[t] = true
		return nil, nil		
	end

	local function IterateKeys(toIterate)
		local tbl = next(tablestack) or {}		
		tablestack[tbl] = nil
		
		local data = toIterate or emptyTbl
		
		local iterator, t, state = pairs(data)
		
		tbl.iterator   = iterator
		tbl.t          = t
		
		return KeyOnlyIter, tbl, state
	end
	
	--- Returns iterator that provides registered text ids with given tag.
	--
	-- @param questId Id of quest for which to iterate the faction ids.
	function QFG:IterateFactionIds(questId)
		UnpackQuest(questId)
		
		return IterateKeys(unpackedQuests[questId])
	end
end

-- setup data
do
quests[20076] = 250 -- Sharptalon's Claw (Orgrimmar)
quests[110072] = 250 -- Riverpaw Gnoll Bounty (Stormwind)
quests[230076] = 250 -- Ursangous' Paw (Orgrimmar)
quests[240076] = 250 -- Shadumbra's Head (Orgrimmar)
quests[250076] = 250 -- Simmer Down Now (Orgrimmar)
quests[350072] = 250 -- Further Concerns (Stormwind)
quests[370072] = 25 -- Find the Lost Guards (Stormwind)
quests[400072] = 10 -- A Fishy Peril (Stormwind)
quests[450072] = 75 -- Discover Rolf's Fate (Stormwind)
quests[460072] = 250 -- Bounty on Murlocs (Stormwind)
quests[470072] = 250 -- Gold Dust Exchange (Stormwind)
quests[520072] = 150 -- Protect the Frontier (Stormwind)
quests[540072] = 75 -- Report to Goldshire (Stormwind)
quests[590072] = 25 -- Cloth and Leather Armor (Stormwind)
quests[600072] = 250 -- Kobold Candles (Stormwind)
quests[620072] = 250 -- The Fargodeep Mine (Stormwind)
quests[710072] = 25 -- Report to Thomas (Stormwind)
quests[760072] = 250 -- The Jasperlode Mine (Stormwind)
quests[830072] = 250 -- Fine Linen Goods (Stormwind)
quests[840072] = 25 -- Back to Billy (Stormwind)
quests[850072] = 25 -- Lost Necklace (Stormwind)
quests[860072] = 350 -- Pie for Billy (Stormwind)
quests[870072] = 250 -- Goldtooth (Stormwind)
quests[880072] = 250 -- Princess Must Die! (Stormwind)
quests[1060072] = 25 -- Young Lovers (Stormwind)
quests[1070072] = 25 -- Note to William (Stormwind)
quests[1110072] = 25 -- Speak with Gramma (Stormwind)
quests[1120072] = 250 -- Collecting Kelp (Stormwind)
quests[1140072] = 250 -- The Escape (Stormwind)
quests[1230072] = 250 -- The Collector (Stormwind)
quests[1470072] = 250 -- Manhunt (Stormwind)
quests[1710469] = 500 -- A Warden of the Alliance (Alliance)
quests[1720067] = 10 -- Children's Week (Horde)
quests[1760072] = 350 -- Wanted: "Hogger" (Stormwind)
quests[1820047] = 250 -- The Troll Menace (Ironforge)
quests[1840072] = 250 -- Furlbrow's Deed (Stormwind)
quests[2160076] = 250 -- Between a Rock and a Thistlefur (Orgrimmar)
quests[2180047] = 350 -- Ice and Fire (Ironforge)
quests[2390072] = 25 -- Westbrook Garrison Needs Help! (Stormwind)
quests[2700072] = 75 -- The Doomed Fleet (Stormwind)
quests[2730047] = 25 -- Resupplying the Excavation (Ironforge)
quests[2910047] = 75 -- The Reports (Ironforge)
quests[2910054] = 75 -- The Reports (Gnomeregan)
quests[3010047] = 75 -- Report to Ironforge (Ironforge)
quests[3020047] = 25 -- Powder to Ironband (Ironforge)
quests[3090047] = 250 -- Protecting the Shipment (Ironforge)
quests[3110047] = 250 -- Return to Marleth (Ironforge)
quests[3110054] = 250 -- Return to Marleth (Gnomeregan)
quests[3130047] = 250 -- Forced to Watch from Afar (Ironforge)
quests[3130054] = 250 -- Forced to Watch from Afar (Gnomeregan)
quests[3140047] = 250 -- Protecting the Herd (Ironforge)
quests[3140054] = 250 -- Protecting the Herd (Gnomeregan)
quests[3150047] = 250 -- The Perfect Stout (Ironforge)
quests[3150054] = 250 -- The Perfect Stout (Gnomeregan)
quests[3170047] = 350 -- Stocking Jetsteam (Ironforge)
quests[3170054] = 350 -- Stocking Jetsteam (Gnomeregan)
quests[3180047] = 25 -- Evershine (Ironforge)
quests[3180054] = 25 -- Evershine (Gnomeregan)
quests[3190047] = 150 -- A Favor for Evershine (Ironforge)
quests[3190054] = 150 -- A Favor for Evershine (Gnomeregan)
quests[3200047] = 350 -- Return to Bellowfiz (Ironforge)
quests[3200054] = 350 -- Return to Bellowfiz (Gnomeregan)
quests[3250072] = 25 -- Armed and Ready (Stormwind)
quests[3320072] = 150 -- Wine Shop Advert (Stormwind)
quests[3330072] = 75 -- Harlan Needs a Resupply (Stormwind)
quests[3340072] = 150 -- Package for Thurman (Stormwind)
quests[3510021] = 250 -- Find OOX-17/TN! (Booty Bay)
quests[3510087] = -250 -- Find OOX-17/TN! (Bloodsail Buccaneers)
quests[3530047] = 250 -- Stormpike's Delivery (Ironforge)
quests[3840047] = 250 -- Beer Basted Boar Ribs (Ironforge)
quests[3840054] = 250 -- Beer Basted Boar Ribs (Gnomeregan)
quests[4000047] = 25 -- Tools for Steelgrill (Ironforge)
quests[4000054] = 25 -- Tools for Steelgrill (Gnomeregan)
quests[4120054] = 250 -- Operation Recombobulation (Gnomeregan)
quests[4120047] = 250 -- Operation Recombobulation (Ironforge)
quests[4320047] = 250 -- Those Blasted Troggs! (Ironforge)
quests[4320054] = 250 -- Those Blasted Troggs! (Gnomeregan)
quests[4330047] = 250 -- The Public Servant (Ironforge)
quests[4330054] = 250 -- The Public Servant (Gnomeregan)
quests[4540047] = 10 -- After the Ambush (Ironforge)
quests[4750069] = 75 -- A Troubling Breeze (Darnassus)
quests[4760069] = 250 -- Gnarlpine Corruption (Darnassus)
quests[4830069] = 250 -- The Relics of Wakening (Darnassus)
quests[4850021] = 250 -- Find OOX-09/HL! (Booty Bay)
quests[4850087] = -250 -- Find OOX-09/HL! (Bloodsail Buccaneers)
quests[4860069] = 350 -- Ursal the Mauler (Darnassus)
quests[4870069] = 250 -- The Road to Darnassus (Darnassus)
quests[4890069] = 500 -- Seek Redemption! (Darnassus)
quests[5430072] = 500 -- The Perenolde Tiara (Stormwind)
quests[6470047] = 350 -- MacKreel's Moonshine (Ironforge)
quests[6480021] = 350 -- Rescue OOX-17/TN! (Booty Bay)
quests[6480087] = -350 -- Rescue OOX-17/TN! (Bloodsail Buccaneers)
quests[7430081] = 250 -- Dangers of the Windfury (Thunder Bluff)
quests[7440081] = 250 -- Preparation for Ceremony (Thunder Bluff)
quests[7490081] = 150 -- The Ravaged Caravan (Thunder Bluff)
quests[7510081] = 150 -- The Ravaged Caravan (Thunder Bluff)
quests[7610081] = 350 -- Swoop Hunting (Thunder Bluff)
quests[7640081] = 150 -- The Venture Co. (Thunder Bluff)
quests[7650081] = 250 -- Supervisor Fizsprocket (Thunder Bluff)
quests[7680081] = 350 -- Gathering Leather (Thunder Bluff)
quests[7690081] = 75 -- Kodo Hide Bag (Thunder Bluff)
quests[7700081] = 250 -- The Demon Scarred Cloak (Thunder Bluff)
quests[7730081] = 250 -- Rite of Wisdom (Thunder Bluff)
quests[8330081] = 150 -- A Sacred Burial (Thunder Bluff)
quests[8340076] = 250 -- Winds in the Desert (Orgrimmar)
quests[8340470] = 250 -- Winds in the Desert (Ratchet)
quests[8340530] = 250 -- Winds in the Desert (Darkspear Trolls)
quests[8350076] = 250 -- Securing the Lines (Orgrimmar)
quests[8350470] = 250 -- Securing the Lines (Ratchet)
quests[8350530] = 250 -- Securing the Lines (Darkspear Trolls)
quests[8360021] = 350 -- Rescue OOX-09/HL! (Booty Bay)
quests[8360087] = -350 -- Rescue OOX-09/HL! (Bloodsail Buccaneers)
quests[8400076] = 75 -- Conscript of the Horde (Orgrimmar)
quests[8400530] = 75 -- Conscript of the Horde (Darkspear Trolls)
quests[8440076] = 250 -- Plainstrider Menace (Orgrimmar)
quests[8450076] = 250 -- The Zhevra (Orgrimmar)
quests[8480068] = 250 -- Fungal Spores (Undercity)
quests[8500076] = 150 -- Kolkar Leaders (Orgrimmar)
quests[8510076] = 150 -- Verog the Dervish (Orgrimmar)
quests[8520076] = 150 -- Hezrul Bloodmark (Orgrimmar)
quests[8550076] = 250 -- Centaur Bracers (Orgrimmar)
quests[8580470] = 250 -- Ignition (Ratchet)
quests[8610081] = 250 -- The Hunter's Way (Thunder Bluff)
quests[8630470] = 350 -- The Escape (Ratchet)
quests[8650470] = 250 -- It's Gotta be the Horn (Ratchet)
quests[8660470] = 250 -- Root Samples (Ratchet)
quests[8670076] = 250 -- Harpy Raiders (Orgrimmar)
quests[8700081] = 150 -- The Forgotten Pools (Thunder Bluff)
quests[8710076] = 250 -- In Defense of Far Watch (Orgrimmar)
quests[8720076] = 250 -- The Far Watch Offensive (Orgrimmar)
quests[8750076] = 250 -- Harpy Lieutenants (Orgrimmar)
quests[8760076] = 350 -- Serena Bloodfeather (Orgrimmar)
quests[8770081] = 250 -- The Stagnant Oasis (Thunder Bluff)
quests[8800081] = 250 -- Altered Beings (Thunder Bluff)
quests[8810076] = 350 -- Echeyakee (Orgrimmar)
quests[8870470] = 150 -- Southsea Freebooters (Ratchet)
quests[8910470] = 250 -- A Captain's Vengeance (Ratchet)
quests[8910067] = 75 -- A Captain's Vengeance (Horde)
quests[8950470] = 250 -- WANTED: Cap'n Garvey (Ratchet)
quests[8990076] = 250 -- Consumed by Hatred (Orgrimmar)
quests[9030076] = 250 -- Hunting the Huntress (Orgrimmar)
quests[9050076] = 250 -- Into the Raptor's Den (Orgrimmar)
quests[9180069] = 250 -- Timberling Seeds (Darnassus)
quests[9190069] = 350 -- Timberling Sprouts (Darnassus)
quests[9220069] = 75 -- Rellian Greenspyre (Darnassus)
quests[9230069] = 350 -- Mossy Tumors (Darnassus)
quests[9270069] = 75 -- The Moss-twined Heart (Darnassus)
quests[9290069] = 250 -- Teldrassil: The Refusal of the Aspects (Darnassus)
quests[9300069] = 250 -- The Glowing Fruit (Darnassus)
quests[9310069] = 250 -- The Shimmering Frond (Darnassus)
quests[9320069] = 250 -- Twisted Hatred (Darnassus)
quests[9330069] = 250 -- Teldrassil: The Coming Dawn (Darnassus)
quests[9350069] = 350 -- The Waters of Teldrassil (Darnassus)
quests[9370069] = 75 -- The Enchanted Glade (Darnassus)
quests[9380069] = 250 -- Mist (Darnassus)
quests[9410069] = 250 -- Planting the Heart (Darnassus)
quests[9970069] = 75 -- Denalan's Earth (Darnassus)
quests[10360087] = 25 -- Avast Ye, Scallywag (Bloodsail Buccaneers)
quests[10360021] = -25 -- Avast Ye, Scallywag (Booty Bay)
quests[12010076] = 250 -- Theramore Spies (Orgrimmar)
quests[12020076] = 250 -- The Theramore Docks (Orgrimmar)
quests[12040072] = 250 -- Mudrock Soup and Bugs (Stormwind)
quests[12050081] = 250 -- Deadmire (Thunder Bluff)
quests[12220072] = 250 -- Stinky's Escape (Stormwind)
quests[12580072] = 250 -- ... and Bugs (Stormwind)
quests[12700470] = 250 -- Stinky's Escape (Ratchet)
quests[13380047] = 250 -- Stormpike's Order (Ironforge)
quests[13930469] = 75 -- Galen's Escape (Alliance)
quests[13930067] = 75 -- Galen's Escape (Horde)
quests[14310076] = 25 -- Alliance Relations (Orgrimmar)
quests[14320076] = 75 -- Alliance Relations (Orgrimmar)
quests[14340068] = 250 -- Befouled by Satyr (Undercity)
quests[14360076] = 250 -- Alliance Relations (Orgrimmar)
quests[14530047] = 75 -- Reclaimers' Business in Desolace (Ironforge)
quests[14540047] = 250 -- The Karnitol Shipwreck (Ironforge)
quests[14550047] = 75 -- The Karnitol Shipwreck (Ironforge)
quests[14560047] = 250 -- The Karnitol Shipwreck (Ironforge)
quests[14670047] = 250 -- Reagents for Reclaimers Inc. (Ironforge)
quests[14680469] = 10 -- Children's Week (Alliance)
quests[15780047] = 250 -- Supplying the Front (Ironforge)
quests[15810069] = 250 -- Elixirs for the Bladeleafs (Darnassus)
quests[15820069] = 250 -- Moonglow Vest (Darnassus)
quests[16180072] = 250 -- Gearing Redridge (Stormwind)
quests[19180076] = 25 -- The Befouled Element (Orgrimmar)
quests[20390054] = 25 -- Find Bingles (Gnomeregan)
quests[21580072] = 25 -- Rest and Relaxation (Stormwind)
quests[21590069] = 25 -- Dolanaar Delivery (Darnassus)
quests[23830076] = 75 -- Simple Parchment (Orgrimmar)
quests[23990069] = 10 -- The Sprouted Fronds (Darnassus)
quests[24380069] = 150 -- The Emerald Dreamcatcher (Darnassus)
quests[24590069] = 350 -- Ferocitas the Dream Eater (Darnassus)
quests[24990069] = 250 -- Oakenscowl (Darnassus)
quests[25180069] = 350 -- Tears of the Moon (Darnassus)
quests[25410069] = 250 -- The Sleeping Druid (Darnassus)
quests[25610069] = 350 -- Druid of the Claw (Darnassus)
quests[29480047] = 250 -- Gnome Improvement (Ironforge)
quests[29480054] = 250 -- Gnome Improvement (Gnomeregan)
quests[29500076] = 250 -- Nogg's Ring Redo (Orgrimmar)
quests[29630047] = 10 -- Portents of Uldum (Ironforge)
quests[29650081] = 10 -- Portents of Uldum (Thunder Bluff)
quests[30870076] = 75 -- Etched Parchment (Orgrimmar)
quests[30880076] = 75 -- Encrypted Parchment (Orgrimmar)
quests[30890076] = 75 -- Rune-Inscribed Parchment (Orgrimmar)
quests[30900076] = 75 -- Tainted Parchment (Orgrimmar)
quests[30910081] = 75 -- Simple Note (Thunder Bluff)
quests[30920081] = 75 -- Etched Note (Thunder Bluff)
quests[30930081] = 75 -- Rune-Inscribed Note (Thunder Bluff)
quests[30940081] = 75 -- Verdant Note (Thunder Bluff)
quests[30950068] = 75 -- Simple Scroll (Undercity)
quests[30960068] = 75 -- Encrypted Scroll (Undercity)
quests[30970068] = 75 -- Hallowed Scroll (Undercity)
quests[30980068] = 75 -- Glyphic Scroll (Undercity)
quests[30990068] = 75 -- Tainted Scroll (Undercity)
quests[31000072] = 75 -- Simple Letter (Stormwind)
quests[31010072] = 75 -- Consecrated Letter (Stormwind)
quests[31020072] = 75 -- Encrypted Letter (Stormwind)
quests[31030072] = 75 -- Hallowed Letter (Stormwind)
quests[31040072] = 75 -- Glyphic Letter (Stormwind)
quests[31050072] = 75 -- Tainted Letter (Stormwind)
quests[31060047] = 75 -- Simple Rune (Ironforge)
quests[31070047] = 75 -- Consecrated Rune (Ironforge)
quests[31080047] = 75 -- Etched Rune (Ironforge)
quests[31090047] = 75 -- Encrypted Rune (Ironforge)
quests[31100047] = 75 -- Hallowed Rune (Ironforge)
quests[31150047] = 75 -- Tainted Rune (Ironforge)
quests[31160069] = 75 -- Simple Sigil (Darnassus)
quests[31170069] = 75 -- Etched Sigil (Darnassus)
quests[31180069] = 75 -- Encrypted Sigil (Darnassus)
quests[31190069] = 75 -- Hallowed Sigil (Darnassus)
quests[31200069] = 75 -- Verdant Sigil (Darnassus)
quests[33610047] = 250 -- A Refugee's Quandary (Ironforge)
quests[33610054] = 250 -- A Refugee's Quandary (Gnomeregan)
quests[40210076] = 350 -- Counterattack! (Orgrimmar)
quests[44940076] = 10 -- March of the Silithid (Orgrimmar)
quests[45120054] = 350 -- A Little Slime Goes a Long Way (Gnomeregan)
quests[45130054] = 350 -- A Little Slime Goes a Long Way (Gnomeregan)
quests[46210087] = 500 -- Avast Ye, Admiral! (Bloodsail Buccaneers)
quests[46210021] = -500 -- Avast Ye, Admiral! (Booty Bay)
quests[49700589] = 250 -- Frostsaber Provisions (Wintersaber Trainers)
quests[50410076] = 250 -- Supplies for the Crossroads (Orgrimmar)
quests[52010589] = 250 -- Winterfall Intrusion (Wintersaber Trainers)
quests[55010470] = 150 -- Bone Collector (Ratchet)
quests[55020067] = 500 -- A Warden of the Horde (Horde)
quests[55610470] = 150 -- Kodo Roundup (Ratchet)
quests[55810081] = 350 -- Portals of the Legion (Thunder Bluff)
quests[57130069] = 250 -- One Shot. One Kill. (Darnassus)
quests[58210470] = 250 -- Bodyguard for Hire (Ratchet)
quests[58920047] = 25 -- Irondeep Supplies (Ironforge)
quests[58930076] = 25 -- Coldtooth Supplies (Orgrimmar)
quests[59430470] = 250 -- Gizelton Caravan (Ratchet)
quests[59810589] = 350 -- Rampaging Giants (Wintersaber Trainers)
quests[60310576] = 1000 -- Runecloth (Timbermaw Hold)
quests[60320576] = 1000 -- Sacred Cloth (Timbermaw Hold)
quests[61420530] = 250 -- Clam Bait (Darkspear Trolls)
quests[61430530] = 250 -- Other Fish to Fry (Darkspear Trolls)
quests[63210068] = 25 -- Supplying Brill (Undercity)
quests[63220068] = 25 -- Michael Garrett (Undercity)
quests[63230068] = 75 -- Ride to the Undercity (Undercity)
quests[63240068] = 350 -- Return to Morris (Undercity)
quests[63410069] = 25 -- To Darnassus (Darnassus)
quests[63420069] = 75 -- An Unexpected Gift (Darnassus)
quests[63430069] = 350 -- Return to Nyoma (Darnassus)
quests[63440069] = 25 -- Reminders of Home (Darnassus)
quests[63610081] = 25 -- A Bundle of Hides (Thunder Bluff)
quests[63620081] = 75 -- Ride to Thunder Bluff (Thunder Bluff)
quests[63630081] = 25 -- Tal the Wind Rider Master (Thunder Bluff)
quests[63640081] = 350 -- Return to Varg (Thunder Bluff)
quests[63650076] = 25 -- Meats to Orgrimmar (Orgrimmar)
quests[63650530] = 25 -- Meats to Orgrimmar (Darkspear Trolls)
quests[63840076] = 75 -- Ride to Orgrimmar (Orgrimmar)
quests[63840530] = 75 -- Ride to Orgrimmar (Darkspear Trolls)
quests[63850076] = 25 -- Doras the Wind Rider Master (Orgrimmar)
quests[63850530] = 25 -- Doras the Wind Rider Master (Darkspear Trolls)
quests[63860076] = 350 -- Return to Razor Hill (Orgrimmar)
quests[63860530] = 350 -- Return to Razor Hill (Darkspear Trolls)
quests[63870047] = 25 -- Honor Students (Ironforge)
quests[63870054] = 25 -- Honor Students (Gnomeregan)
quests[63880047] = 25 -- Gryth Thurden (Ironforge)
quests[63880054] = 25 -- Gryth Thurden (Gnomeregan)
quests[63910047] = 75 -- Ride to Ironforge (Ironforge)
quests[63910054] = 75 -- Ride to Ironforge (Gnomeregan)
quests[63920047] = 350 -- Return to Gremlock (Ironforge)
quests[63920054] = 350 -- Return to Gremlock (Gnomeregan)
quests[64410470] = 250 -- Satyr Horns (Ratchet)
quests[64420530] = 250 -- Naga at the Zoram Strand (Darkspear Trolls)
quests[64620530] = 250 -- Troll Charm (Darkspear Trolls)
quests[64820081] = 350 -- Freedom to Ruul (Thunder Bluff)
quests[65030530] = 250 -- Ashenvale Outrunners (Darkspear Trolls)
quests[65440076] = 350 -- Torek's Assault (Orgrimmar)
quests[66420059] = 75 -- Favor Amongst the Brotherhood, Dark Iron Ore (Thorium Brotherhood)
quests[66430059] = 500 -- Favor Amongst the Brotherhood, Fiery Core (Thorium Brotherhood)
quests[66440059] = 500 -- Favor Amongst the Brotherhood, Lava Core (Thorium Brotherhood)
quests[66450059] = 350 -- Favor Amongst the Brotherhood, Core Leather (Thorium Brotherhood)
quests[66460059] = 500 -- Favor Amongst the Brotherhood, Blood of the Mountain (Thorium Brotherhood)
quests[67010349] = 250 -- Syndicate Emblems (Ravenholdt)
quests[67410076] = 10 -- More Booty! (Orgrimmar)
quests[67810047] = 10 -- More Armor Scraps (Ironforge)
quests[68010076] = 10 -- Lokholar the Ice Lord (Orgrimmar)
quests[68250076] = 10 -- Call of Air - Guse's Fleet (Orgrimmar)
quests[68260076] = 10 -- Call of Air - Jeztor's Fleet (Orgrimmar)
quests[68270076] = 10 -- Call of Air - Mulverick's Fleet (Orgrimmar)
quests[68810047] = 10 -- Ivus the Forest Lord (Ironforge)
quests[69410047] = 10 -- Call of Air - Vipore's Fleet (Ironforge)
quests[69420047] = 10 -- Call of Air - Slidore's Fleet (Ironforge)
quests[69430047] = 10 -- Call of Air - Ichman's Fleet (Ironforge)
quests[69620169] = 25 -- Treats for Great-father Winter (Steamwheedle Cartel)
quests[69820047] = 25 -- Coldtooth Supplies (Ironforge)
quests[69840169] = 250 -- A Smokywood Pastures' Thank You! (Steamwheedle Cartel)
quests[69850076] = 25 -- Irondeep Supplies (Orgrimmar)
quests[70010076] = 10 -- Empty Stables (Orgrimmar)
quests[70020076] = 10 -- Ram Hide Harnesses (Orgrimmar)
quests[70250169] = 25 -- Treats for Greatfather Winter (Steamwheedle Cartel)
quests[70260047] = 10 -- Ram Riding Harnesses (Ironforge)
quests[70270047] = 10 -- Empty Stables (Ironforge)
quests[70450169] = 250 -- A Smokywood Pastures' Thank You! (Steamwheedle Cartel)
quests[70610081] = 500 -- The Feast of Winter Veil (Thunder Bluff)
quests[70630047] = 500 -- The Feast of Winter Veil (Ironforge)
quests[70810730] = 250 -- Alterac Valley Graveyards (Stormpike Guard)
quests[70810047] = 250 -- Alterac Valley Graveyards (Ironforge)
quests[70820729] = 250 -- The Graveyards of Alterac (Frostwolf Clan)
quests[70820076] = 250 -- The Graveyards of Alterac (Orgrimmar)
quests[71010729] = 250 -- Towers and Bunkers (Frostwolf Clan)
quests[71010076] = 250 -- Towers and Bunkers (Orgrimmar)
quests[71020730] = 250 -- Towers and Bunkers (Stormpike Guard)
quests[71020047] = 250 -- Towers and Bunkers (Ironforge)
quests[71210730] = 10 -- The Quartermaster (Stormpike Guard)
quests[71210047] = 10 -- The Quartermaster (Ironforge)
quests[71220730] = 250 -- Capture a Mine (Stormpike Guard)
quests[71220047] = 250 -- Capture a Mine (Ironforge)
quests[71230729] = 10 -- Speak with our Quartermaster (Frostwolf Clan)
quests[71230076] = 10 -- Speak with our Quartermaster (Orgrimmar)
quests[71240729] = 250 -- Capture a Mine (Frostwolf Clan)
quests[71240076] = 250 -- Capture a Mine (Orgrimmar)
quests[71410730] = 500 -- The Battle of Alterac (Stormpike Guard)
quests[71410047] = 500 -- The Battle of Alterac (Ironforge)
quests[71420729] = 500 -- The Battle for Alterac (Frostwolf Clan)
quests[71420076] = 500 -- The Battle for Alterac (Orgrimmar)
quests[71610729] = 250 -- Proving Grounds (Frostwolf Clan)
quests[71610076] = 250 -- Proving Grounds (Orgrimmar)
quests[71620730] = 250 -- Proving Grounds (Stormpike Guard)
quests[71620047] = 250 -- Proving Grounds (Ironforge)
quests[71630729] = 250 -- Rise and Be Recognized (Frostwolf Clan)
quests[71630076] = 250 -- Rise and Be Recognized (Orgrimmar)
quests[71640729] = 250 -- Honored Amongst the Clan (Frostwolf Clan)
quests[71640076] = 250 -- Honored Amongst the Clan (Orgrimmar)
quests[71650729] = 350 -- Earned Reverence (Frostwolf Clan)
quests[71650076] = 350 -- Earned Reverence (Orgrimmar)
quests[71660729] = 500 -- Legendary Heroes (Frostwolf Clan)
quests[71660076] = 500 -- Legendary Heroes (Orgrimmar)
quests[71670729] = 500 -- The Eye of Command (Frostwolf Clan)
quests[71670076] = 500 -- The Eye of Command (Orgrimmar)
quests[71680730] = 250 -- Rise and Be Recognized (Stormpike Guard)
quests[71680047] = 250 -- Rise and Be Recognized (Ironforge)
quests[71690730] = 250 -- Honored Amongst the Guard (Stormpike Guard)
quests[71690047] = 250 -- Honored Amongst the Guard (Ironforge)
quests[71700730] = 350 -- Earned Reverence (Stormpike Guard)
quests[71700047] = 350 -- Earned Reverence (Ironforge)
quests[71710730] = 500 -- Legendary Heroes (Stormpike Guard)
quests[71710047] = 500 -- Legendary Heroes (Ironforge)
quests[71720047] = 500 -- The Eye of Command (Ironforge)
quests[71720730] = 500 -- The Eye of Command (Stormpike Guard)
quests[72230730] = 250 -- Armor Scraps (Stormpike Guard)
quests[72230047] = 250 -- Armor Scraps (Ironforge)
quests[72240729] = 250 -- Enemy Booty (Frostwolf Clan)
quests[72240076] = 250 -- Enemy Booty (Orgrimmar)
quests[73830069] = 150 -- Teldrassil: The Burden of the Kaldorei (Darnassus)
quests[73850076] = 75 -- A Gallon of Blood (Orgrimmar)
quests[73860047] = 75 -- Crystal Cluster (Ironforge)
quests[76040059] = 1000 -- A Binding Contract (Thorium Brotherhood)
quests[77370059] = 625 -- Gaining Even More Acceptance (Thorium Brotherhood)
quests[77810930] = 500 -- The Lord of Blackrock (Exodar)
quests[77810072] = 500 -- The Lord of Blackrock (Stormwind)
quests[77810054] = 500 -- The Lord of Blackrock (Gnomeregan)
quests[77810069] = 500 -- The Lord of Blackrock (Darnassus)
quests[77810047] = 500 -- The Lord of Blackrock (Ironforge)
quests[77820930] = 500 -- The Lord of Blackrock (Exodar)
quests[77820072] = 500 -- The Lord of Blackrock (Stormwind)
quests[77820054] = 500 -- The Lord of Blackrock (Gnomeregan)
quests[77820069] = 500 -- The Lord of Blackrock (Darnassus)
quests[77820047] = 500 -- The Lord of Blackrock (Ironforge)
quests[77830530] = 500 -- The Lord of Blackrock (Darkspear Trolls)
quests[77830068] = 500 -- The Lord of Blackrock (Undercity)
quests[77830081] = 500 -- The Lord of Blackrock (Thunder Bluff)
quests[77830076] = 500 -- The Lord of Blackrock (Orgrimmar)
quests[77830911] = 500 -- The Lord of Blackrock (Silvermoon City)
quests[77840530] = 500 -- The Lord of Blackrock (Darkspear Trolls)
quests[77840068] = 500 -- The Lord of Blackrock (Undercity)
quests[77840081] = 500 -- The Lord of Blackrock (Thunder Bluff)
quests[77840076] = 500 -- The Lord of Blackrock (Orgrimmar)
quests[77840911] = 500 -- The Lord of Blackrock (Silvermoon City)
quests[78630890] = 10 -- Sentinel Basic Care Package (Silverwing Sentinels)
quests[78640890] = 10 -- Sentinel Standard Care Package (Silverwing Sentinels)
quests[78650890] = 10 -- Sentinel Advanced Care Package (Silverwing Sentinels)
quests[78660889] = 10 -- Outrider Basic Care Package (Warsong Outriders)
quests[78670889] = 10 -- Outrider Standard Care Package (Warsong Outriders)
quests[78680889] = 10 -- Outrider Advanced Care Package (Warsong Outriders)
quests[79050909] = 75 -- The Darkmoon Faire (Darkmoon Faire)
quests[79070909] = 350 -- Darkmoon Beast Deck (Darkmoon Faire)
quests[79260909] = 75 -- The Darkmoon Faire (Darkmoon Faire)
quests[79270909] = 350 -- Darkmoon Portals Deck (Darkmoon Faire)
quests[79280909] = 350 -- Darkmoon Warlords Deck (Darkmoon Faire)
quests[79290909] = 350 -- Darkmoon Elementals Deck (Darkmoon Faire)
quests[79370909] = 75 -- Your Fortune Awaits You... (Darkmoon Faire)
quests[79440909] = 75 -- Your Fortune Awaits You... (Darkmoon Faire)
quests[79450909] = 75 -- Your Fortune Awaits You... (Darkmoon Faire)
quests[81050509] = 500 -- The Battle for Arathi Basin! (The League of Arathor)
quests[81140509] = 1000 -- Control Four Bases (The League of Arathor)
quests[81150509] = 500 -- Control Five Bases (The League of Arathor)
quests[81200510] = 500 -- The Battle for Arathi Basin! (The Defilers)
quests[81210510] = 1000 -- Take Four Bases (The Defilers)
quests[81220510] = 500 -- Take Five Bases (The Defilers)
quests[81490469] = 500 -- Honoring a Hero (Alliance)
quests[81500067] = 500 -- Honoring a Hero (Horde)
quests[82600509] = 10 -- Arathor Basic Care Package (The League of Arathor)
quests[82610509] = 10 -- Arathor Standard Care Package (The League of Arathor)
quests[82620509] = 10 -- Arathor Advanced Care Package (The League of Arathor)
quests[82630510] = 10 -- Defiler's Basic Care Package (The Defilers)
quests[82640510] = 10 -- Defiler's Standard Care Package (The Defilers)
quests[82650510] = 10 -- Defiler's Advanced Care Package (The Defilers)
quests[82710730] = 10 -- Hero of the Stormpike (Stormpike Guard)
quests[82710047] = 10 -- Hero of the Stormpike (Ironforge)
quests[82720729] = 10 -- Hero of the Frostwolf (Frostwolf Clan)
quests[82720076] = 10 -- Hero of the Frostwolf (Orgrimmar)
quests[82770609] = 150 -- Deadly Desert Venom (Cenarion Circle)
quests[82780609] = 250 -- Noggle's Last Hope (Cenarion Circle)
quests[82790609] = 250 -- The Twilight Lexicon (Cenarion Circle)
quests[82800609] = 150 -- Securing the Supply Lines (Cenarion Circle)
quests[82810609] = 250 -- Stepping Up Security (Cenarion Circle)
quests[82820609] = 250 -- Noggle's Lost Satchel (Cenarion Circle)
quests[82830609] = 250 -- Wanted - Deathclasp, Terror of the Sands (Cenarion Circle)
quests[82840609] = 250 -- The Twilight Mystery (Cenarion Circle)
quests[82850609] = 75 -- The Deserter (Cenarion Circle)
quests[82870609] = 250 -- A Terrible Purpose (Cenarion Circle)
quests[83040609] = 250 -- Dearest Natalia (Cenarion Circle)
quests[83060609] = 350 -- Into The Maw of Madness (Cenarion Circle)
quests[83070609] = 10 -- Desert Recipe (Cenarion Circle)
quests[83080609] = 250 -- Brann Bronzebeard's Lost Letter (Cenarion Circle)
quests[83090609] = 250 -- Glyph Chasing (Cenarion Circle)
quests[83100609] = 250 -- Breaking the Code (Cenarion Circle)
quests[83110469] = 250 -- Hallow's End Treats for Jesper! (Alliance)
quests[83120067] = 250 -- Hallow's End Treats for Spoops! (Horde)
quests[83130609] = 75 -- Sharing the Knowledge (Cenarion Circle)
quests[83140609] = 150 -- Unraveling the Mystery (Cenarion Circle)
quests[83170609] = 75 -- Kitchen Assistance (Cenarion Circle)
quests[83180609] = 250 -- Secret Communication (Cenarion Circle)
quests[83190609] = 500 -- Encrypted Twilight Texts (Cenarion Circle)
quests[83200609] = 250 -- Twilight Geolords (Cenarion Circle)
quests[83210609] = 350 -- Vyral the Vile (Cenarion Circle)
quests[83250911] = 250 -- Reclaiming Sunstrider Isle (Silvermoon City)
quests[83260911] = 250 -- Unfortunate Measures (Silvermoon City)
quests[83270911] = 10 -- Report to Lanthan Perilon (Silvermoon City)
quests[83280911] = 75 -- Mage Training (Silvermoon City)
quests[83290911] = 25 -- Warrior Training (Silvermoon City)
quests[83300911] = 250 -- Solanian's Belongings (Silvermoon City)
quests[83340911] = 250 -- Aggression (Silvermoon City)
quests[83350911] = 350 -- Felendren the Banished (Silvermoon City)
quests[83360911] = 250 -- A Fistful of Slivers (Silvermoon City)
quests[83380911] = 250 -- Tainted Arcane Sliver (Silvermoon City)
quests[83450911] = 250 -- The Shrine of Dath'Remar (Silvermoon City)
quests[83460911] = 250 -- Thirst Unending (Silvermoon City)
quests[83470911] = 10 -- Aiding the Outrunners (Silvermoon City)
quests[83500911] = 25 -- Completing the Delivery (Silvermoon City)
quests[84460609] = 500 -- Shrouded in Nightmare (Cenarion Circle)
quests[84470609] = 500 -- Waking Legends (Cenarion Circle)
quests[84630911] = 250 -- Unstable Mana Crystals (Silvermoon City)
quests[84680911] = 250 -- Wanted: Thaelis the Hungerer (Silvermoon City)
quests[84700576] = 1400 -- Deadwood Ritual Totem (Timbermaw Hold)
quests[84720911] = 250 -- Major Malfunction (Silvermoon City)
quests[84730911] = 250 -- A Somber Task (Silvermoon City)
quests[84740911] = 250 -- Old Whitebark's Pendant (Silvermoon City)
quests[84750911] = 250 -- The Dead Scar (Silvermoon City)
quests[84760911] = 250 -- Amani Encroachment (Silvermoon City)
quests[84770911] = 250 -- The Spearcrafter's Hammer (Silvermoon City)
quests[84790530] = 75 -- Zul'Marosh (Darkspear Trolls)
quests[84800911] = 250 -- Lost Armaments (Silvermoon City)
quests[84810576] = 500 -- The Root of All Evil (Timbermaw Hold)
quests[84820911] = 250 -- Incriminating Documents (Silvermoon City)
quests[84830911] = 250 -- The Dwarven Spy (Silvermoon City)
quests[84860911] = 250 -- Arcane Instability (Silvermoon City)
quests[84870911] = 250 -- Corrupted Soil (Silvermoon City)
quests[84880911] = 250 -- Unexpected Results (Silvermoon City)
quests[84900911] = 250 -- Powering our Defenses (Silvermoon City)
quests[84910911] = 250 -- Pelt Collection (Silvermoon City)
quests[85440910] = 500 -- Conqueror's Spaulders (Brood of Nozdormu)
quests[85560609] = 250 -- Signet of Unyielding Strength (Cenarion Circle)
quests[85570609] = 250 -- Drape of Unyielding Strength (Cenarion Circle)
quests[85580609] = 250 -- Sickle of Unyielding Strength (Cenarion Circle)
quests[85590910] = 500 -- Conqueror's Greaves (Brood of Nozdormu)
quests[85600910] = 500 -- Conqueror's Legguards (Brood of Nozdormu)
quests[85610910] = 500 -- Conqueror's Crown (Brood of Nozdormu)
quests[85620910] = 500 -- Conqueror's Breastplate (Brood of Nozdormu)
quests[85630911] = 75 -- Warlock Training (Silvermoon City)
quests[85640911] = 75 -- Priest Training (Silvermoon City)
quests[85790910] = 500 -- Mortal Champions (Brood of Nozdormu)
quests[85790609] = 250 -- Mortal Champions (Cenarion Circle)
quests[85920910] = 500 -- Tiara of the Oracle (Brood of Nozdormu)
quests[85930910] = 500 -- Trousers of the Oracle (Brood of Nozdormu)
quests[85940910] = 500 -- Mantle of the Oracle (Brood of Nozdormu)
quests[85950910] = 500 -- Mortal Champions (Brood of Nozdormu)
quests[85950609] = 250 -- Mortal Champions (Cenarion Circle)
quests[85960910] = 500 -- Footwraps of the Oracle (Brood of Nozdormu)
quests[86020910] = 500 -- Stormcaller's Pauldrons (Brood of Nozdormu)
quests[86030910] = 500 -- Vestments of the Oracle (Brood of Nozdormu)
quests[86190067] = 75 -- Morndeep the Elder (Horde)
quests[86190469] = 75 -- Morndeep the Elder (Alliance)
quests[86210910] = 500 -- Stormcaller's Footguards (Brood of Nozdormu)
quests[86220910] = 500 -- Stormcaller's Hauberk (Brood of Nozdormu)
quests[86230910] = 500 -- Stormcaller's Diadem (Brood of Nozdormu)
quests[86240910] = 500 -- Stormcaller's Leggings (Brood of Nozdormu)
quests[86250910] = 500 -- Enigma Shoulderpads (Brood of Nozdormu)
quests[86260910] = 500 -- Striker's Footguards (Brood of Nozdormu)
quests[86270910] = 500 -- Avenger's Breastplate (Brood of Nozdormu)
quests[86280910] = 500 -- Avenger's Crown (Brood of Nozdormu)
quests[86290910] = 500 -- Avenger's Legguards (Brood of Nozdormu)
quests[86300910] = 500 -- Avenger's Pauldrons (Brood of Nozdormu)
quests[86310910] = 500 -- Enigma Leggings (Brood of Nozdormu)
quests[86320910] = 500 -- Enigma Circlet (Brood of Nozdormu)
quests[86330910] = 500 -- Enigma Robes (Brood of Nozdormu)
quests[86340910] = 500 -- Enigma Boots (Brood of Nozdormu)
quests[86350067] = 75 -- Splitrock the Elder (Horde)
quests[86350469] = 75 -- Splitrock the Elder (Alliance)
quests[86360067] = 75 -- Rumblerock the Elder (Horde)
quests[86360469] = 75 -- Rumblerock the Elder (Alliance)
quests[86370910] = 500 -- Deathdealer's Boots (Brood of Nozdormu)
quests[86380910] = 500 -- Deathdealer's Vest (Brood of Nozdormu)
quests[86390910] = 500 -- Deathdealer's Helm (Brood of Nozdormu)
quests[86400910] = 500 -- Deathdealer's Leggings (Brood of Nozdormu)
quests[86410910] = 500 -- Deathdealer's Spaulders (Brood of Nozdormu)
quests[86420067] = 75 -- Silvervein the Elder (Horde)
quests[86420469] = 75 -- Silvervein the Elder (Alliance)
quests[86430067] = 75 -- Highpeak the Elder (Horde)
quests[86430469] = 75 -- Highpeak the Elder (Alliance)
quests[86440067] = 75 -- Stonefort the Elder (Horde)
quests[86440469] = 75 -- Stonefort the Elder (Alliance)
quests[86450067] = 75 -- Obsidian the Elder (Horde)
quests[86450469] = 75 -- Obsidian the Elder (Alliance)
quests[86460067] = 75 -- Hammershout the Elder (Horde)
quests[86460469] = 75 -- Hammershout the Elder (Alliance)
quests[86470067] = 75 -- Bellowrage the Elder (Horde)
quests[86470469] = 75 -- Bellowrage the Elder (Alliance)
quests[86480067] = 75 -- Darkcore the Elder (Horde)
quests[86480469] = 75 -- Darkcore the Elder (Alliance)
quests[86490067] = 75 -- Stormbrow the Elder (Horde)
quests[86490469] = 75 -- Stormbrow the Elder (Alliance)
quests[86500067] = 75 -- Snowcrown the Elder (Horde)
quests[86500469] = 75 -- Snowcrown the Elder (Alliance)
quests[86510067] = 75 -- Ironband the Elder (Horde)
quests[86510469] = 75 -- Ironband the Elder (Alliance)
quests[86520067] = 75 -- Graveborn the Elder (Horde)
quests[86520469] = 75 -- Graveborn the Elder (Alliance)
quests[86530067] = 75 -- Goldwell the Elder (Horde)
quests[86530469] = 75 -- Goldwell the Elder (Alliance)
quests[86540067] = 75 -- Primestone the Elder (Horde)
quests[86540469] = 75 -- Primestone the Elder (Alliance)
quests[86550910] = 500 -- Avenger's Greaves (Brood of Nozdormu)
quests[86560910] = 500 -- Striker's Hauberk (Brood of Nozdormu)
quests[86570910] = 500 -- Striker's Diadem (Brood of Nozdormu)
quests[86580910] = 500 -- Striker's Leggings (Brood of Nozdormu)
quests[86590910] = 500 -- Striker's Pauldrons (Brood of Nozdormu)
quests[86600910] = 500 -- Doomcaller's Footwraps (Brood of Nozdormu)
quests[86610910] = 500 -- Doomcaller's Robes (Brood of Nozdormu)
quests[86620910] = 500 -- Doomcaller's Circlet (Brood of Nozdormu)
quests[86630910] = 500 -- Doomcaller's Trousers (Brood of Nozdormu)
quests[86640910] = 500 -- Doomcaller's Mantle (Brood of Nozdormu)
quests[86650910] = 500 -- Genesis Boots (Brood of Nozdormu)
quests[86660910] = 500 -- Genesis Vest (Brood of Nozdormu)
quests[86670910] = 500 -- Genesis Helm (Brood of Nozdormu)
quests[86680910] = 500 -- Genesis Trousers (Brood of Nozdormu)
quests[86690910] = 500 -- Genesis Shoulderpads (Brood of Nozdormu)
quests[86700067] = 75 -- Runetotem the Elder (Horde)
quests[86700469] = 75 -- Runetotem the Elder (Alliance)
quests[86710067] = 75 -- Ragetotem the Elder (Horde)
quests[86710469] = 75 -- Ragetotem the Elder (Alliance)
quests[86720067] = 75 -- Stonespire the Elder (Horde)
quests[86720469] = 75 -- Stonespire the Elder (Alliance)
quests[86730067] = 75 -- Bloodhoof the Elder (Horde)
quests[86730469] = 75 -- Bloodhoof the Elder (Alliance)
quests[86740067] = 75 -- Winterhoof the Elder (Horde)
quests[86740469] = 75 -- Winterhoof the Elder (Alliance)
quests[86750067] = 75 -- Skychaser the Elder (Horde)
quests[86750469] = 75 -- Skychaser the Elder (Alliance)
quests[86760067] = 75 -- Wildmane the Elder (Horde)
quests[86760469] = 75 -- Wildmane the Elder (Alliance)
quests[86770067] = 75 -- Darkhorn the Elder (Horde)
quests[86770469] = 75 -- Darkhorn the Elder (Alliance)
quests[86780067] = 75 -- Wheathoof the Elder (Horde)
quests[86780469] = 75 -- Wheathoof the Elder (Alliance)
quests[86790067] = 75 -- Grimtotem the Elder (Horde)
quests[86790469] = 75 -- Grimtotem the Elder (Alliance)
quests[86800067] = 75 -- Windtotem the Elder (Horde)
quests[86800469] = 75 -- Windtotem the Elder (Alliance)
quests[86810067] = 75 -- Thunderhorn the Elder (Horde)
quests[86810469] = 75 -- Thunderhorn the Elder (Alliance)
quests[86820067] = 75 -- Skyseer the Elder (Horde)
quests[86820469] = 75 -- Skyseer the Elder (Alliance)
quests[86830067] = 75 -- Dawnstrider the Elder (Horde)
quests[86830469] = 75 -- Dawnstrider the Elder (Alliance)
quests[86840067] = 75 -- Dreamseer the Elder (Horde)
quests[86840469] = 75 -- Dreamseer the Elder (Alliance)
quests[86850067] = 75 -- Mistwalker the Elder (Horde)
quests[86850469] = 75 -- Mistwalker the Elder (Alliance)
quests[86860067] = 75 -- High Mountain the Elder (Horde)
quests[86860469] = 75 -- High Mountain the Elder (Alliance)
quests[86880067] = 75 -- Windrun the Elder (Horde)
quests[86880469] = 75 -- Windrun the Elder (Alliance)
quests[86890609] = 250 -- Shroud of Infinite Wisdom (Cenarion Circle)
quests[86900609] = 250 -- Cloak of the Gathering Storm (Cenarion Circle)
quests[86910609] = 250 -- Drape of Vaulted Secrets (Cenarion Circle)
quests[86920609] = 250 -- Cloak of Unending Life (Cenarion Circle)
quests[86930609] = 250 -- Cloak of Veiled Shadows (Cenarion Circle)
quests[86940609] = 250 -- Shroud of Unspoken Names (Cenarion Circle)
quests[86950609] = 250 -- Cape of Eternal Justice (Cenarion Circle)
quests[86960609] = 250 -- Cloak of the Unseen Path (Cenarion Circle)
quests[86970609] = 250 -- Ring of Infinite Wisdom (Cenarion Circle)
quests[86980609] = 250 -- Ring of the Gathering Storm (Cenarion Circle)
quests[86990609] = 250 -- Band of Vaulted Secrets (Cenarion Circle)
quests[87000609] = 250 -- Band of Unending Life (Cenarion Circle)
quests[87010609] = 250 -- Band of Veiled Shadows (Cenarion Circle)
quests[87020609] = 250 -- Ring of Unspoken Names (Cenarion Circle)
quests[87030609] = 250 -- Ring of Eternal Justice (Cenarion Circle)
quests[87040609] = 500 -- Signet of the Unseen Path (Cenarion Circle)
quests[87050609] = 250 -- Gavel of Infinite Wisdom (Cenarion Circle)
quests[87060609] = 250 -- Hammer of the Gathering Storm (Cenarion Circle)
quests[87070609] = 250 -- Blade of Vaulted Secrets (Cenarion Circle)
quests[87080609] = 250 -- Mace of Unending Life (Cenarion Circle)
quests[87090609] = 250 -- Dagger of Veiled Shadows (Cenarion Circle)
quests[87100609] = 250 -- Kris of Unspoken Names (Cenarion Circle)
quests[87110609] = 250 -- Blade of Eternal Justice (Cenarion Circle)
quests[87120609] = 250 -- Scythe of the Unseen Path (Cenarion Circle)
quests[87130067] = 75 -- Starsong the Elder (Horde)
quests[87130469] = 75 -- Starsong the Elder (Alliance)
quests[87140067] = 75 -- Moonstrike the Elder (Horde)
quests[87140469] = 75 -- Moonstrike the Elder (Alliance)
quests[87150067] = 75 -- Bladeleaf the Elder (Horde)
quests[87150469] = 75 -- Bladeleaf the Elder (Alliance)
quests[87160067] = 75 -- Starglade the Elder (Horde)
quests[87160469] = 75 -- Starglade the Elder (Alliance)
quests[87170067] = 75 -- Moonwarden the Elder (Horde)
quests[87170469] = 75 -- Moonwarden the Elder (Alliance)
quests[87180067] = 75 -- Bladeswift the Elder (Horde)
quests[87180469] = 75 -- Bladeswift the Elder (Alliance)
quests[87190067] = 75 -- Bladesing the Elder (Horde)
quests[87190469] = 75 -- Bladesing the Elder (Alliance)
quests[87200067] = 75 -- Skygleam the Elder (Horde)
quests[87200469] = 75 -- Skygleam the Elder (Alliance)
quests[87210067] = 75 -- Starweave the Elder (Horde)
quests[87210469] = 75 -- Starweave the Elder (Alliance)
quests[87220067] = 75 -- Meadowrun the Elder (Horde)
quests[87220469] = 75 -- Meadowrun the Elder (Alliance)
quests[87230067] = 75 -- Nightwind the Elder (Horde)
quests[87230469] = 75 -- Nightwind the Elder (Alliance)
quests[87240067] = 75 -- Morningdew the Elder (Horde)
quests[87240469] = 75 -- Morningdew the Elder (Alliance)
quests[87250067] = 75 -- Riversong the Elder (Horde)
quests[87250469] = 75 -- Riversong the Elder (Alliance)
quests[87260067] = 75 -- Brightspear the Elder (Horde)
quests[87260469] = 75 -- Brightspear the Elder (Alliance)
quests[87270067] = 75 -- Farwhisper the Elder (Horde)
quests[87270469] = 75 -- Farwhisper the Elder (Alliance)
quests[87470910] = 250 -- The Path of the Protector (Brood of Nozdormu)
quests[87480910] = 250 -- The Path of the Protector (Brood of Nozdormu)
quests[87490910] = 250 -- The Path of the Protector (Brood of Nozdormu)
quests[87500910] = 250 -- The Path of the Protector (Brood of Nozdormu)
quests[87510910] = 250 -- The Protector of Kalimdor (Brood of Nozdormu)
quests[87520910] = 250 -- The Path of the Conqueror (Brood of Nozdormu)
quests[87530910] = 250 -- The Path of the Conqueror (Brood of Nozdormu)
quests[87540910] = 250 -- The Path of the Conqueror (Brood of Nozdormu)
quests[87550910] = 250 -- The Path of the Conqueror (Brood of Nozdormu)
quests[87560910] = 250 -- The Qiraji Conqueror (Brood of Nozdormu)
quests[87570910] = 250 -- The Path of the Invoker (Brood of Nozdormu)
quests[87580910] = 250 -- The Path of the Invoker (Brood of Nozdormu)
quests[87590910] = 250 -- The Path of the Invoker (Brood of Nozdormu)
quests[87600910] = 250 -- The Path of the Invoker (Brood of Nozdormu)
quests[87610910] = 250 -- The Grand Invoker (Brood of Nozdormu)
quests[87840910] = 1000 -- Secrets of the Qiraji (Brood of Nozdormu)
quests[87890910] = 500 -- Imperial Qiraji Armaments (Brood of Nozdormu)
quests[87900910] = 500 -- Imperial Qiraji Regalia (Brood of Nozdormu)
quests[87910609] = 500 -- The Fall of Ossirian (Cenarion Circle)
quests[87910910] = 500 -- The Fall of Ossirian (Brood of Nozdormu)
quests[88010609] = 500 -- C'Thun's Legacy (Cenarion Circle)
quests[88010910] = 500 -- C'Thun's Legacy (Brood of Nozdormu)
quests[88020609] = 500 -- The Savior of Kalimdor (Cenarion Circle)
quests[88020910] = 1000 -- The Savior of Kalimdor (Brood of Nozdormu)
quests[88600072] = 250 -- New Year Celebrations! (Stormwind)
quests[88610081] = 250 -- New Year Celebrations! (Thunder Bluff)
quests[88660067] = 75 -- Bronzebeard the Elder (Horde)
quests[88660469] = 75 -- Bronzebeard the Elder (Alliance)
quests[88670609] = 75 -- Lunar Fireworks (Cenarion Circle)
quests[88680609] = 250 -- Elune's Blessing (Cenarion Circle)
quests[88840911] = 250 -- Fish Heads, Fish Heads... (Silvermoon City)
quests[88850911] = 350 -- The Ring of Mmmrrrggglll (Silvermoon City)
quests[88860911] = 250 -- Grimscale Pirates! (Silvermoon City)
quests[88870911] = 250 -- Captain Kelisendra's Lost Rutters (Silvermoon City)
quests[88880911] = 10 -- The Magister's Apprentice (Silvermoon City)
quests[88890911] = 150 -- Deactivating the Spire (Silvermoon City)
quests[88900911] = 250 -- Word from the Spire (Silvermoon City)
quests[88910911] = 250 -- Abandoned Investigations (Silvermoon City)
quests[88920911] = 250 -- Situation at Sunsail Anchorage (Silvermoon City)
quests[88940911] = 250 -- Cleaning up the Grounds (Silvermoon City)
quests[88950911] = 25 -- Delivery to the North Sanctum (Silvermoon City)
quests[90350911] = 10 -- Roadside Ambush (Silvermoon City)
quests[90620911] = 75 -- Soaked Pages (Silvermoon City)
quests[90640911] = 25 -- Taking the Fall (Silvermoon City)
quests[90660911] = 250 -- Swift Discipline (Silvermoon City)
quests[90670911] = 250 -- The Party Never Ends (Silvermoon City)
quests[90760911] = 250 -- Wretched Ringleader (Silvermoon City)
quests[91190911] = 250 -- Malfunction at the West Sanctum (Silvermoon City)
quests[91300911] = 25 -- Goods from Silvermoon City (Silvermoon City)
quests[91330911] = 75 -- Fly to Silvermoon City (Silvermoon City)
quests[91340911] = 25 -- Skymistress Gloaming (Silvermoon City)
quests[91350911] = 250 -- Return to Sathiel (Silvermoon City)
quests[91380922] = 1000 -- Suncrown Village (Tranquillien)
quests[91390922] = 1000 -- Goldenmist Village (Tranquillien)
quests[91400922] = 3000 -- Windrunner Village (Tranquillien)
quests[91430911] = 250 -- Dealing with Zeb'Sora (Silvermoon City)
quests[91440911] = 75 -- Missing in the Ghostlands (Silvermoon City)
quests[91470922] = 1000 -- The Fallen Courier (Tranquillien)
quests[91480922] = 250 -- Delivery to Tranquillien (Tranquillien)
quests[91490922] = 1000 -- The Plagued Coast (Tranquillien)
quests[91500922] = 1000 -- Salvaging the Past (Tranquillien)
quests[91510922] = 250 -- The Sanctum of the Sun (Tranquillien)
quests[91520922] = 1000 -- Tomber's Supplies (Tranquillien)
quests[91550922] = 1000 -- Down the Dead Scar (Tranquillien)
quests[91560922] = 3000 -- Wanted: Knucklerot and Luzran (Tranquillien)
quests[91570922] = 1000 -- Forgotten Rituals (Tranquillien)
quests[91580911] = 250 -- Bearers of the Plague (Silvermoon City)
quests[91590911] = 250 -- Curbing the Plague (Silvermoon City)
quests[91600922] = 1000 -- Investigate An'daroth (Tranquillien)
quests[91610911] = 250 -- The Traitor's Shadow (Silvermoon City)
quests[91620911] = 250 -- Hints of the Past (Silvermoon City)
quests[91630922] = 1000 -- Into Occupied Territory (Tranquillien)
quests[91640922] = 3000 -- Captives at Deatholme (Tranquillien)
quests[91660922] = 250 -- Deliver the Plans to An'telas (Tranquillien)
quests[91670922] = 8000 -- The Traitor's Destruction (Tranquillien)
quests[91690922] = 1000 -- Deactivate An'owyn (Tranquillien)
quests[91700922] = 3000 -- Dar'Khan's Lieutenants (Tranquillien)
quests[91710922] = 1000 -- Culinary Crunch (Tranquillien)
quests[91720922] = 250 -- Report to Magister Kaendris (Tranquillien)
quests[91730922] = 1000 -- Retaking Windrunner Spire (Tranquillien)
quests[91740922] = 1000 -- Vanquishing Aquantion (Tranquillien)
quests[91750922] = 1000 -- The Lady's Necklace (Tranquillien)
quests[91760922] = 2000 -- The Twin Ziggurats (Tranquillien)
quests[91800068] = 250 -- Journey to Undercity (Undercity)
quests[91920922] = 1000 -- Trouble at the Underlight Mines (Tranquillien)
quests[91930922] = 2000 -- Investigate the Amani Catacombs (Tranquillien)
quests[91990922] = 1000 -- Troll Juju (Tranquillien)
quests[92070922] = 1000 -- Underlight Ore Samples (Tranquillien)
quests[92120911] = 250 -- Escape from the Catacombs (Silvermoon City)
quests[92140911] = 250 -- Shadowpine Weaponry (Silvermoon City)
quests[92150911] = 350 -- Bring Me Kel'gash's Head! (Silvermoon City)
quests[92160922] = 500 -- Rotting Hearts (Tranquillien)
quests[92170922] = 500 -- More Rotting Hearts (Tranquillien)
quests[92180922] = 500 -- Spinal Dust (Tranquillien)
quests[92190922] = 500 -- More Spinal Dust (Tranquillien)
quests[92200922] = 3000 -- War on Deatholme (Tranquillien)
quests[92500910] = 500 -- Frame of Atiesh (Brood of Nozdormu)
quests[92500529] = 500 -- Frame of Atiesh (Argent Dawn)
quests[92510910] = 500 -- Atiesh, the Befouled Greatstaff (Brood of Nozdormu)
quests[92510529] = 500 -- Atiesh, the Befouled Greatstaff (Argent Dawn)
quests[92520911] = 250 -- Defending Fairbreeze Village (Silvermoon City)
quests[92530911] = 25 -- Runewarden Deryan (Silvermoon City)
quests[92540911] = 75 -- The Wayward Apprentice (Silvermoon City)
quests[92550911] = 75 -- Research Notes (Silvermoon City)
quests[92560911] = 75 -- Fairbreeze Village (Silvermoon City)
quests[92570910] = 500 -- Atiesh, Greatstaff of the Guardian (Brood of Nozdormu)
quests[92570529] = 500 -- Atiesh, Greatstaff of the Guardian (Argent Dawn)
quests[92580911] = 25 -- The Scorched Grove (Silvermoon City)
quests[92590021] = 500 -- Traitor to the Bloodsail (Booty Bay)
quests[92590087] = -500 -- Traitor to the Bloodsail (Bloodsail Buccaneers)
quests[92660577] = 500 -- Making Amends (Everlook)
quests[92660087] = -500 -- Making Amends (Bloodsail Buccaneers)
quests[92670470] = 500 -- Mending Old Wounds (Ratchet)
quests[92670087] = -500 -- Mending Old Wounds (Bloodsail Buccaneers)
quests[92680369] = 500 -- War at Sea (Gadgetzan)
quests[92680087] = -500 -- War at Sea (Bloodsail Buccaneers)
quests[92690910] = 500 -- Atiesh, Greatstaff of the Guardian (Brood of Nozdormu)
quests[92690529] = 500 -- Atiesh, Greatstaff of the Guardian (Argent Dawn)
quests[92700910] = 500 -- Atiesh, Greatstaff of the Guardian (Brood of Nozdormu)
quests[92700529] = 500 -- Atiesh, Greatstaff of the Guardian (Argent Dawn)
quests[92710910] = 500 -- Atiesh, Greatstaff of the Guardian (Brood of Nozdormu)
quests[92710529] = 500 -- Atiesh, Greatstaff of the Guardian (Argent Dawn)
quests[92720087] = 10 -- Dressing the Part (Bloodsail Buccaneers)
quests[92720021] = -250 -- Dressing the Part (Booty Bay)
quests[92740911] = 250 -- Spirits of the Drowned (Silvermoon City)
quests[92750922] = 1000 -- A Little Dash of Seasoning (Tranquillien)
quests[92760911] = 250 -- Attack on Zeb'Tela (Silvermoon City)
quests[92770911] = 250 -- Assault on Zeb'Nowa (Silvermoon City)
quests[92790930] = 75 -- You Survived! (Exodar)
quests[92800930] = 250 -- Replenishing the Healing Crystals (Exodar)
quests[92810922] = 1000 -- Clearing the Way (Tranquillien)
quests[92830930] = 350 -- Rescue the Survivors! (Exodar)
quests[92870930] = 75 -- Paladin Training (Exodar)
quests[92880930] = 75 -- Hunter Training (Exodar)
quests[92890930] = 75 -- Warrior Training (Exodar)
quests[92900930] = 75 -- Mage Training (Exodar)
quests[92910930] = 75 -- Priest Training (Exodar)
quests[92930930] = 250 -- What Must Be Done... (Exodar)
quests[92940930] = 250 -- Healing the Lake (Exodar)
quests[93030930] = 350 -- Inoculation (Exodar)
quests[93050930] = 250 -- Spare Parts (Exodar)
quests[93090930] = 25 -- The Missing Scout (Exodar)
quests[93110930] = 500 -- Blood Elf Spy (Exodar)
quests[93120930] = 25 -- The Emitter (Exodar)
quests[93130930] = 350 -- Travel to Azure Watch (Exodar)
quests[93140930] = 25 -- Word from Azure Watch (Exodar)
quests[93150922] = 1000 -- Anok'suten (Tranquillien)
quests[93270922] = 250 -- The Forsaken (Tranquillien)
quests[93280911] = 75 -- Hero of the Sin'dorei (Silvermoon City)
quests[93290922] = 250 -- The Forsaken (Tranquillien)
quests[93400911] = 250 -- The Great Fissure (Silvermoon City)
quests[93450947] = 250 -- Preparing the Salve (Thrallmar)
quests[93520911] = 250 -- Darnassian Intrusions (Silvermoon City)
quests[93550946] = 250 -- A Job for an Intelligent Man (Honor Hold)
quests[93580911] = 10 -- Ranger Sareyn (Silvermoon City)
quests[93600911] = 250 -- Amani Invasion (Silvermoon City)
quests[93630911] = 250 -- Warning Fairbreeze Village (Silvermoon City)
quests[93660911] = 250 -- In Need of Felblood (Silvermoon City)
quests[93690930] = 250 -- Replenishing the Healing Crystals (Exodar)
quests[93700911] = 250 -- The Cleansing Must Be Stopped (Silvermoon City)
quests[93710930] = 25 -- Botanist Taerix (Exodar)
quests[93720942] = 250 -- Demonic Contamination (Cenarion Expedition)
quests[93730942] = 250 -- Missing Missive (Cenarion Expedition)
quests[93740911] = 250 -- Arelion's Journal (Silvermoon City)
quests[93750911] = 250 -- The Road to Falcon Watch (Silvermoon City)
quests[93760911] = 250 -- A Pilgrim's Plight (Silvermoon City)
quests[93810911] = 250 -- Trueflight Arrows (Silvermoon City)
quests[93830930] = 250 -- An Ambitious Plan (Exodar)
quests[93850946] = 250 -- Rampaging Ravagers (Honor Hold)
quests[93870911] = 250 -- Source of the Corruption (Silvermoon City)
quests[93900930] = 25 -- In Search of Sedai (Exodar)
quests[93910911] = 250 -- Marking the Path (Silvermoon City)
quests[93920911] = 75 -- Rogue Training (Silvermoon City)
quests[93930911] = 75 -- Hunter Training (Silvermoon City)
quests[93940911] = 10 -- Where's Wyllithen? (Silvermoon City)
quests[93950911] = 10 -- Saltheril's Haven (Silvermoon City)
quests[93960911] = 250 -- Magic of the Arakkoa (Silvermoon City)
quests[93970911] = 250 -- Birds of a Feather (Silvermoon City)
quests[93980930] = 250 -- Deadly Predators (Exodar)
quests[93990930] = 250 -- Cruel Taskmasters (Exodar)
quests[94000947] = 250 -- The Assassin (Thrallmar)
quests[94010947] = 250 -- A Strange Weapon (Thrallmar)
quests[94050947] = 75 -- The Warchief's Mandate (Thrallmar)
quests[94060947] = 250 -- The Mag'har (Thrallmar)
quests[94060941] = 250 -- The Mag'har (The Mag'har)
quests[94070947] = 250 -- Through the Dark Portal (Thrallmar)
quests[94090930] = 25 -- Urgent Delivery! (Exodar)
quests[94100941] = 350 -- A Spirit Guide (The Mag'har)
quests[94170946] = 250 -- The Arakkoa Threat (Honor Hold)
quests[94190609] = 250 -- Scouring the Desert (Cenarion Circle)
quests[94200946] = 250 -- The Finest Down (Honor Hold)
quests[94210930] = 75 -- Shaman Training (Exodar)
quests[94220609] = 250 -- Scouring the Desert (Cenarion Circle)
quests[94230930] = 25 -- Return to Obadei (Exodar)
quests[94260930] = 250 -- The Pools of Aggonar (Exodar)
quests[94270930] = 250 -- Cleansing the Waters (Exodar)
quests[94300930] = 250 -- Sha'naar Relics (Exodar)
quests[94360911] = 250 -- Bloodscalp Insight (Silvermoon City)
quests[94370911] = 250 -- Twilight of the Dawn Runner (Silvermoon City)
quests[94420941] = 250 -- A Debilitating Sickness (The Mag'har)
quests[94470941] = 250 -- Administering the Salve (The Mag'har)
quests[94520930] = 250 -- Red Snapper - Very Tasty! (Exodar)
quests[94530930] = 10 -- Find Acteon! (Exodar)
quests[94540930] = 150 -- The Great Moongraze Hunt (Exodar)
quests[94550930] = 150 -- Strange Findings (Exodar)
quests[94560930] = 250 -- Nightstalker Clean Up, Isle 2... (Exodar)
quests[94570911] = 250 -- An Unusual Patron (Silvermoon City)
quests[94630930] = 250 -- Medicinal Purpose (Exodar)
quests[94660911] = 350 -- Wanted: Blacktalon the Savage (Silvermoon City)
quests[94720911] = 350 -- Arelion's Mistress (Silvermoon City)
quests[94730930] = 250 -- An Alternative Alternative (Exodar)
quests[94900930] = 350 -- The Rock Flayer Matriarch (Exodar)
quests[95050930] = 150 -- The Prophecy of Velen (Exodar)
quests[95060930] = 250 -- A Small Start (Exodar)
quests[95060072] = 75 -- A Small Start (Stormwind)
quests[95120072] = 150 -- Cookie's Jumbo Gumbo (Stormwind)
quests[95120930] = 150 -- Cookie's Jumbo Gumbo (Exodar)
quests[95130069] = 250 -- Reclaiming the Ruins (Darnassus)
quests[95130930] = 150 -- Reclaiming the Ruins (Exodar)
quests[95140069] = 10 -- Rune Covered Tablet (Darnassus)
quests[95140930] = 10 -- Rune Covered Tablet (Exodar)
quests[95150069] = 500 -- Warlord Sriss'tiz (Darnassus)
quests[95150930] = 250 -- Warlord Sriss'tiz (Exodar)
quests[95230047] = 250 -- Precious and Fragile Things Need Special Handling (Ironforge)
quests[95230930] = 150 -- Precious and Fragile Things Need Special Handling (Exodar)
quests[95270069] = 250 -- All That Remains (Darnassus)
quests[95270930] = 150 -- All That Remains (Exodar)
quests[95280069] = 250 -- A Cry For Help (Darnassus)
quests[95280930] = 150 -- A Cry For Help (Exodar)
quests[95300072] = 250 -- I've Got a Plant (Stormwind)
quests[95300930] = 150 -- I've Got a Plant (Exodar)
quests[95310072] = 500 -- Tree's Company (Stormwind)
quests[95310930] = 250 -- Tree's Company (Exodar)
quests[95370930] = 500 -- Show Gnomercy (Exodar)
quests[95430930] = 10 -- Atonement (Exodar)
quests[95440930] = 500 -- The Prophecy of Akida (Exodar)
quests[95450930] = 350 -- The Seer's Relic (Exodar)
quests[95480054] = 250 -- Pilfered Equipment (Gnomeregan)
quests[95490054] = 250 -- Artifacts of the Blacksilt (Gnomeregan)
quests[95500069] = 250 -- A Map to Where? (Darnassus)
quests[95570930] = 250 -- Deciphering the Book (Exodar)
quests[95590930] = 25 -- Stillpine Hold (Exodar)
quests[95610069] = 250 -- Nolkai's Words (Darnassus)
quests[95630946] = 250 -- Gaining Mirren's Trust (Honor Hold)
quests[95670930] = 250 -- Know Thine Enemy (Exodar)
quests[95690930] = 250 -- Containing the Threat (Exodar)
quests[95740930] = 250 -- Victims of Corruption (Exodar)
quests[95760930] = 250 -- Cruelfin's Necklace (Exodar)
quests[95780930] = 250 -- Searching for Galaen (Exodar)
quests[95790930] = 250 -- Galaen's Fate (Exodar)
quests[95800930] = 250 -- The Bear Necessities (Exodar)
quests[95810930] = 250 -- Learning from the Crystals (Exodar)
quests[95820930] = 250 -- Strength of One (Exodar)
quests[95840930] = 250 -- The Second Sample (Exodar)
quests[95850930] = 250 -- The Final Sample (Exodar)
quests[95940930] = 250 -- Signs of the Legion (Exodar)
quests[96020930] = 250 -- Deliver Them From Evil... (Exodar)
quests[96030930] = 75 -- Beds, Bandages, and Beyond (Exodar)
quests[96040930] = 75 -- On the Wings of a Hippogryph (Exodar)
quests[96050930] = 75 -- Hippogryph Master Stephanos (Exodar)
quests[96060930] = 250 -- Return to Caregiver Chellan (Exodar)
quests[96120930] = 350 -- A Hearty Thanks! (Exodar)
quests[96160930] = 250 -- Bandits! (Exodar)
quests[96200930] = 150 -- The Missing Survey Team (Exodar)
quests[96210068] = 75 -- Envoy to the Horde (Undercity)
quests[96220930] = 250 -- Warn Your People (Exodar)
quests[96230930] = 10 -- Coming of Age (Exodar)
quests[96240930] = 250 -- A Favorite Treat (Exodar)
quests[96250930] = 25 -- Elekks Are Serious Business (Exodar)
quests[96260076] = 250 -- Meeting the Warchief (Orgrimmar)
quests[96270911] = 500 -- Allegiance to the Horde (Silvermoon City)
quests[96280930] = 250 -- Salvaging the Data (Exodar)
quests[96290930] = 250 -- Catch and Release (Exodar)
quests[96300967] = 250 -- Medivh's Journal (The Violet Eye)
quests[96310967] = 75 -- A Colleague's Aid (The Violet Eye)
quests[96320069] = 75 -- Newfound Allies (Darnassus)
quests[96340930] = 250 -- Alien Predators (Exodar)
quests[96370967] = 250 -- Kalynna's Request (The Violet Eye)
quests[96380967] = 10 -- In Good Hands (The Violet Eye)
quests[96390967] = 10 -- Kamsis (The Violet Eye)
quests[96400967] = 250 -- The Shade of Aran (The Violet Eye)
quests[96410930] = 250 -- Irradiated Crystal Shards (Exodar)
quests[96430930] = 250 -- Constrictor Vines (Exodar)
quests[96440967] = 3000 -- Nightbane (The Violet Eye)
quests[96450967] = 250 -- The Master's Terrace (The Violet Eye)
quests[96460930] = 250 -- WANTED: Deathclaw (Exodar)
quests[96470930] = 250 -- Culling the Flutterers (Exodar)
quests[96480930] = 250 -- Mac'Aree Mushroom Menagerie (Exodar)
quests[96490930] = 250 -- Ysera's Tears (Exodar)
quests[96630930] = 250 -- The Kessel Run (Exodar)
quests[96660930] = 250 -- Declaration of Power (Exodar)
quests[96670930] = 250 -- Saving Princess Stillpine (Exodar)
quests[96680930] = 10 -- Report to Exarch Admetius (Exodar)
quests[96690930] = 250 -- The Missing Expedition (Exodar)
quests[96700930] = 250 -- They're Alive! Maybe... (Exodar)
quests[96720072] = 25 -- The Bloodcurse Legacy (Stormwind)
quests[96740072] = 250 -- The Bloodcursed Naga (Stormwind)
quests[96760911] = 75 -- Paladin Training (Silvermoon City)
quests[96800967] = 250 -- Digging Up the Past (The Violet Eye)
quests[96820072] = 250 -- The Hopeless Ones... (Stormwind)
quests[96830072] = 250 -- Ending the Bloodcurse (Stormwind)
quests[96870069] = 250 -- Restoring Sanctity (Darnassus)
quests[96880069] = 250 -- Into the Dream (Darnassus)
quests[96890069] = 500 -- Razormaw (Darnassus)
quests[96930930] = 10 -- What Argus Means to Me (Exodar)
quests[96940930] = 250 -- Blood Watch (Exodar)
quests[96960930] = 10 -- Translations... (Exodar)
quests[96970942] = 75 -- Watcher Leesa'oh (Cenarion Expedition)
quests[96980930] = 250 -- Audience with the Prophet (Exodar)
quests[96990930] = 25 -- Truth or Fiction (Exodar)
quests[97000930] = 250 -- I Shoot Magic Into the Darkness (Exodar)
quests[97010942] = 250 -- Observing the Sporelings (Cenarion Expedition)
quests[97020942] = 250 -- A Question of Gluttony (Cenarion Expedition)
quests[97030930] = 250 -- The Cryo-Core (Exodar)
quests[97040911] = 25 -- Slain by the Wretched (Silvermoon City)
quests[97050911] = 75 -- Package Recovery (Silvermoon City)
quests[97060930] = 150 -- Galaen's Journal - The Fate of Vindicator Saruan (Exodar)
quests[97080942] = 250 -- Familiar Fungi (Cenarion Expedition)
quests[97090942] = 350 -- Stealing Back the Mushrooms (Cenarion Expedition)
quests[97110930] = 250 -- Matis the Cruel (Exodar)
quests[97160942] = 250 -- Disturbance at Umbrafen Lake (Cenarion Expedition)
quests[97180942] = 150 -- As the Crow Flies (Cenarion Expedition)
quests[97200942] = 350 -- Balance Must Be Preserved (Cenarion Expedition)
quests[97240942] = 75 -- Warning the Cenarion Circle (Cenarion Expedition)
quests[97260970] = 250 -- Now That We're Friends... (Sporeggar)
quests[97270970] = 750 -- Now That We're Still Friends... (Sporeggar)
quests[97280942] = 350 -- A Warm Welcome (Cenarion Expedition)
quests[97290970] = 350 -- Fhwoor Smash! (Sporeggar)
quests[97300942] = 350 -- Leader of the Darkcrest (Cenarion Expedition)
quests[97310942] = 250 -- Drain Schematics (Cenarion Expedition)
quests[97320942] = 250 -- Return to the Marsh (Cenarion Expedition)
quests[97390970] = 250 -- The Sporelings' Plight (Sporeggar)
quests[97400930] = 250 -- The Sun Gate (Exodar)
quests[97410930] = 150 -- Critters of the Void (Exodar)
quests[97420970] = 250 -- More Spore Sacs (Sporeggar)
quests[97430970] = 250 -- Natural Enemies (Sporeggar)
quests[97440970] = 250 -- More Tendrils! (Sporeggar)
quests[97460930] = 250 -- Limits of Physical Exhaustion (Exodar)
quests[97470942] = 250 -- The Umbrafen Tribe (Cenarion Expedition)
quests[97480930] = 250 -- Don't Drink the Water (Exodar)
quests[97520942] = 250 -- Escape from Umbrafen (Cenarion Expedition)
quests[97530930] = 10 -- What We Know... (Exodar)
quests[97560930] = 150 -- What We Don't Know... (Exodar)
quests[97580922] = 250 -- Return to Arcanist Vandril (Tranquillien)
quests[97590930] = 500 -- Ending Their World (Exodar)
quests[97600930] = 25 -- Vindicator's Rest (Exodar)
quests[97610930] = 350 -- Clearing the Way (Exodar)
quests[97620930] = 500 -- The Unwritten Prophecy (Exodar)
quests[97690530] = 250 -- There's No Explanation for Fashion (Darkspear Trolls)
quests[97700530] = 250 -- Menacing Marshfangs (Darkspear Trolls)
quests[97710530] = 250 -- Searching for Scout Jyoba (Darkspear Trolls)
quests[97720530] = 250 -- Jyoba's Report (Darkspear Trolls)
quests[97730530] = 250 -- No More Mushrooms! (Darkspear Trolls)
quests[97740530] = 250 -- Thick Hydra Scales (Darkspear Trolls)
quests[97770930] = 250 -- Fulgor Spores (Exodar)
quests[97790930] = 250 -- Intercepting the Message (Exodar)
quests[97800930] = 250 -- Umbrafen Eel Filets (Exodar)
quests[97810930] = 250 -- Too Many Mouths to Feed (Exodar)
quests[97820930] = 250 -- The Dead Mire (Exodar)
quests[97830930] = 250 -- An Unnatural Drought (Exodar)
quests[97840942] = 250 -- Identify Plant Parts (Cenarion Expedition)
quests[97850942] = 75 -- Blessings of the Ancients (Cenarion Expedition)
quests[97860930] = 250 -- The Boha'mu Ruins (Exodar)
quests[97870930] = 250 -- Idols of the Feralfen (Exodar)
quests[97880942] = 250 -- A Damp, Dark Place (Cenarion Expedition)
quests[97900930] = 250 -- Diaphanous Wings (Exodar)
quests[97910930] = 250 -- Menacing Marshfangs (Exodar)
quests[97980930] = 250 -- Blood Elf Plans (Exodar)
quests[97990930] = 250 -- Botanical Legwork (Exodar)
quests[98010930] = 250 -- Gathering the Reagents (Exodar)
quests[98020942] = 250 -- Plants of Zangarmarsh (Cenarion Expedition)
quests[98030930] = 250 -- Messenger to the Feralfen (Exodar)
quests[98060970] = 250 -- Fertile Spores (Sporeggar)
quests[98070970] = 250 -- More Fertile Spores (Sporeggar)
quests[98080970] = 250 -- Glowcap Mushrooms (Sporeggar)
quests[98090970] = 250 -- More Glowcaps (Sporeggar)
quests[98110911] = 75 -- Friend of the Sin'dorei (Silvermoon City)
quests[98120068] = 75 -- Envoy to the Horde (Undercity)
quests[98130076] = 250 -- Meeting the Warchief (Orgrimmar)
quests[98140530] = 250 -- Burstcap Mushrooms, Mon! (Darkspear Trolls)
quests[98160530] = 250 -- Have You Ever Seen One of These? (Darkspear Trolls)
quests[98170942] = 350 -- Leader of the Bloodscale (Cenarion Expedition)
quests[98220530] = 250 -- Impending Attack (Darkspear Trolls)
quests[98230530] = 250 -- Us or Them (Darkspear Trolls)
quests[98240967] = 250 -- Arcane Disturbances (The Violet Eye)
quests[98250967] = 250 -- Restless Activity (The Violet Eye)
quests[98260967] = 75 -- Contact from Dalaran (The Violet Eye)
quests[98270930] = 250 -- Withered Basidium (Exodar)
quests[98280530] = 250 -- Withered Basidium (Darkspear Trolls)
quests[98290967] = 75 -- Khadgar (The Violet Eye)
quests[98300978] = 250 -- Stinger Venom (Kurenai)
quests[98310967] = 250 -- Entry Into Karazhan (The Violet Eye)
quests[98320967] = 350 -- The Second and Third Fragments (The Violet Eye)
quests[98330978] = 250 -- Lines of Communication (Kurenai)
quests[98340978] = 250 -- Natural Armor (Kurenai)
quests[98350978] = 250 -- Ango'rosh Encroachment (Kurenai)
quests[98360967] = 350 -- The Master's Touch (The Violet Eye)
quests[98370967] = 500 -- Return to Khadgar (The Violet Eye)
quests[98380967] = 75 -- The Violet Eye (The Violet Eye)
quests[98390978] = 250 -- Overlord Gorefist (Kurenai)
quests[98400967] = 250 -- Assessing the Situation (The Violet Eye)
quests[98410530] = 250 -- Stinging the Stingers (Darkspear Trolls)
quests[98420530] = 250 -- The Sharpest Blades (Darkspear Trolls)
quests[98430967] = 250 -- Keanna's Log (The Violet Eye)
quests[98440967] = 250 -- A Demonic Presence (The Violet Eye)
quests[98450530] = 250 -- Angling to Beat the Competition (Darkspear Trolls)
quests[98460530] = 250 -- Spirits of the Feralfen (Darkspear Trolls)
quests[98470530] = 250 -- A Spirit Ally? (Darkspear Trolls)
quests[98480072] = 250 -- Secrets of the Daggerfen (Stormwind)
quests[98600967] = 3000 -- The New Directive (The Violet Eye)
quests[98630941] = 250 -- Vile Idolatry (The Mag'har)
quests[98640941] = 25 -- The Missing War Party (The Mag'har)
quests[98650941] = 250 -- Once Were Warriors (The Mag'har)
quests[98660941] = 75 -- He Will Walk The Earth... (The Mag'har)
quests[98670941] = 250 -- Murkblood Leaders... (The Mag'har)
quests[98680941] = 350 -- The Totem of Kar'dash (The Mag'har)
quests[98690978] = 25 -- The Throne of the Elements (Kurenai)
quests[98700941] = 25 -- The Throne of the Elements (The Mag'har)
quests[98710978] = 250 -- Murkblood Invaders (Kurenai)
quests[98720941] = 250 -- Murkblood Invaders (The Mag'har)
quests[98730978] = 250 -- Ortor My Old Friend... (Kurenai)
quests[98740978] = 250 -- Stopping the Spread (Kurenai)
quests[98750942] = 500 -- Uncatalogued Species (Cenarion Expedition)
quests[98770922] = 250 -- A Restorative Draught (Tranquillien)
quests[98780978] = 250 -- Solving the Problem (Kurenai)
quests[98790978] = 350 -- The Totem of Kar'dash (Kurenai)
quests[98820933] = 250 -- Stealing from Thieves (The Consortium)
quests[98830933] = 250 -- More Crystal Fragments (The Consortium)
quests[98880941] = 250 -- The Impotent Leader (The Mag'har)
quests[98890941] = 500 -- Don't Kill the Fat One (The Mag'har)
quests[98900941] = 10 -- Success! (The Mag'har)
quests[98910941] = 250 -- Because Kilrath is a Coward (The Mag'har)
quests[98920933] = 250 -- More Obsidian Warbeads (The Consortium)
quests[98930933] = 250 -- Obsidian Warbeads (The Consortium)
quests[98940942] = 250 -- Safeguarding the Watchers (Cenarion Expedition)
quests[98950942] = 250 -- The Dying Balance (Cenarion Expedition)
quests[98960047] = 250 -- Blacksting's Bane (Ironforge)
quests[98980530] = 250 -- Nothin' Says Lovin' Like a Big Stinger (Darkspear Trolls)
quests[98990530] = 250 -- A Job Undone (Darkspear Trolls)
quests[99000933] = 250 -- Gava'xi (The Consortium)
quests[99010047] = 250 -- Unfinished Business (Ironforge)
quests[99020978] = 250 -- The Terror of Marshlight Lake (Kurenai)
quests[99030530] = 250 -- The Biggest of Them All (Darkspear Trolls)
quests[99040530] = 250 -- Pursuing Terrorclaw (Darkspear Trolls)
quests[99050978] = 250 -- Maktu's Revenge (Kurenai)
quests[99060941] = 250 -- Message in a Battle (The Mag'har)
quests[99070941] = 350 -- An Audacious Advance (The Mag'har)
quests[99100941] = 250 -- Standards and Practices (The Mag'har)
quests[99110942] = 250 -- The Count of the Marshes (Cenarion Expedition)
quests[99130933] = 10 -- The Consortium Needs You! (The Consortium)
quests[99140933] = 250 -- A Head Full of Ivory (The Consortium)
quests[99150933] = 250 -- More Heads Full of Ivory (The Consortium)
quests[99160941] = 250 -- Bleeding Hollow Supply Crates (The Mag'har)
quests[99170978] = 250 -- Do My Eyes Deceive Me (Kurenai)
quests[99180978] = 500 -- Not On My Watch! (Kurenai)
quests[99190970] = 250 -- Sporeggar (Sporeggar)
quests[99200978] = 10 -- Mo'mor the Breaker (Kurenai)
quests[99210978] = 250 -- The Ruins of Burning Blade (Kurenai)
quests[99220978] = 350 -- The Twin Clefts of Nagrand (Kurenai)
quests[99230978] = 250 -- HELP! (Kurenai)
quests[99240978] = 250 -- Corki's Gone Missing Again! (Kurenai)
quests[99250933] = 250 -- Matters of Security (The Consortium)
quests[99330978] = 350 -- Message to Telaar (Kurenai)
quests[99340941] = 350 -- Message to Garadar (The Mag'har)
quests[99350941] = 250 -- Wanted: Giselda the Crone (The Mag'har)
quests[99360978] = 250 -- Wanted: Giselda the Crone (Kurenai)
quests[99370941] = 350 -- Wanted: Durn the Hungerer (The Mag'har)
quests[99380978] = 350 -- Wanted: Durn the Hungerer (Kurenai)
quests[99390941] = 250 -- Wanted: Zorbo the Advisor (The Mag'har)
quests[99400978] = 250 -- Wanted: Zorbo the Advisor (Kurenai)
quests[99440941] = 10 -- Missing Mag'hari Procession (The Mag'har)
quests[99450941] = 250 -- War on the Warmaul (The Mag'har)
quests[99460941] = 350 -- Cho'war the Pillager (The Mag'har)
quests[99480941] = 350 -- Finding the Survivors (The Mag'har)
quests[99510942] = 350 -- It's Watching You! (Cenarion Expedition)
quests[99540978] = 250 -- Corki's Ransom (Kurenai)
quests[99550978] = 350 -- Cho'war the Pillager (Kurenai)
quests[99560978] = 350 -- The Ravaged Caravan (Kurenai)
quests[99570942] = 25 -- What's Wrong at Cenarion Thicket? (Cenarion Expedition)
quests[99600942] = 25 -- What's Wrong at Cenarion Thicket? (Cenarion Expedition)
quests[99610942] = 25 -- What's Wrong at Cenarion Thicket? (Cenarion Expedition)
quests[99680942] = 250 -- Strange Energy (Cenarion Expedition)
quests[99710942] = 250 -- Clues in the Thicket (Cenarion Expedition)
quests[99900942] = 250 -- Investigate Tuurem (Cenarion Expedition)
quests[100050942] = 500 -- Letting Earthbinder Tavgren Know (Cenarion Expedition)
quests[100060942] = 500 -- Letting Earthbinder Tavgren Know (Cenarion Expedition)
quests[100170932] = 250 -- Strained Supplies (The Aldor)
quests[100190932] = 250 -- More Venom Sacs (The Aldor)
quests[100200932] = 350 -- A Cure for Zahlia (The Aldor)
quests[100210932] = 250 -- Restoring the Light (The Aldor)
quests[100240934] = 250 -- Voren'thal's Visions (The Scryers)
quests[100250934] = 250 -- More Basilisk Eyes (The Scryers)
quests[100301011] = 250 -- Recover the Bones (Lower City)
quests[100311011] = 250 -- Helping the Lost Find Their Way (Lower City)
quests[100440941] = 25 -- A Visit With the Greatmother (The Mag'har)
quests[100450941] = 350 -- Material Components (The Mag'har)
quests[100470930] = 250 -- The Path of Glory (Exodar)
quests[100500946] = 250 -- Unyielding Souls (Honor Hold)
quests[100550946] = 250 -- Waste Not, Want Not (Honor Hold)
quests[100570946] = 250 -- Looking to the Leadership (Honor Hold)
quests[100580946] = 250 -- An Old Gift (Honor Hold)
quests[100650930] = 250 -- Cutting a Path (Exodar)
quests[100660930] = 250 -- Oh, the Tangled Webs They Weave (Exodar)
quests[100670930] = 250 -- Fouled Water Spirits (Exodar)
quests[100680911] = 250 -- Frost Nova (Silvermoon City)
quests[100690911] = 250 -- Ways of the Light (Silvermoon City)
quests[100700911] = 250 -- Steady Shot (Silvermoon City)
quests[100710911] = 250 -- Evisceration (Silvermoon City)
quests[100720911] = 250 -- Learning the Word (Silvermoon City)
quests[100730911] = 250 -- Corruption (Silvermoon City)
quests[100780946] = 250 -- Laying Waste to the Unwanted (Honor Hold)
quests[100790946] = 250 -- When This Mine's a-Rockin' (Honor Hold)
quests[100810941] = 10 -- To Meet Mother Kashur (The Mag'har)
quests[100820941] = 250 -- The Agitated Ancestors (The Mag'har)
quests[100850941] = 250 -- A Visit With The Ancestors (The Mag'har)
quests[100860947] = 250 -- I Work... For the Horde! (Thrallmar)
quests[100870947] = 250 -- Burn It Up... For the Horde! (Thrallmar)
quests[100960942] = 250 -- Saving the Sporeloks (Cenarion Expedition)
quests[100990946] = 250 -- The Mastermind (Honor Hold)
quests[101010941] = 500 -- When Spirits Speak (The Mag'har)
quests[101010935] = 250 -- When Spirits Speak (The Sha'tar)
quests[101020941] = 250 -- A Secret Revealed (The Mag'har)
quests[101020935] = 250 -- A Secret Revealed (The Sha'tar)
quests[101060946] = 150 -- Hellfire Fortifications (Honor Hold)
quests[101070941] = 150 -- Diplomatic Measures (The Mag'har)
quests[101080978] = 150 -- Diplomatic Measures (Kurenai)
quests[101100947] = 150 -- Hellfire Fortifications (Thrallmar)
quests[101150978] = 250 -- Daggerfen Deviance (Kurenai)
quests[101160978] = 250 -- Wanted: Chieftain Mummaki (Kurenai)
quests[101170530] = 250 -- Wanted: Chieftain Mummaki (Darkspear Trolls)
quests[101180530] = 250 -- Message to the Daggerfen (Darkspear Trolls)
quests[101190946] = 250 -- Through the Dark Portal (Honor Hold)
quests[101200947] = 250 -- Arrival in Outland (Thrallmar)
quests[101210947] = 250 -- Eradicate the Burning Legion (Thrallmar)
quests[101230947] = 250 -- Felspark Ravine (Thrallmar)
quests[101240947] = 10 -- Forward Base: Reaver's Fall (Thrallmar)
quests[101290947] = 350 -- Mission: Gateways Murketh and Shaadraz (Thrallmar)
quests[101320942] = 350 -- Colossal Menace (Cenarion Expedition)
quests[101340942] = 75 -- Crimson Crystal Clue (Cenarion Expedition)
quests[101360947] = 500 -- Cruel's Intentions (Thrallmar)
quests[101400946] = 250 -- Journey to Honor Hold (Honor Hold)
quests[101410946] = 250 -- The Legion Reborn (Honor Hold)
quests[101420946] = 250 -- The Path of Anguish (Honor Hold)
quests[101430946] = 10 -- Expedition Point (Honor Hold)
quests[101440946] = 250 -- Disrupt Their Reinforcements (Honor Hold)
quests[101460946] = 350 -- Mission: The Murketh and Shaadraz Gateways (Honor Hold)
quests[101590942] = 250 -- Keep Thornfang Hill Clear! (Cenarion Expedition)
quests[101620947] = 250 -- Mission: The Abyssal Shelf (Thrallmar)
quests[101630946] = 250 -- Mission: The Abyssal Shelf (Honor Hold)
quests[101660911] = 250 -- Whitebark's Memory (Silvermoon City)
quests[101670941] = 350 -- Auchindoun... (The Mag'har)
quests[101670935] = 350 -- Auchindoun... (The Sha'tar)
quests[101680941] = 350 -- What the Soul Sees (The Mag'har)
quests[101680935] = 350 -- What the Soul Sees (The Sha'tar)
quests[101700941] = 10 -- Return to the Greatmother (The Mag'har)
quests[101710941] = 10 -- The Inconsolable Chieftain (The Mag'har)
quests[101720941] = 1000 -- There Is No Hope (The Mag'har)
quests[101890934] = 250 -- Manaforge B'naar (The Scryers)
quests[101930934] = 250 -- High Value Targets (The Scryers)
quests[101940934] = 10 -- Stealth Flight (The Scryers)
quests[101970934] = 250 -- A Convincing Disguise (The Scryers)
quests[101980934] = 150 -- Information Gathering (The Scryers)
quests[102020934] = 75 -- A Defector (The Scryers)
quests[102040934] = 250 -- Bloodgem Crystals (The Scryers)
quests[102050933] = 250 -- Warp-Raider Nesaad (The Consortium)
quests[102080947] = 250 -- Disrupt Their Reinforcements (Thrallmar)
quests[102100935] = 10 -- A'dal (The Sha'tar)
quests[102200947] = 250 -- Make Them Listen (Thrallmar)
quests[102270935] = 10 -- I See Dead Draenei (The Sha'tar)
quests[102280935] = 25 -- Ezekiel (The Sha'tar)
quests[102290947] = 10 -- Decipher the Tome (Thrallmar)
quests[102300947] = 250 -- The Battle Horn (Thrallmar)
quests[102310935] = 250 -- What Book? I Don't See Any Book. (The Sha'tar)
quests[102410932] = 250 -- Distraction at Manaforge B'naar (The Aldor)
quests[102420947] = 25 -- Spinebreaker Post (Thrallmar)
quests[102430932] = 75 -- Naaru Technology (The Aldor)
quests[102450932] = 75 -- B'naar Console Transcription (The Aldor)
quests[102460932] = 250 -- Attack on Manaforge Coruu (The Aldor)
quests[102500947] = 250 -- Bloody Vengeance (Thrallmar)
quests[102510935] = 75 -- The Master's Grand Design? (The Sha'tar)
quests[102520935] = 250 -- Vision of the Dead (The Sha'tar)
quests[102530935] = 500 -- Levixus the Soul Caller (The Sha'tar)
quests[102550942] = 250 -- Testing the Antidote (Cenarion Expedition)
quests[102580947] = 250 -- Honor the Fallen (Thrallmar)
quests[102620933] = 250 -- A Heap of Ethereals (The Consortium)
quests[102630933] = 10 -- Assisting the Consortium (The Consortium)
quests[102640933] = 10 -- Assisting the Consortium (The Consortium)
quests[102650933] = 250 -- Consortium Crystal Collection (The Consortium)
quests[102660933] = 25 -- Request for Assistance (The Consortium)
quests[102670933] = 250 -- Rightful Repossession (The Consortium)
quests[102680933] = 75 -- An Audience with the Prince (The Consortium)
quests[102690933] = 250 -- Triangulation Point One (The Consortium)
quests[102700933] = 250 -- A Not-So-Modest Proposal (The Consortium)
quests[102750933] = 250 -- Triangulation Point Two (The Consortium)
quests[102760933] = 500 -- Full Triangle (The Consortium)
quests[102770989] = 250 -- The Caverns of Time (Keepers of Time)
quests[102780947] = 250 -- The Warp Rifts (Thrallmar)
quests[102790989] = 10 -- To The Master's Lair (Keepers of Time)
quests[102800935] = 1000 -- Special Delivery to Shattrath City (The Sha'tar)
quests[102820989] = 10 -- Old Hillsbrad (Keepers of Time)
quests[102860911] = 150 -- Arelion's Secret (Silvermoon City)
quests[102870911] = 25 -- The Mistress Revealed (Silvermoon City)
quests[102880946] = 250 -- Arrival in Outland (Honor Hold)
quests[102890947] = 250 -- Journey to Thrallmar (Thrallmar)
quests[102900933] = 250 -- In Search of Farahlite (The Consortium)
quests[102930933] = 250 -- Hitting the Motherlode (The Consortium)
quests[102940947] = 250 -- Void Ridge (Thrallmar)
quests[102950947] = 250 -- From the Abyss (Thrallmar)
quests[102970989] = 500 -- The Opening of the Dark Portal (Keepers of Time)
quests[102990932] = 250 -- Shutting Down Manaforge B'naar (The Aldor)
quests[103020930] = 250 -- Volatile Mutations (Exodar)
quests[103030930] = 250 -- The Blood Elves (Exodar)
quests[103040930] = 75 -- Vindicator Aldar (Exodar)
quests[103080933] = 250 -- Another Heap of Ethereals (The Consortium)
quests[103100933] = 350 -- Sabotage the Warp-Gate! (The Consortium)
quests[103110933] = 10 -- Drijya Needs Your Help (The Consortium)
quests[103130932] = 250 -- Measuring Warp Energies (The Aldor)
quests[103150933] = 250 -- Neutralizing the Nethermancers (The Consortium)
quests[103180933] = 250 -- Dealing with the Overmaster (The Consortium)
quests[103210932] = 250 -- Shutting Down Manaforge Coruu (The Aldor)
quests[103220932] = 250 -- Shutting Down Manaforge Duro (The Aldor)
quests[103230932] = 250 -- Shutting Down Manaforge Ara (The Aldor)
quests[103240930] = 150 -- The Great Moongraze Hunt (Exodar)
quests[103250932] = 250 -- Marks of Kil'jaeden (The Aldor)
quests[103260932] = 250 -- More Marks of Kil'jaeden (The Aldor)
quests[103270932] = 25 -- Single Mark of Kil'jaeden (The Aldor)
quests[103280932] = 250 -- Sunfury Briefings (The Aldor)
quests[103290934] = 250 -- Shutting Down Manaforge B'naar (The Scryers)
quests[103300934] = 250 -- Shutting Down Manaforge Coruu (The Scryers)
quests[103350933] = 250 -- Surveying the Ruins (The Consortium)
quests[103360933] = 250 -- The Minions of Culuthas (The Consortium)
quests[103380934] = 250 -- Shutting Down Manaforge Duro (The Scryers)
quests[103390933] = 250 -- The Ethereum (The Consortium)
quests[103410934] = 250 -- Kick Them While They're Down (The Scryers)
quests[103450933] = 250 -- The Flesh Lies... (The Consortium)
quests[103480933] = 250 -- New Opportunities (The Consortium)
quests[103490942] = 10 -- The Earthbinder (Cenarion Expedition)
quests[103510942] = 350 -- Natural Remedies (Cenarion Expedition)
quests[103530933] = 250 -- Arconus the Insatiable (The Consortium)
quests[103550930] = 250 -- Withered Flesh (Exodar)
quests[103650934] = 250 -- Shutting Down Manaforge Ara (The Scryers)
quests[103800932] = 250 -- A Dark Pact (The Aldor)
quests[103810932] = 150 -- Aldor No More (The Aldor)
quests[103820946] = 25 -- Go to the Front (Honor Hold)
quests[103840933] = 75 -- Ethereum Data (The Consortium)
quests[103850933] = 250 -- Potential for Brain Damage = High (The Consortium)
quests[103880947] = 75 -- Return to Thrallmar (Thrallmar)
quests[103890947] = 250 -- The Agony and the Darkness (Thrallmar)
quests[103900947] = 250 -- Forge Camp: Mageddon (Thrallmar)
quests[103910947] = 250 -- Cannons of Rage (Thrallmar)
quests[103920947] = 250 -- Doorway to the Abyss (Thrallmar)
quests[103930947] = 250 -- Vile Plans (Thrallmar)
quests[103940946] = 250 -- Disruption - Forge Camp: Mageddon (Honor Hold)
quests[103950946] = 250 -- The Dark Missive (Honor Hold)
quests[103960946] = 250 -- Enemy of my Enemy... (Honor Hold)
quests[103970946] = 250 -- Invasion Point: Annihilator (Honor Hold)
quests[103990946] = 250 -- The Heart of Darkness (Honor Hold)
quests[104000946] = 500 -- Overlord (Honor Hold)
quests[104050933] = 250 -- S-A-B-O-T-A-G-E (The Consortium)
quests[104060933] = 250 -- Delivering the Message (The Consortium)
quests[104070932] = 350 -- Socrethar's Shadow (The Aldor)
quests[104080933] = 500 -- Nexus-King Salhadaar (The Consortium)
quests[104090932] = 1000 -- Deathblow to the Legion (The Aldor)
quests[104100932] = 150 -- Ishanah's Help (The Aldor)
quests[104110933] = 250 -- Electro-Shock Goodness! (The Consortium)
quests[104120934] = 250 -- Firewing Signets (The Scryers)
quests[104130933] = 250 -- The Horrors of Pollution (The Consortium)
quests[104140934] = 25 -- Single Firewing Signet (The Scryers)
quests[104150934] = 250 -- More Firewing Signets (The Scryers)
quests[104160934] = 350 -- Synthesis of Power (The Scryers)
quests[104170933] = 150 -- Run a Diagnostic! (The Consortium)
quests[104180933] = 250 -- Deal With the Saboteurs (The Consortium)
quests[104190934] = 350 -- Arcane Tomes (The Scryers)
quests[104200932] = 350 -- A Cleansing Light (The Aldor)
quests[104210932] = 350 -- Fel Armaments (The Aldor)
quests[104220933] = 250 -- Captain Tyralius (The Consortium)
quests[104250933] = 250 -- Escape from the Staging Grounds (The Consortium)
quests[104260942] = 250 -- Flora of the Eco-Domes (Cenarion Expedition)
quests[104270942] = 350 -- Creatures of the Eco-Domes (Cenarion Expedition)
quests[104290942] = 250 -- When Nature Goes Too Far (Cenarion Expedition)
quests[104310932] = 25 -- Outside Assistance (The Aldor)
quests[104320934] = 250 -- Damning Evidence (The Scryers)
quests[104360933] = 250 -- All Clear! (The Consortium)
quests[104370933] = 250 -- Recipe for Destruction (The Consortium)
quests[104380933] = 250 -- On Nethery Wings (The Consortium)
quests[104390933] = 500 -- Dimensius the All-Devouring (The Consortium)
quests[104420942] = 10 -- Helping the Cenarion Post (Cenarion Expedition)
quests[104430942] = 10 -- Helping the Cenarion Post (Cenarion Expedition)
quests[104490947] = 25 -- Apothecary Zelana (Thrallmar)
quests[104500947] = 250 -- Bonechewer Blood (Thrallmar)
quests[104600990] = 250 -- Defender's Pledge (The Scale of the Sands)
quests[104610990] = 250 -- Restorer's Pledge (The Scale of the Sands)
quests[104620990] = 250 -- Champion's Pledge (The Scale of the Sands)
quests[104630990] = 250 -- Sage's Pledge (The Scale of the Sands)
quests[104640990] = 350 -- Sage's Vow (The Scale of the Sands)
quests[104650990] = 350 -- Restorer's Vow (The Scale of the Sands)
quests[104660990] = 350 -- Champion's Vow (The Scale of the Sands)
quests[104670990] = 350 -- Defender's Vow (The Scale of the Sands)
quests[104680990] = 350 -- Sage's Oath (The Scale of the Sands)
quests[104690990] = 350 -- Restorer's Oath (The Scale of the Sands)
quests[104700990] = 350 -- Champion's Oath (The Scale of the Sands)
quests[104710990] = 350 -- Defender's Oath (The Scale of the Sands)
quests[104720990] = 350 -- Sage's Covenant (The Scale of the Sands)
quests[104730990] = 350 -- Restorer's Covenant (The Scale of the Sands)
quests[104740990] = 350 -- Champion's Covenant (The Scale of the Sands)
quests[104750990] = 350 -- Defender's Covenant (The Scale of the Sands)
quests[104760978] = 250 -- Fierce Enemies (Kurenai)
quests[104770978] = 500 -- More Warbeads (Kurenai)
quests[104780941] = 500 -- More Warbeads! (The Mag'har)
quests[104790941] = 250 -- Proving Your Strength (The Mag'har)
quests[104820946] = 250 -- Fel Orc Scavengers (Honor Hold)
quests[104830946] = 25 -- Ill Omens (Honor Hold)
quests[104840946] = 250 -- Cursed Talismans (Honor Hold)
quests[104850946] = 250 -- Warlord of the Bleeding Hollow (Honor Hold)
quests[105070934] = 1000 -- Turning Point (The Scryers)
quests[105080934] = 350 -- A Gift for Voren'thal (The Scryers)
quests[105090934] = 150 -- Bound for Glory (The Scryers)
quests[105380947] = 250 -- Boiling Blood (Thrallmar)
quests[105510932] = 3500 -- Allegiance to the Aldor (The Aldor)
quests[105520934] = 3500 -- Allegiance to the Scryers (The Scryers)
quests[105530934] = 10 -- Voren'thal the Seer (The Scryers)
quests[105540932] = 10 -- Ishanah (The Aldor)
quests[105670942] = 250 -- Creating the Pendant (Cenarion Expedition)
quests[105680932] = 250 -- Tablets of Baa'ri (The Aldor)
quests[105710932] = 250 -- Oronu the Elder (The Aldor)
quests[105740932] = 250 -- The Ashtongue Corruptors (The Aldor)
quests[105750932] = 75 -- The Warden's Cage (The Aldor)
quests[105870932] = 250 -- Karabor Training Grounds (The Aldor)
quests[106070942] = 250 -- Whispers of the Raven God (Cenarion Expedition)
quests[106190932] = 250 -- The Ashtongue Tribe (The Aldor)
quests[106221012] = 250 -- Proof of Allegiance (Ashtongue Deathsworn)
quests[106281012] = 25 -- Akama (Ashtongue Deathsworn)
quests[106370932] = 250 -- A Necessary Distraction (The Aldor)
quests[106400932] = 75 -- Altruis (The Aldor)
quests[106500932] = 75 -- Return to the Aldor (The Aldor)
quests[106510932] = 1000 -- Varedis Must Be Stopped (The Aldor)
quests[106520934] = 10 -- Behind Enemy Lines (The Scryers)
quests[106530932] = 250 -- Marks of Sargeras (The Aldor)
quests[106540932] = 250 -- More Marks of Sargeras (The Aldor)
quests[106550932] = 25 -- Single Mark of Sargeras (The Aldor)
quests[106560934] = 250 -- Sunfury Signets (The Scryers)
quests[106580934] = 250 -- More Sunfury Signets (The Scryers)
quests[106590934] = 25 -- Single Sunfury Signet (The Scryers)
quests[106820942] = 250 -- A Time for Negotiation... (Cenarion Expedition)
quests[106830934] = 250 -- Tablets of Baa'ri (The Scryers)
quests[106840934] = 250 -- Oronu the Elder (The Scryers)
quests[106850934] = 250 -- The Ashtongue Corruptors (The Scryers)
quests[106860934] = 75 -- The Warden's Cage (The Scryers)
quests[106870934] = 250 -- Karabor Training Grounds (The Scryers)
quests[106880934] = 250 -- A Necessary Distraction (The Scryers)
quests[106890934] = 75 -- Altruis (The Scryers)
quests[106910934] = 75 -- Return to the Scryers (The Scryers)
quests[106920934] = 1000 -- Varedis Must Be Stopped (The Scryers)
quests[107040935] = 1000 -- How to Break Into the Arcatraz (The Sha'tar)
quests[107051012] = 250 -- Seer Udalo (Ashtongue Deathsworn)
quests[107061012] = 150 -- A Mysterious Portent (Ashtongue Deathsworn)
quests[107071012] = 250 -- The Ata'mal Terrace (Ashtongue Deathsworn)
quests[107080935] = 1000 -- Akama's Promise (The Sha'tar)
quests[107081012] = 1000 -- Akama's Promise (Ashtongue Deathsworn)
quests[107130942] = 250 -- ...and a Time for Action (Cenarion Expedition)
quests[107190942] = 10 -- Did You Get The Note? (Cenarion Expedition)
quests[107220942] = 250 -- Meeting at the Blackwing Coven (Cenarion Expedition)
quests[107250967] = 350 -- Eminence Among the Violet Eye (The Violet Eye)
quests[107260967] = 350 -- Eminence Among the Violet Eye (The Violet Eye)
quests[107270967] = 350 -- Eminence Among the Violet Eye (The Violet Eye)
quests[107280967] = 350 -- Eminence Among the Violet Eye (The Violet Eye)
quests[107290967] = 250 -- Path of the Violet Mage (The Violet Eye)
quests[107300967] = 250 -- Path of the Violet Restorer (The Violet Eye)
quests[107310967] = 250 -- Path of the Violet Assassin (The Violet Eye)
quests[107320967] = 250 -- Path of the Violet Protector (The Violet Eye)
quests[107330967] = 350 -- Down the Violet Path (The Violet Eye)
quests[107340967] = 350 -- Down the Violet Path (The Violet Eye)
quests[107350967] = 350 -- Down the Violet Path (The Violet Eye)
quests[107360967] = 350 -- Down the Violet Path (The Violet Eye)
quests[107380967] = 350 -- Distinguished Service (The Violet Eye)
quests[107390967] = 350 -- Distinguished Service (The Violet Eye)
quests[107400967] = 350 -- Distinguished Service (The Violet Eye)
quests[107410967] = 350 -- Distinguished Service (The Violet Eye)
quests[107480942] = 350 -- Maxnar Must Die! (Cenarion Expedition)
quests[107530942] = 150 -- Culling the Wild (Cenarion Expedition)
quests[107540946] = 250 -- Entry Into the Citadel (Honor Hold)
quests[107620946] = 25 -- Grand Master Dumphry (Honor Hold)
quests[107630946] = 250 -- Dumphry's Request (Honor Hold)
quests[107640946] = 350 -- Hotter than Hell (Honor Hold)
quests[107700942] = 250 -- Little Embers (Cenarion Expedition)
quests[107710942] = 250 -- From the Ashes (Cenarion Expedition)
quests[107810935] = 500 -- Battle of the Crimson Watch (The Sha'tar)
quests[107920947] = 250 -- Zeth'Gor Must Burn! (Thrallmar)
quests[107930935] = 150 -- The Journal of Val'zareq: Portends of War (The Sha'tar)
quests[108070934] = 250 -- The Ashtongue Broken (The Scryers)
quests[108090947] = 250 -- Wanted: Worg Master Kruush (Thrallmar)
quests[108100942] = 25 -- Damaged Mask (Cenarion Expedition)
quests[108120942] = 25 -- Mystery Mask (Cenarion Expedition)
quests[108130947] = 75 -- The Eyes of Grillok (Thrallmar)
quests[108160932] = 250 -- Reclaiming Holy Grounds (The Aldor)
quests[108170934] = 250 -- The Great Retribution (The Scryers)
quests[108190942] = 250 -- Felsworn Gas Mask (Cenarion Expedition)
quests[108200942] = 150 -- Deceive thy Enemy (Cenarion Expedition)
quests[108210942] = 250 -- You're Fired! (Cenarion Expedition)
quests[108220934] = 25 -- Single Sunfury Signet (The Scryers)
quests[108230934] = 250 -- More Sunfury Signets (The Scryers)
quests[108240934] = 250 -- Sunfury Signets (The Scryers)
quests[108260932] = 250 -- Marks of Sargeras (The Aldor)
quests[108270932] = 250 -- More Marks of Sargeras (The Aldor)
quests[108280932] = 25 -- Single Mark of Sargeras (The Aldor)
quests[108340947] = 350 -- Grillok "Darkeye" (Thrallmar)
quests[108350947] = 25 -- Apothecary Antonivich (Thrallmar)
quests[108380947] = 350 -- The Demoniac Scryer (Thrallmar)
quests[108391011] = 250 -- Veil Skith: Darkstone of Terokk (Lower City)
quests[108401011] = 250 -- The Tomb of Lights (Lower City)
quests[108421011] = 250 -- Vengeful Souls (Lower City)
quests[108471011] = 250 -- The Eyes of Skettis (Lower City)
quests[108481011] = 250 -- Veil Rhaze: Unliving Evil (Lower City)
quests[108491011] = 25 -- Seek Out Kirrik (Lower City)
quests[108521011] = 250 -- Missing Friends (Lower City)
quests[108550933] = 250 -- Fel Reavers, No Thanks! (The Consortium)
quests[108560933] = 250 -- The Best Defense (The Consortium)
quests[108570933] = 350 -- Teleport This! (The Consortium)
quests[108611011] = 250 -- Veil Lithic: Preemptive Strike (Lower City)
quests[108621011] = 75 -- Surrender to the Horde (Lower City)
quests[108631011] = 75 -- Secrets of the Arakkoa (Lower City)
quests[108640947] = 250 -- A Burden of Souls (Thrallmar)
quests[108701015] = 42000 -- Ally of the Netherwing (Netherwing)
quests[108730935] = 250 -- Taken in the Night (The Sha'tar)
quests[108741011] = 250 -- Veil Shalas: Signal Fires (Lower City)
quests[108750947] = 25 -- Report to Nazgrel (Thrallmar)
quests[108760947] = 1000 -- The Foot of the Citadel (Thrallmar)
quests[108770935] = 250 -- The Dread Relic (The Sha'tar)
quests[108781011] = 250 -- Before Darkness Falls (Lower City)
quests[108791011] = 500 -- The Skettis Offensive (Lower City)
quests[108801011] = 75 -- Cabal Orders (Lower City)
quests[108811011] = 350 -- The Shadow Tomb (Lower City)
quests[108830935] = 10 -- The Tempest Key (The Sha'tar)
quests[108840935] = 500 -- Trial of the Naaru: Mercy (The Sha'tar)
quests[108850935] = 500 -- Trial of the Naaru: Strength (The Sha'tar)
quests[108860935] = 500 -- Trial of the Naaru: Tenacity (The Sha'tar)
quests[108871011] = 250 -- Escaping the Tomb (Lower City)
quests[108880935] = 500 -- Trial of the Naaru: Magtheridon (The Sha'tar)
quests[108891011] = 150 -- Return to Shattrath (Lower City)
quests[108930942] = 250 -- Longtail is the Lynchpin (Cenarion Expedition)
quests[108940942] = 10 -- Wyrmskull Watcher (Cenarion Expedition)
quests[108950946] = 250 -- Zeth'Gor Must Burn! (Honor Hold)
quests[108960942] = 250 -- The Infested Protectors (Cenarion Expedition)
quests[108981011] = 350 -- Skywing (Lower City)
quests[109040942] = 250 -- Harvesting the Fel Ammunition (Cenarion Expedition)
quests[109090946] = 250 -- Fel Spirits (Honor Hold)
quests[109100942] = 250 -- Death's Door (Cenarion Expedition)
quests[109110942] = 250 -- Fire At Will! (Cenarion Expedition)
quests[109120942] = 500 -- The Hound-Master (Cenarion Expedition)
quests[109131011] = 250 -- An Improper Burial (Lower City)
quests[109141011] = 250 -- A Hero Is Needed (Lower City)
quests[109151011] = 250 -- The Fallen Exarch (Lower City)
quests[109160946] = 75 -- Digging for Prayer Beads (Honor Hold)
quests[109171011] = 250 -- The Outcast's Plight (Lower City)
quests[109181011] = 250 -- More Feathers (Lower City)
quests[109200935] = 250 -- For the Fallen (The Sha'tar)
quests[109210935] = 250 -- Terokkarantula (The Sha'tar)
quests[109221011] = 250 -- Digging Through Bones (Lower City)
quests[109230935] = 350 -- Evil Draws Near (The Sha'tar)
quests[109260935] = 250 -- Return to Sha'tari Base Camp (The Sha'tar)
quests[109291011] = 250 -- Fumping (Lower City)
quests[109301011] = 350 -- The Big Bone Worm (Lower City)
quests[109350946] = 250 -- The Exorcism of Colonel Jules (Honor Hold)
quests[109370946] = 1000 -- Drill the Drillmaster (Honor Hold)
quests[109380909] = 350 -- Darkmoon Blessings Deck (Darkmoon Faire)
quests[109390909] = 350 -- Darkmoon Storms Deck (Darkmoon Faire)
quests[109400909] = 350 -- Darkmoon Furies Deck (Darkmoon Faire)
quests[109410909] = 350 -- Darkmoon Lunacy Deck (Darkmoon Faire)
quests[109421011] = 10 -- Children's Week (Lower City)
quests[109431011] = 10 -- Children's Week (Lower City)
quests[109441012] = 350 -- The Secret Compromised (Ashtongue Deathsworn)
quests[109451011] = 250 -- Hch'uu and the Mushroom People (Lower City)
quests[109461012] = 500 -- Ruse of the Ashtongue (Ashtongue Deathsworn)
quests[109471012] = 500 -- An Artifact From the Past (Ashtongue Deathsworn)
quests[109481012] = 75 -- The Hostage Soul (Ashtongue Deathsworn)
quests[109491012] = 150 -- Entry Into the Black Temple (Ashtongue Deathsworn)
quests[109490935] = 150 -- Entry Into the Black Temple (The Sha'tar)
quests[109501011] = 250 -- Auchindoun and the Ring of Observance (Lower City)
quests[109511011] = 250 -- A Trip to the Dark Portal (Lower City)
quests[109521011] = 250 -- A Trip to the Dark Portal (Lower City)
quests[109531011] = 250 -- Visit the Throne of the Elements (Lower City)
quests[109541011] = 250 -- Jheel is at Aeris Landing! (Lower City)
quests[109561011] = 75 -- The Seat of the Naaru (Lower City)
quests[109571012] = 500 -- Redemption of the Ashtongue (Ashtongue Deathsworn)
quests[109581012] = 350 -- Seek Out the Ashtongue (Ashtongue Deathsworn)
quests[109591012] = 500 -- The Fall of the Betrayer (Ashtongue Deathsworn)
quests[109621011] = 250 -- Time to Visit the Caverns (Lower City)
quests[109631011] = 250 -- Time to Visit the Caverns (Lower City)
quests[109661011] = 500 -- Back to the Orphanage (Lower City)
quests[109671011] = 500 -- Back to the Orphanage (Lower City)
quests[109681011] = 75 -- Call on the Farseer (Lower City)
quests[109690933] = 10 -- Seek Out Ameer (The Consortium)
quests[109700933] = 350 -- A Mission of Mercy (The Consortium)
quests[109710933] = 250 -- Ethereum Secrets (The Consortium)
quests[109720933] = 250 -- Ethereum Prisoner I.D. Catalogue (The Consortium)
quests[109730933] = 500 -- A Thousand Worlds (The Consortium)
quests[109740933] = 500 -- Stasis Chambers of Bash'ir (The Consortium)
quests[109750933] = 500 -- Purging the Chambers of Bash'ir (The Consortium)
quests[109760933] = 500 -- The Mark of the Nexus-King (The Consortium)
quests[109770933] = 500 -- Stasis Chambers of the Mana-Tombs (The Consortium)
quests[109810933] = 500 -- Nexus-Prince Shaffar's Personal Chamber (The Consortium)
quests[109850935] = 250 -- A Distraction for Akama (The Sha'tar)
quests[110020946] = 1000 -- The Fall of Magtheridon (Honor Hold)
quests[110030947] = 1000 -- The Fall of Magtheridon (Thrallmar)
quests[110041031] = 250 -- World of Shadows (Sha'tari Skyguard)
quests[110051031] = 250 -- Secrets of the Talonpriests (Sha'tari Skyguard)
quests[110061031] = 150 -- More Shadow Dust (Sha'tari Skyguard)
quests[110070935] = 1000 -- Kael'thas and the Verdant Sphere (The Sha'tar)
quests[110081031] = 350 -- Fires Over Skettis (Sha'tari Skyguard)
quests[110091038] = 25 -- Ogre Heaven (Ogri'la)
quests[110101031] = 500 -- Bombing Run (Sha'tari Skyguard)
quests[110101038] = 500 -- Bombing Run (Ogri'la)
quests[110131015] = 250 -- In Service of the Illidari (Netherwing)
quests[110151015] = 250 -- Netherwing Crystals (Netherwing)
quests[110161015] = 250 -- Nethermine Flayer Hide (Netherwing)
quests[110171015] = 250 -- Netherdust Pollen (Netherwing)
quests[110181015] = 250 -- Nethercite Ore (Netherwing)
quests[110201015] = 250 -- A Slow Death (Netherwing)
quests[110211031] = 75 -- Ishaal's Almanac (Sha'tari Skyguard)
quests[110231031] = 500 -- Bomb Them Again! (Sha'tari Skyguard)
quests[110231038] = 500 -- Bomb Them Again! (Ogri'la)
quests[110241031] = 75 -- An Ally in Lower City (Sha'tari Skyguard)
quests[110251038] = 250 -- The Crystals (Ogri'la)
quests[110261038] = 350 -- Banish the Demons (Ogri'la)
quests[110261031] = 350 -- Banish the Demons (Sha'tari Skyguard)
quests[110281031] = 75 -- Countdown to Doom (Sha'tari Skyguard)
quests[110291031] = 150 -- A Shabby Disguise (Sha'tari Skyguard)
quests[110301038] = 250 -- Our Boy Wants To Be A Skyguard Ranger (Ogri'la)
quests[110351015] = 250 -- The Not-So-Friendly Skies... (Netherwing)
quests[110380932] = 10 -- Assist Exarch Orelis (The Aldor)
quests[110390934] = 10 -- Report to Spymaster Thalodien (The Scryers)
quests[110411015] = 250 -- A Job Unfinished... (Netherwing)
quests[110491015] = 350 -- The Great Netherwing Egg Hunt (Netherwing)
quests[110501015] = 250 -- Accepting All Eggs (Netherwing)
quests[110511038] = 350 -- Banish More Demons (Ogri'la)
quests[110531015] = 250 -- Rise, Overseer! (Netherwing)
quests[110541015] = 250 -- Overseeing and You: Making the Right Choices (Netherwing)
quests[110551015] = 350 -- The Booterang: A Cure For The Common Worthless Peon (Netherwing)
quests[110561031] = 75 -- Hazzik's Bargain (Sha'tari Skyguard)
quests[110581038] = 250 -- An Apexis Relic (Ogri'la)
quests[110581031] = 250 -- An Apexis Relic (Sha'tari Skyguard)
quests[110591038] = 500 -- Guardian of the Monument (Ogri'la)
quests[110591031] = 500 -- Guardian of the Monument (Sha'tari Skyguard)
quests[110611038] = 250 -- A Father's Duty (Ogri'la)
quests[110621031] = 10 -- The Skyguard Outpost (Sha'tari Skyguard)
quests[110641015] = 350 -- Dragonmaw Race: The Ballad of Oldie McOld (Netherwing)
quests[110651031] = 350 -- Wrangle Some Aether Rays! (Sha'tari Skyguard)
quests[110651038] = 350 -- Wrangle Some Aether Rays! (Ogri'la)
quests[110661031] = 350 -- Wrangle More Aether Rays! (Sha'tari Skyguard)
quests[110661038] = 350 -- Wrangle More Aether Rays! (Ogri'la)
quests[110671015] = 350 -- Dragonmaw Race: Trope the Filth-Belcher (Netherwing)
quests[110681015] = 500 -- Dragonmaw Race: Corlok the Vet (Netherwing)
quests[110691015] = 500 -- Dragonmaw Race: Wing Commander Ichman (Netherwing)
quests[110701015] = 500 -- Dragonmaw Race: Wing Commander Mulverick (Netherwing)
quests[110711015] = 1000 -- Dragonmaw Race: Captain Skyshatter (Netherwing)
quests[110721031] = 350 -- Adversarial Blood (Sha'tari Skyguard)
quests[110731031] = 1000 -- Terokk's Downfall (Sha'tari Skyguard)
quests[110741031] = 350 -- Tokens of the Descendants (Sha'tari Skyguard)
quests[110751015] = 75 -- The Netherwing Mines (Netherwing)
quests[110761015] = 350 -- Picking Up The Pieces... (Netherwing)
quests[110771015] = 350 -- Dragons are the Least of Our Problems (Netherwing)
quests[110781031] = 500 -- To Rule The Skies (Sha'tari Skyguard)
quests[110791038] = 500 -- A Fel Whip For Gahk (Ogri'la)
quests[110801038] = 350 -- The Relic's Emanation (Ogri'la)
quests[110811015] = 350 -- The Great Murkblood Revolt (Netherwing)
quests[110821015] = 500 -- Seeker of Truth (Netherwing)
quests[110831015] = 350 -- Crazed and Confused (Netherwing)
quests[110841015] = 500 -- Stand Tall, Captain! (Netherwing)
quests[110851031] = 150 -- Escape from Skettis (Sha'tari Skyguard)
quests[110861015] = 500 -- Disrupting the Twilight Portal (Netherwing)
quests[110891015] = 500 -- The Soul Cannon of Reth'hedron (Netherwing)
quests[110901015] = 500 -- Subdue the Subduer (Netherwing)
quests[110911038] = 10 -- A Special Thank You (Ogri'la)
quests[110921015] = 500 -- Hail, Commander! (Netherwing)
quests[110931031] = 250 -- Hungry Nether Rays (Sha'tari Skyguard)
quests[110941015] = 250 -- Kill Them All! (Netherwing)
quests[110961031] = 250 -- Threat from Above (Sha'tari Skyguard)
quests[110971015] = 500 -- The Deadliest Trap Ever Laid (Netherwing)
quests[110981031] = 10 -- To Skettis! (Sha'tari Skyguard)
quests[110991015] = 250 -- Kill Them All! (Netherwing)
quests[111011015] = 500 -- The Deadliest Trap Ever Laid (Netherwing)
quests[111021031] = 500 -- Bombing Run (Sha'tari Skyguard)
quests[111021038] = 500 -- Bombing Run (Ogri'la)
quests[111081015] = 1000 -- Lord Illidan Stormrage (Netherwing)
quests[111191031] = 10 -- Assault on Bash'ir Landing! (Sha'tari Skyguard)
quests[111290081] = 250 -- Kyle's Gone Missing! (Thunder Bluff)
quests[111531050] = 250 -- Break the Blockade (Valiance Expedition)
quests[111541050] = 250 -- Scare the Guano Out of Them! (Valiance Expedition)
quests[111551050] = 250 -- Shoveltusk Soup Again? (Valiance Expedition)
quests[111571050] = 250 -- The Clutches of Evil (Valiance Expedition)
quests[111671052] = 250 -- The New Plague (Horde Expedition)
quests[111701067] = 250 -- Test at Sea (The Hand of Vengeance)
quests[111751068] = 10 -- My Daughter (Explorers' League)
quests[111761068] = 10 -- See to the Operations (Explorers' League)
quests[111871050] = 10 -- Mage-Lieutenant Malister (Valiance Expedition)
quests[111881050] = 250 -- Two Wrongs... (Valiance Expedition)
quests[111901050] = 250 -- One Size Does Not Fit All (Valiance Expedition)
quests[111991050] = 10 -- Report to Scout Knowles (Valiance Expedition)
quests[112021050] = 250 -- Mission: Eternal Flame (Valiance Expedition)
quests[112080369] = 10 -- Delivery for Drazzit (Gadgetzan)
quests[112110369] = 10 -- Help for Mudsprocket (Gadgetzan)
quests[112181068] = 250 -- Danger! Explosives! (Explorers' League)
quests[112211067] = 150 -- Reports from the Field (The Hand of Vengeance)
quests[112241068] = 250 -- Send Them Packing (Explorers' League)
quests[112271067] = 250 -- Let Them Eat Crow (The Hand of Vengeance)
quests[112291052] = 25 -- The Windrunner Fleet (Horde Expedition)
quests[112301052] = 150 -- Ambushed! (Horde Expedition)
quests[112311050] = 250 -- Of Keys and Cages (Valiance Expedition)
quests[112321067] = 250 -- Guide Our Sights (The Hand of Vengeance)
quests[112331067] = 350 -- Landing the Killing Blow (The Hand of Vengeance)
quests[112341067] = 75 -- Report to Anselm (The Hand of Vengeance)
quests[112351050] = 250 -- Dealing With Gjalerbron (Valiance Expedition)
quests[112361050] = 350 -- Necro Overlord Mezhen (Valiance Expedition)
quests[112371050] = 250 -- Gjalerbron Attack Plans (Valiance Expedition)
quests[112381050] = 350 -- The Frost Wyrm and its Master (Valiance Expedition)
quests[112391050] = 250 -- In Service to the Light (Valiance Expedition)
quests[112401068] = 250 -- Leader of the Deranged (Explorers' League)
quests[112411052] = 350 -- Trail of Fire (Horde Expedition)
quests[112420072] = 250 -- Free at Last! (Stormwind)
quests[112431050] = 250 -- If Valgarde Falls... (Valiance Expedition)
quests[112441050] = 250 -- Rescuing the Rescuers (Valiance Expedition)
quests[112451050] = 250 -- Towers of Certain Doom (Valiance Expedition)
quests[112461050] = 250 -- Gruesome, But Necessary (Valiance Expedition)
quests[112471050] = 250 -- Burn Skorn, Burn! (Valiance Expedition)
quests[112481050] = 10 -- Operation: Skornful Wrath (Valiance Expedition)
quests[112491050] = 350 -- Stop the Ascension! (Valiance Expedition)
quests[112501050] = 350 -- All Hail the Conqueror of Skorn! (Valiance Expedition)
quests[112531052] = 150 -- Sniff Out the Enemy (Horde Expedition)
quests[112541052] = 75 -- The Dragonskin Map (Horde Expedition)
quests[112551050] = 250 -- Prisoners of Wyrmskull (Valiance Expedition)
quests[112561064] = 10 -- Skorn Must Fall! (The Taunka)
quests[112571064] = 250 -- Gruesome, But Necessary (The Taunka)
quests[112581064] = 250 -- Burn Skorn, Burn! (The Taunka)
quests[112591064] = 250 -- Towers of Certain Doom (The Taunka)
quests[112601064] = 350 -- Stop the Ascension! (The Taunka)
quests[112611064] = 350 -- The Conqueror of Skorn! (The Taunka)
quests[112631064] = 250 -- Dealing With Gjalerbron (The Taunka)
quests[112641064] = 350 -- Necro Overlord Mezhen (The Taunka)
quests[112651064] = 250 -- Of Keys and Cages (The Taunka)
quests[112661064] = 250 -- Gjalerbron Attack Plans (The Taunka)
quests[112671064] = 350 -- The Frost Wyrm and its Master (The Taunka)
quests[112681064] = 250 -- The Walking Dead (The Taunka)
quests[112691050] = 250 -- Down to the Wire (Valiance Expedition)
quests[112701067] = 250 -- War is Hell (The Hand of Vengeance)
quests[112711064] = 250 -- Hasty Preparations (The Taunka)
quests[112721052] = 500 -- A Score to Settle (Horde Expedition)
quests[112731050] = 25 -- The Human League (Valiance Expedition)
quests[112741050] = 25 -- Zedd's Probably Dead (Valiance Expedition)
quests[112761050] = 150 -- And Then There Were Two... (Valiance Expedition)
quests[112771050] = 250 -- The Depths of Depravity (Valiance Expedition)
quests[112781050] = 250 -- Return to Valgarde (Valiance Expedition)
quests[112791067] = 250 -- Green Eggs and Whelps (The Hand of Vengeance)
quests[112801067] = 250 -- Draconis Gastritis (The Hand of Vengeance)
quests[112821052] = 250 -- A Lesson in Fear (Horde Expedition)
quests[112831067] = 250 -- Baleheim Bodycount (The Hand of Vengeance)
quests[112841050] = 250 -- The Yeti Next Door (Valiance Expedition)
quests[112851052] = 150 -- Baleheim Must Burn! (Horde Expedition)
quests[112861064] = 250 -- The Artifacts of Steel Gate (The Taunka)
quests[112871064] = 10 -- Find Sage Mistwalker (The Taunka)
quests[112881050] = 250 -- The Shining Light (Valiance Expedition)
quests[112891050] = 75 -- Guided by Honor (Valiance Expedition)
quests[112891106] = 75 -- Guided by Honor (Argent Crusade)
quests[112901050] = 250 -- Dragonflayer Battle Plans (Valiance Expedition)
quests[112911050] = 25 -- To Westguard Keep! (Valiance Expedition)
quests[112951052] = 10 -- The Offensive Begins (Horde Expedition)
quests[112961064] = 250 -- Rivenwood Captives (The Taunka)
quests[112981067] = 250 -- What's in That Brew? (The Hand of Vengeance)
quests[112991050] = 10 -- The Ring of Judgment (Valiance Expedition)
quests[113001050] = 250 -- Stunning Defeat at the Ring (Valiance Expedition)
quests[113011067] = 250 -- Brains! Brains! Brains! (The Hand of Vengeance)
quests[113031067] = 350 -- The Ambush (The Hand of Vengeance)
quests[113041067] = 10 -- New Agamand (The Hand of Vengeance)
quests[113051067] = 250 -- A Tailor-Made Formula (The Hand of Vengeance)
quests[113061067] = 150 -- Apply Heat and Stir (The Hand of Vengeance)
quests[113071067] = 250 -- Field Test (The Hand of Vengeance)
quests[113091052] = 250 -- Parts for the Job (Horde Expedition)
quests[113101052] = 250 -- Warning: Some Assembly Required (Horde Expedition)
quests[113111064] = 250 -- Suppressing the Elements (The Taunka)
quests[113171064] = 250 -- The Cleansing (The Taunka)
quests[113221050] = 250 -- The Cleansing (Valiance Expedition)
quests[113241064] = 350 -- Alpha Worg (The Taunka)
quests[113261050] = 350 -- Alpha Worg (Valiance Expedition)
quests[113271050] = 250 -- Mission: Package Retrieval (Valiance Expedition)
quests[113281050] = 10 -- Mission: Forsaken Intel (Valiance Expedition)
quests[113291050] = 250 -- I'll Try Anything! (Valiance Expedition)
quests[113301050] = 75 -- Absholutely... Thish Will Work! (Valiance Expedition)
quests[113311050] = 10 -- You Tell Him ...Hic! (Valiance Expedition)
quests[113321050] = 350 -- Mission: Plague This! (Valiance Expedition)
quests[113331050] = 250 -- Into the World of Spirits (Valiance Expedition)
quests[113431050] = 250 -- The Echo of Ymiron (Valiance Expedition)
quests[113441050] = 250 -- Anguish of Nifflevar (Valiance Expedition)
quests[113461068] = 250 -- The Book of Runes (Explorers' League)
quests[113481068] = 250 -- The Rune of Command (Explorers' League)
quests[113491068] = 250 -- Mastering the Runes (Explorers' League)
quests[113501064] = 250 -- The Book of Runes (The Taunka)
quests[113511064] = 250 -- Mastering the Runes (The Taunka)
quests[113521064] = 250 -- The Rune of Command (The Taunka)
quests[113551050] = 250 -- March of the Giants (Valiance Expedition)
quests[113560072] = 10 -- Costumed Orphan Matron (Stormwind)
quests[113581050] = 250 -- The Lodestone (Valiance Expedition)
quests[113591050] = 350 -- Demolishing Megalith (Valiance Expedition)
quests[113600072] = 75 -- Fire Brigade Practice (Stormwind)
quests[113610076] = 75 -- Fire Training (Orgrimmar)
quests[113640947] = 250 -- Wanted: Shattered Hand Centurions (Thrallmar)
quests[113640946] = 250 -- Wanted: Shattered Hand Centurions (Honor Hold)
quests[113640933] = 250 -- Wanted: Shattered Hand Centurions (The Consortium)
quests[113651064] = 250 -- March of the Giants (The Taunka)
quests[113661064] = 250 -- The Lodestone (The Taunka)
quests[113671064] = 350 -- Demolishing Megalith (The Taunka)
quests[113710933] = 250 -- Wanted: Coilfang Myrmidons (The Consortium)
quests[113710942] = 250 -- Wanted: Coilfang Myrmidons (Cenarion Expedition)
quests[113760933] = 250 -- Wanted: Malicious Instructors (The Consortium)
quests[113761011] = 250 -- Wanted: Malicious Instructors (Lower City)
quests[113830933] = 250 -- Wanted: Rift Lords (The Consortium)
quests[113830989] = 250 -- Wanted: Rift Lords (Keepers of Time)
quests[113850933] = 250 -- Wanted: Sunseeker Channelers (The Consortium)
quests[113850935] = 250 -- Wanted: Sunseeker Channelers (The Sha'tar)
quests[113890933] = 250 -- Wanted: Arcatraz Sentinels (The Consortium)
quests[113890935] = 250 -- Wanted: Arcatraz Sentinels (The Sha'tar)
quests[113901068] = 250 -- I've Got a Flying Machine! (Explorers' League)
quests[113911068] = 250 -- Steel Gate Patrol (Explorers' League)
quests[113931068] = 10 -- Where is Explorer Jaren? (Explorers' League)
quests[113941068] = 250 -- And You Thought Murlocs Smelled Bad! (Explorers' League)
quests[113951068] = 10 -- It's a Scourge Device (Explorers' League)
quests[113961068] = 350 -- Bring Down Those Shields (Explorers' League)
quests[113971067] = 250 -- And You Thought Murlocs Smelled Bad! (The Hand of Vengeance)
quests[113981067] = 10 -- It's a Scourge Device (The Hand of Vengeance)
quests[113991067] = 350 -- Bring Down Those Shields (The Hand of Vengeance)
quests[114030076] = 250 -- Free at Last! (Orgrimmar)
quests[114061050] = 10 -- Everything Must Be Ready (Valiance Expedition)
quests[114101050] = 250 -- The One That Got Away (Valiance Expedition)
quests[114181050] = 150 -- We Call Him Steelfeather (Valiance Expedition)
quests[114201050] = 250 -- The Path to Payback (Valiance Expedition)
quests[114211050] = 350 -- It Goes to 11... (Valiance Expedition)
quests[114231052] = 250 -- The Enemy's Legacy (Horde Expedition)
quests[114241052] = 250 -- Shield Hill (Horde Expedition)
quests[114261050] = 250 -- Locating the Mechanism (Valiance Expedition)
quests[114271050] = 10 -- Meet Lieutenant Icehammer... (Valiance Expedition)
quests[114291050] = 250 -- Drop It then Rock It! (Valiance Expedition)
quests[114301050] = 250 -- Harpoon Master Yavus (Valiance Expedition)
quests[114321050] = 250 -- Sleeping Giants (Valiance Expedition)
quests[114331064] = 250 -- Sleeping Giants (The Taunka)
quests[114361050] = 10 -- Let's Go Surfing Now (Valiance Expedition)
quests[114390047] = 75 -- Fire Brigade Practice (Ironforge)
quests[114400930] = 75 -- Fire Brigade Practice (Exodar)
quests[114431050] = 250 -- Daggercap Divin' (Valiance Expedition)
quests[114481050] = 10 -- The Explorers' League Outpost (Valiance Expedition)
quests[114490068] = 75 -- Fire Training (Undercity)
quests[114500911] = 75 -- Fire Training (Silvermoon City)
quests[114531064] = 350 -- The Slumbering King (The Taunka)
quests[114561073] = 250 -- Feeding the Survivors (The Kalu'ak)
quests[114571073] = 250 -- Arming Kamagua (The Kalu'ak)
quests[114581073] = 350 -- Avenge Iskaal (The Kalu'ak)
quests[114601050] = 250 -- Trust is Earned (Valiance Expedition)
quests[114651050] = 250 -- The Ransacked Caravan (Valiance Expedition)
quests[114651068] = 250 -- The Ransacked Caravan (Explorers' League)
quests[114681050] = 250 -- Falcon Versus Hawk (Valiance Expedition)
quests[114681068] = 250 -- Falcon Versus Hawk (Explorers' League)
quests[114701050] = 350 -- There Exists No Honor Among Birds (Valiance Expedition)
quests[114701068] = 350 -- There Exists No Honor Among Birds (Explorers' League)
quests[114721073] = 250 -- The Way to His Heart... (The Kalu'ak)
quests[114741050] = 10 -- Problems on the High Bluff (Valiance Expedition)
quests[114741068] = 10 -- Problems on the High Bluff (Explorers' League)
quests[114751050] = 250 -- Tools to Get the Job Done (Valiance Expedition)
quests[114751068] = 250 -- Tools to Get the Job Done (Explorers' League)
quests[114771050] = 250 -- Out of My Element? (Valiance Expedition)
quests[114771068] = 250 -- Out of My Element? (Explorers' League)
quests[114781050] = 25 -- Outpost Over Yonder... (Valiance Expedition)
quests[114781068] = 25 -- Outpost Over Yonder... (Explorers' League)
quests[114811077] = 10 -- Crisis at the Sunwell (Shattered Sun Offensive)
quests[114821077] = 10 -- Duty Calls (Shattered Sun Offensive)
quests[114831068] = 250 -- We Can Rebuild It (Explorers' League)
quests[114831050] = 250 -- We Can Rebuild It (Valiance Expedition)
quests[114841068] = 250 -- We Have the Technology (Explorers' League)
quests[114841050] = 250 -- We Have the Technology (Valiance Expedition)
quests[114851068] = 10 -- Iron Rune Constructs and You: Rocket Jumping (Explorers' League)
quests[114851050] = 10 -- Iron Rune Constructs and You: Rocket Jumping (Valiance Expedition)
quests[114891068] = 10 -- Iron Rune Constructs and You: Collecting Data (Explorers' League)
quests[114891050] = 10 -- Iron Rune Constructs and You: Collecting Data (Valiance Expedition)
quests[114901077] = 250 -- The Scryer's Scryer (Shattered Sun Offensive)
quests[114911068] = 10 -- Iron Rune Constructs and You: The Bluff (Explorers' League)
quests[114911050] = 10 -- Iron Rune Constructs and You: The Bluff (Valiance Expedition)
quests[114941068] = 350 -- Lightning Infused Relics (Explorers' League)
quests[114941050] = 350 -- Lightning Infused Relics (Valiance Expedition)
quests[114951068] = 350 -- The Delicate Sound of Thunder (Explorers' League)
quests[114951050] = 350 -- The Delicate Sound of Thunder (Valiance Expedition)
quests[114961077] = 150 -- The Sanctum Wards (Shattered Sun Offensive)
quests[115000933] = 250 -- Wanted: Sisters of Torment (The Consortium)
quests[115001077] = 250 -- Wanted: Sisters of Torment (Shattered Sun Offensive)
quests[115011068] = 10 -- News From the East (Explorers' League)
quests[115011050] = 10 -- News From the East (Valiance Expedition)
quests[115020978] = 250 -- In Defense of Halaa (Kurenai)
quests[115030941] = 250 -- Enemies, Old and New (The Mag'har)
quests[115041073] = 250 -- The Dead Rise! (The Kalu'ak)
quests[115071073] = 75 -- Elder Atuik and Kamagua (The Kalu'ak)
quests[115081073] = 10 -- Grezzix Spindlesnap (The Kalu'ak)
quests[115091073] = 25 -- Street "Cred" (The Kalu'ak)
quests[115101073] = 250 -- "Scoodles" (The Kalu'ak)
quests[115111073] = 250 -- The Staff of Storm's Fury (The Kalu'ak)
quests[115121073] = 250 -- The Frozen Heart of Isuldof (The Kalu'ak)
quests[115131077] = 250 -- Intercepting the Mana Cells (Shattered Sun Offensive)
quests[115141077] = 250 -- Maintaining the Sunwell Portal (Shattered Sun Offensive)
quests[115151077] = 250 -- Blood for Blood (Shattered Sun Offensive)
quests[115161077] = 250 -- Blast the Gateway (Shattered Sun Offensive)
quests[115171077] = 10 -- Report to Nasuun (Shattered Sun Offensive)
quests[115191073] = 150 -- The Lost Shield of the Aesirites (The Kalu'ak)
quests[115201077] = 350 -- Discovering Your Roots (Shattered Sun Offensive)
quests[115211077] = 350 -- Rediscovering Your Roots (Shattered Sun Offensive)
quests[115231077] = 150 -- Arm the Wards! (Shattered Sun Offensive)
quests[115241077] = 150 -- Erratic Behavior (Shattered Sun Offensive)
quests[115251077] = 150 -- Further Conversions (Shattered Sun Offensive)
quests[115261077] = 75 -- The Missing Magistrix (Shattered Sun Offensive)
quests[115271073] = 150 -- Mutiny on the Mercy (The Kalu'ak)
quests[115291073] = 150 -- Sorlof's Booty (The Kalu'ak)
quests[115301073] = 250 -- The Shield of the Aesirites (The Kalu'ak)
quests[115321077] = 150 -- Distraction at the Dead Scar (Shattered Sun Offensive)
quests[115331077] = 150 -- The Air Strikes Must Continue (Shattered Sun Offensive)
quests[115341077] = 10 -- Report to Nasuun (Shattered Sun Offensive)
quests[115351077] = 250 -- Making Ready (Shattered Sun Offensive)
quests[115361077] = 250 -- Don't Stop Now.... (Shattered Sun Offensive)
quests[115371077] = 250 -- The Battle Must Go On (Shattered Sun Offensive)
quests[115381077] = 250 -- The Battle for the Sun's Reach Armory (Shattered Sun Offensive)
quests[115391077] = 250 -- Taking the Harbor (Shattered Sun Offensive)
quests[115401077] = 250 -- Crush the Dawnblade (Shattered Sun Offensive)
quests[115411077] = 250 -- Disrupt the Greengill Coast (Shattered Sun Offensive)
quests[115421077] = 250 -- Intercept the Reinforcements (Shattered Sun Offensive)
quests[115431077] = 250 -- Keeping the Enemy at Bay (Shattered Sun Offensive)
quests[115441077] = 350 -- Ata'mal Armaments (Shattered Sun Offensive)
quests[115451077] = 150 -- A Charitable Donation (Shattered Sun Offensive)
quests[115461077] = 250 -- Open for Business (Shattered Sun Offensive)
quests[115471077] = 250 -- Know Your Ley Lines (Shattered Sun Offensive)
quests[115481077] = 150 -- Your Continued Support (Shattered Sun Offensive)
quests[115491077] = 500 -- A Magnanimous Benefactor (Shattered Sun Offensive)
quests[115501077] = 25 -- Enter, the Deceiver... (Shattered Sun Offensive)
quests[115600942] = 250 -- Oh Noes, the Tadpoles! (Cenarion Expedition)
quests[115650942] = 10 -- The Spare Suit (Cenarion Expedition)
quests[115660942] = 350 -- Surrender... Not! (Cenarion Expedition)
quests[115671073] = 250 -- The Ancient Armor of the Kvaldir (The Kalu'ak)
quests[115681073] = 250 -- A Return to Resting (The Kalu'ak)
quests[115700942] = 250 -- Escape from the Winterfin Caverns (Cenarion Expedition)
quests[115710942] = 250 -- Learning to Communicate (Cenarion Expedition)
quests[115721073] = 500 -- Return to Atuik (The Kalu'ak)
quests[115731073] = 25 -- Orfus of Kamagua (The Kalu'ak)
quests[115761090] = 150 -- Monitoring the Rift: Cleftcliff Anomaly (Kirin Tor)
quests[115821090] = 150 -- Monitoring the Rift: Sundered Chasm (Kirin Tor)
quests[115851085] = 10 -- Hellscream's Vigil (Warsong Offensive)
quests[115861085] = 10 -- Hellscream's Vigil (Warsong Offensive)
quests[115871090] = 250 -- Prison Break (Kirin Tor)
quests[115901090] = 250 -- Abduction (Kirin Tor)
quests[115911064] = 10 -- Report to Steeljaw's Caravan (The Taunka)
quests[115921085] = 350 -- We Strike! (Warsong Offensive)
quests[115921064] = 350 -- We Strike! (The Taunka)
quests[115931085] = 250 -- The Honored Dead (Warsong Offensive)
quests[115941085] = 250 -- Put Them to Rest (Warsong Offensive)
quests[115951085] = 10 -- The Defense of Warsong Hold (Warsong Offensive)
quests[115961085] = 10 -- The Defense of Warsong Hold (Warsong Offensive)
quests[115971085] = 10 -- The Defense of Warsong Hold (Warsong Offensive)
quests[115981085] = 250 -- Taking Back Mightstone Quarry (Warsong Offensive)
quests[116001050] = 150 -- The Late William Allerton (Valiance Expedition)
quests[116011050] = 75 -- Lost and Found (Valiance Expedition)
quests[116021085] = 250 -- Cutting Off the Source (Warsong Offensive)
quests[116031050] = 150 -- In Wine, Truth (Valiance Expedition)
quests[116041050] = 25 -- A Deserter (Valiance Expedition)
quests[116051073] = 250 -- The Honored Ancestors (The Kalu'ak)
quests[116061085] = 250 -- Patience is a Virtue that We Don't Need (Warsong Offensive)
quests[116071073] = 250 -- The Lost Spirits (The Kalu'ak)
quests[116081085] = 250 -- Bury Those Cockroaches! (Warsong Offensive)
quests[116091073] = 250 -- Picking Up the Pieces (The Kalu'ak)
quests[116101073] = 350 -- Leading the Ancestors Home (The Kalu'ak)
quests[116111085] = 250 -- Taken by the Scourge (Warsong Offensive)
quests[116121073] = 250 -- Reclaiming the Quarry (The Kalu'ak)
quests[116131073] = 350 -- Karuk's Oath (The Kalu'ak)
quests[116141085] = 25 -- Untold Truths (Warsong Offensive)
quests[116151085] = 25 -- Nerub'ar Secrets (Warsong Offensive)
quests[116161085] = 25 -- Message to Hellscream (Warsong Offensive)
quests[116171073] = 350 -- Hampering Their Escape (The Kalu'ak)
quests[116181085] = 25 -- Reinforcements Incoming... (Warsong Offensive)
quests[116191073] = 250 -- Gamel the Cruel (The Kalu'ak)
quests[116201073] = 150 -- A Father's Words (The Kalu'ak)
quests[116231073] = 250 -- A Visit to the Curator (The Kalu'ak)
quests[116261073] = 500 -- The Emissary (The Kalu'ak)
quests[116281064] = 250 -- Shrouds of the Scourge (The Taunka)
quests[116291085] = 10 -- Return to the Spirit Talker (Warsong Offensive)
quests[116301064] = 250 -- The Bad Earth (The Taunka)
quests[116311085] = 10 -- Vision of Air (Warsong Offensive)
quests[116321085] = 250 -- What the Cold Wind Brings... (Warsong Offensive)
quests[116331064] = 250 -- Blending In (The Taunka)
quests[116341085] = 10 -- Wind Master To'bor (Warsong Offensive)
quests[116351085] = 250 -- Farseer Grimwalker's Spirit (Warsong Offensive)
quests[116361085] = 10 -- Magic Carpet Ride (Warsong Offensive)
quests[116371085] = 250 -- Kaganishu (Warsong Offensive)
quests[116381085] = 350 -- Return My Remains (Warsong Offensive)
quests[116391052] = 250 -- Revenge Upon Magmoth (Horde Expedition)
quests[116401064] = 250 -- Words of Power (The Taunka)
quests[116411064] = 250 -- A Courageous Strike (The Taunka)
quests[116421085] = 10 -- Tank Ain't Gonna Fix Itself (Warsong Offensive)
quests[116431085] = 250 -- Mobu's Pneumatic Tank Transjigamarig (Warsong Offensive)
quests[116441085] = 250 -- Super Strong Metal Plates! (Warsong Offensive)
quests[116451050] = 250 -- Dirty, Stinkin' Snobolds! (Valiance Expedition)
quests[116461090] = 25 -- The Borean Inquisition (Kirin Tor)
quests[116471064] = 250 -- Neutralizing the Cauldrons (The Taunka)
quests[116481090] = 150 -- The Art of Persuasion (Kirin Tor)
quests[116501050] = 250 -- Just a Few More Things... (Valiance Expedition)
quests[116511085] = 25 -- Tanks a lot... (Warsong Offensive)
quests[116521085] = 500 -- The Plains of Nasam (Warsong Offensive)
quests[116531050] = 250 -- Hah... You're Not So Big Now! (Valiance Expedition)
quests[116541064] = 150 -- The Spire of Blood (The Taunka)
quests[116551073] = 250 -- Into the Mist (The Kalu'ak)
quests[116561073] = 250 -- Burn in Effigy (The Kalu'ak)
quests[116581050] = 250 -- Plan B (Valiance Expedition)
quests[116591064] = 250 -- Shatter the Orbs! (The Taunka)
quests[116601073] = 250 -- Horn of the Ancient Mariner (The Kalu'ak)
quests[116611073] = 250 -- Orabus the Helmsman (The Kalu'ak)
quests[116621073] = 250 -- Seek Out Karuk! (The Kalu'ak)
quests[116631090] = 10 -- Sharing Intelligence (Kirin Tor)
quests[116641085] = 350 -- Escaping the Mist (Warsong Offensive)
quests[116701050] = 350 -- It Was The Orcs, Honest! (Valiance Expedition)
quests[116711090] = 250 -- A Race Against Time (Kirin Tor)
quests[116721050] = 10 -- Enlistment Day (Valiance Expedition)
quests[116731050] = 350 -- Get Me Outa Here! (Valiance Expedition)
quests[116741064] = 10 -- Sage Highmesa is Missing (The Taunka)
quests[116751064] = 250 -- A Proper Death (The Taunka)
quests[116761085] = 250 -- Merciful Freedom (Warsong Offensive)
quests[116771064] = 250 -- Stop the Plague (The Taunka)
quests[116781064] = 250 -- Find Bristlehorn (The Taunka)
quests[116801090] = 10 -- Taking Wing (Kirin Tor)
quests[116811090] = 250 -- Rescuing Evanor (Kirin Tor)
quests[116821090] = 75 -- Dragonspeak (Kirin Tor)
quests[116831064] = 250 -- Fallen Necropolis (The Taunka)
quests[116841064] = 250 -- Scouting the Sinkholes (The Taunka)
quests[116851064] = 250 -- The Heart of the Elements (The Taunka)
quests[116861085] = 250 -- The Warsong Farms (Warsong Offensive)
quests[116871064] = 250 -- The Doctor and the Lich-Lord (The Taunka)
quests[116881085] = 250 -- Damned Filthy Swine (Warsong Offensive)
quests[116891064] = 350 -- Return with the Bad News (The Taunka)
quests[116901085] = 250 -- Bring 'Em Back Alive (Warsong Offensive)
quests[116921050] = 10 -- Check in With Bixie (Valiance Expedition)
quests[116931050] = 250 -- Oh Great... Plagued Magnataur! (Valiance Expedition)
quests[116941050] = 250 -- There's Something Going On In Those Caves (Valiance Expedition)
quests[116951064] = 250 -- The Horn of Elemental Fury (The Taunka)
quests[116971050] = 250 -- Rats, Tinky Went into the Necropolis! (Valiance Expedition)
quests[116981050] = 250 -- Might As Well Wipe Out the Scourge (Valiance Expedition)
quests[116991050] = 250 -- I'm Stuck in this Damned Cage... But Not For Long! (Valiance Expedition)
quests[117001050] = 10 -- Let Bixie Know (Valiance Expedition)
quests[117011050] = 350 -- Back to the Airstrip (Valiance Expedition)
quests[117031085] = 10 -- Get to Getry (Warsong Offensive)
quests[117051085] = 500 -- Foolish Endeavors (Warsong Offensive)
quests[117061064] = 350 -- The Collapse (The Taunka)
quests[117071050] = 10 -- Distress Call (Valiance Expedition)
quests[117081050] = 10 -- The Mechagnomes (Valiance Expedition)
quests[117091085] = 10 -- Nork Bloodfrenzy's Charge (Warsong Offensive)
quests[117101050] = 250 -- What's the Matter with the Transmatter? (Valiance Expedition)
quests[117111085] = 150 -- Coward Delivery... Under 30 Minutes or it's Free (Warsong Offensive)
quests[117121050] = 250 -- Re-Cursive (Valiance Expedition)
quests[117131050] = 250 -- Scouting the Sinkholes (Valiance Expedition)
quests[117141085] = 250 -- Vermin Extermination (Warsong Offensive)
quests[117151050] = 250 -- Fueling the Project (Valiance Expedition)
quests[117161085] = 250 -- The Wondrous Bloodspore (Warsong Offensive)
quests[117171085] = 250 -- Pollen from the Source (Warsong Offensive)
quests[117181050] = 250 -- A Bot in Mammoth's Clothing (Valiance Expedition)
quests[117191085] = 10 -- A Suitable Test Subject (Warsong Offensive)
quests[117201085] = 10 -- The Invasion of Gammoth (Warsong Offensive)
quests[117211085] = 250 -- Gammothra the Tormentor (Warsong Offensive)
quests[117221085] = 150 -- Trophies of Gammoth (Warsong Offensive)
quests[117231050] = 350 -- Deploy the Shake-n-Quake! (Valiance Expedition)
quests[117241085] = 250 -- Massive Moth Omelet? (Warsong Offensive)
quests[117251050] = 10 -- Finding Pilot Tailspin (Valiance Expedition)
quests[117261050] = 250 -- A Little Bit of Spice (Valiance Expedition)
quests[117271050] = 10 -- A Time for Heroes (Valiance Expedition)
quests[117281050] = 250 -- Lupus Pupus (Valiance Expedition)
quests[117291050] = 10 -- The Ultrasonic Screwdriver (Valiance Expedition)
quests[117301050] = 350 -- Master and Servant (Valiance Expedition)
quests[117331090] = 150 -- Traversing the Rift (Kirin Tor)
quests[117881050] = 350 -- Lefty Loosey, Righty Tighty (Valiance Expedition)
quests[117891050] = 150 -- A Soldier in Need (Valiance Expedition)
quests[117901050] = 75 -- Cultists Among Us (Valiance Expedition)
quests[117911050] = 10 -- Notify Arlos (Valiance Expedition)
quests[117921050] = 250 -- Enemies of the Light (Valiance Expedition)
quests[117931050] = 25 -- Further Investigation (Valiance Expedition)
quests[117941050] = 250 -- The Hunt is On (Valiance Expedition)
quests[117951050] = 250 -- Emergency Protocol: Section 8.2, Paragraph C (Valiance Expedition)
quests[117961050] = 250 -- Emergency Protocol: Section 8.2, Paragraph D (Valiance Expedition)
quests[117971050] = 250 -- The Siege (Valiance Expedition)
quests[117981050] = 350 -- The Gearmaster (Valiance Expedition)
quests[118640942] = 10 -- A Mission Statement (Cenarion Expedition)
quests[118650942] = 250 -- Unfit for Death (Cenarion Expedition)
quests[118660942] = 250 -- Ears of Our Enemies (Cenarion Expedition)
quests[118670942] = 150 -- Can't Get Ear-nough... (Cenarion Expedition)
quests[118680942] = 250 -- The Culler Cometh (Cenarion Expedition)
quests[118690942] = 250 -- Happy as a Clam (Cenarion Expedition)
quests[118700942] = 10 -- The Abandoned Reach (Cenarion Expedition)
quests[118710942] = 250 -- Not On Our Watch (Cenarion Expedition)
quests[118720942] = 250 -- The Nefarious Clam Master... (Cenarion Expedition)
quests[118731050] = 350 -- Give Fizzcrank the News (Valiance Expedition)
quests[118751077] = 250 -- Gaining the Advantage (Shattered Sun Offensive)
quests[118760942] = 250 -- Help Those That Cannot Help Themselves (Cenarion Expedition)
quests[118771077] = 250 -- Sunfury Attack Plans (Shattered Sun Offensive)
quests[118780942] = 150 -- Khu'nok Will Know (Cenarion Expedition)
quests[118790942] = 350 -- Kaw the Mammoth Destroyer (Cenarion Expedition)
quests[118801077] = 150 -- The Multiphase Survey (Shattered Sun Offensive)
quests[118840942] = 350 -- Nedar, Lord of Rhinos... (Cenarion Expedition)
quests[118851031] = 350 -- Adversarial Blood (Sha'tari Skyguard)
quests[118881064] = 10 -- Ride to Taunka'le Village (The Taunka)
quests[118891050] = 250 -- Death From Above (Valiance Expedition)
quests[118901064] = 250 -- What Are They Up To? (The Taunka)
quests[118920942] = 500 -- The Assassination of Harold Lane (Cenarion Expedition)
quests[118931064] = 250 -- The Power of the Elements (The Taunka)
quests[118951064] = 250 -- Master the Storm (The Taunka)
quests[118961064] = 250 -- Weakness to Lightning (The Taunka)
quests[118971050] = 250 -- Plug the Sinkholes (Valiance Expedition)
quests[118981064] = 350 -- Breaking Through (The Taunka)
quests[118991064] = 250 -- Souls of the Decursed (The Taunka)
quests[119001090] = 250 -- Reading the Meters (Kirin Tor)
quests[119011050] = 150 -- Military? What Military? (Valiance Expedition)
quests[119021050] = 75 -- Pernicious Evidence (Valiance Expedition)
quests[119031050] = 250 -- It's Time for Action (Valiance Expedition)
quests[119041050] = 250 -- Fruits of Our Labor (Valiance Expedition)
quests[119051090] = 350 -- Postponing the Inevitable (Kirin Tor)
quests[119061064] = 250 -- Cleaning Up the Pools (The Taunka)
quests[119071064] = 350 -- The Sub-Chieftains (The Taunka)
quests[119081050] = 150 -- Reference Material (Valiance Expedition)
quests[119091064] = 350 -- Defeat the Gearmaster (The Taunka)
quests[119101090] = 250 -- Secrets of the Ancients (Kirin Tor)
quests[119111090] = 250 -- Quickening (Kirin Tor)
quests[119121090] = 250 -- Nuts for Berries (Kirin Tor)
quests[119131050] = 250 -- Take No Chances (Valiance Expedition)
quests[119141090] = 250 -- Keep the Secret Safe (Kirin Tor)
quests[119161085] = 500 -- Hellscream's Champion (Warsong Offensive)
quests[119181091] = 250 -- Basic Training (The Wyrmrest Accord)
quests[119191091] = 250 -- Drake Hunt (The Wyrmrest Accord)
quests[119201050] = 75 -- Cultists Among Us (Valiance Expedition)
quests[119281050] = 10 -- Farshire (Valiance Expedition)
quests[119291064] = 10 -- The Fall of Taunka'le Village (The Taunka)
quests[119291085] = 10 -- The Fall of Taunka'le Village (Warsong Offensive)
quests[119301064] = 250 -- Across Transborea (The Taunka)
quests[119301085] = 250 -- Across Transborea (Warsong Offensive)
quests[119311091] = 250 -- Cracking the Code (The Wyrmrest Accord)
quests[119321050] = 75 -- Cowards and Fools (Valiance Expedition)
quests[119361091] = 250 -- Hatching a Plan (The Wyrmrest Accord)
quests[119381050] = 250 -- Buying Some Time (Valiance Expedition)
quests[119401091] = 250 -- Drake Hunt (The Wyrmrest Accord)
quests[119411091] = 250 -- Puzzling... (The Wyrmrest Accord)
quests[119421050] = 250 -- Words of Power (Valiance Expedition)
quests[119431091] = 250 -- The Cell (The Wyrmrest Accord)
quests[119441050] = 250 -- Surrounded! (Valiance Expedition)
quests[119451073] = 250 -- Preparing for the Worst (The Kalu'ak)
quests[119461091] = 10 -- Keristrasza (The Wyrmrest Accord)
quests[119491073] = 250 -- Not Without a Fight! (The Kalu'ak)
quests[119501073] = 250 -- Muahit's Wisdom (The Kalu'ak)
quests[119511091] = 250 -- Bait and Switch (The Wyrmrest Accord)
quests[119561050] = 250 -- Finding the Phylactery (Valiance Expedition)
quests[119571091] = 250 -- Saragosa's End (The Wyrmrest Accord)
quests[119581073] = 250 -- Let Nothing Go To Waste (The Kalu'ak)
quests[119591073] = 350 -- Slay Loguhn (The Kalu'ak)
quests[119601073] = 500 -- Planning for the Future (The Kalu'ak)
quests[119611073] = 150 -- Spirits Watch Over Us (The Kalu'ak)
quests[119621050] = 25 -- One Last Delivery (Valiance Expedition)
quests[119631050] = 25 -- Weapons for Farshire (Valiance Expedition)
quests[119651050] = 250 -- Call to Arms! (Valiance Expedition)
quests[119671091] = 75 -- Mustering the Reds (The Wyrmrest Accord)
quests[119681073] = 250 -- The Tides Turn (The Kalu'ak)
quests[119691091] = 250 -- Springing the Trap (The Wyrmrest Accord)
quests[119731091] = 250 -- Prisoner of War (The Wyrmrest Accord)
quests[119751011] = 250 -- Now, When I Grow Up... (Lower City)
quests[119771064] = 10 -- A Tauren Among Taunka (The Taunka)
quests[119771085] = 10 -- A Tauren Among Taunka (Warsong Offensive)
quests[119781064] = 250 -- Into the Fold (The Taunka)
quests[119781085] = 250 -- Into the Fold (Warsong Offensive)
quests[119791064] = 10 -- The Taunka and the Tauren (The Taunka)
quests[119791085] = 10 -- The Taunka and the Tauren (Warsong Offensive)
quests[119801085] = 250 -- Pride of the Horde (Warsong Offensive)
quests[119831064] = 150 -- Blood Oath of the Horde (The Taunka)
quests[119831085] = 150 -- Blood Oath of the Horde (Warsong Offensive)
quests[119951090] = 10 -- Your Presence is Required at Stars' Rest (Kirin Tor)
quests[119961090] = 10 -- Your Presence is Required at Agmar's Hammer (Kirin Tor)
quests[119991090] = 250 -- Rifle the Bodies (Kirin Tor)
quests[120001090] = 250 -- Rifle the Bodies (Kirin Tor)
quests[120041090] = 350 -- Prevent the Accord (Kirin Tor)
quests[120051090] = 350 -- Prevent the Accord (Kirin Tor)
quests[120081085] = 150 -- Agmar's Hammer (Warsong Offensive)
quests[120091073] = 250 -- Tua'kea's Crab Traps (The Kalu'ak)
quests[120111073] = 10 -- Signs of Big Watery Trouble (The Kalu'ak)
quests[120161073] = 250 -- The Bait (The Kalu'ak)
quests[120171073] = 350 -- Meat on the Hook (The Kalu'ak)
quests[120191050] = 500 -- Last Rites (Valiance Expedition)
quests[120261064] = 250 -- The Damaged Journal (The Taunka)
quests[120281073] = 250 -- Spiritual Insight (The Kalu'ak)
quests[120301073] = 250 -- Elder Mana'loa (The Kalu'ak)
quests[120311073] = 250 -- Freedom for the Lingering (The Kalu'ak)
quests[120321073] = 500 -- Conversing With the Depths (The Kalu'ak)
quests[120331085] = 150 -- Message from the West (Warsong Offensive)
quests[120341085] = 10 -- Victory Nears... (Warsong Offensive)
quests[120351050] = 250 -- Repurposed Technology (Valiance Expedition)
quests[120361085] = 250 -- From the Depths of Azjol-Nerub (Warsong Offensive)
quests[120391085] = 250 -- Black Blood of Yogg-Saron (Warsong Offensive)
quests[120411052] = 150 -- The Lost Empire (Horde Expedition)
quests[120481085] = 250 -- Scourge Armaments (Warsong Offensive)
quests[120531085] = 250 -- The Might of the Horde (Warsong Offensive)
quests[120541064] = 250 -- Deciphering the Journal (The Taunka)
quests[120551090] = 75 -- A Strange Device (Kirin Tor)
quests[120561085] = 250 -- Marked for Death: High Cultist Zangus (Warsong Offensive)
quests[120571085] = 250 -- The Flesh-Bound Tome (Warsong Offensive)
quests[120581064] = 250 -- The Runic Prophecies (The Taunka)
quests[120591090] = 75 -- A Strange Device (Kirin Tor)
quests[120601090] = 250 -- Projections and Plans (Kirin Tor)
quests[120611090] = 250 -- Projections and Plans (Kirin Tor)
quests[120631052] = 150 -- Strength of Icemist (Horde Expedition)
quests[120631064] = 150 -- Strength of Icemist (The Taunka)
quests[120641052] = 250 -- Chains of the Anub'ar (Horde Expedition)
quests[120641064] = 250 -- Chains of the Anub'ar (The Taunka)
quests[120651090] = 250 -- The Focus on the Beach (Kirin Tor)
quests[120661090] = 250 -- The Focus on the Beach (Kirin Tor)
quests[120671050] = 250 -- A Letter for Home (Valiance Expedition)
quests[120691052] = 500 -- Return of the High Chief (Horde Expedition)
quests[120691064] = 500 -- Return of the High Chief (The Taunka)
quests[120711085] = 10 -- Attack by Air!  (Warsong Offensive)
quests[120721085] = 250 -- Blightbeasts be Damned! (Warsong Offensive)
quests[120731085] = 250 -- Pounding the Iron (Warsong Offensive)
quests[120831090] = 250 -- Atop the Woodlands (Kirin Tor)
quests[120841090] = 250 -- Atop the Woodlands (Kirin Tor)
quests[120851085] = 250 -- A Letter for Home (Warsong Offensive)
quests[120861050] = 250 -- The Son of Karkut (Valiance Expedition)
quests[120881050] = 25 -- Thassarian, the Death Knight (Valiance Expedition)
quests[120891085] = 350 -- Wanted: Magister Keldonus (Warsong Offensive)
quests[120901052] = 350 -- Wanted: Gigantaur (Horde Expedition)
quests[120911085] = 350 -- Wanted: Dreadtalon (Warsong Offensive)
quests[120921050] = 250 -- Strengthen the Ancients (Valiance Expedition)
quests[120951085] = 10 -- To Dragon's Fall (Warsong Offensive)
quests[120961052] = 250 -- Strengthen the Ancients (Horde Expedition)
quests[120971085] = 500 -- Sarathstra, Scourge of the North (Warsong Offensive)
quests[120981090] = 250 -- Search Indu'le Village (Kirin Tor)
quests[121001085] = 250 -- Containing the Rot (Warsong Offensive)
quests[121011085] = 10 -- The Good Doctor... (Warsong Offensive)
quests[121021085] = 250 -- In Search of the Ruby Lilac (Warsong Offensive)
quests[121041085] = 10 -- Return to Soar (Warsong Offensive)
quests[121051050] = 75 -- Descent into Darkness (Valiance Expedition)
quests[121061090] = 250 -- Search Indu'le Village (Kirin Tor)
quests[121071090] = 350 -- The End of the Line (Kirin Tor)
quests[121091050] = 25 -- Report to Gryan Stoutmantle... Again (Valiance Expedition)
quests[121101090] = 350 -- The End of the Line (Kirin Tor)
quests[121111085] = 250 -- Where the Wild Things Roam (Warsong Offensive)
quests[121151085] = 10 -- Koltira and the Language of Death (Warsong Offensive)
quests[121171073] = 10 -- Travel to Moa'ki Harbor (The Kalu'ak)
quests[121181073] = 10 -- Travel to Moa'ki Harbor (The Kalu'ak)
quests[121191091] = 25 -- Gaining an Audience (The Wyrmrest Accord)
quests[121221091] = 25 -- Gaining an Audience (The Wyrmrest Accord)
quests[121231091] = 10 -- Informing the Queen (The Wyrmrest Accord)
quests[121241091] = 10 -- Informing the Queen (The Wyrmrest Accord)
quests[121251085] = 250 -- In Service of Blood (Warsong Offensive)
quests[121261085] = 350 -- In Service of the Unholy (Warsong Offensive)
quests[121271085] = 250 -- In Service of Frost (Warsong Offensive)
quests[121321085] = 250 -- The Power to Destroy (Warsong Offensive)
quests[121361085] = 500 -- The Translated Tome (Warsong Offensive)
quests[121401085] = 250 -- All Hail Roanauk! (Warsong Offensive)
quests[121421050] = 250 -- Pest Control (Valiance Expedition)
quests[121431050] = 250 -- Canyon Chase (Valiance Expedition)
quests[121441085] = 250 -- Pest Control (Warsong Offensive)
quests[121451085] = 250 -- Canyon Chase (Warsong Offensive)
quests[121461091] = 250 -- Disturbing Implications (The Wyrmrest Accord)
quests[121471091] = 250 -- Disturbing Implications (The Wyrmrest Accord)
quests[121481091] = 250 -- One of a Kind (The Wyrmrest Accord)
quests[121491091] = 250 -- Mighty Magnataur (The Wyrmrest Accord)
quests[121501091] = 350 -- Reclusive Runemaster (The Wyrmrest Accord)
quests[121511091] = 350 -- Wanton Warlord (The Wyrmrest Accord)
quests[121571050] = 25 -- The Lost Courier (Valiance Expedition)
quests[121651064] = 250 -- An Intriguing Plan (The Taunka)
quests[121661050] = 250 -- The Liquid Fire of Elune (Valiance Expedition)
quests[121671050] = 250 -- Kill the Cultists (Valiance Expedition)
quests[121681050] = 10 -- The Favor of Zangus (Valiance Expedition)
quests[121691050] = 350 -- The High Cultist (Valiance Expedition)
quests[121701052] = 250 -- Blackriver Brawl (Horde Expedition)
quests[121711050] = 10 -- Of Traitors and Treason (Valiance Expedition)
quests[121721090] = 250 -- Attunement to Dalaran (Kirin Tor)
quests[121731090] = 250 -- Attunement to Dalaran (Kirin Tor)
quests[121741050] = 10 -- High Commander Halford Wyrmbane (Valiance Expedition)
quests[121751052] = 250 -- Gray Worg Hides (Horde Expedition)
quests[121761052] = 250 -- A Minor Substitution (Horde Expedition)
quests[121771052] = 75 -- Jun'ik's Coverup (Horde Expedition)
quests[121781052] = 75 -- Delivery to Krenna (Horde Expedition)
quests[121811067] = 10 -- Give it a Name (The Hand of Vengeance)
quests[121821067] = 10 -- To Venomspite! (The Hand of Vengeance)
quests[121881067] = 250 -- The Forsaken Blight and You: How Not to Die (The Hand of Vengeance)
quests[121891067] = 25 -- Imbeciles Abound! (The Hand of Vengeance)
quests[121951064] = 250 -- The Unexpected 'Guest' (The Taunka)
quests[121961064] = 250 -- From the Ground Up (The Taunka)
quests[121971064] = 250 -- We Have the Power (The Taunka)
quests[121981064] = 250 -- ... Or Maybe We Don't (The Taunka)
quests[121991064] = 350 -- Bringing Down the Iron Thane (The Taunka)
quests[122001067] = 250 -- Emerald Dragon Tears (The Hand of Vengeance)
quests[122051067] = 250 -- Wanted: The Scarlet Onslaught (The Hand of Vengeance)
quests[122061067] = 10 -- Blighted Last Rites (The Hand of Vengeance)
quests[122071085] = 250 -- Vordrassil's Fall (Warsong Offensive)
quests[122091067] = 250 -- Materiel Plunder (The Hand of Vengeance)
quests[122111067] = 250 -- Let Them Not Rise! (The Hand of Vengeance)
quests[122121050] = 250 -- Replenishing the Storehouse (Valiance Expedition)
quests[122131085] = 250 -- The Darkness Beneath (Warsong Offensive)
quests[122141067] = 250 -- Fresh Remounts (The Hand of Vengeance)
quests[122151050] = 150 -- Them or Us! (Valiance Expedition)
quests[122161050] = 250 -- Take Their Rear! (Valiance Expedition)
quests[122171050] = 250 -- Eagle Eyes (Valiance Expedition)
quests[122181067] = 250 -- Spread the Good Word (The Hand of Vengeance)
quests[122191050] = 250 -- The Failed World Tree (Valiance Expedition)
quests[122211067] = 150 -- The Forsaken Blight (The Hand of Vengeance)
quests[122221050] = 250 -- Secrets of the Flamebinders (Valiance Expedition)
quests[122231050] = 250 -- Thinning the Ranks (Valiance Expedition)
quests[122241085] = 25 -- The Kor'kron Vanguard! (Warsong Offensive)
quests[122251050] = 75 -- Mmm... Amberseeds! (Valiance Expedition)
quests[122261050] = 250 -- Just Passing Through (Valiance Expedition)
quests[122271050] = 150 -- Doing Your Duty (Valiance Expedition)
quests[122291085] = 350 -- A Possible Link (Warsong Offensive)
quests[122301067] = 250 -- Stealing from the Siegesmiths (The Hand of Vengeance)
quests[122311085] = 150 -- The Bear God's Offspring (Warsong Offensive)
quests[122321067] = 250 -- Bombard the Ballistae (The Hand of Vengeance)
quests[122341067] = 250 -- Need to Know (The Hand of Vengeance)
quests[122351050] = 10 -- Naxxramas and the Fall of Wintergarde (Valiance Expedition)
quests[122361085] = 500 -- Ursoc, the Bear God (Warsong Offensive)
quests[122371050] = 250 -- Flight of the Wintergarde Defender (Valiance Expedition)
quests[122391067] = 250 -- The Spy in New Hearthglen (The Hand of Vengeance)
quests[122401067] = 250 -- A Means to an End (The Hand of Vengeance)
quests[122411085] = 250 -- Destroy the Sapling (Warsong Offensive)
quests[122421085] = 250 -- Vordrassil's Seeds (Warsong Offensive)
quests[122431067] = 250 -- Fire Upon the Waters (The Hand of Vengeance)
quests[122451067] = 250 -- No Mercy for the Captured (The Hand of Vengeance)
quests[122461050] = 350 -- A Possible Link (Valiance Expedition)
quests[122471085] = 150 -- Children of Ursoc (Warsong Offensive)
quests[122481050] = 250 -- Vordrassil's Sapling (Valiance Expedition)
quests[122491050] = 500 -- Ursoc, the Bear God (Valiance Expedition)
quests[122501050] = 250 -- Vordrassil's Seeds (Valiance Expedition)
quests[122511050] = 10 -- Return to the High Commander (Valiance Expedition)
quests[122521067] = 250 -- Torture the Torturer (The Hand of Vengeance)
quests[122531050] = 250 -- Rescue from Town Square (Valiance Expedition)
quests[122541067] = 250 -- Without a Prayer (The Hand of Vengeance)
quests[122551050] = 250 -- The Thane of Voldrune (Valiance Expedition)
quests[122561085] = 250 -- The Flamebinders' Secrets (Warsong Offensive)
quests[122571085] = 250 -- A Show of Strength (Warsong Offensive)
quests[122581050] = 250 -- The Fate of the Dead (Valiance Expedition)
quests[122591085] = 250 -- The Thane of Voldrune (Warsong Offensive)
quests[122601067] = 250 -- The Perfect Dissemblance (The Hand of Vengeance)
quests[122611091] = 250 -- No Place to Run (The Wyrmrest Accord)
quests[122621091] = 250 -- No One to Save You (The Wyrmrest Accord)
quests[122631091] = 250 -- The Best of Intentions (The Wyrmrest Accord)
quests[122641091] = 250 -- Culling the Damned (The Wyrmrest Accord)
quests[122651091] = 250 -- Defiling the Defilers (The Wyrmrest Accord)
quests[122661091] = 10 -- Tales of Destruction (The Wyrmrest Accord)
quests[122671091] = 350 -- Neltharion's Flame (The Wyrmrest Accord)
quests[122691050] = 250 -- Not In Our Mine (Valiance Expedition)
quests[122701085] = 250 -- Shred the Alliance (Warsong Offensive)
quests[122711067] = 75 -- The Rod of Compulsion (The Hand of Vengeance)
quests[122721050] = 250 -- The Bleeding Ore (Valiance Expedition)
quests[122731067] = 350 -- The Denouncement (The Hand of Vengeance)
quests[122741067] = 250 -- A Fall From Grace (The Hand of Vengeance)
quests[122761050] = 250 -- The Search for Slinkin (Valiance Expedition)
quests[122771050] = 350 -- Leave Nothing to Chance (Valiance Expedition)
quests[122801085] = 250 -- Making Repairs (Warsong Offensive)
quests[122811050] = 10 -- Understanding the Scourge War Machine (Valiance Expedition)
quests[122821050] = 250 -- Imprints on the Past (Valiance Expedition)
quests[122831067] = 250 -- The Truth Will Out (The Hand of Vengeance)
quests[122841085] = 250 -- Keep 'Em on Their Heels (Warsong Offensive)
quests[122851067] = 350 -- Do Unto Others (The Hand of Vengeance)
quests[122871050] = 25 -- Orik Trueheart and the Forgotten Shore (Valiance Expedition)
quests[122881085] = 250 -- Overwhelmed! (Warsong Offensive)
quests[122901050] = 250 -- The Murkweed Elixir (Valiance Expedition)
quests[122901106] = 250 -- The Murkweed Elixir (Argent Crusade)
quests[122911050] = 250 -- The Forgotten Tale (Valiance Expedition)
quests[122911106] = 250 -- The Forgotten Tale (Argent Crusade)
quests[122921050] = 250 -- Local Support (Valiance Expedition)
quests[122931050] = 25 -- Close the Deal (Valiance Expedition)
quests[122941050] = 25 -- A Tentative Pact (Valiance Expedition)
quests[122951050] = 250 -- An Exercise in Diplomacy (Valiance Expedition)
quests[122971050] = 10 -- Of Traitors and Treason (Valiance Expedition)
quests[122981050] = 10 -- High Commander Halford Wyrmbane (Valiance Expedition)
quests[123011050] = 500 -- The Truth Shall Set Us Free (Valiance Expedition)
quests[123011106] = 500 -- The Truth Shall Set Us Free (Argent Crusade)
quests[123021050] = 250 -- Words of Warning (Valiance Expedition)
quests[123031067] = 250 -- Funding the War Effort (The Hand of Vengeance)
quests[123041067] = 250 -- Beachfront Property (The Hand of Vengeance)
quests[123051050] = 25 -- Parting Thoughts (Valiance Expedition)
quests[123081050] = 250 -- Escape from Silverbrook (Valiance Expedition)
quests[123091050] = 10 -- Find Durkon! (Valiance Expedition)
quests[123101050] = 250 -- A Swift Response (Valiance Expedition)
quests[123111050] = 250 -- The Noble's Crypt (Valiance Expedition)
quests[123121050] = 25 -- Secrets of the Scourge (Valiance Expedition)
quests[123191050] = 10 -- Mystery of the Tome (Valiance Expedition)
quests[123201050] = 10 -- Understanding the Language of Death (Valiance Expedition)
quests[123211050] = 250 -- A Righteous Sermon (Valiance Expedition)
quests[123251050] = 25 -- Into Hostile Territory (Valiance Expedition)
quests[123261050] = 350 -- Steamtank Surprise (Valiance Expedition)
quests[123721091] = 250 -- Defending Wyrmrest Temple (The Wyrmrest Accord)
quests[124131085] = 250 -- Attack on Silverbrook (Warsong Offensive)
quests[124151064] = 250 -- The Horse Hollerer (The Taunka)
quests[124161091] = 250 -- Heated Battle (The Wyrmrest Accord)
quests[124171091] = 250 -- Return to the Earth (The Wyrmrest Accord)
quests[124181091] = 250 -- Through Fields of Flame (The Wyrmrest Accord)
quests[124191091] = 350 -- The Fate of the Ruby Dragonshrine (The Wyrmrest Accord)
quests[124221085] = 25 -- Tactical Clemency (Warsong Offensive)
quests[124231085] = 75 -- Mikhail's Journal (Warsong Offensive)
quests[124241085] = 75 -- Gorgonna (Warsong Offensive)
quests[124311085] = 500 -- The Conquest Pit: Final Showdown (Warsong Offensive)
quests[124351091] = 10 -- Report to Lord Afrasastrasz (The Wyrmrest Accord)
quests[124361085] = 250 -- Supplemental Income (Warsong Offensive)
quests[124381050] = 350 -- Wanted: Kreug Oathbreaker (Valiance Expedition)
quests[124391050] = 10 -- A Disturbance In The West (Valiance Expedition)
quests[124401050] = 10 -- To Stars' Rest! (Valiance Expedition)
quests[124411050] = 350 -- Wanted: High Shaman Bloodpaw (Valiance Expedition)
quests[124421050] = 350 -- Wanted: Onslaught Commander Iustus (Valiance Expedition)
quests[124441050] = 250 -- Blackriver Skirmish (Valiance Expedition)
quests[124471091] = 10 -- The Obsidian Dragonshrine (The Wyrmrest Accord)
quests[124481091] = 250 -- Heated Battle (The Wyrmrest Accord)
quests[124491091] = 250 -- Return to the Earth (The Wyrmrest Accord)
quests[124501091] = 250 -- Through Fields of Flame (The Wyrmrest Accord)
quests[124511064] = 25 -- Onward to Camp Oneqwah (The Taunka)
quests[124531085] = 150 -- Eyes Above (Warsong Offensive)
quests[124541091] = 250 -- Cycle of Life (The Wyrmrest Accord)
quests[124551050] = 250 -- Scattered To The Wind (Valiance Expedition)
quests[124561091] = 350 -- The Plume of Alystros (The Wyrmrest Accord)
quests[124571050] = 350 -- The Chain Gun And You (Valiance Expedition)
quests[124581091] = 250 -- Seeds of the Lashers (The Wyrmrest Accord)
quests[124591091] = 500 -- That Which Creates Can Also Destroy (The Wyrmrest Accord)
quests[124601091] = 10 -- Report to the Ruby Dragonshrine (The Wyrmrest Accord)
quests[124611091] = 10 -- Report to the Ruby Dragonshrine (The Wyrmrest Accord)
quests[124621050] = 250 -- Breaking Off A Piece (Valiance Expedition)
quests[124631050] = 75 -- Plunderbeard Must Be Found! (Valiance Expedition)
quests[124641050] = 350 -- My Old Enemy (Valiance Expedition)
quests[124651050] = 250 -- Plunderbeard's Journal (Valiance Expedition)
quests[124661050] = 25 -- Chasing Icestorm: The 7th Legion Front (Valiance Expedition)
quests[124671050] = 350 -- Chasing Icestorm: Thel'zan's Phylactery (Valiance Expedition)
quests[124701091] = 350 -- Mystery of the Infinite (The Wyrmrest Accord)
quests[124711073] = 150 -- Cruelty of the Kvaldir (The Kalu'ak)
quests[124721050] = 10 -- Finality (Valiance Expedition)
quests[124731050] = 250 -- An End And A Beginning (Valiance Expedition)
quests[124741050] = 500 -- To Fordragon Hold! (Valiance Expedition)
quests[124751050] = 250 -- What Secrets Men Hide (Valiance Expedition)
quests[124771050] = 250 -- The Path of Redemption (Valiance Expedition)
quests[124781050] = 350 -- Frostmourne Cavern (Valiance Expedition)
quests[124811052] = 350 -- Adding Injury to Insult (Horde Expedition)
quests[124821067] = 250 -- Against Nifflevar (The Hand of Vengeance)
quests[124861085] = 10 -- To Bor'gorok Outpost, Quickly! (Warsong Offensive)
quests[124871085] = 10 -- To Conquest Hold, But Be Careful! (Warsong Offensive)
quests[124881067] = 10 -- The High Executor Needs You (The Hand of Vengeance)
quests[124951050] = 10 -- Audience With The Dragon Queen (Valiance Expedition)
quests[124951091] = 10 -- Audience With The Dragon Queen (The Wyrmrest Accord)
quests[124961085] = 10 -- Audience With The Dragon Queen (Warsong Offensive)
quests[124961091] = 10 -- Audience With The Dragon Queen (The Wyrmrest Accord)
quests[124981091] = 350 -- On Ruby Wings (The Wyrmrest Accord)
quests[124991050] = 500 -- Return To Angrathar (Valiance Expedition)
quests[125001085] = 500 -- Return To Angrathar (Warsong Offensive)
quests[125031106] = 250 -- Defend the Stand (Argent Crusade)
quests[125041106] = 250 -- Argent Crusade, We Are Leaving! (Argent Crusade)
quests[125051106] = 10 -- New Orders for Sergeant Stackhammer (Argent Crusade)
quests[125081106] = 250 -- Mopping Up (Argent Crusade)
quests[125111050] = 25 -- The Hills Have Us (Valiance Expedition)
quests[125121106] = 250 -- Leave No One Behind (Argent Crusade)
quests[125170909] = 25 -- Rogues Deck (Darkmoon Faire)
quests[125180909] = 25 -- Mages Deck (Darkmoon Faire)
quests[125271106] = 250 -- Gluttonous Lurkers (Argent Crusade)
quests[125281104] = 350 -- Playing Along (Frenzyheart Tribe)
quests[125291104] = 350 -- The Ape Hunter's Slave (Frenzyheart Tribe)
quests[125301104] = 500 -- Tormenting the Softknuckles (Frenzyheart Tribe)
quests[125311104] = 500 -- The Underground Menace (Frenzyheart Tribe)
quests[125321104] = 350 -- Flown the Coop! (Frenzyheart Tribe)
quests[125331104] = 350 -- The Wasp Hunter's Apprentice (Frenzyheart Tribe)
quests[125341104] = 500 -- The Sapphire Queen (Frenzyheart Tribe)
quests[125351104] = 350 -- Mischief in the Making (Frenzyheart Tribe)
quests[125361104] = 350 -- A Rough Ride (Frenzyheart Tribe)
quests[125371104] = 500 -- Lightning Definitely Strikes Twice (Frenzyheart Tribe)
quests[125381104] = 350 -- The Mist Isn't Listening (Frenzyheart Tribe)
quests[125391104] = 350 -- Hoofing It (Frenzyheart Tribe)
quests[125421106] = 25 -- The Call Of The Crusade (Argent Crusade)
quests[125451106] = 250 -- The Cleansing Of Jintha'kalar (Argent Crusade)
quests[125521106] = 250 -- Death to the Necromagi (Argent Crusade)
quests[125531106] = 250 -- Skimmer Spinnerets (Argent Crusade)
quests[125541106] = 350 -- Malas the Corrupter (Argent Crusade)
quests[125551106] = 250 -- A Tangled Skein (Argent Crusade)
quests[125621106] = 250 -- The Drakkari Do Not Need Water Elementals! (Argent Crusade)
quests[125661064] = 10 -- Help for Camp Winterhoof (The Taunka)
quests[125701105] = 350 -- Fortunate Misunderstandings (The Oracles)
quests[125711105] = 350 -- Make the Bad Snake Go Away (The Oracles)
quests[125721105] = 350 -- Gods like Shiny Things (The Oracles)
quests[125731105] = 350 -- Making Peace (The Oracles)
quests[125741105] = 250 -- Back So Soon? (The Oracles)
quests[125751105] = 350 -- The Lost Mistwhisper Treasure (The Oracles)
quests[125761105] = 350 -- Forced Hand (The Oracles)
quests[125771105] = 250 -- Home Time! (The Oracles)
quests[125781105] = 250 -- The Angry Gorloc (The Oracles)
quests[125791105] = 350 -- Lifeblood of the Mosswalker Shrine (The Oracles)
quests[125801105] = 350 -- The Mosswalker Savior (The Oracles)
quests[125831106] = 250 -- Crashed Sprayer (Argent Crusade)
quests[125841106] = 250 -- Pure Evil (Argent Crusade)
quests[125871106] = 350 -- Troll Patrol (Argent Crusade)
quests[125961106] = 250 -- Pa'Troll (Argent Crusade)
quests[125971106] = 75 -- Something for the Pain (Argent Crusade)
quests[125981106] = 75 -- Throwing Down (Argent Crusade)
quests[125991106] = 75 -- Creature Comforts (Argent Crusade)
quests[126041106] = 500 -- Congratulations! (Argent Crusade)
quests[126061106] = 250 -- Cocooned! (Argent Crusade)
quests[126091106] = 250 -- Stocking the Shelves (Argent Crusade)
quests[126101106] = 250 -- Clipping Their Wings (Argent Crusade)
quests[126291098] = 250 -- You Can Run, But You Can't Hide (Knights of the Ebon Blade)
quests[126311098] = 150 -- An Invitation, of Sorts... (Knights of the Ebon Blade)
quests[126331098] = 150 -- Darkness Calling (Knights of the Ebon Blade)
quests[126371098] = 75 -- Near Miss (Knights of the Ebon Blade)
quests[126381098] = 75 -- Close Call (Knights of the Ebon Blade)
quests[126431098] = 250 -- Silver Lining (Knights of the Ebon Blade)
quests[126481098] = 150 -- Dressing Down (Knights of the Ebon Blade)
quests[126491098] = 150 -- Suit Up! (Knights of the Ebon Blade)
quests[126611098] = 250 -- Infiltrating Voltarus (Knights of the Ebon Blade)
quests[126691098] = 250 -- So Far, So Bad (Knights of the Ebon Blade)
quests[126761098] = 250 -- Sabotage (Knights of the Ebon Blade)
quests[126771098] = 250 -- Hazardous Materials (Knights of the Ebon Blade)
quests[127021104] = 250 -- Chicken Party! (Frenzyheart Tribe)
quests[127031104] = 350 -- Kartak's Rampage (Frenzyheart Tribe)
quests[127041105] = 250 -- Appeasing the Great Rain Stone (The Oracles)
quests[127051105] = 350 -- Will of the Titans (The Oracles)
quests[127131098] = 500 -- Betrayal (Knights of the Ebon Blade)
quests[127261105] = 250 -- Song of Wind and Water (The Oracles)
quests[127281090] = 350 -- Monitoring the Rift: Winterfin Cavern (Kirin Tor)
quests[127321104] = 250 -- The Heartblood's Strength (Frenzyheart Tribe)
quests[127341104] = 250 -- Rejek: First Blood (Frenzyheart Tribe)
quests[127351105] = 250 -- A Cleansing Song (The Oracles)
quests[127361105] = 250 -- Song of Reflection (The Oracles)
quests[127371105] = 250 -- Song of Fecundity (The Oracles)
quests[127401106] = 250 -- Parachutes for the Argent Crusade (Argent Crusade)
quests[127411104] = 250 -- Strength of the Tempest (Frenzyheart Tribe)
quests[127581104] = 250 -- A Hero's Headgear (Frenzyheart Tribe)
quests[127591104] = 350 -- Tools of War (Frenzyheart Tribe)
quests[127601104] = 350 -- Secret Strength of the Frenzyheart (Frenzyheart Tribe)
quests[127611105] = 350 -- Mastery of the Crystals (The Oracles)
quests[127621105] = 350 -- Power of the Great Ones (The Oracles)
quests[127661050] = 10 -- Speak with your Ambassador (Valiance Expedition)
quests[127671085] = 10 -- Speak with your Ambassador (Warsong Offensive)
quests[127681091] = 10 -- The Steward of Wyrmrest Temple (The Wyrmrest Accord)
quests[127691091] = 10 -- The Steward of Wyrmrest Temple (The Wyrmrest Accord)
quests[127901090] = 10 -- Learning to Leave and Return: the Magical Way (Kirin Tor)
quests[127911090] = 10 -- The Magical Kingdom of Dalaran (Kirin Tor)
quests[127941090] = 10 -- The Magical Kingdom of Dalaran (Kirin Tor)
quests[127961090] = 10 -- The Magical Kingdom of Dalaran (Kirin Tor)
quests[127980909] = 25 -- Swords Deck (Darkmoon Faire)
quests[128011098] = 3000 -- The Light of Dawn (Knights of the Ebon Blade)
quests[128061098] = 10 -- To the Rise with all Due Haste! (Knights of the Ebon Blade)
quests[128071098] = 10 -- The Story Thus Far... (Knights of the Ebon Blade)
quests[128101098] = 250 -- Blood in the Water (Knights of the Ebon Blade)
quests[128131098] = 250 -- From Their Corpses, Rise! (Knights of the Ebon Blade)
quests[128141098] = 250 -- You'll Need a Gryphon (Knights of the Ebon Blade)
quests[128151098] = 250 -- No Fly Zone (Knights of the Ebon Blade)
quests[128381098] = 250 -- Intelligence Gathering (Knights of the Ebon Blade)
quests[128391098] = 250 -- The Grand (Admiral's) Plan (Knights of the Ebon Blade)
quests[128401098] = 250 -- In Strict Confidence (Knights of the Ebon Blade)
quests[128471098] = 350 -- Second Chances (Knights of the Ebon Blade)
quests[128521098] = 350 -- The Admiral Revealed (Knights of the Ebon Blade)
quests[128551068] = 250 -- Sniffing Out the Perpetrator (Explorers' League)
quests[128571106] = 350 -- Wanted: Ragemane's Flipper (Argent Crusade)
quests[128581068] = 250 -- Pieces to the Puzzle (Explorers' League)
quests[128591106] = 250 -- This Just In: Fire Still Hot! (Argent Crusade)
quests[128601068] = 250 -- Data Mining (Explorers' League)
quests[128631126] = 10 -- Offering Thanks (The Frostborn)
quests[128641126] = 250 -- Missing Scouts (The Frostborn)
quests[128651126] = 250 -- Loyal Companions (The Frostborn)
quests[128661126] = 250 -- Stemming the Aggressors (The Frostborn)
quests[128671126] = 250 -- Baby Stealers (The Frostborn)
quests[128681126] = 350 -- Sirana Iceshriek (The Frostborn)
quests[128691126] = 250 -- Pushed Too Far (The Frostborn)
quests[128701068] = 250 -- Ancient Relics (Explorers' League)
quests[128711126] = 10 -- Aid from the Explorers' League (The Frostborn)
quests[128721068] = 250 -- Norgannon's Shell (Explorers' League)
quests[128731126] = 10 -- The Frostborn King (The Frostborn)
quests[128741126] = 250 -- Fervor of the Frostborn (The Frostborn)
quests[128751126] = 250 -- An Experienced Guide (The Frostborn)
quests[128761126] = 250 -- Unwelcome Guests (The Frostborn)
quests[128771126] = 250 -- The Lonesome Watcher (The Frostborn)
quests[128791126] = 350 -- Fury of the Frostborn King (The Frostborn)
quests[128801068] = 10 -- The Master Explorer (Explorers' League)
quests[128821085] = 250 -- Ancient Relics (Warsong Offensive)
quests[128831106] = 250 -- Orders From Drakuru (Argent Crusade)
quests[128841098] = 25 -- The Ebon Watch (Knights of the Ebon Blade)
quests[128871098] = 250 -- It's All Fun and Games (Knights of the Ebon Blade)
quests[128911098] = 250 -- I Have an Idea, But First... (Knights of the Ebon Blade)
quests[128921098] = 250 -- It's All Fun and Games (Knights of the Ebon Blade)
quests[128931098] = 350 -- Free Your Mind (Knights of the Ebon Blade)
quests[128941106] = 25 -- Crusader Forward Camp (Argent Crusade)
quests[128951085] = 250 -- The Missing Bronzebeard (Warsong Offensive)
quests[128981098] = 10 -- The Shadow Vault (Knights of the Ebon Blade)
quests[128991098] = 10 -- The Shadow Vault (Knights of the Ebon Blade)
quests[129011106] = 250 -- Making Something Out Of Nothing (Argent Crusade)
quests[129021106] = 150 -- In Search Of Answers (Argent Crusade)
quests[129031106] = 250 -- That's What Friends Are For... (Argent Crusade)
quests[129041106] = 350 -- Light Won't Grant Me Vengeance (Argent Crusade)
quests[129091085] = 10 -- The Nose Knows (Warsong Offensive)
quests[129131085] = 10 -- Speak Orcish, Man! (Warsong Offensive)
quests[129141106] = 250 -- Gymer's Salvation (Argent Crusade)
quests[129151119] = 22000 -- Mending Fences (The Sons of Hodir)
quests[129161106] = 75 -- Our Only Hope (Argent Crusade)
quests[129171085] = 250 -- Speaking with the Wind's Voice (Warsong Offensive)
quests[129191106] = 500 -- The Storm King's Vengeance (Argent Crusade)
quests[129241119] = 250 -- Forging an Alliance (The Sons of Hodir)
quests[129381098] = 10 -- The Duke (Knights of the Ebon Blade)
quests[129391098] = 250 -- Honor Challenge (Knights of the Ebon Blade)
quests[129431098] = 250 -- Shadow Vault Decree (Knights of the Ebon Blade)
quests[129491098] = 250 -- Get the Key (Knights of the Ebon Blade)
quests[129511098] = 25 -- Let the Baron Know (Knights of the Ebon Blade)
quests[129551098] = 250 -- Eliminate the Competition (Knights of the Ebon Blade)
quests[129561119] = 22000 -- A Spark of Hope (The Sons of Hodir)
quests[129581090] = 25 -- Shipment: Blood Jade Amulet  (Kirin Tor)
quests[129591090] = 25 -- Shipment: Glowing Ivory Figurine (Kirin Tor)
quests[129601090] = 25 -- Shipment: Wicked Sun Brooch (Kirin Tor)
quests[129611090] = 25 -- Shipment: Intricate Bone Figurine (Kirin Tor)
quests[129621090] = 25 -- Shipment: Bright Armor Relic (Kirin Tor)
quests[129631090] = 25 -- Shipment: Shifting Sun Curio  (Kirin Tor)
quests[129661119] = 75 -- You Can't Miss Him (The Sons of Hodir)
quests[129671119] = 75 -- Battling the Elements (The Sons of Hodir)
quests[129731068] = 500 -- The Brothers Bronzebeard (Explorers' League)
quests[129731126] = 500 -- The Brothers Bronzebeard (The Frostborn)
quests[129751119] = 250 -- In Memoriam (The Sons of Hodir)
quests[129761119] = 75 -- A Monument to the Fallen (The Sons of Hodir)
quests[129771119] = 350 -- Blowing Hodir's Horn (The Sons of Hodir)
quests[129811119] = 350 -- Hot and Cold (The Sons of Hodir)
quests[129821098] = 250 -- Ebon Blade Prisoners (Knights of the Ebon Blade)
quests[129851119] = 250 -- Forging a Head (The Sons of Hodir)
quests[129871119] = 150 -- Mounting Hodir's Helm (The Sons of Hodir)
quests[129921098] = 250 -- Crush Dem Vrykuls! (Knights of the Ebon Blade)
quests[129941119] = 350 -- Spy Hunter (The Sons of Hodir)
quests[129951098] = 250 -- Leave Our Mark (Knights of the Ebon Blade)
quests[129991098] = 10 -- The Bone Witch (Knights of the Ebon Blade)
quests[130001085] = 250 -- Emergency Measures (Warsong Offensive)
quests[130011119] = 250 -- Raising Hodir's Spear (The Sons of Hodir)
quests[130031119] = 500 -- Thrusting Hodir's Spear (The Sons of Hodir)
quests[130061119] = 350 -- Polishing the Helm (The Sons of Hodir)
quests[130081106] = 250 -- Scourge Tactics (Argent Crusade)
quests[130101119] = 350 -- Krolmir, Hammer of Storms (The Sons of Hodir)
quests[130111119] = 250 -- Jormuttar is Soo Fat... (The Sons of Hodir)
quests[130120067] = 75 -- Sardis the Elder (Horde)
quests[130120469] = 75 -- Sardis the Elder (Alliance)
quests[130130067] = 75 -- Beldak the Elder (Horde)
quests[130130469] = 75 -- Beldak the Elder (Alliance)
quests[130140067] = 75 -- Morthie the Elder (Horde)
quests[130140469] = 75 -- Morthie the Elder (Alliance)
quests[130150067] = 75 -- Fargal the Elder (Horde)
quests[130150469] = 75 -- Fargal the Elder (Alliance)
quests[130160067] = 75 -- Northal the Elder (Horde)
quests[130160469] = 75 -- Northal the Elder (Alliance)
quests[130170067] = 75 -- Jarten the Elder (Horde)
quests[130170469] = 75 -- Jarten the Elder (Alliance)
quests[130180067] = 75 -- Sandrene the Elder (Horde)
quests[130180469] = 75 -- Sandrene the Elder (Alliance)
quests[130190067] = 75 -- Thoim the Elder (Horde)
quests[130190469] = 75 -- Thoim the Elder (Alliance)
quests[130200067] = 75 -- Stonebeard the Elder (Horde)
quests[130200469] = 75 -- Stonebeard the Elder (Alliance)
quests[130210067] = 75 -- Igasho the Elder (Horde)
quests[130210469] = 75 -- Igasho the Elder (Alliance)
quests[130220067] = 75 -- Nurgen the Elder (Horde)
quests[130220469] = 75 -- Nurgen the Elder (Alliance)
quests[130230067] = 75 -- Kilias the Elder (Horde)
quests[130230469] = 75 -- Kilias the Elder (Alliance)
quests[130240067] = 75 -- Wanikaya the Elder (Horde)
quests[130240469] = 75 -- Wanikaya the Elder (Alliance)
quests[130250067] = 75 -- Lunaro the Elder (Horde)
quests[130250469] = 75 -- Lunaro the Elder (Alliance)
quests[130260067] = 75 -- Bluewolf the Elder (Horde)
quests[130260469] = 75 -- Bluewolf the Elder (Alliance)
quests[130270067] = 75 -- Tauros the Elder (Horde)
quests[130270469] = 75 -- Tauros the Elder (Alliance)
quests[130280067] = 75 -- Graymane the Elder (Horde)
quests[130280469] = 75 -- Graymane the Elder (Alliance)
quests[130290067] = 75 -- Pamuya the Elder (Horde)
quests[130290469] = 75 -- Pamuya the Elder (Alliance)
quests[130300067] = 75 -- Whurain the Elder (Horde)
quests[130300469] = 75 -- Whurain the Elder (Alliance)
quests[130310067] = 75 -- Skywarden the Elder (Horde)
quests[130310469] = 75 -- Skywarden the Elder (Alliance)
quests[130320067] = 75 -- Muraco the Elder (Horde)
quests[130320469] = 75 -- Muraco the Elder (Alliance)
quests[130330067] = 75 -- Arp the Elder (Horde)
quests[130330469] = 75 -- Arp the Elder (Alliance)
quests[130341085] = 250 -- The Witness and the Hero (Warsong Offensive)
quests[130361037] = 10 -- Honor Above All Else (Alliance Vanguard)
quests[130371085] = 250 -- Memories of Stormhoof (Warsong Offensive)
quests[130381085] = 250 -- Distortions in Time (Warsong Offensive)
quests[130391106] = 250 -- Defending The Vanguard (Argent Crusade)
quests[130401106] = 250 -- Curing The Incurable (Argent Crusade)
quests[130421098] = 250 -- Deep in the Bowels of The Underhalls (Knights of the Ebon Blade)
quests[130431098] = 350 -- The Sum is Greater than the Parts (Knights of the Ebon Blade)
quests[130451106] = 350 -- Into The Wild Green Yonder (Argent Crusade)
quests[130461119] = 350 -- Feeding Arngrim (The Sons of Hodir)
quests[130481085] = 250 -- Where Time Went Wrong (Warsong Offensive)
quests[130541085] = 75 -- The Missing Tracker (Warsong Offensive)
quests[130561085] = 250 -- There's Always Time for Revenge (Warsong Offensive)
quests[130581085] = 350 -- Changing the Wind's Course (Warsong Offensive)
quests[130591098] = 350 -- Revenge for the Vargul (Knights of the Ebon Blade)
quests[130650067] = 75 -- Ohanzee the Elder (Horde)
quests[130650469] = 75 -- Ohanzee the Elder (Alliance)
quests[130660067] = 75 -- Yurauk the Elder (Horde)
quests[130660469] = 75 -- Yurauk the Elder (Alliance)
quests[130670067] = 75 -- Chogan'gada the Elder (Horde)
quests[130670469] = 75 -- Chogan'gada the Elder (Alliance)
quests[130681106] = 10 -- A Tale of Valor (Argent Crusade)
quests[130691098] = 250 -- Shoot 'Em Up (Knights of the Ebon Blade)
quests[130701106] = 10 -- A Cold Front Approaches (Argent Crusade)
quests[130711098] = 250 -- Vile Like Fire! (Knights of the Ebon Blade)
quests[130721106] = 250 -- A Hero Remains (Argent Crusade)
quests[130731106] = 150 -- The Keeper's Favor (Argent Crusade)
quests[130751106] = 250 -- The Boon of Remulos (Argent Crusade)
quests[130761106] = 10 -- Time Yet Remains (Argent Crusade)
quests[130771106] = 150 -- The Touch of an Aspect (Argent Crusade)
quests[130791106] = 250 -- The Boon of Alexstrasza (Argent Crusade)
quests[130801106] = 10 -- Hope Yet Remains (Argent Crusade)
quests[130811106] = 150 -- The Will of the Naaru (Argent Crusade)
quests[130821106] = 250 -- The Boon of A'dal (Argent Crusade)
quests[130831106] = 500 -- Light Within the Darkness (Argent Crusade)
quests[130841098] = 250 -- Vandalizing Jotunheim (Knights of the Ebon Blade)
quests[130851098] = 10 -- Vaelen Has Returned (Knights of the Ebon Blade)
quests[130861106] = 500 -- The Last Line Of Defense (Argent Crusade)
quests[130871050] = 75 -- Northern Cooking (Valiance Expedition)
quests[130881050] = 75 -- Northern Cooking (Valiance Expedition)
quests[130891067] = 75 -- Northern Cooking (The Hand of Vengeance)
quests[130901085] = 75 -- Northern Cooking (Warsong Offensive)
quests[130911098] = 250 -- The Art of Being a Water Terror (Knights of the Ebon Blade)
quests[130921098] = 250 -- Reading the Bones (Knights of the Ebon Blade)
quests[130931098] = 10 -- Reading the Bones (Knights of the Ebon Blade)
quests[130941090] = 350 -- Have They No Shame? (Kirin Tor)
quests[130951090] = 350 -- Have They No Shame? (Kirin Tor)
quests[131001090] = 150 -- Infused Mushroom Meatloaf (Kirin Tor)
quests[131011090] = 150 -- Convention at the Legerdemain (Kirin Tor)
quests[131021090] = 150 -- Sewer Stew (Kirin Tor)
quests[131031090] = 150 -- Cheese for Glowergold (Kirin Tor)
quests[131041106] = 10 -- Once More Unto The Breach, Hero (Argent Crusade)
quests[131041098] = 10 -- Once More Unto The Breach, Hero (Knights of the Ebon Blade)
quests[131051106] = 10 -- Once More Unto The Breach, Hero (Argent Crusade)
quests[131051098] = 10 -- Once More Unto The Breach, Hero (Knights of the Ebon Blade)
quests[131061098] = 10 -- Blackwatch (Knights of the Ebon Blade)
quests[131071090] = 150 -- Mustard Dogs! (Kirin Tor)
quests[131101106] = 250 -- The Restless Dead (Argent Crusade)
quests[131121090] = 150 -- Infused Mushroom Meatloaf (Kirin Tor)
quests[131131090] = 150 -- Convention at the Legerdemain (Kirin Tor)
quests[131141090] = 150 -- Sewer Stew (Kirin Tor)
quests[131151090] = 150 -- Cheese for Glowergold (Kirin Tor)
quests[131161090] = 150 -- Mustard Dogs! (Kirin Tor)
quests[131171098] = 150 -- Where Are They Coming From? (Knights of the Ebon Blade)
quests[131181106] = 250 -- The Purging Of Scourgeholme (Argent Crusade)
quests[131181098] = 250 -- The Purging Of Scourgeholme (Knights of the Ebon Blade)
quests[131191098] = 250 -- Destroying the Altars (Knights of the Ebon Blade)
quests[131201098] = 250 -- Death's Gaze (Knights of the Ebon Blade)
quests[131211098] = 250 -- Through the Eye (Knights of the Ebon Blade)
quests[131221106] = 250 -- The Scourgestone (Argent Crusade)
quests[131221098] = 250 -- The Scourgestone (Knights of the Ebon Blade)
quests[131241091] = 350 -- The Struggle Persists (The Wyrmrest Accord)
quests[131251106] = 500 -- The Air Stands Still (Argent Crusade)
quests[131251098] = 500 -- The Air Stands Still (Knights of the Ebon Blade)
quests[131261091] = 350 -- A Unified Front (The Wyrmrest Accord)
quests[131271091] = 350 -- Mage-Lord Urom (The Wyrmrest Accord)
quests[131281091] = 350 -- A Wing and a Prayer (The Wyrmrest Accord)
quests[131301106] = 250 -- The Stone That Started A Revolution (Argent Crusade)
quests[131331098] = 250 -- Find the Ancient Hero (Knights of the Ebon Blade)
quests[131341098] = 250 -- Spill Their Blood (Knights of the Ebon Blade)
quests[131351106] = 250 -- It Could Kill Us All (Argent Crusade)
quests[131361106] = 250 -- Jagged Shards (Argent Crusade)
quests[131371098] = 350 -- Not-So-Honorable Combat (Knights of the Ebon Blade)
quests[131381106] = 250 -- I'm Smelting... Smelting! (Argent Crusade)
quests[131391106] = 250 -- Into The Frozen Heart Of Northrend (Argent Crusade)
quests[131401106] = 250 -- The Runesmiths of Malykriss (Argent Crusade)
quests[131411106] = 250 -- The Battle For Crusaders' Pinnacle (Argent Crusade)
quests[131421098] = 500 -- Banshee's Revenge (Knights of the Ebon Blade)
quests[131431098] = 250 -- New Recruit (Knights of the Ebon Blade)
quests[131441098] = 250 -- Killing Two Scourge With One Skeleton (Knights of the Ebon Blade)
quests[131451098] = 250 -- The Vile Hold (Knights of the Ebon Blade)
quests[131461098] = 250 -- Generosity Abounds (Knights of the Ebon Blade)
quests[131471098] = 250 -- Matchmaker (Knights of the Ebon Blade)
quests[131481090] = 25 -- Necklace Repair (Kirin Tor)
quests[131521098] = 250 -- A Visit to the Doctor (Knights of the Ebon Blade)
quests[131551098] = 75 -- Vereth the Cunning (Knights of the Ebon Blade)
quests[131571106] = 500 -- The Crusaders' Pinnacle (Argent Crusade)
quests[131601098] = 250 -- Stunning View (Knights of the Ebon Blade)
quests[131611098] = 350 -- The Rider of the Unholy (Knights of the Ebon Blade)
quests[131621098] = 350 -- The Rider of Frost (Knights of the Ebon Blade)
quests[131631098] = 350 -- The Rider of Blood (Knights of the Ebon Blade)
quests[131641098] = 500 -- The Fate of Bloodbane (Knights of the Ebon Blade)
quests[131651098] = 10 -- Taking Back Acherus (Knights of the Ebon Blade)
quests[131661098] = 500 -- The Battle For The Ebon Hold (Knights of the Ebon Blade)
quests[131681098] = 10 -- Parting Gifts (Knights of the Ebon Blade)
quests[131691098] = 250 -- An Undead's Best Friend (Knights of the Ebon Blade)
quests[131701098] = 250 -- Honor is for the Weak (Knights of the Ebon Blade)
quests[131711098] = 250 -- From Whence They Came (Knights of the Ebon Blade)
quests[131721098] = 250 -- Seeds of Chaos (Knights of the Ebon Blade)
quests[131741098] = 250 -- Amidst the Confusion (Knights of the Ebon Blade)
quests[131880072] = 6000 -- Where Kings Walk (Stormwind)
quests[131890076] = 6000 -- Warchief's Blessing (Orgrimmar)
quests[132111106] = 250 -- By Fire Be Purged (Argent Crusade)
quests[132121098] = 250 -- He's Gone to Pieces (Knights of the Ebon Blade)
quests[132201098] = 250 -- Putting Olakin Back Together Again (Knights of the Ebon Blade)
quests[132211106] = 250 -- I'm Not Dead Yet! (Argent Crusade)
quests[132241085] = 25 -- Orgrim's Hammer (Warsong Offensive)
quests[132251050] = 25 -- The Skybreaker (Valiance Expedition)
quests[132261106] = 25 -- Judgment Day Comes! (Argent Crusade)
quests[132271106] = 25 -- Judgment Day Comes! (Argent Crusade)
quests[132351098] = 350 -- The Flesh Giant Champion (Knights of the Ebon Blade)
quests[132581085] = 75 -- Opportunity (Warsong Offensive)
quests[132801050] = 250 -- King of the Mountain (Valiance Expedition)
quests[132831085] = 250 -- King of the Mountain (Warsong Offensive)
quests[132841050] = 250 -- Assault by Ground (Valiance Expedition)
quests[132851085] = 350 -- Forging the Keystone (Warsong Offensive)
quests[132931085] = 10 -- Get to Ymirheim! (Warsong Offensive)
quests[132961050] = 10 -- Get to Ymirheim! (Valiance Expedition)
quests[133001106] = 250 -- Slaves to Saronite (Argent Crusade)
quests[133011085] = 250 -- Assault by Ground (Warsong Offensive)
quests[133021106] = 250 -- Slaves to Saronite (Argent Crusade)
quests[133061098] = 250 -- Raise the Barricades (Knights of the Ebon Blade)
quests[133071098] = 250 -- Bloodspattered Banners (Knights of the Ebon Blade)
quests[133091050] = 250 -- Assault by Air (Valiance Expedition)
quests[133101085] = 250 -- Assault by Air (Warsong Offensive)
quests[133110909] = 25 -- Demons Deck (Darkmoon Faire)
quests[133121098] = 350 -- The Ironwall Rampart (Knights of the Ebon Blade)
quests[133131085] = 250 -- Blinding the Eyes in the Sky (Warsong Offensive)
quests[133141050] = 250 -- Get the Message (Valiance Expedition)
quests[133161098] = 350 -- The Guardians of Corp'rethar (Knights of the Ebon Blade)
quests[133240909] = 350 -- Darkmoon Prisms Deck (Darkmoon Faire)
quests[133250909] = 350 -- Darkmoon Chaos Deck (Darkmoon Faire)
quests[133260909] = 350 -- Darkmoon Nobles Deck (Darkmoon Faire)
quests[133270909] = 350 -- Darkmoon Undeath Deck (Darkmoon Faire)
quests[133281098] = 350 -- Shatter the Shards (Knights of the Ebon Blade)
quests[133291098] = 250 -- Before the Gate of Horror (Knights of the Ebon Blade)
quests[133301085] = 250 -- Blood of the Chosen (Warsong Offensive)
quests[133311085] = 250 -- Keeping the Alliance Blind (Warsong Offensive)
quests[133321098] = 250 -- Raise the Barricades (Knights of the Ebon Blade)
quests[133331050] = 250 -- Capture More Dispatches (Valiance Expedition)
quests[133341098] = 250 -- Bloodspattered Banners (Knights of the Ebon Blade)
quests[133351098] = 250 -- Before the Gate of Horror (Knights of the Ebon Blade)
quests[133361050] = 250 -- Blood of the Chosen (Valiance Expedition)
quests[133371098] = 350 -- The Ironwall Rampart (Knights of the Ebon Blade)
quests[133381098] = 350 -- The Guardians of Corp'rethar (Knights of the Ebon Blade)
quests[133391098] = 350 -- Shatter the Shards (Knights of the Ebon Blade)
quests[133401085] = 25 -- Joining the Assault (Warsong Offensive)
quests[133411050] = 25 -- Joining the Assault (Valiance Expedition)
quests[133431091] = 350 -- Mystery of the Infinite, Redux (The Wyrmrest Accord)
quests[133481085] = 75 -- Futility (Warsong Offensive)
quests[133491085] = 250 -- Cradle of the Frostbrood (Warsong Offensive)
quests[133621050] = 75 -- Knowledge is a Terrible Burden (Valiance Expedition)
quests[133641098] = 500 -- Tirion's Gambit (Knights of the Ebon Blade)
quests[133641085] = 500 -- Tirion's Gambit (Warsong Offensive)
quests[133641106] = 500 -- Tirion's Gambit (Argent Crusade)
quests[133641050] = 500 -- Tirion's Gambit (Valiance Expedition)
quests[133721091] = 500 -- The Key to the Focusing Iris (The Wyrmrest Accord)
quests[133751091] = 500 -- The Heroic Key to the Focusing Iris (The Wyrmrest Accord)
quests[133841091] = 500 -- Judgment at the Eye of Eternity (The Wyrmrest Accord)
quests[133851091] = 1000 -- Heroic Judgment at the Eye of Eternity (The Wyrmrest Accord)
quests[133861050] = 75 -- Exploiting an Opening (Valiance Expedition)
quests[133871050] = 250 -- Securing the Perimeter (Valiance Expedition)
quests[133921050] = 75 -- Return to the Surface (Valiance Expedition)
quests[133961050] = 75 -- Futility (Valiance Expedition)
quests[133971050] = 250 -- Sindragosa's Fall (Valiance Expedition)
quests[134031098] = 500 -- Tirion's Gambit (Knights of the Ebon Blade)
quests[134031085] = 500 -- Tirion's Gambit (Warsong Offensive)
quests[134031106] = 500 -- Tirion's Gambit (Argent Crusade)
quests[134031050] = 500 -- Tirion's Gambit (Valiance Expedition)
quests[134080946] = 150 -- Hellfire Fortifications (Honor Hold)
quests[134090947] = 150 -- Hellfire Fortifications (Thrallmar)
quests[134100946] = 150 -- Hellfire Fortifications (Honor Hold)
quests[134110947] = 150 -- Hellfire Fortifications (Thrallmar)
quests[134121091] = 25 -- Corastrasza (The Wyrmrest Accord)
quests[134131091] = 350 -- Aces High! (The Wyrmrest Accord)
quests[134141091] = 250 -- Aces High! (The Wyrmrest Accord)
quests[134171068] = 500 -- The Brothers Bronzebeard (Explorers' League)
quests[134171126] = 500 -- The Brothers Bronzebeard (The Frostborn)
quests[134181050] = 25 -- Preparations for War (Valiance Expedition)
quests[134191085] = 25 -- Preparations for War (Warsong Offensive)
quests[134201119] = 350 -- Everfrost (The Sons of Hodir)
quests[134211119] = 350 -- Remember Everfrost! (The Sons of Hodir)
quests[134261085] = 10 -- Xarantaur, the Witness (Warsong Offensive)
quests[134290935] = 250 -- A Distraction for Akama (The Sha'tar)
quests[134300935] = 500 -- Trial of the Naaru: Magtheridon (The Sha'tar)
quests[134320990] = 3000 -- The Vials of Eternity (The Scale of the Sands)
quests[134821106] = 250 -- Let's Get Out of Here (Argent Crusade)
quests[135040069] = 250 -- Shatterspear Laborers (Darnassus)
quests[135050069] = 250 -- Remnants of the Highborne (Darnassus)
quests[135060069] = 250 -- Reason to Worry (Darnassus)
quests[135070069] = 250 -- Denying Manpower (Darnassus)
quests[135080069] = 250 -- Swift Response (Darnassus)
quests[135090069] = 250 -- War Supplies (Darnassus)
quests[135100069] = 250 -- Timely Arrival (Darnassus)
quests[135110069] = 250 -- One Bitter Wish (Darnassus)
quests[135120069] = 250 -- Strategic Strikes (Darnassus)
quests[135130069] = 250 -- On the Brink (Darnassus)
quests[135140069] = 250 -- The Ancients' Ire (Darnassus)
quests[135150069] = 350 -- Ending the Threat (Darnassus)
quests[135180069] = 250 -- The Last Wave of Survivors (Darnassus)
quests[135190069] = 250 -- The Twilight's Hammer (Darnassus)
quests[135200069] = 250 -- The Boon of the Seas (Darnassus)
quests[135210069] = 250 -- Buzzbox 413 (Darnassus)
quests[135220069] = 250 -- Threat from the Water (Darnassus)
quests[135270069] = 250 -- No Accounting for Taste (Darnassus)
quests[135280069] = 250 -- Buzzbox 723 (Darnassus)
quests[135290069] = 250 -- The Corruption's Source (Darnassus)
quests[135371134] = 250 -- A Taste for Grouper (Gilneas)
quests[135420069] = 250 -- Against the Wind (Darnassus)
quests[135430069] = 250 -- Three Hammers to Break (Darnassus)
quests[135470069] = 250 -- Coaxing the Spirits (Darnassus)
quests[135540069] = 250 -- A Cure In The Dark (Darnassus)
quests[135570069] = 250 -- Bearer of Good Fortune (Darnassus)
quests[135580069] = 350 -- Call Down the Thunder (Darnassus)
quests[135591119] = 325 -- Hodir's Tribute (The Sons of Hodir)
quests[135600069] = 250 -- An Ocean Not So Deep (Darnassus)
quests[135610069] = 250 -- Solace for the Highborne (Darnassus)
quests[135620069] = 250 -- The Final Flame of Bashal'Aran (Darnassus)
quests[135630069] = 250 -- A Love Eternal (Darnassus)
quests[135640069] = 250 -- A Lost Companion (Darnassus)
quests[135650069] = 250 -- Twice Removed (Darnassus)
quests[135660069] = 250 -- Ritual Materials (Darnassus)
quests[135690069] = 250 -- The Ritual Bond (Darnassus)
quests[135700069] = 350 -- Remembrance of Auberdine (Darnassus)
quests[135730069] = 10 -- Malfurion's Return (Darnassus)
quests[135780069] = 250 -- Aroom's Farewell (Darnassus)
quests[135790069] = 10 -- Protector of Ameth'Aran (Darnassus)
quests[135810069] = 75 -- The Blackwood Pledge (Darnassus)
quests[135830069] = 75 -- The Wildkin's Oath (Darnassus)
quests[135870069] = 75 -- The Waking Nightmare (Darnassus)
quests[135880069] = 250 -- The Eye of All Storms (Darnassus)
quests[135890069] = 150 -- The Shatterspear Invaders (Darnassus)
quests[135900069] = 250 -- The Front Line (Darnassus)
quests[135910069] = 350 -- Disturbing Connections (Darnassus)
quests[135920072] = 250 -- A Valiant's Field Training (Stormwind)
quests[135921094] = 250 -- A Valiant's Field Training (The Silver Covenant)
quests[135930072] = 10 -- Valiant Of Stormwind (Stormwind)
quests[135940069] = 250 -- Don't Forget the Horde (Darnassus)
quests[135950069] = 250 -- Of Their Own Design (Darnassus)
quests[135960069] = 250 -- Twilight Plans (Darnassus)
quests[135980069] = 250 -- Unsavory Remedies (Darnassus)
quests[135990069] = 150 -- Grimclaw's Return (Darnassus)
quests[136000072] = 250 -- A Worthy Weapon (Stormwind)
quests[136001094] = 250 -- A Worthy Weapon (The Silver Covenant)
quests[136010069] = 25 -- In Aid of the Refugees (Darnassus)
quests[136020069] = 250 -- Naga of the Strand (Darnassus)
quests[136030072] = 250 -- A Blade Fit For A Champion (Stormwind)
quests[136031094] = 250 -- A Blade Fit For A Champion (The Silver Covenant)
quests[136050047] = 350 -- The Last Refugee (Ironforge)
quests[136050069] = 350 -- The Last Refugee (Darnassus)
quests[136120076] = 250 -- Mor'shan Defense (Orgrimmar)
quests[136130076] = 250 -- Rescue the Fallen (Orgrimmar)
quests[136150076] = 250 -- Empty Quivers (Orgrimmar)
quests[136160072] = 250 -- The Edge Of Winter (Stormwind)
quests[136161094] = 250 -- The Edge Of Winter (The Silver Covenant)
quests[136170069] = 10 -- West to the Strand (Darnassus)
quests[136180076] = 250 -- Find Gorat! (Orgrimmar)
quests[136190076] = 250 -- Final Report (Orgrimmar)
quests[136200076] = 10 -- To Dinah, at Once! (Orgrimmar)
quests[136210076] = 350 -- Gorat's Vengeance (Orgrimmar)
quests[136230069] = 25 -- Delivery for Orendil (Darnassus)
quests[136240069] = 250 -- A Squad of Your Own (Darnassus)
quests[136251094] = 250 -- Learning The Reins (The Silver Covenant)
quests[136260069] = 250 -- Respect for the Fallen (Darnassus)
quests[136280076] = 250 -- Got Wood? (Orgrimmar)
quests[136300069] = 250 -- The Reason Why (Darnassus)
quests[136320069] = 250 -- Rear Guard Duty (Darnassus)
quests[136331106] = 250 -- The Black Knight of Westfall? (Argent Crusade)
quests[136341106] = 250 -- The Black Knight of Silverpine? (Argent Crusade)
quests[136350047] = 150 -- South Gate Status Report (Ironforge)
quests[136360047] = 150 -- Stormpike's Orders (Ironforge)
quests[136380047] = 250 -- The Captured Mountaineer (Ironforge)
quests[136390047] = 150 -- Resupplying the Excavation (Ironforge)
quests[136400076] = 250 -- Management Material (Orgrimmar)
quests[136411106] = 250 -- The Seer's Crystal (Argent Crusade)
quests[136420069] = 250 -- Bathed in Light (Darnassus)
quests[136431106] = 250 -- The Stories Dead Men Tell (Argent Crusade)
quests[136440069] = 350 -- That Which Has Risen (Darnassus)
quests[136450069] = 25 -- To Raene Wolfrunner (Darnassus)
quests[136460069] = 10 -- Astranaar Bound (Darnassus)
quests[136470047] = 250 -- Joining the Hunt (Ironforge)
quests[136500047] = 250 -- Keep Your Hands Off The Goods! (Ironforge)
quests[136510076] = 250 -- Needs a Little Lubrication (Orgrimmar)
quests[136520047] = 250 -- Defense of the Stonewrought Dam (Ironforge)
quests[136530076] = 150 -- Crisis at Splintertree (Orgrimmar)
quests[136550047] = 250 -- Explorers' League Document (2 of 6) (Ironforge)
quests[136560047] = 250 -- Explorers' League Document (1 of 6) (Ironforge)
quests[136570047] = 250 -- Explorers' League Document (3 of 6) (Ironforge)
quests[136580047] = 250 -- Explorers' League Document (4 of 6) (Ironforge)
quests[136590047] = 250 -- Explorers' League Document (6 of 6) (Ironforge)
quests[136600047] = 250 -- Explorers' League Document (5 of 6) (Ironforge)
quests[136610047] = 350 -- Heartfelt Appreciation (Ironforge)
quests[136620059] = 15 -- Gaining Acceptance (Thorium Brotherhood)
quests[136631106] = 250 -- The Black Knight's Orders (Argent Crusade)
quests[136641106] = 350 -- The Black Knight's Fall (Argent Crusade)
quests[136650072] = 250 -- The Grand Melee (Stormwind)
quests[136651094] = 250 -- The Grand Melee (The Silver Covenant)
quests[136661094] = 250 -- A Blade Fit For A Champion (The Silver Covenant)
quests[136671094] = 10 -- The Argent Tournament (The Silver Covenant)
quests[136681124] = 10 -- The Argent Tournament (The Sunreavers)
quests[136691094] = 250 -- A Worthy Weapon (The Silver Covenant)
quests[136701094] = 250 -- The Edge Of Winter (The Silver Covenant)
quests[136711094] = 250 -- Training In The Field (The Silver Covenant)
quests[136721094] = 10 -- Up To The Challenge (The Silver Covenant)
quests[136731124] = 250 -- A Blade Fit For A Champion (The Sunreavers)
quests[136741124] = 250 -- A Worthy Weapon (The Sunreavers)
quests[136751124] = 250 -- The Edge Of Winter (The Sunreavers)
quests[136761124] = 250 -- Training In The Field (The Sunreavers)
quests[136771124] = 250 -- Learning The Reins (The Sunreavers)
quests[136781124] = 10 -- Up To The Challenge (The Sunreavers)
quests[136791094] = 250 -- The Aspirant's Challenge (The Silver Covenant)
quests[136801124] = 250 -- The Aspirant's Challenge (The Sunreavers)
quests[136821106] = 250 -- Threat From Above (Argent Crusade)
quests[136821094] = 250 -- Threat From Above (The Silver Covenant)
quests[136830930] = 250 -- Stopping the Rituals (Exodar)
quests[136840072] = 10 -- A Valiant Of Stormwind (Stormwind)
quests[136841094] = 10 -- A Valiant Of Stormwind (The Silver Covenant)
quests[136850047] = 10 -- A Valiant Of Ironforge (Ironforge)
quests[136851094] = 10 -- A Valiant Of Ironforge (The Silver Covenant)
quests[136880054] = 10 -- A Valiant Of Gnomeregan (Gnomeregan)
quests[136881094] = 10 -- A Valiant Of Gnomeregan (The Silver Covenant)
quests[136890069] = 10 -- A Valiant Of Darnassus (Darnassus)
quests[136891094] = 10 -- A Valiant Of Darnassus (The Silver Covenant)
quests[136900930] = 10 -- A Valiant Of The Exodar (Exodar)
quests[136901094] = 10 -- A Valiant Of The Exodar (The Silver Covenant)
quests[136910076] = 10 -- A Valiant Of Orgrimmar (Orgrimmar)
quests[136911124] = 10 -- A Valiant Of Orgrimmar (The Sunreavers)
quests[136930530] = 10 -- A Valiant Of Sen'jin (Darkspear Trolls)
quests[136931124] = 10 -- A Valiant Of Sen'jin (The Sunreavers)
quests[136940081] = 10 -- A Valiant Of Thunder Bluff (Thunder Bluff)
quests[136941124] = 10 -- A Valiant Of Thunder Bluff (The Sunreavers)
quests[136950068] = 10 -- A Valiant Of Undercity (Undercity)
quests[136951124] = 10 -- A Valiant Of Undercity (The Sunreavers)
quests[136960911] = 10 -- A Valiant Of Silvermoon (Silvermoon City)
quests[136961124] = 10 -- A Valiant Of Silvermoon (The Sunreavers)
quests[136970076] = 250 -- The Valiant's Charge (Orgrimmar)
quests[136971124] = 250 -- The Valiant's Charge (The Sunreavers)
quests[136980069] = 350 -- Explosives Shredding (Darnassus)
quests[136990072] = 250 -- The Valiant's Challenge (Stormwind)
quests[136991094] = 250 -- The Valiant's Challenge (The Silver Covenant)
quests[137021106] = 350 -- A Champion Rises (Argent Crusade)
quests[137030047] = 10 -- Valiant Of Ironforge (Ironforge)
quests[137040054] = 10 -- Valiant Of Gnomeregan (Gnomeregan)
quests[137050930] = 10 -- Valiant Of The Exodar (Exodar)
quests[137060069] = 10 -- Valiant Of Darnassus (Darnassus)
quests[137070076] = 10 -- Valiant Of Orgrimmar (Orgrimmar)
quests[137080530] = 10 -- Valiant Of Sen'jin (Darkspear Trolls)
quests[137090081] = 10 -- Valiant Of Thunder Bluff (Thunder Bluff)
quests[137100068] = 10 -- Valiant Of Undercity (Undercity)
quests[137110911] = 10 -- Valiant Of Silvermoon (Silvermoon City)
quests[137120076] = 150 -- To the Rescue! (Orgrimmar)
quests[137130047] = 250 -- The Valiant's Challenge (Ironforge)
quests[137131094] = 250 -- The Valiant's Challenge (The Silver Covenant)
quests[137140047] = 250 -- The Valiant's Charge (Ironforge)
quests[137141094] = 250 -- The Valiant's Charge (The Silver Covenant)
quests[137150054] = 250 -- The Valiant's Charge (Gnomeregan)
quests[137151094] = 250 -- The Valiant's Charge (The Silver Covenant)
quests[137160930] = 250 -- The Valiant's Charge (Exodar)
quests[137161094] = 250 -- The Valiant's Charge (The Silver Covenant)
quests[137170069] = 250 -- The Valiant's Charge (Darnassus)
quests[137171094] = 250 -- The Valiant's Charge (The Silver Covenant)
quests[137180072] = 250 -- The Valiant's Charge (Stormwind)
quests[137181094] = 250 -- The Valiant's Charge (The Silver Covenant)
quests[137190530] = 250 -- The Valiant's Charge (Darkspear Trolls)
quests[137191124] = 250 -- The Valiant's Charge (The Sunreavers)
quests[137200081] = 250 -- The Valiant's Charge (Thunder Bluff)
quests[137201124] = 250 -- The Valiant's Charge (The Sunreavers)
quests[137210068] = 250 -- The Valiant's Charge (Undercity)
quests[137211124] = 250 -- The Valiant's Charge (The Sunreavers)
quests[137220072] = 250 -- The Valiant's Charge (Stormwind)
quests[137221124] = 250 -- The Valiant's Charge (The Sunreavers)
quests[137230054] = 250 -- The Valiant's Challenge (Gnomeregan)
quests[137231094] = 250 -- The Valiant's Challenge (The Silver Covenant)
quests[137240930] = 250 -- The Valiant's Challenge (Exodar)
quests[137241094] = 250 -- The Valiant's Challenge (The Silver Covenant)
quests[137250069] = 250 -- The Valiant's Challenge (Darnassus)
quests[137251094] = 250 -- The Valiant's Challenge (The Silver Covenant)
quests[137260076] = 250 -- The Valiant's Challenge (Orgrimmar)
quests[137261124] = 250 -- The Valiant's Challenge (The Sunreavers)
quests[137270530] = 250 -- The Valiant's Challenge (Darkspear Trolls)
quests[137271124] = 250 -- The Valiant's Challenge (The Sunreavers)
quests[137280081] = 250 -- The Valiant's Challenge (Thunder Bluff)
quests[137281124] = 250 -- The Valiant's Challenge (The Sunreavers)
quests[137290068] = 250 -- The Valiant's Challenge (Undercity)
quests[137291124] = 250 -- The Valiant's Challenge (The Sunreavers)
quests[137300076] = 250 -- Playing With Felfire (Orgrimmar)
quests[137310911] = 250 -- The Valiant's Challenge (Silvermoon City)
quests[137311124] = 250 -- The Valiant's Challenge (The Sunreavers)
quests[137321106] = 350 -- A Champion Rises (Argent Crusade)
quests[137331106] = 350 -- A Champion Rises (Argent Crusade)
quests[137341106] = 350 -- A Champion Rises (Argent Crusade)
quests[137351106] = 350 -- A Champion Rises (Argent Crusade)
quests[137361106] = 350 -- A Champion Rises (Argent Crusade)
quests[137371106] = 350 -- A Champion Rises (Argent Crusade)
quests[137381106] = 350 -- A Champion Rises (Argent Crusade)
quests[137391106] = 350 -- A Champion Rises (Argent Crusade)
quests[137401106] = 350 -- A Champion Rises (Argent Crusade)
quests[137410047] = 250 -- A Blade Fit For A Champion (Ironforge)
quests[137411094] = 250 -- A Blade Fit For A Champion (The Silver Covenant)
quests[137420047] = 250 -- A Worthy Weapon (Ironforge)
quests[137421094] = 250 -- A Worthy Weapon (The Silver Covenant)
quests[137430047] = 250 -- The Edge Of Winter (Ironforge)
quests[137431094] = 250 -- The Edge Of Winter (The Silver Covenant)
quests[137440047] = 250 -- A Valiant's Field Training (Ironforge)
quests[137441094] = 250 -- A Valiant's Field Training (The Silver Covenant)
quests[137450047] = 250 -- The Grand Melee (Ironforge)
quests[137451094] = 250 -- The Grand Melee (The Silver Covenant)
quests[137460054] = 250 -- A Blade Fit For A Champion (Gnomeregan)
quests[137461094] = 250 -- A Blade Fit For A Champion (The Silver Covenant)
quests[137470054] = 250 -- A Worthy Weapon (Gnomeregan)
quests[137471094] = 250 -- A Worthy Weapon (The Silver Covenant)
quests[137480054] = 250 -- The Edge Of Winter (Gnomeregan)
quests[137481094] = 250 -- The Edge Of Winter (The Silver Covenant)
quests[137490054] = 250 -- A Valiant's Field Training (Gnomeregan)
quests[137491094] = 250 -- A Valiant's Field Training (The Silver Covenant)
quests[137500054] = 250 -- The Grand Melee (Gnomeregan)
quests[137501094] = 250 -- The Grand Melee (The Silver Covenant)
quests[137510076] = 25 -- Tell No One! (Orgrimmar)
quests[137520930] = 250 -- A Blade Fit For A Champion (Exodar)
quests[137521094] = 250 -- A Blade Fit For A Champion (The Silver Covenant)
quests[137530930] = 250 -- A Worthy Weapon (Exodar)
quests[137531094] = 250 -- A Worthy Weapon (The Silver Covenant)
quests[137540930] = 250 -- The Edge Of Winter (Exodar)
quests[137541094] = 250 -- The Edge Of Winter (The Silver Covenant)
quests[137550930] = 250 -- A Valiant's Field Training (Exodar)
quests[137551094] = 250 -- A Valiant's Field Training (The Silver Covenant)
quests[137560930] = 250 -- The Grand Melee (Exodar)
quests[137561094] = 250 -- The Grand Melee (The Silver Covenant)
quests[137570069] = 250 -- A Blade Fit For A Champion (Darnassus)
quests[137571094] = 250 -- A Blade Fit For A Champion (The Silver Covenant)
quests[137580069] = 250 -- A Worthy Weapon (Darnassus)
quests[137581094] = 250 -- A Worthy Weapon (The Silver Covenant)
quests[137590069] = 250 -- The Edge Of Winter (Darnassus)
quests[137591094] = 250 -- The Edge Of Winter (The Silver Covenant)
quests[137600069] = 250 -- A Valiant's Field Training (Darnassus)
quests[137601094] = 250 -- A Valiant's Field Training (The Silver Covenant)
quests[137610069] = 250 -- The Grand Melee (Darnassus)
quests[137611094] = 250 -- The Grand Melee (The Silver Covenant)
quests[137620076] = 250 -- A Blade Fit For A Champion (Orgrimmar)
quests[137621124] = 250 -- A Blade Fit For A Champion (The Sunreavers)
quests[137630076] = 250 -- A Worthy Weapon (Orgrimmar)
quests[137631124] = 250 -- A Worthy Weapon (The Sunreavers)
quests[137640076] = 250 -- The Edge Of Winter (Orgrimmar)
quests[137641124] = 250 -- The Edge Of Winter (The Sunreavers)
quests[137650076] = 250 -- A Valiant's Field Training (Orgrimmar)
quests[137651124] = 250 -- A Valiant's Field Training (The Sunreavers)
quests[137660069] = 250 -- Closure is Only Natural (Darnassus)
quests[137670076] = 250 -- The Grand Melee (Orgrimmar)
quests[137671124] = 250 -- The Grand Melee (The Sunreavers)
quests[137680530] = 250 -- A Blade Fit For A Champion (Darkspear Trolls)
quests[137681124] = 250 -- A Blade Fit For A Champion (The Sunreavers)
quests[137690530] = 250 -- A Worthy Weapon (Darkspear Trolls)
quests[137691124] = 250 -- A Worthy Weapon (The Sunreavers)
quests[137700530] = 250 -- The Edge Of Winter (Darkspear Trolls)
quests[137701124] = 250 -- The Edge Of Winter (The Sunreavers)
quests[137710530] = 250 -- A Valiant's Field Training (Darkspear Trolls)
quests[137711124] = 250 -- A Valiant's Field Training (The Sunreavers)
quests[137720530] = 250 -- The Grand Melee (Darkspear Trolls)
quests[137721124] = 250 -- The Grand Melee (The Sunreavers)
quests[137730081] = 250 -- A Blade Fit For A Champion (Thunder Bluff)
quests[137731124] = 250 -- A Blade Fit For A Champion (The Sunreavers)
quests[137740081] = 250 -- A Worthy Weapon (Thunder Bluff)
quests[137741124] = 250 -- A Worthy Weapon (The Sunreavers)
quests[137750081] = 250 -- The Edge Of Winter (Thunder Bluff)
quests[137751124] = 250 -- The Edge Of Winter (The Sunreavers)
quests[137760081] = 250 -- A Valiant's Field Training (Thunder Bluff)
quests[137761124] = 250 -- A Valiant's Field Training (The Sunreavers)
quests[137770081] = 250 -- The Grand Melee (Thunder Bluff)
quests[137771124] = 250 -- The Grand Melee (The Sunreavers)
quests[137780068] = 250 -- A Blade Fit For A Champion (Undercity)
quests[137781124] = 250 -- A Blade Fit For A Champion (The Sunreavers)
quests[137790068] = 250 -- A Worthy Weapon (Undercity)
quests[137791124] = 250 -- A Worthy Weapon (The Sunreavers)
quests[137800068] = 250 -- The Edge Of Winter (Undercity)
quests[137801124] = 250 -- The Edge Of Winter (The Sunreavers)
quests[137810068] = 250 -- A Valiant's Field Training (Undercity)
quests[137811124] = 250 -- A Valiant's Field Training (The Sunreavers)
quests[137820068] = 250 -- The Grand Melee (Undercity)
quests[137821124] = 250 -- The Grand Melee (The Sunreavers)
quests[137830911] = 250 -- A Blade Fit For A Champion (Silvermoon City)
quests[137831124] = 250 -- A Blade Fit For A Champion (The Sunreavers)
quests[137840911] = 250 -- A Worthy Weapon (Silvermoon City)
quests[137841124] = 250 -- A Worthy Weapon (The Sunreavers)
quests[137850911] = 250 -- The Edge Of Winter (Silvermoon City)
quests[137851124] = 250 -- The Edge Of Winter (The Sunreavers)
quests[137860911] = 250 -- A Valiant's Field Training (Silvermoon City)
quests[137861124] = 250 -- A Valiant's Field Training (The Sunreavers)
quests[137870911] = 250 -- The Grand Melee (Silvermoon City)
quests[137871124] = 250 -- The Grand Melee (The Sunreavers)
quests[137881098] = 250 -- Threat From Above (Knights of the Ebon Blade)
quests[137881094] = 250 -- Threat From Above (The Silver Covenant)
quests[137891106] = 250 -- Taking Battle To The Enemy (Argent Crusade)
quests[137891094] = 250 -- Taking Battle To The Enemy (The Silver Covenant)
quests[137901106] = 250 -- Among the Champions (Argent Crusade)
quests[137901094] = 250 -- Among the Champions (The Silver Covenant)
quests[137911098] = 250 -- Taking Battle To The Enemy (Knights of the Ebon Blade)
quests[137911094] = 250 -- Taking Battle To The Enemy (The Silver Covenant)
quests[137920069] = 250 -- The Shadewalker (Darnassus)
quests[137931098] = 250 -- Among the Champions (Knights of the Ebon Blade)
quests[137931094] = 250 -- Among the Champions (The Silver Covenant)
quests[137941106] = 10 -- Eadric the Pure (Argent Crusade)
quests[137951098] = 10 -- The Scourgebane (Knights of the Ebon Blade)
quests[137960069] = 350 -- The Forest Heart (Darnassus)
quests[137970076] = 150 -- Dirty Deeds (Orgrimmar)
quests[138010076] = 250 -- Dead Elves Walking (Orgrimmar)
quests[138030076] = 75 -- Blood of the Weak (Orgrimmar)
quests[138050076] = 250 -- Pierce Their Heart! (Orgrimmar)
quests[138060076] = 250 -- Demon Duty (Orgrimmar)
quests[138080076] = 75 -- Mission Improbable (Orgrimmar)
quests[138091106] = 250 -- Threat From Above (Argent Crusade)
quests[138091124] = 250 -- Threat From Above (The Sunreavers)
quests[138101106] = 250 -- Taking Battle To The Enemy (Argent Crusade)
quests[138101124] = 250 -- Taking Battle To The Enemy (The Sunreavers)
quests[138111106] = 250 -- Among the Champions (Argent Crusade)
quests[138111124] = 250 -- Among the Champions (The Sunreavers)
quests[138121098] = 250 -- Threat From Above (Knights of the Ebon Blade)
quests[138121124] = 250 -- Threat From Above (The Sunreavers)
quests[138131098] = 250 -- Taking Battle To The Enemy (Knights of the Ebon Blade)
quests[138131124] = 250 -- Taking Battle To The Enemy (The Sunreavers)
quests[138141098] = 250 -- Among the Champions (Knights of the Ebon Blade)
quests[138141124] = 250 -- Among the Champions (The Sunreavers)
quests[138150076] = 250 -- Making Stumps (Orgrimmar)
quests[138281094] = 75 -- Mastery Of Melee (The Silver Covenant)
quests[138291124] = 75 -- Mastery Of Melee (The Sunreavers)
quests[138301090] = 250 -- The Ghostfish (Kirin Tor)
quests[138310069] = 250 -- A Troubling Prescription (Darnassus)
quests[138321090] = 250 -- Jewel Of The Sewers (Kirin Tor)
quests[138331090] = 250 -- Blood Is Thicker (Kirin Tor)
quests[138341090] = 250 -- Dangerously Delicious (Kirin Tor)
quests[138351094] = 75 -- Mastery Of The Shield-Breaker (The Silver Covenant)
quests[138361090] = 250 -- Disarmed! (Kirin Tor)
quests[138371094] = 75 -- Mastery Of The Charge (The Silver Covenant)
quests[138381124] = 75 -- Mastery Of The Shield-Breaker (The Sunreavers)
quests[138391124] = 75 -- Mastery Of The Charge (The Sunreavers)
quests[138410076] = 25 -- All Apologies (Orgrimmar)
quests[138440069] = 250 -- The Looting of Althalaxx (Darnassus)
quests[138451090] = 250 -- Sealed Vial of Poison (Kirin Tor)
quests[138461106] = 100 -- Contributin' To The Cause (Argent Crusade)
quests[138470072] = 250 -- At The Enemy's Gates (Stormwind)
quests[138471094] = 250 -- At The Enemy's Gates (The Silver Covenant)
quests[138480076] = 150 -- Bad News Bear-er (Orgrimmar)
quests[138490069] = 250 -- Astranaar's Burning! (Darnassus)
quests[138510047] = 250 -- At The Enemy's Gates (Ironforge)
quests[138511094] = 250 -- At The Enemy's Gates (The Silver Covenant)
quests[138520054] = 250 -- At The Enemy's Gates (Gnomeregan)
quests[138521094] = 250 -- At The Enemy's Gates (The Silver Covenant)
quests[138530069] = 350 -- Return Fire (Darnassus)
quests[138540930] = 250 -- At The Enemy's Gates (Exodar)
quests[138541094] = 250 -- At The Enemy's Gates (The Silver Covenant)
quests[138550069] = 250 -- At The Enemy's Gates (Darnassus)
quests[138551094] = 250 -- At The Enemy's Gates (The Silver Covenant)
quests[138560076] = 250 -- At The Enemy's Gates (Orgrimmar)
quests[138561124] = 250 -- At The Enemy's Gates (The Sunreavers)
quests[138570530] = 250 -- At The Enemy's Gates (Darkspear Trolls)
quests[138571124] = 250 -- At The Enemy's Gates (The Sunreavers)
quests[138580081] = 250 -- At The Enemy's Gates (Thunder Bluff)
quests[138581124] = 250 -- At The Enemy's Gates (The Sunreavers)
quests[138590911] = 250 -- At The Enemy's Gates (Silvermoon City)
quests[138591124] = 250 -- At The Enemy's Gates (The Sunreavers)
quests[138600068] = 250 -- At The Enemy's Gates (Undercity)
quests[138601124] = 250 -- At The Enemy's Gates (The Sunreavers)
quests[138611106] = 250 -- Battle Before The Citadel (Argent Crusade)
quests[138611094] = 250 -- Battle Before The Citadel (The Silver Covenant)
quests[138621106] = 250 -- Battle Before The Citadel (Argent Crusade)
quests[138621124] = 250 -- Battle Before The Citadel (The Sunreavers)
quests[138631098] = 250 -- Battle Before The Citadel (Knights of the Ebon Blade)
quests[138631124] = 250 -- Battle Before The Citadel (The Sunreavers)
quests[138641098] = 250 -- Battle Before The Citadel (Knights of the Ebon Blade)
quests[138641094] = 250 -- Battle Before The Citadel (The Silver Covenant)
quests[138650076] = 250 -- Wet Work (Orgrimmar)
quests[138670069] = 250 -- Culling the Furbolg Threat (Darnassus)
quests[138680069] = 250 -- Corrupting Influence? (Darnassus)
quests[138690069] = 250 -- Recover the Remains (Darnassus)
quests[138700076] = 25 -- As Good as it Gets (Orgrimmar)
quests[138710076] = 250 -- Security! (Orgrimmar)
quests[138721134] = 10 -- Worgen Wisdom (Gilneas)
quests[138730076] = 150 -- Sheelah's Last Wish (Orgrimmar)
quests[138741134] = 250 -- A Not-So-Charmed Life (Gilneas)
quests[138750076] = 250 -- Gurtar's Request (Orgrimmar)
quests[138830076] = 250 -- Lousy Pieces of Ship (Orgrimmar)
quests[138900076] = 250 -- Keep the Fires Burning (Orgrimmar)
quests[138920069] = 250 -- Leave No Tracks (Darnassus)
quests[138930069] = 25 -- Soggoth and Kronn (Darnassus)
quests[138960047] = 75 -- Unearthed Knowledge (Ironforge)
quests[138970069] = 500 -- The Battle for Darkshore (Darnassus)
quests[138980069] = 250 -- The Tides Turn Against Us (Darnassus)
quests[139000069] = 500 -- The Offering to Azshara (Darnassus)
quests[139010076] = 250 -- Deep Despair (Orgrimmar)
quests[139130069] = 250 -- They Took Our Gnomes (Darnassus)
quests[139130054] = 250 -- They Took Our Gnomes (Gnomeregan)
quests[139190069] = 75 -- A Trip to the Moonwell (Darnassus)
quests[139200076] = 250 -- Before You Go... (Orgrimmar)
quests[139210069] = 25 -- He Who Would Be Forgiven (Darnassus)
quests[139220069] = 250 -- In the Hands of the Perverse (Darnassus)
quests[139230076] = 10 -- To Hellscream's Watch (Orgrimmar)
quests[139240069] = 350 -- All's Well (Darnassus)
quests[139280069] = 250 -- Recover the Fallen (Darnassus)
quests[139350069] = 350 -- Defend the Tree! (Darnassus)
quests[139360076] = 10 -- Tweedle's Dumb (Orgrimmar)
quests[139420076] = 250 -- Set Us Up the Bomb (Orgrimmar)
quests[139430076] = 250 -- Breathing Room (Orgrimmar)
quests[139440076] = 25 -- Small Hands, Short Fuse (Orgrimmar)
quests[139450069] = 250 -- Resident Danger (Darnassus)
quests[139460069] = 250 -- Nature's Reprisal (Darnassus)
quests[139470076] = 250 -- Blastranaar! (Orgrimmar)
quests[139490076] = 250 -- Crossroads Caravan Pickup (Orgrimmar)
quests[139530069] = 75 -- Naga In Our Midst (Darnassus)
quests[139620076] = 250 -- Stalemate (Orgrimmar)
quests[139640069] = 25 -- To the Spire (Darnassus)
quests[139650069] = 25 -- Check in on the Edunes (Darnassus)
quests[139670076] = 250 -- Thinning the... Herd? (Orgrimmar)
quests[139740076] = 150 -- Tweedle's Tiny Package (Orgrimmar)
quests[139760069] = 250 -- Three Friends of the Forest (Darnassus)
quests[139770076] = 10 -- Mass Production (Orgrimmar)
quests[139790069] = 250 -- The Goblin Braintrust (Darnassus)
quests[139800076] = 250 -- They're Out There! (Orgrimmar)
quests[139810069] = 250 -- They Set Them Up The Bomb! (Darnassus)
quests[139820069] = 250 -- In a Bind (Darnassus)
quests[139830076] = 250 -- Building Your Own Coffin (Orgrimmar)
quests[139850069] = 250 -- Clear the Shrine (Darnassus)
quests[139870069] = 350 -- The Last Stand (Darnassus)
quests[139890069] = 350 -- King of the Foulweald (Darnassus)
quests[140011133] = 250 -- Goblin Escape Pods (Bilgewater Cartel)
quests[140061133] = 350 -- Read the Manual (Bilgewater Cartel)
quests[140071133] = 75 -- Steady Shot (Bilgewater Cartel)
quests[140081133] = 75 -- Frost Nova (Bilgewater Cartel)
quests[140091133] = 75 -- Learning the Word (Bilgewater Cartel)
quests[140101133] = 75 -- Eviscerate (Bilgewater Cartel)
quests[140111133] = 75 -- Primal Strike (Bilgewater Cartel)
quests[140121133] = 75 -- Corruption (Bilgewater Cartel)
quests[140131133] = 75 -- Charge (Bilgewater Cartel)
quests[140141133] = 250 -- Get Our Stuff Back! (Bilgewater Cartel)
quests[140161106] = 250 -- The Black Knight's Curse (Argent Crusade)
quests[140171106] = 250 -- The Black Knight's Fate (Argent Crusade)
quests[140180069] = 25 -- He Who Would Be Forgiven (Darnassus)
quests[140191133] = 250 -- Monkey Business (Bilgewater Cartel)
quests[140211133] = 250 -- Miner Troubles (Bilgewater Cartel)
quests[140220072] = 10 -- Pilgrim's Bounty (Stormwind)
quests[140230072] = 75 -- Spice Bread Stuffing (Stormwind)
quests[140240047] = 75 -- Pumpkin Pie (Ironforge)
quests[140280072] = 75 -- Cranberry Chutney (Stormwind)
quests[140300069] = 75 -- They're Ravenous In Darnassus (Darnassus)
quests[140311133] = 250 -- Capturing the Unknown (Bilgewater Cartel)
quests[140330072] = 75 -- Candied Sweet Potatoes (Stormwind)
quests[140350069] = 75 -- Slow-roasted Turkey (Darnassus)
quests[140360068] = 10 -- Pilgrim's Bounty (Undercity)
quests[140370068] = 75 -- Spice Bread Stuffing (Undercity)
quests[140390069] = 250 -- Home of the Kaldorei (Darnassus)
quests[140400076] = 75 -- Pumpkin Pie (Orgrimmar)
quests[140410081] = 75 -- Cranberry Chutney (Thunder Bluff)
quests[140430076] = 75 -- Candied Sweet Potatoes (Orgrimmar)
quests[140440068] = 75 -- Undersupplied in the Undercity (Undercity)
quests[140470076] = 75 -- Slow-roasted Turkey (Orgrimmar)
quests[140480047] = 150 -- Can't Get Enough Turkey (Ironforge)
quests[140510047] = 150 -- Don't Forget The Stuffing! (Ironforge)
quests[140530072] = 150 -- We're Out of Cranberry Chutney Again? (Stormwind)
quests[140540069] = 150 -- Easy As Pie (Darnassus)
quests[140550072] = 150 -- She Says Potato (Stormwind)
quests[140580068] = 150 -- She Says Potato (Undercity)
quests[140590068] = 150 -- We're Out of Cranberry Chutney Again? (Undercity)
quests[140600081] = 150 -- Easy As Pie (Thunder Bluff)
quests[140610076] = 150 -- Can't Get Enough Turkey (Orgrimmar)
quests[140620076] = 150 -- Don't Forget The Stuffing! (Orgrimmar)
quests[140691133] = 250 -- Good Help is Hard to Find (Bilgewater Cartel)
quests[140701133] = 250 -- Do it Yourself (Bilgewater Cartel)
quests[140711133] = 250 -- Rolling with my Homies (Bilgewater Cartel)
quests[140751133] = 250 -- Trouble in the Mines (Bilgewater Cartel)
quests[140781134] = 250 -- Lockdown! (Gilneas)
quests[140911134] = 250 -- Something's Amiss (Gilneas)
quests[140931134] = 350 -- All Hell Breaks Loose (Gilneas)
quests[140941134] = 350 -- Salvage the Supplies (Gilneas)
quests[140981134] = 350 -- Evacuate the Merchant Square (Gilneas)
quests[140991134] = 250 -- Royal Orders (Gilneas)
quests[141031090] = 10 -- Titanium Powder (Kirin Tor)
quests[141091133] = 250 -- The New You (Bilgewater Cartel)
quests[141101133] = 250 -- The New You (Bilgewater Cartel)
quests[141131133] = 250 -- Life of the Party (Bilgewater Cartel)
quests[141151133] = 250 -- Pirate Party Crashers (Bilgewater Cartel)
quests[141161133] = 10 -- The Uninvited Guest (Bilgewater Cartel)
quests[141170076] = 250 -- The Eyes of Ashenvale (Orgrimmar)
quests[141180076] = 250 -- Venison for the Troops (Orgrimmar)
quests[141201133] = 10 -- A Bazillion Macaroons?! (Bilgewater Cartel)
quests[141211133] = 250 -- Robbing Hoods (Bilgewater Cartel)
quests[141221133] = 250 -- The Great Bank Heist (Bilgewater Cartel)
quests[141231133] = 250 -- Waltz Right In (Bilgewater Cartel)
quests[141241133] = 250 -- Liberate the Kaja'mite (Bilgewater Cartel)
quests[141251133] = 250 -- 447 (Bilgewater Cartel)
quests[141261133] = 350 -- Life Savings (Bilgewater Cartel)
quests[141270076] = 150 -- Return of the Highborne? (Orgrimmar)
quests[141280911] = 150 -- Return of the Highborne? (Silvermoon City)
quests[141291133] = 250 -- Runaway Shredder! (Bilgewater Cartel)
quests[141341133] = 250 -- The Captain's Logs (Bilgewater Cartel)
quests[141351133] = 250 -- Up a Tree (Bilgewater Cartel)
quests[141381133] = 10 -- Taking Care of Business (Bilgewater Cartel)
quests[141461133] = 250 -- Defend the Gates! (Bilgewater Cartel)
quests[141511090] = 250 -- Cardinal Ruby (Kirin Tor)
quests[141531133] = 250 -- Life of the Party (Bilgewater Cartel)
quests[141541134] = 250 -- By the Skin of His Teeth (Gilneas)
quests[141551133] = 350 -- Arborcide (Bilgewater Cartel)
quests[141571134] = 10 -- Old Divisions (Gilneas)
quests[141611133] = 250 -- Basilisk Bashin' (Bilgewater Cartel)
quests[141621133] = 25 -- Report to Horzak (Bilgewater Cartel)
quests[141651133] = 250 -- Stone Cold (Bilgewater Cartel)
quests[141840911] = 25 -- My Time Has Passed (Silvermoon City)
quests[141880911] = 10 -- Avenge Furien! (Silvermoon City)
quests[141890911] = 75 -- Translation (Silvermoon City)
quests[141901133] = 250 -- The Perfect Prism (Bilgewater Cartel)
quests[141910911] = 250 -- Furien's Footsteps (Silvermoon City)
quests[141930072] = 250 -- Nothing a Couple of Melons Won't Fix (Stormwind)
quests[141941133] = 250 -- Refleshification (Bilgewater Cartel)
quests[141950911] = 150 -- All Becoming Clearer (Silvermoon City)
quests[141960911] = 250 -- Firestarter (Silvermoon City)
quests[141971133] = 250 -- A Quota to Meet (Bilgewater Cartel)
quests[142010911] = 250 -- A Thousand Stories in the Sand (Silvermoon City)
quests[142021133] = 250 -- Survey the Lakeshore (Bilgewater Cartel)
quests[142041134] = 250 -- From the Shadows (Gilneas)
quests[142091133] = 150 -- Gunk in the Trunk (Bilgewater Cartel)
quests[142121134] = 150 -- Sacrifices (Gilneas)
quests[142141134] = 25 -- Message to Greymane (Gilneas)
quests[142150911] = 250 -- Memories of the Dead (Silvermoon City)
quests[142160911] = 250 -- Mystery of the Sarcen Stone (Silvermoon City)
quests[142181134] = 250 -- By Blood and Ash (Gilneas)
quests[142201134] = 500 -- This Is the End (Gilneas)
quests[142211134] = 75 -- Never Surrender, Sometimes Retreat (Gilneas)
quests[142221134] = 250 -- Last Stand (Gilneas)
quests[142331133] = 10 -- Orcs Can Write? (Bilgewater Cartel)
quests[142341133] = 10 -- The Enemy of My Enemy (Bilgewater Cartel)
quests[142340076] = 10 -- The Enemy of My Enemy (Orgrimmar)
quests[142351133] = 10 -- The Vicious Vale (Bilgewater Cartel)
quests[142350076] = 10 -- The Vicious Vale (Orgrimmar)
quests[142361133] = 250 -- Weed Whacker (Bilgewater Cartel)
quests[142360076] = 250 -- Weed Whacker (Orgrimmar)
quests[142371133] = 10 -- Forward Movement (Bilgewater Cartel)
quests[142370076] = 10 -- Forward Movement (Orgrimmar)
quests[142381133] = 250 -- Infrared = Infradead (Bilgewater Cartel)
quests[142380076] = 250 -- Infrared = Infradead (Orgrimmar)
quests[142391133] = 10 -- Don't Go Into the Light! (Bilgewater Cartel)
quests[142401133] = 10 -- To the Cliffs (Bilgewater Cartel)
quests[142400076] = 10 -- To the Cliffs (Orgrimmar)
quests[142411133] = 250 -- Get to the Gyrochoppa! (Bilgewater Cartel)
quests[142410076] = 250 -- Get to the Gyrochoppa! (Orgrimmar)
quests[142421133] = 75 -- Precious Cargo (Bilgewater Cartel)
quests[142420076] = 75 -- Precious Cargo (Orgrimmar)
quests[142431133] = 350 -- Warchief's Revenge (Bilgewater Cartel)
quests[142430076] = 350 -- Warchief's Revenge (Orgrimmar)
quests[142441133] = 10 -- Up, Up & Away! (Bilgewater Cartel)
quests[142451133] = 10 -- It's a Town-In-A-Box (Bilgewater Cartel)
quests[142481133] = 10 -- Help Wanted (Bilgewater Cartel)
quests[142570609] = 250 -- Slitherblade Slaughter (Cenarion Circle)
quests[142581133] = 250 -- Mortar the Point (Bilgewater Cartel)
quests[142600609] = 75 -- Going Deep (Cenarion Circle)
quests[142611133] = 250 -- Ice Cold (Bilgewater Cartel)
quests[142620076] = 250 -- To Gut a Fish (Orgrimmar)
quests[142651134] = 250 -- Your Instructor (Gilneas)
quests[142661134] = 75 -- Charge (Gilneas)
quests[142680609] = 250 -- Deep Impact (Cenarion Circle)
quests[142691134] = 250 -- Someone's Looking for You (Gilneas)
quests[142710076] = 150 -- Report to Twocrush (Orgrimmar)
quests[142721134] = 75 -- Eviscerate (Gilneas)
quests[142731134] = 250 -- Shady Associates (Gilneas)
quests[142741134] = 75 -- Corruption (Gilneas)
quests[142751134] = 250 -- Someone's Keeping Track of You (Gilneas)
quests[142761134] = 75 -- Steady Shot (Gilneas)
quests[142771134] = 250 -- Arcane Inquiries (Gilneas)
quests[142781134] = 250 -- Seek the Sister (Gilneas)
quests[142791134] = 75 -- Learning the Word (Gilneas)
quests[142801134] = 250 -- The Winds Know Your Name... Apparently (Gilneas)
quests[142811134] = 75 -- Frost Nova (Gilneas)
quests[142820609] = 250 -- Mystery Solved (Cenarion Circle)
quests[142831134] = 75 -- Moonfire (Gilneas)
quests[142851134] = 25 -- Safety in Numbers (Gilneas)
quests[142861134] = 25 -- Safety in Numbers (Gilneas)
quests[142871134] = 25 -- Safety in Numbers (Gilneas)
quests[142881134] = 25 -- Safety in Numbers (Gilneas)
quests[142891134] = 25 -- Safety in Numbers (Gilneas)
quests[142901134] = 25 -- Safety in Numbers (Gilneas)
quests[142911134] = 25 -- Safety in Numbers (Gilneas)
quests[142920609] = 75 -- The Enemy of Our Enemy (Cenarion Circle)
quests[142931134] = 75 -- Save Krennan Aranas (Gilneas)
quests[142941134] = 10 -- Time to Regroup (Gilneas)
quests[142950076] = 350 -- Sisters of the Sea (Orgrimmar)
quests[142971133] = 250 -- Pro-liberation (Bilgewater Cartel)
quests[143010609] = 250 -- Return and Report (Cenarion Circle)
quests[143020609] = 10 -- Official Assessment (Cenarion Circle)
quests[143031133] = 10 -- Back to Aggra (Bilgewater Cartel)
quests[143030076] = 10 -- Back to Aggra (Orgrimmar)
quests[143040609] = 250 -- Blood Theory (Cenarion Circle)
quests[143050609] = 250 -- A Time to Reap (Cenarion Circle)
quests[143060609] = 250 -- Pulling Weeds (Cenarion Circle)
quests[143070609] = 250 -- Stubborn Winds (Cenarion Circle)
quests[143081133] = 250 -- When Science Attacks (Bilgewater Cartel)
quests[143090609] = 250 -- Calming the Kodo (Cenarion Circle)
quests[143101133] = 250 -- Segmentation Fault: Core Dumped (Bilgewater Cartel)
quests[143110609] = 250 -- Taking Part (Cenarion Circle)
quests[143120609] = 10 -- An Introduction Is In Order (Cenarion Circle)
quests[143131134] = 25 -- Among Humans Again (Gilneas)
quests[143140609] = 250 -- Breakout! (Cenarion Circle)
quests[143160609] = 250 -- Cenarion Property (Cenarion Circle)
quests[143191134] = 25 -- Further Treatment (Gilneas)
quests[143201134] = 250 -- In Need of Ingredients (Gilneas)
quests[143211134] = 250 -- Invasion (Gilneas)
quests[143221133] = 250 -- Bad Science! Bad! (Bilgewater Cartel)
quests[143261133] = 10 -- Meet Me Up Top (Bilgewater Cartel)
quests[143260076] = 10 -- Meet Me Up Top (Orgrimmar)
quests[143320609] = 250 -- Time For Change (Cenarion Circle)
quests[143361134] = 75 -- Kill or Be Killed (Gilneas)
quests[143430609] = 75 -- Maurin's Concoction (Cenarion Circle)
quests[143460609] = 250 -- Cleansing Our Crevasse (Cenarion Circle)
quests[143471134] = 250 -- Hold the Line (Gilneas)
quests[143481134] = 250 -- You Can't Take 'Em Alone (Gilneas)
quests[143600911] = 250 -- Nothing a Couple of Melons Won't Fix (Silvermoon City)
quests[143661134] = 150 -- Holding Steady (Gilneas)
quests[143671134] = 75 -- The Allens' Storm Cellar (Gilneas)
quests[143681134] = 250 -- Save the Children! (Gilneas)
quests[143691134] = 250 -- Unleash the Beast (Gilneas)
quests[143701133] = 250 -- Mysterious Azsharite (Bilgewater Cartel)
quests[143711133] = 250 -- A Gigantic Snack (Bilgewater Cartel)
quests[143740081] = 350 -- Portals of the Legion (Thunder Bluff)
quests[143751134] = 250 -- Last Chance at Humanity (Gilneas)
quests[143771133] = 250 -- Befriending Giants (Bilgewater Cartel)
quests[143810609] = 250 -- Cleansing Our Crevasse (Cenarion Circle)
quests[143821134] = 250 -- Two By Sea (Gilneas)
quests[143831133] = 250 -- The Terrible Tinkers of the Ruined Reaches (Bilgewater Cartel)
quests[143851133] = 250 -- Azsharite Experiment Number One (Bilgewater Cartel)
quests[143861134] = 150 -- Leader of the Pack (Gilneas)
quests[143881133] = 250 -- Azsharite Experiment Number Two (Bilgewater Cartel)
quests[143940092] = 10000000 -- Death to Agogridon (Gelkis Clan Centaur)
quests[143940609] = 700 -- Death to Agogridon (Cenarion Circle)
quests[143940093] = 10000000 -- Death to Agogridon (Magram Clan Centaur)
quests[143951134] = 250 -- Gasping for Breath (Gilneas)
quests[143961134] = 75 -- As the Land Shatters (Gilneas)
quests[143971134] = 25 -- Evacuation (Gilneas)
quests[143981134] = 75 -- Grandma Wahl (Gilneas)
quests[143991134] = 150 -- Grandma's Lost It Alright (Gilneas)
quests[144001134] = 75 -- I Can't Wear This (Gilneas)
quests[144011134] = 150 -- Grandma's Cat (Gilneas)
quests[144021134] = 75 -- Ready to Go (Gilneas)
quests[144031134] = 75 -- The Hayward Brothers (Gilneas)
quests[144041134] = 150 -- Not Quite Shipshape (Gilneas)
quests[144051134] = 75 -- Escape By Sea (Gilneas)
quests[144061134] = 10 -- The Crowley Orchard (Gilneas)
quests[144071133] = 75 -- Azshara Blues (Bilgewater Cartel)
quests[144121134] = 150 -- Washed Up (Gilneas)
quests[144161134] = 150 -- The Hungry Ettin (Gilneas)
quests[144231133] = 250 -- Dozercism (Bilgewater Cartel)
quests[144241133] = 150 -- Need More Science (Bilgewater Cartel)
quests[144300911] = 150 -- Hacking the Construct (Silvermoon City)
quests[144310076] = 250 -- The Blackmaw Scar (Orgrimmar)
quests[144320076] = 250 -- A Pale Brew (Orgrimmar)
quests[144330076] = 250 -- Diplomacy by Another Means (Orgrimmar)
quests[144350076] = 350 -- The Blackmaw Doublecross (Orgrimmar)
quests[144360081] = 250 -- Dwarven Digging (Thunder Bluff)
quests[144380081] = 250 -- Sharing the Land (Thunder Bluff)
quests[144390081] = 75 -- Journey into Thunder Bluff (Thunder Bluff)
quests[144451133] = 10 -- Farewell, For Now (Bilgewater Cartel)
quests[144520081] = 10 -- Rite of Strength (Thunder Bluff)
quests[144550081] = 250 -- Stop the Thorncallers (Thunder Bluff)
quests[144560081] = 250 -- Rite of Courage (Thunder Bluff)
quests[144580081] = 25 -- Go to Adana (Thunder Bluff)
quests[144590081] = 250 -- The Battleboars (Thunder Bluff)
quests[144600081] = 350 -- Rite of Honor (Thunder Bluff)
quests[144610081] = 250 -- Feed of Evil (Thunder Bluff)
quests[144621133] = 10 -- Where's My Head? (Bilgewater Cartel)
quests[144640076] = 250 -- Lightning Strike Assassination (Orgrimmar)
quests[144651134] = 25 -- To Greymane Manor (Gilneas)
quests[144661134] = 25 -- The King's Observatory (Gilneas)
quests[144671134] = 25 -- Alas, Gilneas! (Gilneas)
quests[144681133] = 75 -- Another Warm Body (Bilgewater Cartel)
quests[144691133] = 250 -- Hand-me-downs (Bilgewater Cartel)
quests[144701133] = 250 -- Military Breakthrough (Bilgewater Cartel)
quests[144711133] = 250 -- First Degree Mortar (Bilgewater Cartel)
quests[144721133] = 250 -- In The Face! (Bilgewater Cartel)
quests[144731133] = 250 -- It's Our Problem Now (Bilgewater Cartel)
quests[144741133] = 250 -- Goblin Escape Pods (Bilgewater Cartel)
quests[144751133] = 150 -- Grounded! (Bilgewater Cartel)
quests[144761133] = 250 -- Rigged to Blow (Bilgewater Cartel)
quests[144770076] = 250 -- Push the Button! (Orgrimmar)
quests[144781133] = 75 -- Operation Fishgut (Bilgewater Cartel)
quests[144791133] = 150 -- There Are Many Like It (Bilgewater Cartel)
quests[144801133] = 250 -- Extermination (Bilgewater Cartel)
quests[144821135] = 250 -- Call of Duty (The Earthen Ring)
quests[144841133] = 250 -- Head of the Snake (Bilgewater Cartel)
quests[144851133] = 250 -- Ticker Required (Bilgewater Cartel)
quests[144861133] = 250 -- Handling the Goods (Bilgewater Cartel)
quests[144871133] = 250 -- Still Beating Heart (Bilgewater Cartel)
quests[204400081] = 250 -- Poison Water (Thunder Bluff)
quests[204410081] = 150 -- Rite of Vision (Thunder Bluff)
quests[237330081] = 150 -- Rites of the Earthmother (Thunder Bluff)
quests[242150081] = 150 -- Rite of the Winds (Thunder Bluff)
quests[244280072] = 500 -- A Most Puzzling Circumstance (Stormwind)
quests[244290076] = 500 -- A Most Puzzling Circumstance (Orgrimmar)
quests[244301133] = 250 -- Blacken the Skies (Bilgewater Cartel)
quests[244321135] = 250 -- Sea Legs (The Earthen Ring)
quests[244330076] = 250 -- Let Them Feast on Fear (Orgrimmar)
quests[244351133] = 250 -- Mop Up (Bilgewater Cartel)
quests[244361133] = 250 -- Halo Drops (Bilgewater Cartel)
quests[244371133] = 250 -- First Come, First Served (Bilgewater Cartel)
quests[244381134] = 250 -- Exodus (Gilneas)
quests[244390076] = 350 -- The Conquest of Azshara (Orgrimmar)
quests[244400081] = 250 -- Winterhoof Cleansing (Thunder Bluff)
quests[244410081] = 250 -- Thunderhorn Totem (Thunder Bluff)
quests[244421106] = 250 -- Battle Plans Of The Kvaldir (Argent Crusade)
quests[244481133] = 250 -- Field Promotion (Bilgewater Cartel)
quests[244491133] = 500 -- Shore Leave (Bilgewater Cartel)
quests[244521133] = 350 -- Profitability Scouting (Bilgewater Cartel)
quests[244531133] = 75 -- Private Chat (Bilgewater Cartel)
quests[244551133] = 75 -- Rapid Deployment (Bilgewater Cartel)
quests[244560081] = 250 -- Thunderhorn Cleansing (Thunder Bluff)
quests[244570081] = 25 -- Rite of Vision (Thunder Bluff)
quests[244581133] = 150 -- A Hello to Arms (Bilgewater Cartel)
quests[244681134] = 250 -- Stranded at the Marsh (Gilneas)
quests[244690047] = 250 -- Hold the Line! (Ironforge)
quests[244700047] = 250 -- Give 'em What-For (Ironforge)
quests[244710047] = 250 -- Aid for the Wounded (Ironforge)
quests[244721134] = 250 -- Introductions Are in Order (Gilneas)
quests[244730047] = 10 -- Lockdown in Anvilmar (Ironforge)
quests[244740047] = 250 -- First Things First: We're Gonna Need Some Beer (Ironforge)
quests[244750047] = 250 -- All the Other Stuff (Ironforge)
quests[244770047] = 250 -- Dwarven Artifacts (Ironforge)
quests[244831134] = 75 -- Stormglen (Gilneas)
quests[244841134] = 250 -- Pest Control (Gilneas)
quests[244860047] = 250 -- Make Hay While the Sun Shines (Ironforge)
quests[244870047] = 10 -- Whitebeard Needs Ye (Ironforge)
quests[244881133] = 250 -- The Replacements (Bilgewater Cartel)
quests[244890047] = 250 -- Trolling for Information (Ironforge)
quests[244900047] = 10 -- A Trip to Ironforge (Ironforge)
quests[244910047] = 10 -- Follow that Gyro-Copter! (Ironforge)
quests[244910054] = 10 -- Follow that Gyro-Copter! (Gnomeregan)
quests[244920047] = 150 -- Pack Your Bags (Ironforge)
quests[244920054] = 150 -- Pack Your Bags (Gnomeregan)
quests[244930047] = 10 -- Don't Forget About Us (Ironforge)
quests[244940047] = 75 -- Empowered Rune (Ironforge)
quests[244951134] = 250 -- Pieces of the Past (Gilneas)
quests[244960047] = 75 -- Arcane Rune (Ironforge)
quests[244971133] = 75 -- Airborne Again (Bilgewater Cartel)
quests[244981050] = 500 -- The Path to the Citadel (Valiance Expedition)
quests[244991050] = 500 -- Echoes of Tortured Souls (Valiance Expedition)
quests[245001050] = 500 -- Wrath of the Lich King (Valiance Expedition)
quests[245011134] = 250 -- Queen-Sized Troubles (Gilneas)
quests[245021133] = 250 -- Necessary Roughness (Bilgewater Cartel)
quests[245031133] = 250 -- Fourth and Goal (Bilgewater Cartel)
quests[245040081] = 250 -- Clear the High Road (Thunder Bluff)
quests[245050081] = 10 -- Holdout at Hunter's Hill (Thunder Bluff)
quests[245061085] = 25 -- Inside the Frozen Citadel (Warsong Offensive)
quests[245071085] = 500 -- The Path to the Citadel (Warsong Offensive)
quests[245111085] = 500 -- Echoes of Tortured Souls (Warsong Offensive)
quests[245120076] = 250 -- Warriors' Redemption (Orgrimmar)
quests[245130076] = 250 -- Eye for an Eye (Orgrimmar)
quests[245140081] = 250 -- Raptor Scraps (Thunder Bluff)
quests[245150081] = 250 -- Signals in the Sky (Thunder Bluff)
quests[245170076] = 250 -- Put the Fear in Them (Orgrimmar)
quests[245180076] = 150 -- The Low Road (Orgrimmar)
quests[245190076] = 250 -- Stalling the Survey (Orgrimmar)
quests[245201133] = 10 -- Give Sassy the News (Bilgewater Cartel)
quests[245230081] = 250 -- Wildmane Totem (Thunder Bluff)
quests[245240081] = 350 -- Wildmane Cleansing (Thunder Bluff)
quests[245250081] = 250 -- Keeping the Dogs at Bay (Thunder Bluff)
quests[245260047] = 75 -- Filling Up the Spellbook (Ironforge)
quests[245270047] = 75 -- Your Path Begins Here (Ironforge)
quests[245280047] = 75 -- The Power of the Light (Ironforge)
quests[245290081] = 250 -- Bad to Worse (Thunder Bluff)
quests[245300047] = 75 -- Oh, A Hunter's Life For Me (Ironforge)
quests[245310047] = 75 -- Getting Battle-Ready (Ironforge)
quests[245320047] = 75 -- Evisceratin' the Enemy (Ironforge)
quests[245330047] = 75 -- Words of Power (Ironforge)
quests[245340081] = 250 -- Speaking Their Language (Thunder Bluff)
quests[245390081] = 250 -- When Plants Attack (Thunder Bluff)
quests[245420609] = 250 -- A Curious Bloom (Cenarion Circle)
quests[245430081] = 250 -- A Family Divided (Thunder Bluff)
quests[245451156] = 350 -- The Sacred and the Corrupt (The Ashen Verdict)
quests[245460081] = 250 -- A Line in the Dirt (Thunder Bluff)
quests[245471156] = 350 -- A Feast of Souls (The Ashen Verdict)
quests[245481156] = 350 -- The Splintered Throne (The Ashen Verdict)
quests[245491156] = 3000 -- Shadowmourne... (The Ashen Verdict)
quests[245500081] = 75 -- Journey into Thunder Bluff (Thunder Bluff)
quests[245510076] = 250 -- Meet the New Boss (Orgrimmar)
quests[245520081] = 250 -- Lion's Pride (Thunder Bluff)
quests[245531094] = 500 -- The Purification of Quel'Delar (The Silver Covenant)
quests[245641124] = 500 -- The Purification of Quel'Delar (The Sunreavers)
quests[245671133] = 10 -- Report for Tryouts (Bilgewater Cartel)
quests[245680081] = 350 -- To the Walls! (Thunder Bluff)
quests[245690076] = 250 -- Siegebreaker (Orgrimmar)
quests[245720076] = 250 -- Taking Back Taurajo (Orgrimmar)
quests[245730081] = 250 -- Honoring the Dead (Thunder Bluff)
quests[245751134] = 150 -- Liberation Day (Gilneas)
quests[245770076] = 75 -- Desolation Hold Inspection (Orgrimmar)
quests[245780069] = 75 -- The Blackwald (Darnassus)
quests[245791090] = 250 -- Sartharion Must Die! (Kirin Tor)
quests[245801090] = 250 -- Anub'Rekhan Must Die! (Kirin Tor)
quests[245811090] = 250 -- Noth the Plaguebringer Must Die! (Kirin Tor)
quests[245821090] = 250 -- Instructor Razuvious Must Die! (Kirin Tor)
quests[245831090] = 250 -- Patchwerk Must Die! (Kirin Tor)
quests[245841090] = 250 -- Malygos Must Die! (Kirin Tor)
quests[245851090] = 250 -- Flame Leviathan Must Die! (Kirin Tor)
quests[245861090] = 250 -- Razorscale Must Die! (Kirin Tor)
quests[245871090] = 250 -- Ignis the Furnace Master Must Die! (Kirin Tor)
quests[245881090] = 250 -- XT-002 Deconstructor Must Die! (Kirin Tor)
quests[245891090] = 250 -- Lord Jaraxxus Must Die! (Kirin Tor)
quests[245901090] = 250 -- Lord Marrowgar Must Die! (Kirin Tor)
quests[245910076] = 250 -- Changing of the Gar'dul (Orgrimmar)
quests[245921134] = 250 -- Betrayal at Tempest's Reach (Gilneas)
quests[245931134] = 250 -- Neither Human Nor Beast (Gilneas)
quests[245941124] = 500 -- The Purification of Quel'Delar (The Sunreavers)
quests[245951094] = 500 -- The Purification of Quel'Delar (The Silver Covenant)
quests[245961124] = 500 -- The Purification of Quel'Delar (The Sunreavers)
quests[245981124] = 500 -- The Purification of Quel'Delar (The Sunreavers)
quests[246021134] = 250 -- Laid to Rest (Gilneas)
quests[246070530] = 75 -- The Rise of the Darkspear (Darkspear Trolls)
quests[246161134] = 250 -- Losing Your Tail (Gilneas)
quests[246171134] = 250 -- Tal'doren, the Wild Home (Gilneas)
quests[246180076] = 250 -- Claim the Battle Scar (Orgrimmar)
quests[246190068] = 250 -- Parts-is-Parts (Undercity)
quests[246200068] = 250 -- Hearts-is-Hearts (Undercity)
quests[246210068] = 350 -- Smarts-is-Smarts (Undercity)
quests[246220530] = 75 -- A Troll's Truest Companion (Darkspear Trolls)
quests[246230530] = 250 -- Saving the Young (Darkspear Trolls)
quests[246240530] = 250 -- Mercy for the Lost (Darkspear Trolls)
quests[246250530] = 250 -- Consort of the Sea Witch (Darkspear Trolls)
quests[246260530] = 250 -- Young and Vicious (Darkspear Trolls)
quests[246271134] = 250 -- At Our Doorstep (Gilneas)
quests[246280069] = 150 -- Preparations (Darnassus)
quests[246310081] = 250 -- Flightmare (Thunder Bluff)
quests[246320081] = 75 -- Tauren Vengeance (Thunder Bluff)
quests[246340076] = 250 -- Intelligence Warfare (Orgrimmar)
quests[246370076] = 350 -- The Butcher of Taurajo (Orgrimmar)
quests[246390530] = 250 -- The Basics: Hitting Things (Darkspear Trolls)
quests[246400530] = 150 -- The Arts of a Warrior (Darkspear Trolls)
quests[246410530] = 250 -- A Rough Start (Darkspear Trolls)
quests[246420530] = 250 -- Proving Pit (Darkspear Trolls)
quests[246430530] = 75 -- More Than Expected (Darkspear Trolls)
quests[246461134] = 150 -- Take Back What's Ours (Gilneas)
quests[246540076] = 250 -- Silithissues (Orgrimmar)
quests[246671135] = 75 -- Firestone Point (The Earthen Ring)
quests[246711133] = 250 -- Cluster Cluck (Bilgewater Cartel)
quests[246721134] = 150 -- Onwards and Upwards (Gilneas)
quests[246731134] = 75 -- Return to Stormglen (Gilneas)
quests[246741134] = 150 -- Slaves to No One (Gilneas)
quests[246751134] = 350 -- Last Meal (Gilneas)
quests[246761134] = 250 -- Push Them Out (Gilneas)
quests[246771134] = 75 -- Flank the Forsaken (Gilneas)
quests[246781134] = 250 -- Knee-Deep (Gilneas)
quests[246791134] = 250 -- Patriarch's Blessing (Gilneas)
quests[246801134] = 10 -- Keel Harbor (Gilneas)
quests[246811134] = 350 -- They Have Allies, But So Do We (Gilneas)
quests[246840081] = 250 -- A Weezil in the Henhouse (Thunder Bluff)
quests[246850081] = 250 -- Dwarf Fortress (Thunder Bluff)
quests[246860609] = 250 -- Carried on the Waves (Cenarion Circle)
quests[246870609] = 250 -- Bouquets of Death (Cenarion Circle)
quests[246890609] = 10 -- Flowing to the North (Cenarion Circle)
quests[247101050] = 500 -- Deliverance from the Pit (Valiance Expedition)
quests[247121085] = 500 -- Deliverance from the Pit (Warsong Offensive)
quests[247190609] = 250 -- Claws of White (Cenarion Circle)
quests[247370369] = 250 -- Super Sticky (Gadgetzan)
quests[247411133] = 250 -- Trading Up (Bilgewater Cartel)
quests[247431156] = 250 -- Shadow's Edge (The Ashen Verdict)
quests[247441133] = 350 -- The Biggest Egg Ever (Bilgewater Cartel)
quests[247470081] = 350 -- Sabotage! (Thunder Bluff)
quests[247481156] = 3000 -- The Lich King's Last Stand (The Ashen Verdict)
quests[247491156] = 350 -- Unholy Infusion (The Ashen Verdict)
quests[247500530] = 75 -- The Rise of the Darkspear (Darkspear Trolls)
quests[247510530] = 250 -- The Basics: Hitting Things (Darkspear Trolls)
quests[247520530] = 150 -- The Arts of a Mage (Darkspear Trolls)
quests[247530530] = 250 -- A Rough Start (Darkspear Trolls)
quests[247540530] = 250 -- Proving Pit (Darkspear Trolls)
quests[247550530] = 75 -- More Than Expected (Darkspear Trolls)
quests[247561156] = 350 -- Blood Infusion (The Ashen Verdict)
quests[247571156] = 350 -- Frost Infusion (The Ashen Verdict)
quests[247580530] = 75 -- The Rise of the Darkspear (Darkspear Trolls)
quests[247590530] = 250 -- The Basics: Hitting Things (Darkspear Trolls)
quests[247600530] = 150 -- The Arts of a Shaman (Darkspear Trolls)
quests[247610530] = 250 -- A Rough Start (Darkspear Trolls)
quests[247620530] = 250 -- Proving Pit (Darkspear Trolls)
quests[247630530] = 75 -- More Than Expected (Darkspear Trolls)
quests[247640530] = 75 -- The Rise of the Darkspear (Darkspear Trolls)
quests[247650530] = 250 -- The Basics: Hitting Things (Darkspear Trolls)
quests[247660530] = 150 -- The Arts of a Druid (Darkspear Trolls)
quests[247670530] = 250 -- A Rough Start (Darkspear Trolls)
quests[247680530] = 250 -- Proving Pit (Darkspear Trolls)
quests[247690530] = 75 -- More Than Expected (Darkspear Trolls)
quests[247700530] = 75 -- The Rise of the Darkspear (Darkspear Trolls)
quests[247710530] = 250 -- The Basics: Hitting Things (Darkspear Trolls)
quests[247720530] = 150 -- The Arts of a Rogue (Darkspear Trolls)
quests[247730530] = 250 -- A Rough Start (Darkspear Trolls)
quests[247740530] = 250 -- Proving Pit (Darkspear Trolls)
quests[247750530] = 75 -- More Than Expected (Darkspear Trolls)
quests[247760530] = 75 -- The Rise of the Darkspear (Darkspear Trolls)
quests[247770530] = 250 -- The Basics: Hitting Things (Darkspear Trolls)
quests[247780530] = 150 -- The Arts of a Hunter (Darkspear Trolls)
quests[247790530] = 250 -- A Rough Start (Darkspear Trolls)
quests[247800530] = 250 -- Proving Pit (Darkspear Trolls)
quests[247810530] = 75 -- More Than Expected (Darkspear Trolls)
quests[247820530] = 75 -- The Rise of the Darkspear (Darkspear Trolls)
quests[247830530] = 250 -- The Basics: Hitting Things (Darkspear Trolls)
quests[247840530] = 150 -- Learnin' tha Word (Darkspear Trolls)
quests[247850530] = 250 -- A Rough Start (Darkspear Trolls)
quests[247860530] = 250 -- Proving Pit (Darkspear Trolls)
quests[247870530] = 75 -- More Than Expected (Darkspear Trolls)
quests[248021085] = 500 -- Wrath of the Lich King (Warsong Offensive)
quests[248120530] = 250 -- No More Mercy (Darkspear Trolls)
quests[248130530] = 250 -- Territorial Fetish (Darkspear Trolls)
quests[248140530] = 500 -- An Ancient Enemy (Darkspear Trolls)
quests[248151156] = 250 -- Choose Your Path (The Ashen Verdict)
quests[248161133] = 250 -- Who's Top of the Food Chain Now? (Bilgewater Cartel)
quests[248171133] = 350 -- A Goblin in Shark's Clothing (Bilgewater Cartel)
quests[248231156] = 250 -- Path of Destruction (The Ashen Verdict)
quests[248251156] = 250 -- Path of Wisdom (The Ashen Verdict)
quests[248261156] = 250 -- Path of Vengeance (The Ashen Verdict)
quests[248271156] = 250 -- Path of Courage (The Ashen Verdict)
quests[248281156] = 250 -- Path of Destruction (The Ashen Verdict)
quests[248291156] = 250 -- Path of Destruction (The Ashen Verdict)
quests[248301156] = 250 -- Path of Wisdom (The Ashen Verdict)
quests[248311156] = 250 -- Path of Wisdom (The Ashen Verdict)
quests[248321156] = 250 -- Path of Vengeance (The Ashen Verdict)
quests[248331156] = 250 -- Path of Vengeance (The Ashen Verdict)
quests[248341156] = 250 -- Path of Courage (The Ashen Verdict)
quests[248351156] = 250 -- Path of Courage (The Ashen Verdict)
quests[248520081] = 250 -- Our Tribe, Imprisoned (Thunder Bluff)
quests[248540942] = 10 -- Finding Stormclaw (Cenarion Expedition)
quests[248550609] = 250 -- Aberrant Flora (Cenarion Circle)
quests[248561133] = 10 -- Invasion Imminent! (Bilgewater Cartel)
quests[248581133] = 250 -- Bilgewater Cartel Represent (Bilgewater Cartel)
quests[248591133] = 250 -- Naga Hide (Bilgewater Cartel)
quests[248610081] = 25 -- Last Rites, First Rites (Thunder Bluff)
quests[248620072] = 250 -- Running the Gauntlet (Stormwind)
quests[248630072] = 250 -- Breaking the Siege (Stormwind)
quests[248641133] = 250 -- Irresistible Pool Pony (Bilgewater Cartel)
quests[248681133] = 350 -- Surrender or Else! (Bilgewater Cartel)
quests[248971133] = 10 -- Get Back to Town (Bilgewater Cartel)
quests[249011133] = 250 -- Town-In-A-Box: Under Attack (Bilgewater Cartel)
quests[249021134] = 350 -- The Hunt For Sylvanas (Gilneas)
quests[249031134] = 150 -- Vengeance or Survival (Gilneas)
quests[249041134] = 500 -- The Battle for Gilneas City (Gilneas)
quests[249061133] = 250 -- Seaside Salvage (Bilgewater Cartel)
quests[249101133] = 250 -- Rocket Rescue (Bilgewater Cartel)
quests[249121156] = 250 -- Empowerment (The Ashen Verdict)
quests[249130930] = 350 -- Remember the Light (Exodar)
quests[249141156] = 250 -- Personal Property (The Ashen Verdict)
quests[249151156] = 250 -- Mograine's Reunion (The Ashen Verdict)
quests[249161156] = 250 -- Jaina's Locket (The Ashen Verdict)
quests[249171156] = 250 -- Muradin's Lament (The Ashen Verdict)
quests[249181156] = 250 -- Sylvanas' Vengeance (The Ashen Verdict)
quests[249191156] = 250 -- The Lightbringer's Redemption (The Ashen Verdict)
quests[249201134] = 350 -- Slowing the Inevitable (Gilneas)
quests[249210072] = 75 -- Report to Aubrey (Stormwind)
quests[249241133] = 10 -- Oomlot Village (Bilgewater Cartel)
quests[249251133] = 250 -- Free the Captives (Bilgewater Cartel)
quests[249281133] = 250 -- To The Ground! (Bilgewater Cartel)
quests[249291133] = 250 -- Send a Message (Bilgewater Cartel)
quests[249301134] = 150 -- While You're At It (Gilneas)
quests[249320169] = 250 -- Cutting Losses (Steamwheedle Cartel)
quests[249330169] = 250 -- Chicken of the Desert (Steamwheedle Cartel)
quests[249340072] = 250 -- Repel Boarders! (Stormwind)
quests[249371133] = 10 -- Oomlot Dealt With (Bilgewater Cartel)
quests[249380072] = 75 -- The Guns of Northwatch (Stormwind)
quests[249390072] = 250 -- Run Out the Guns (Stormwind)
quests[249401133] = 10 -- Up the Volcano (Bilgewater Cartel)
quests[249410072] = 250 -- Langridge Shot (Stormwind)
quests[249421133] = 250 -- Zombies vs. Super Booster Rocket Boots (Bilgewater Cartel)
quests[249430072] = 250 -- Re-Take the Courtyard (Stormwind)
quests[249441134] = 250 -- Back in the Fight (Gilneas)
quests[249451133] = 250 -- Three Little Pygmies (Bilgewater Cartel)
quests[249461133] = 250 -- Rockin' Powder (Bilgewater Cartel)
quests[249480072] = 250 -- Lifting the Siege (Stormwind)
quests[249491133] = 250 -- Booty Duty (Bilgewater Cartel)
quests[249510169] = 250 -- A Great Idea (Steamwheedle Cartel)
quests[249521133] = 10 -- Rocket Boot Boost (Bilgewater Cartel)
quests[249530169] = 350 -- Just Trying to Kill Some Bugs (Steamwheedle Cartel)
quests[249541133] = 250 -- Children of a Turtle God (Bilgewater Cartel)
quests[249551133] = 250 -- Un-Chartered (Bilgewater Cartel)
quests[249561134] = 350 -- Silencing Rageroar (Gilneas)
quests[249571133] = 250 -- Get The Centipaarty Started (Bilgewater Cartel)
quests[249581133] = 350 -- Volcanoth! (Bilgewater Cartel)
quests[249590068] = 10 -- Fresh out of the Grave (Undercity)
quests[249600068] = 150 -- The Wakening (Undercity)
quests[249610068] = 150 -- The Truth of the Grave (Undercity)
quests[249620068] = 75 -- Trail-Worn Scroll (Undercity)
quests[249631133] = 250 -- Maul 'Em With Kindness (Bilgewater Cartel)
quests[249640068] = 75 -- The Thrill of the Hunt (Undercity)
quests[249650068] = 75 -- Magic Training (Undercity)
quests[249660068] = 75 -- Of Light and Shadows (Undercity)
quests[249670068] = 75 -- Stab! (Undercity)
quests[249680068] = 75 -- Dark Deeds (Undercity)
quests[249690068] = 75 -- Charging into Battle (Undercity)
quests[249700068] = 10 -- No Better Than the Zombies (Undercity)
quests[249710068] = 500 -- Assault on the Rotbrain Encampment (Undercity)
quests[249720068] = 150 -- Vital Intelligence (Undercity)
quests[249730068] = 350 -- Night Web's Hollow (Undercity)
quests[249740068] = 250 -- Ever So Lonely (Undercity)
quests[249750068] = 250 -- Fields of Grief (Undercity)
quests[249760068] = 250 -- Variety is the Spice of Death (Undercity)
quests[249770068] = 25 -- Johaan's Experiment (Undercity)
quests[249780068] = 250 -- Reaping the Reapers (Undercity)
quests[249790068] = 350 -- A Scarlet Letter (Undercity)
quests[249800068] = 250 -- The Scarlet Palisade (Undercity)
quests[249810068] = 250 -- A Thorn in our Side (Undercity)
quests[249820068] = 10 -- The New Forsaken (Undercity)
quests[249830068] = 25 -- Forsaken Duties (Undercity)
quests[249880068] = 250 -- The Chill of Death (Undercity)
quests[249890068] = 75 -- Return to the Magistrate (Undercity)
quests[249900068] = 250 -- Darkhound Pounding (Undercity)
quests[249910068] = 10 -- Garren's Haunt (Undercity)
quests[249920068] = 75 -- Escaped From Gilneas (Undercity)
quests[249930068] = 250 -- Annihilate the Worgen (Undercity)
quests[249940068] = 250 -- Doom Weed (Undercity)
quests[249950068] = 250 -- Off the Scales (Undercity)
quests[249960068] = 250 -- Holland's Experiment (Undercity)
quests[249970068] = 250 -- Graverobbers (Undercity)
quests[249980068] = 250 -- Maggot Eye (Undercity)
quests[249990068] = 250 -- Planting the Seed of Fear (Undercity)
quests[250000072] = 250 -- Teegan's People (Stormwind)
quests[250011133] = 250 -- Sandscraper (Bilgewater Cartel)
quests[250020072] = 250 -- Scavenged (Stormwind)
quests[250030068] = 250 -- The Family Crypt (Undercity)
quests[250040068] = 350 -- The Mills Overrun (Undercity)
quests[250050068] = 25 -- Speak with Sevren (Undercity)
quests[250060068] = 350 -- The Grasp Weakens (Undercity)
quests[250070068] = 25 -- East... Always to the East (Undercity)
quests[250080072] = 250 -- Filthy Animals (Stormwind)
quests[250090068] = 250 -- At War With The Scarlet Crusade (Undercity)
quests[250100068] = 75 -- A Deadly New Ally (Undercity)
quests[250110068] = 10 -- To Bigger and Better Things (Undercity)
quests[250120068] = 150 -- Take to the Skies (Undercity)
quests[250130068] = 250 -- A Little Oomph (Undercity)
quests[250150072] = 250 -- Make 'em Squeal (Stormwind)
quests[250220072] = 250 -- A Failure to Communicate (Stormwind)
quests[250231133] = 10 -- Old Friends (Bilgewater Cartel)
quests[250230076] = 10 -- Old Friends (Orgrimmar)
quests[250241133] = 250 -- Repel the Paratroopers (Bilgewater Cartel)
quests[250240076] = 250 -- Repel the Paratroopers (Orgrimmar)
quests[250270072] = 350 -- You Flicked a Fine Vine to Leaf Me (Stormwind)
quests[250280609] = 250 -- Trouble From the Ground Up (Cenarion Circle)
quests[250290068] = 250 -- Deaths in the Family (Undercity)
quests[250310068] = 25 -- Head for the Mills (Undercity)
quests[250340072] = 75 -- To the Front! (Stormwind)
quests[250350530] = 25 -- Breaking the Line (Darkspear Trolls)
quests[250360072] = 150 -- Teegan's Troubles (Stormwind)
quests[250370530] = 250 -- Crab Fishin' (Darkspear Trolls)
quests[250380068] = 75 -- Gordo's Task (Undercity)
quests[250390068] = 75 -- Have You Seen Anything Weird Out There? (Undercity)
quests[250410072] = 250 -- Lion's Pride (Stormwind)
quests[250420072] = 250 -- Marching On Our Stomachs (Stormwind)
quests[250430072] = 250 -- Fields of Blood (Stormwind)
quests[250440072] = 250 -- Diplomacy By Another Means (Stormwind)
quests[250450072] = 250 -- A Line in the Dirt (Stormwind)
quests[250460068] = 350 -- A Daughter's Embrace (Undercity)
quests[250480054] = 250 -- Seaside Salvage (Gnomeregan)
quests[250490054] = 250 -- Puddle Stomping (Gnomeregan)
quests[250500054] = 250 -- Rocket Rescue (Gnomeregan)
quests[250510076] = 250 -- Darkmist Extermination (Orgrimmar)
quests[250540054] = 250 -- Lootin' Plunder (Gnomeregan)
quests[250560068] = 250 -- Grisly Grizzlies (Undercity)
quests[250570072] = 250 -- Clap 'Em In Irons (Stormwind)
quests[250581133] = 250 -- Mine Disposal, the Goblin Way (Bilgewater Cartel)
quests[250590072] = 250 -- The Taurajo Briefs (Stormwind)
quests[250620054] = 250 -- What We Came For (Gnomeregan)
quests[250640530] = 10 -- Moraya (Darkspear Trolls)
quests[250650054] = 250 -- You Too, Brute? (Gnomeregan)
quests[250661133] = 350 -- The Pride of Kezan (Bilgewater Cartel)
quests[250720169] = 250 -- A Few Good Goblins (Steamwheedle Cartel)
quests[250730530] = 10 -- Sen'jin Village (Darkspear Trolls)
quests[250740072] = 75 -- Meet Me at Triumph (Stormwind)
quests[250750054] = 250 -- Pick-a-Part (Gnomeregan)
quests[250790054] = 250 -- Powder Play (Gnomeregan)
quests[250800054] = 250 -- Batteries Not Yet Included (Gnomeregan)
quests[250810072] = 350 -- Claim the Battlescar (Stormwind)
quests[250820072] = 250 -- Working the Bugs Out (Stormwind)
quests[250861135] = 75 -- Firestone Point (The Earthen Ring)
quests[250870047] = 75 -- Report to Twinbraid (Ironforge)
quests[250890068] = 75 -- Beyond the Graves (Undercity)
quests[250900068] = 150 -- A Putrid Task (Undercity)
quests[250931133] = 250 -- The Heads of the SI:7 (Bilgewater Cartel)
quests[250930076] = 250 -- The Heads of the SI:7 (Orgrimmar)
quests[250981133] = 10 -- The Warchief Wants You (Bilgewater Cartel)
quests[250980076] = 10 -- The Warchief Wants You (Orgrimmar)
quests[250991133] = 10 -- Borrow Bastia (Bilgewater Cartel)
quests[250990076] = 10 -- Borrow Bastia (Orgrimmar)
quests[251001133] = 10 -- Let's Ride (Bilgewater Cartel)
quests[251020047] = 250 -- No Bull (Ironforge)
quests[251040047] = 75 -- Digsite Status (Ironforge)
quests[251060047] = 250 -- Hands Off Me Dig! (Ironforge)
quests[251080047] = 250 -- Feegly the Exiled (Ironforge)
quests[251091133] = 75 -- The Gallywix Labor Mine (Bilgewater Cartel)
quests[251101133] = 250 -- Kaja'Cola Gives You IDEAS! (TM) (Bilgewater Cartel)
quests[251110369] = 250 -- Scavengers Scavenged (Gadgetzan)
quests[251120369] = 250 -- Butcherbot (Gadgetzan)
quests[251150369] = 250 -- Blisterpaw Butchery (Gadgetzan)
quests[251180047] = 250 -- Looking for Lurkers (Ironforge)
quests[251200047] = 10 -- Marley's Final Flight (Ironforge)
quests[251210054] = 75 -- Momentum (Gnomeregan)
quests[251221133] = 250 -- Morale Boost (Bilgewater Cartel)
quests[251231133] = 250 -- Throw It On the Ground! (Bilgewater Cartel)
quests[251251133] = 25 -- Light at the End of the Tunnel (Bilgewater Cartel)
quests[251260076] = 250 -- Cutting Teeth (Orgrimmar)
quests[251270076] = 250 -- Sting of the Scorpid (Orgrimmar)
quests[251280076] = 25 -- Hana'zua (Orgrimmar)
quests[251280530] = 25 -- Hana'zua (Darkspear Trolls)
quests[251290076] = 250 -- Sarkoth (Orgrimmar)
quests[251290530] = 250 -- Sarkoth (Darkspear Trolls)
quests[251300076] = 25 -- Back to the Den (Orgrimmar)
quests[251300530] = 25 -- Back to the Den (Darkspear Trolls)
quests[251310076] = 350 -- Vile Familiars (Orgrimmar)
quests[251320076] = 500 -- Burning Blade Medallion (Orgrimmar)
quests[251330076] = 75 -- Report to Sen'jin Village (Orgrimmar)
quests[251340076] = 350 -- Lazy Peons (Orgrimmar)
quests[251350076] = 350 -- Thazz'ril's Pick (Orgrimmar)
quests[251360076] = 500 -- Galgar's Cactus Apple Surprise (Orgrimmar)
quests[251380530] = 75 -- Glyphic Parchment (Darkspear Trolls)
quests[251390076] = 75 -- Steady Shot (Orgrimmar)
quests[251410076] = 75 -- Eviscerate (Orgrimmar)
quests[251430076] = 75 -- Primal Strike (Orgrimmar)
quests[251450076] = 75 -- Corruption (Orgrimmar)
quests[251470076] = 75 -- Charge (Orgrimmar)
quests[251490530] = 75 -- Frost Nova (Darkspear Trolls)
quests[251510047] = 250 -- He Needs Ale! (Ironforge)
quests[251520076] = 75 -- Your Place In The World (Orgrimmar)
quests[251530047] = 350 -- Bael'dun Rescue (Ironforge)
quests[251630047] = 25 -- Who Did This? (Ironforge)
quests[251641135] = 350 -- Backed Into a Corner (The Earthen Ring)
quests[251650076] = 250 -- Never Trust a Big Barb and a Smile (Orgrimmar)
quests[251650530] = 250 -- Never Trust a Big Barb and a Smile (Darkspear Trolls)
quests[251670076] = 250 -- Breaking the Chain (Orgrimmar)
quests[251670530] = 350 -- Breaking the Chain (Darkspear Trolls)
quests[251680076] = 250 -- Purge the Valley (Orgrimmar)
quests[251680530] = 350 -- Purge the Valley (Darkspear Trolls)
quests[251690076] = 350 -- The War of Northwatch Aggression (Orgrimmar)
quests[251690530] = 500 -- The War of Northwatch Aggression (Darkspear Trolls)
quests[251700076] = 250 -- Cleaning Up the Coastline (Orgrimmar)
quests[251700530] = 250 -- Cleaning Up the Coastline (Darkspear Trolls)
quests[251710076] = 150 -- Riding On (Orgrimmar)
quests[251710530] = 150 -- Riding On (Darkspear Trolls)
quests[251720076] = 250 -- Invaders in Our Home (Orgrimmar)
quests[251730076] = 250 -- From Bad to Worse (Orgrimmar)
quests[251730530] = 250 -- From Bad to Worse (Darkspear Trolls)
quests[251740047] = 75 -- Twinbraid Needs to Know (Ironforge)
quests[251750047] = 75 -- The Bad News (Ironforge)
quests[251760076] = 250 -- Exploiting the Situation (Orgrimmar)
quests[251761133] = 250 -- Exploiting the Situation (Bilgewater Cartel)
quests[251770076] = 250 -- Storming the Beaches (Orgrimmar)
quests[251770530] = 250 -- Storming the Beaches (Darkspear Trolls)
quests[251780076] = 250 -- Shipwreck Searching (Orgrimmar)
quests[251781133] = 250 -- Shipwreck Searching (Bilgewater Cartel)
quests[251790076] = 250 -- Loss Reduction (Orgrimmar)
quests[251790530] = 250 -- Loss Reduction (Darkspear Trolls)
quests[251820047] = 250 -- Kill 'em All! (Ironforge)
quests[251830047] = 350 -- ...And Bury the Rest (Ironforge)
quests[251841133] = 10 -- Wild Mine Cart Ride (Bilgewater Cartel)
quests[251850047] = 75 -- The Deed Is Done (Ironforge)
quests[251860072] = 350 -- Futile Resistance (Stormwind)
quests[251880076] = 350 -- Watershed Patrol (Orgrimmar)
quests[251890076] = 250 -- Spirits Be Praised (Orgrimmar)
quests[251900076] = 250 -- Raggaran's Rage (Orgrimmar)
quests[251910054] = 75 -- Survey the Destruction (Gnomeregan)
quests[251920076] = 250 -- Raggaran's Fury (Orgrimmar)
quests[251930076] = 250 -- Lost But Not Forgotten (Orgrimmar)
quests[251940530] = 250 -- Unbidden Visitors (Darkspear Trolls)
quests[251950530] = 350 -- That's the End of That Raptor (Darkspear Trolls)
quests[251960076] = 25 -- The Dranosh'ar Blockade (Orgrimmar)
quests[252001133] = 250 -- Shredder Shutdown (Bilgewater Cartel)
quests[252011133] = 250 -- The Ultimate Footbomb Uniform (Bilgewater Cartel)
quests[252021133] = 350 -- The Fastest Way to His Heart (Bilgewater Cartel)
quests[252031133] = 350 -- What Kind of Name is Chip, Anyway? (Bilgewater Cartel)
quests[252041133] = 250 -- Release the Valves (Bilgewater Cartel)
quests[252050076] = 250 -- The Wolf and The Kodo (Orgrimmar)
quests[252060076] = 350 -- Ignoring the Warnings (Orgrimmar)
quests[252071133] = 250 -- Good-bye, Sweet Oil (Bilgewater Cartel)
quests[252090076] = 250 -- The Gordunni Threat (Orgrimmar)
quests[252100081] = 250 -- Signs of Change (Thunder Bluff)
quests[252110047] = 250 -- Cleaning Hovel (Ironforge)
quests[252131133] = 10 -- The Slave Pits (Bilgewater Cartel)
quests[252141133] = 250 -- Escape Velocity (Bilgewater Cartel)
quests[252151135] = 250 -- A Distracting Scent (The Earthen Ring)
quests[252171135] = 250 -- Totem Modification (The Earthen Ring)
quests[252181135] = 250 -- Undersea Inflation (The Earthen Ring)
quests[252191135] = 250 -- Don't be Shellfish (The Earthen Ring)
quests[252201135] = 250 -- Slippery Threat (The Earthen Ring)
quests[252211135] = 250 -- Rundown (The Earthen Ring)
quests[252231158] = 250 -- Trial By Fire (Guardians of Hyjal)
quests[252241158] = 250 -- In Bloom (Guardians of Hyjal)
quests[252270076] = 10 -- Thonk (Orgrimmar)
quests[252271133] = 10 -- Thonk (Bilgewater Cartel)
quests[252300081] = 250 -- More Than Illness (Thunder Bluff)
quests[252320076] = 250 -- The Burning Blade (Orgrimmar)
quests[252320530] = 250 -- The Burning Blade (Darkspear Trolls)
quests[252331158] = 250 -- End of the Supply Line (Guardians of Hyjal)
quests[252341158] = 250 -- In the Rear With the Gear (Guardians of Hyjal)
quests[252350072] = 250 -- Vulgar Vul'Gol (Stormwind)
quests[252360076] = 250 -- Thunder Down Under (Orgrimmar)
quests[252370081] = 250 -- Tears of Stone (Thunder Bluff)
quests[252391156] = 250 -- Path of Might (The Ashen Verdict)
quests[252401156] = 250 -- Path of Might (The Ashen Verdict)
quests[252410081] = 250 -- The Land, Corrupted (Thunder Bluff)
quests[252421156] = 250 -- Path of Might (The Ashen Verdict)
quests[252431133] = 350 -- She Loves Me, She Loves Me NOT! (Bilgewater Cartel)
quests[252441133] = 350 -- What Kind of Name is Candy, Anyway? (Bilgewater Cartel)
quests[252500081] = 250 -- Sealing the Dream (Thunder Bluff)
quests[252511133] = 500 -- Final Confrontation (Bilgewater Cartel)
quests[252520076] = 250 -- Rulers of Dire Maul (Orgrimmar)
quests[252551158] = 250 -- Harrying the Hunters (Guardians of Hyjal)
quests[252560076] = 10 -- Sent for Help (Orgrimmar)
quests[252570076] = 150 -- Ghislania (Orgrimmar)
quests[252580076] = 250 -- Griswold Hanniston (Orgrimmar)
quests[252590076] = 350 -- Gaur Icehorn (Orgrimmar)
quests[252600076] = 250 -- Fizzled (Orgrimmar)
quests[252610076] = 10 -- Margoz (Orgrimmar)
quests[252610530] = 10 -- Margoz (Darkspear Trolls)
quests[252620076] = 250 -- Skull Rock (Orgrimmar)
quests[252620530] = 250 -- Skull Rock (Darkspear Trolls)
quests[252630076] = 75 -- Arnak Fireblade (Orgrimmar)
quests[252630530] = 75 -- Arnak Fireblade (Darkspear Trolls)
quests[252640076] = 75 -- Ak'Zeloth (Orgrimmar)
quests[252640530] = 75 -- Ak'Zeloth (Darkspear Trolls)
quests[252651133] = 25 -- Victory! (Bilgewater Cartel)
quests[252650076] = 25 -- Victory! (Orgrimmar)
quests[252661133] = 25 -- Warchief's Emissary (Bilgewater Cartel)
quests[252660076] = 25 -- Warchief's Emissary (Orgrimmar)
quests[252671133] = 350 -- Message for Vol'jin (Bilgewater Cartel)
quests[252670076] = 350 -- Message for Vol'jin (Orgrimmar)
quests[252681158] = 10 -- The Voice of Goldrinn (Guardians of Hyjal)
quests[252691158] = 10 -- The Voice of Lo'Gosh (Guardians of Hyjal)
quests[252701158] = 250 -- Howling Mad (Guardians of Hyjal)
quests[252711158] = 250 -- Goldrinn's Ferocity (Guardians of Hyjal)
quests[252721158] = 250 -- Lycanthoth the Corruptor (Guardians of Hyjal)
quests[252731158] = 250 -- Lycanthoth the Corruptor (Guardians of Hyjal)
quests[252741158] = 250 -- Signed in Blood (Guardians of Hyjal)
quests[252751133] = 10 -- Report to the Labor Captain (Bilgewater Cartel)
quests[252750076] = 10 -- Report to the Labor Captain (Orgrimmar)
quests[252761158] = 75 -- Your New Identity (Guardians of Hyjal)
quests[252771158] = 250 -- Cleaning House (Guardians of Hyjal)
quests[252781158] = 250 -- Cleaning House (Guardians of Hyjal)
quests[252791158] = 10 -- The Shrine Reclaimed (Guardians of Hyjal)
quests[252801158] = 10 -- The Shrine Reclaimed (Guardians of Hyjal)
quests[252811135] = 250 -- Pay It Forward (The Earthen Ring)
quests[252840076] = 250 -- Feeding the Fear (Orgrimmar)
quests[252911158] = 75 -- Twilight Training (Guardians of Hyjal)
quests[252920081] = 75 -- Next of Kin (Thunder Bluff)
quests[252941158] = 250 -- Walking the Dog (Guardians of Hyjal)
quests[252961158] = 250 -- Gather the Intelligence (Guardians of Hyjal)
quests[252971158] = 250 -- From the Mouth of Madness (Guardians of Hyjal)
quests[252981158] = 250 -- Free Your Mind, the Rest Follows (Guardians of Hyjal)
quests[252991158] = 250 -- Mental Training: Speaking the Truth to Power (Guardians of Hyjal)
quests[253001158] = 75 -- The Eye of Twilight (Guardians of Hyjal)
quests[253011158] = 75 -- Mastering Puppets (Guardians of Hyjal)
quests[253031158] = 250 -- Elementary! (Guardians of Hyjal)
quests[253081158] = 250 -- Seeds of Discord (Guardians of Hyjal)
quests[253091158] = 250 -- Spiritual Training: Mercy is for the Weak (Guardians of Hyjal)
quests[253101158] = 350 -- The Greater of Two Evils (Guardians of Hyjal)
quests[253111158] = 250 -- Twilight Territory (Guardians of Hyjal)
quests[253121158] = 75 -- Return to Duskwhisper (Guardians of Hyjal)
quests[253141158] = 250 -- Speech Writing for Dummies (Guardians of Hyjal)
quests[253151158] = 500 -- Graduation Speech (Guardians of Hyjal)
quests[253161158] = 150 -- As Hyjal Burns (Guardians of Hyjal)
quests[253171158] = 250 -- Protect the World Tree (Guardians of Hyjal)
quests[253191158] = 250 -- War on the Twilight's Hammer (Guardians of Hyjal)
quests[253201158] = 75 -- The Captured Scout (Guardians of Hyjal)
quests[253211158] = 150 -- Twilight Captivity (Guardians of Hyjal)
quests[253231158] = 250 -- Flamebreaker (Guardians of Hyjal)
quests[253241158] = 75 -- A Prisoner of Interest (Guardians of Hyjal)
quests[253251158] = 150 -- Through the Dream (Guardians of Hyjal)
quests[253281158] = 250 -- Gar'gol's Gotta Go (Guardians of Hyjal)
quests[253290076] = 350 -- Might of the Stonemaul (Orgrimmar)
quests[253301158] = 150 -- Waste of Flesh (Guardians of Hyjal)
quests[253321158] = 250 -- Get Me Outta Here! (Guardians of Hyjal)
quests[253331134] = 350 -- Might of the Sentinels (Gilneas)
quests[253341135] = 150 -- The Looming Threat (The Earthen Ring)
quests[253360530] = 250 -- Testing the Vessel (Darkspear Trolls)
quests[253370530] = 250 -- Hippogryph Muisek (Darkspear Trolls)
quests[253380530] = 250 -- Treant Muisek (Darkspear Trolls)
quests[253390081] = 150 -- Vengeance on the Northspring (Thunder Bluff)
quests[253400081] = 350 -- Dark Heart (Thunder Bluff)
quests[253410076] = 250 -- The Gordunni Orb (Orgrimmar)
quests[253450530] = 250 -- Faerie Dragon Muisek (Darkspear Trolls)
quests[253460530] = 250 -- Mountain Giant Muisek (Darkspear Trolls)
quests[253521158] = 250 -- Sweeping the Shelf (Guardians of Hyjal)
quests[253531158] = 250 -- Lightning in a Bottle (Guardians of Hyjal)
quests[253541158] = 250 -- Sweeping the Shelf (Guardians of Hyjal)
quests[253551158] = 250 -- Lightning in a Bottle (Guardians of Hyjal)
quests[253560081] = 25 -- To the Summit (Thunder Bluff)
quests[253570072] = 250 -- Buy Us Some Time (Stormwind)
quests[253601135] = 250 -- Crabby Patrons (The Earthen Ring)
quests[253610076] = 250 -- A New Cloak's Sheen (Orgrimmar)
quests[253620076] = 250 -- A Grim Discovery (Orgrimmar)
quests[253630081] = 150 -- War on the Woodpaw (Thunder Bluff)
quests[253640081] = 150 -- Alpha Strike (Thunder Bluff)
quests[253650081] = 75 -- Woodpaw Investigation (Thunder Bluff)
quests[253660081] = 250 -- The Battle Plans (Thunder Bluff)
quests[253670067] = 250 -- Zukk'ash Infestation (Horde)
quests[253690067] = 350 -- Stinglasher (Horde)
quests[253701158] = 250 -- Inciting the Elements (Guardians of Hyjal)
quests[253711135] = 350 -- The Abyssal Ride (The Earthen Ring)
quests[253721158] = 350 -- Aessina's Miracle (Guardians of Hyjal)
quests[253730081] = 150 -- The Hilltop Threat (Thunder Bluff)
quests[253740081] = 350 -- Sasquatch Sighting (Thunder Bluff)
quests[253750081] = 150 -- Taming The Tamers (Thunder Bluff)
quests[253811158] = 250 -- Fighting Fire With ... Anything (Guardians of Hyjal)
quests[253821158] = 250 -- Disrupting the Rituals (Guardians of Hyjal)
quests[253830081] = 350 -- Ysondre's Farewell (Thunder Bluff)
quests[253851158] = 250 -- Save the Wee Animals (Guardians of Hyjal)
quests[253860076] = 75 -- To Stonemaul Hold (Orgrimmar)
quests[253870081] = 150 -- To Camp Mojache (Thunder Bluff)
quests[253910530] = 250 -- Weapons of Spirit (Darkspear Trolls)
quests[253921158] = 250 -- Oh, Deer! (Guardians of Hyjal)
quests[253940069] = 250 -- More Than Illness (Darnassus)
quests[253950047] = 250 -- The Stolen Keg (Ironforge)
quests[253960069] = 250 -- Tears of Stone (Darnassus)
quests[253970069] = 250 -- The Land, Corrupted (Darnassus)
quests[253980069] = 350 -- Sealing the Dream (Darnassus)
quests[253990069] = 250 -- Hatecrest Forces (Darnassus)
quests[254000069] = 250 -- The Gordunni Threat (Darnassus)
quests[254011134] = 250 -- The Gordunni Orb (Gilneas)
quests[254041158] = 150 -- If You're Not Against Us... (Guardians of Hyjal)
quests[254060069] = 250 -- Gordok Guards (Darnassus)
quests[254070069] = 250 -- Forces of Nature: Wisps (Darnassus)
quests[254081158] = 250 -- Seeds of Their Demise (Guardians of Hyjal)
quests[254090069] = 250 -- Forces of Nature: Hippogryphs (Darnassus)
quests[254100069] = 250 -- Forces of Nature: Treants (Darnassus)
quests[254111158] = 150 -- A New Master (Guardians of Hyjal)
quests[254121158] = 250 -- The Name Never Spoken (Guardians of Hyjal)
quests[254200047] = 250 -- Ancient Obstacles (Ironforge)
quests[254210047] = 350 -- The Grand Tablet (Ironforge)
quests[254241158] = 150 -- Return to Alysra (Guardians of Hyjal)
quests[254260069] = 250 -- War on the Woodpaw (Darnassus)
quests[254271134] = 150 -- Alpha Strike (Gilneas)
quests[254281158] = 250 -- Black Heart of Flame (Guardians of Hyjal)
quests[254290069] = 250 -- Zukk'ash Infestation (Darnassus)
quests[254301158] = 75 -- Emerald Allies (Guardians of Hyjal)
quests[254310069] = 350 -- Stinglasher (Darnassus)
quests[254320069] = 150 -- It's Not "Ogre" Yet (Darnassus)
quests[254330069] = 350 -- Sasquatch Sighting (Darnassus)
quests[254340069] = 250 -- Taming The Tamers (Darnassus)
quests[254360069] = 250 -- Spiteful Sisters (Darnassus)
quests[254380069] = 350 -- Ysondre's Farewell (Darnassus)
quests[254391135] = 250 -- Vengeful Heart (The Earthen Ring)
quests[254401135] = 250 -- Fathom-Lord Zin'jatar (The Earthen Ring)
quests[254411135] = 250 -- Vortex (The Earthen Ring)
quests[254421135] = 250 -- A Pearl of Wisdom (The Earthen Ring)
quests[254470069] = 250 -- Signs of Change (Darnassus)
quests[254480069] = 150 -- The Northspring Menace (Darnassus)
quests[254490069] = 250 -- The Mark of Quality (Darnassus)
quests[254500069] = 250 -- Improved Quality (Darnassus)
quests[254510069] = 250 -- Pristine Yeti Hide (Darnassus)
quests[254520081] = 250 -- The Mark of Quality (Thunder Bluff)
quests[254530081] = 250 -- Improved Quality (Thunder Bluff)
quests[254540081] = 250 -- Perfect Yeti Hide (Thunder Bluff)
quests[254580069] = 350 -- General Skessesh (Darnassus)
quests[254601158] = 250 -- The Earth Rises (Guardians of Hyjal)
quests[254621158] = 250 -- The Bears Up There (Guardians of Hyjal)
quests[254641158] = 350 -- The Return of Baron Geddon (Guardians of Hyjal)
quests[254650169] = 350 -- Still With The Zapped Giants (Steamwheedle Cartel)
quests[254660169] = 350 -- Even More Fuel for the Zapping (Steamwheedle Cartel)
quests[254680069] = 250 -- Forces of Nature: Faerie Dragons (Darnassus)
quests[254690069] = 250 -- Forces of Nature: Mountain Giants (Darnassus)
quests[254711135] = 10 -- Across the Great Divide (The Earthen Ring)
quests[254721158] = 250 -- The Flameseer's Staff (Guardians of Hyjal)
quests[254731133] = 10 -- Kaja'Cola (Bilgewater Cartel)
quests[254750021] = 250 -- Find OOX-22/FE! (Booty Bay)
quests[254750087] = -250 -- Find OOX-22/FE! (Bloodsail Buccaneers)
quests[254760021] = 350 -- Rescue OOX-22/FE! (Booty Bay)
quests[254760087] = -350 -- Rescue OOX-22/FE! (Bloodsail Buccaneers)
quests[254771135] = 250 -- Better Late Than Dead (The Earthen Ring)
quests[254780081] = 25 -- To the Summit (Thunder Bluff)
quests[254790069] = 25 -- To New Thalanaar (Darnassus)
quests[254810069] = 25 -- To New Thalanaar (Darnassus)
quests[254860069] = 250 -- The Grimtotem are Coming (Darnassus)
quests[254870081] = 250 -- The Grimtotem are Coming (Thunder Bluff)
quests[254880054] = 25 -- Two If By Boat (Gnomeregan)
quests[254891133] = 25 -- Two If By Boat (Bilgewater Cartel)
quests[254901158] = 250 -- Smashing Through Ashes (Guardians of Hyjal)
quests[254911158] = 250 -- Durable Seeds (Guardians of Hyjal)
quests[254921158] = 250 -- Firebreak (Guardians of Hyjal)
quests[254931158] = 250 -- Fresh Bait (Guardians of Hyjal)
quests[254941158] = 250 -- A Champion's Collar (Guardians of Hyjal)
quests[254961158] = 350 -- Grudge Match (Guardians of Hyjal)
quests[254971135] = 250 -- Back in the Saddle (The Earthen Ring)
quests[254981135] = 250 -- Shark Bait (The Earthen Ring)
quests[254991158] = 250 -- Agility Training: Run Like Hell! (Guardians of Hyjal)
quests[255021158] = 250 -- Prepping the Soil (Guardians of Hyjal)
quests[255040054] = 25 -- Do Me a Favor? (Gnomeregan)
quests[255051133] = 25 -- Do Me a Favor? (Bilgewater Cartel)
quests[255071158] = 250 -- Hell's Shells (Guardians of Hyjal)
quests[255091158] = 250 -- Physical Training: Forced Labor (Guardians of Hyjal)
quests[255141158] = 250 -- Breaking the Bonds (Guardians of Hyjal)
quests[255150054] = 250 -- Down in the Deeps (Gnomeregan)
quests[255161133] = 250 -- Down in the Deeps (Bilgewater Cartel)
quests[255170054] = 250 -- Bar Fight! (Gnomeregan)
quests[255181133] = 250 -- Bar Fight! (Bilgewater Cartel)
quests[255191158] = 250 -- Children of Tortolla (Guardians of Hyjal)
quests[255201158] = 10 -- An Ancient Awakens (Guardians of Hyjal)
quests[255210369] = 250 -- I'm With Scorpid (Gadgetzan)
quests[255220369] = 250 -- Gargantapid (Gadgetzan)
quests[255231158] = 250 -- Flight in the Firelands (Guardians of Hyjal)
quests[255240054] = 250 -- In the Outhouse (Gnomeregan)
quests[255251158] = 250 -- Wave One (Guardians of Hyjal)
quests[255261133] = 250 -- In the Outhouse (Bilgewater Cartel)
quests[255311158] = 250 -- Twilight Riot (Guardians of Hyjal)
quests[255320054] = 250 -- Pirate Accuracy Increasing (Gnomeregan)
quests[255331133] = 250 -- Pirate Accuracy Increasing (Bilgewater Cartel)
quests[255341133] = 250 -- Going Off-Task (Bilgewater Cartel)
quests[255401135] = 250 -- Bellies Await (The Earthen Ring)
quests[255411133] = 250 -- Filling Our Pockets (Bilgewater Cartel)
quests[255420054] = 250 -- Circle the Wagons... er, Boats (Gnomeregan)
quests[255431133] = 250 -- Circle the Wagons... er, Boats (Bilgewater Cartel)
quests[255441158] = 250 -- Wave Two (Guardians of Hyjal)
quests[255450072] = 250 -- To Arms! (Stormwind)
quests[255460072] = 250 -- Traveling on Our Stomachs (Stormwind)
quests[255470072] = 250 -- On Our Own Terms (Stormwind)
quests[255481158] = 250 -- Might of the Firelord (Guardians of Hyjal)
quests[255491158] = 250 -- The Sanctum of the Prophets (Guardians of Hyjal)
quests[255501158] = 350 -- Magma Monarch (Guardians of Hyjal)
quests[255511158] = 250 -- The Firelord (Guardians of Hyjal)
quests[255521158] = 250 -- Brood of Evil (Guardians of Hyjal)
quests[255531158] = 350 -- Death to the Broodmother (Guardians of Hyjal)
quests[255541158] = 250 -- Secrets of the Flame (Guardians of Hyjal)
quests[255551158] = 350 -- The Gatekeeper (Guardians of Hyjal)
quests[255580072] = 250 -- All or Nothing (Stormwind)
quests[255590047] = 250 -- Laying Claim (Ironforge)
quests[255601158] = 250 -- Egg Wave (Guardians of Hyjal)
quests[255610054] = 250 -- Circle the Wagons... er, Boats (Gnomeregan)
quests[255621133] = 250 -- Circle the Wagons... er, Boats (Bilgewater Cartel)
quests[255640072] = 250 -- Stormwind Elite Aquatic and Land Forces (Stormwind)
quests[255650047] = 250 -- Fragments of Language (Ironforge)
quests[255741158] = 250 -- Flames from Above (Guardians of Hyjal)
quests[255751158] = 250 -- Forged of Shadow and Flame (Guardians of Hyjal)
quests[255761158] = 250 -- Rage of the Wolf Ancient (Guardians of Hyjal)
quests[255771158] = 250 -- Crushing the Cores (Guardians of Hyjal)
quests[255781158] = 250 -- Return to Nordrassil (Guardians of Hyjal)
quests[255841158] = 75 -- The Return of the Ancients (Guardians of Hyjal)
quests[255850054] = 250 -- Quiet the Cannons (Gnomeregan)
quests[255861133] = 250 -- Quiet the Cannons (Bilgewater Cartel)
quests[255880054] = 250 -- A Little Payback (Gnomeregan)
quests[255891133] = 250 -- A Little Payback (Bilgewater Cartel)
quests[255900054] = 250 -- Where's Wizzle? (Gnomeregan)
quests[255951135] = 250 -- Something Edible (The Earthen Ring)
quests[255961133] = 250 -- Where's Synge? (Bilgewater Cartel)
quests[255971158] = 150 -- Commander Jarod Shadowsong (Guardians of Hyjal)
quests[255991158] = 250 -- Cindermaul, the Portal Master (Guardians of Hyjal)
quests[256001158] = 350 -- Forgemaster Pyrendius (Guardians of Hyjal)
quests[256011158] = 75 -- Head of the Class (Guardians of Hyjal)
quests[256070069] = 250 -- Ze Gnomecorder (Darnassus)
quests[256070054] = 250 -- Ze Gnomecorder (Gnomeregan)
quests[256081158] = 350 -- Slash and Burn (Guardians of Hyjal)
quests[256090054] = 250 -- Sunken Treasure (Gnomeregan)
quests[256101133] = 250 -- Sunken Treasure (Bilgewater Cartel)
quests[256111158] = 25 -- Return from the Firelands (Guardians of Hyjal)
quests[256121158] = 25 -- Return from the Firelands (Guardians of Hyjal)
quests[256130069] = 10 -- Do Yourself a Favor (Darnassus)
quests[256130054] = 10 -- Do Yourself a Favor (Gnomeregan)
quests[256140069] = 250 -- The Only Way Down is in a Body Bag (Darnassus)
quests[256140054] = 250 -- The Only Way Down is in a Body Bag (Gnomeregan)
quests[256150069] = 10 -- Return to Stardust (Darnassus)
quests[256150054] = 10 -- Return to Stardust (Gnomeregan)
quests[256160069] = 250 -- Hellscream's Legacy (Darnassus)
quests[256160054] = 250 -- Hellscream's Legacy (Gnomeregan)
quests[256171158] = 25 -- Into the Maw! (Guardians of Hyjal)
quests[256181158] = 25 -- Into the Maw! (Guardians of Hyjal)
quests[256210069] = 25 -- Field Test: Gnomecorder (Darnassus)
quests[256210054] = 25 -- Field Test: Gnomecorder (Gnomeregan)
quests[256220069] = 250 -- Burn, Baby, Burn! (Darnassus)
quests[256220054] = 250 -- Burn, Baby, Burn! (Gnomeregan)
quests[256231158] = 25 -- Into the Maw! (Guardians of Hyjal)
quests[256241158] = 25 -- Into the Maw! (Guardians of Hyjal)
quests[256261135] = 500 -- Visions of the Past: Rise from the Deep (The Earthen Ring)
quests[256270054] = 350 -- Two-Tusk Takedown (Gnomeregan)
quests[256281133] = 350 -- Two-Tusk Takedown (Bilgewater Cartel)
quests[256301158] = 25 -- The Fires of Mount Hyjal (Guardians of Hyjal)
quests[256400069] = 10 -- Bombs Away: Windshear Mine! (Darnassus)
quests[256400054] = 10 -- Bombs Away: Windshear Mine! (Gnomeregan)
quests[256410530] = 75 -- The Flow of Muisek (Darkspear Trolls)
quests[256420069] = 250 -- Don't Look Them in the Eyes (Darnassus)
quests[256420054] = 250 -- Don't Look Them in the Eyes (Gnomeregan)
quests[256441158] = 250 -- The Twilight Egg (Guardians of Hyjal)
quests[256450081] = 250 -- Return to Sage Palerunner (Thunder Bluff)
quests[256460069] = 250 -- Windshear Mine Cleanup (Darnassus)
quests[256460054] = 250 -- Windshear Mine Cleanup (Gnomeregan)
quests[256470069] = 150 -- Illegible Orc Letter (Darnassus)
quests[256470054] = 150 -- Illegible Orc Letter (Gnomeregan)
quests[256480076] = 10 -- Beyond Durotar (Orgrimmar)
quests[256490069] = 150 -- Minx'll Fix It (Darnassus)
quests[256490054] = 150 -- Minx'll Fix It (Gnomeregan)
quests[256500069] = 150 -- Orders from High Command (Darnassus)
quests[256500054] = 150 -- Orders from High Command (Gnomeregan)
quests[256520069] = 350 -- Commandeer That Balloon! (Darnassus)
quests[256520054] = 350 -- Commandeer That Balloon! (Gnomeregan)
quests[256531158] = 10 -- The Ancients are With Us (Guardians of Hyjal)
quests[256540069] = 350 -- Dark Heart (Darnassus)
quests[256551158] = 250 -- The Wormwing Problem (Guardians of Hyjal)
quests[256561158] = 250 -- Scrambling for Eggs (Guardians of Hyjal)
quests[256620069] = 250 -- Free Our Sisters (Darnassus)
quests[256631158] = 75 -- An Offering for Aviana (Guardians of Hyjal)
quests[256641158] = 250 -- A Prayer and a Wing (Guardians of Hyjal)
quests[256651158] = 10 -- A Plea From Beyond (Guardians of Hyjal)
quests[256670047] = 250 -- Culling the Wendigos (Ironforge)
quests[256670054] = 250 -- Culling the Wendigos (Gnomeregan)
quests[256680047] = 250 -- Pilfered Supplies (Ironforge)
quests[256680054] = 250 -- Pilfered Supplies (Gnomeregan)
quests[256690069] = 350 -- Rumble in the Lumber... Mill (Darnassus)
quests[256710069] = 350 -- Thinning the Horde (Darnassus)
quests[256730069] = 350 -- Just Ask Alice (Darnassus)
quests[256740076] = 250 -- Futile Pride (Orgrimmar)
quests[256750076] = 250 -- Ogre Combat (Orgrimmar)
quests[256760076] = 75 -- Heartstrike (Orgrimmar)
quests[256770076] = 250 -- It's All Mine (Orgrimmar)
quests[256780076] = 250 -- Pick Your Fate (Orgrimmar)
quests[256790076] = 25 -- Into the Mountain (Orgrimmar)
quests[256800076] = 250 -- That's Not Us (Orgrimmar)
quests[256800530] = 250 -- That's Not Us (Darkspear Trolls)
quests[256810076] = 350 -- Some People Just Need Killing (Orgrimmar)
quests[256810530] = 350 -- Some People Just Need Killing (Darkspear Trolls)
quests[256820076] = 350 -- Mission Complete (Orgrimmar)
quests[256830076] = 250 -- The Dreadmaul Furnace (Orgrimmar)
quests[256830911] = 250 -- The Dreadmaul Furnace (Silvermoon City)
quests[256840911] = 25 -- Attune the Bloodstone (Silvermoon City)
quests[256850911] = 250 -- The First Step (Silvermoon City)
quests[256860911] = 25 -- Blood Ritual (Silvermoon City)
quests[256870911] = 250 -- Not Just Any Body (Silvermoon City)
quests[256880911] = 500 -- The Altar of Storms (Silvermoon City)
quests[256891133] = 25 -- Time is Short (Bilgewater Cartel)
quests[256900911] = 250 -- A Bloodmage's Gotta Eat Too (Silvermoon City)
quests[256911133] = 250 -- The Charred Granite of the Dark Portal (Bilgewater Cartel)
quests[256921133] = 250 -- The Vile Blood of Demons (Bilgewater Cartel)
quests[256931133] = 75 -- Enhancing the Stone (Bilgewater Cartel)
quests[256940076] = 350 -- Protecting Our Rear (Orgrimmar)
quests[256950076] = 250 -- Watching Our Back (Orgrimmar)
quests[256960911] = 75 -- The Sunveil Excursion (Silvermoon City)
quests[256970911] = 250 -- The Amulet of Allistarj (Silvermoon City)
quests[256980911] = 250 -- The Amulet of Sevine (Silvermoon City)
quests[256990911] = 250 -- The Amulet of Grol (Silvermoon City)
quests[257010067] = 1000 -- You Are Rakh'likh, Demon (Horde)
quests[257080469] = 350 -- Our Fallen Friends (Alliance)
quests[257090469] = 250 -- Curtail the Darktail (Alliance)
quests[257100469] = 150 -- Minor Distractions (Alliance)
quests[257110469] = 350 -- Eliminate the Okril'lon (Alliance)
quests[257120469] = 250 -- Nethergarde Reigns (Alliance)
quests[257130469] = 250 -- Remove Their Arms (Alliance)
quests[257140469] = 10 -- Watcher Mahar Ba (Alliance)
quests[257150469] = 250 -- A Closer Look (Alliance)
quests[257160469] = 250 -- Cultists at our Doorstep (Alliance)
quests[257170911] = 350 -- The Demons and the Druid (Silvermoon City)
quests[257180911] = 10 -- How Best to Proceed (Silvermoon City)
quests[257190911] = 350 -- Demoniac Vessel (Silvermoon City)
quests[257200911] = 500 -- The Downfall of Marl Wormthorn (Silvermoon City)
quests[257210047] = 250 -- Fight the Flood (Ironforge)
quests[257220047] = 250 -- Sedimentary, My Dear (Ironforge)
quests[257230047] = 250 -- Thresh Out of Luck (Ironforge)
quests[257240047] = 250 -- Frostmane Aggression (Ironforge)
quests[257240054] = 250 -- Frostmane Aggression (Gnomeregan)
quests[257250047] = 250 -- Fenbush Berries (Ironforge)
quests[257260047] = 250 -- A Dumpy Job (Ironforge)
quests[257270047] = 250 -- Drungeld Glowerglare (Ironforge)
quests[257280069] = 500 -- Mr. P's Wild Ride (Darnassus)
quests[257290054] = 250 -- Gerenzo the Traitor (Gnomeregan)
quests[257300047] = 500 -- BEWARE OF CRAGJAW! (Ironforge)
quests[257311158] = 150 -- A Bird in Hand (Guardians of Hyjal)
quests[257330047] = 250 -- Get Out Of Here, Stalkers (Ironforge)
quests[257340047] = 250 -- Down In Thelgen Rock (Ironforge)
quests[257350047] = 250 -- Incendicite Ore (Ironforge)
quests[257360047] = 250 -- The Floodsurge Core (Ironforge)
quests[257390069] = 250 -- If the Horde Don't Get You... (Darnassus)
quests[257390072] = 250 -- If the Horde Don't Get You... (Stormwind)
quests[257401158] = 10 -- Fact-Finding Mission (Guardians of Hyjal)
quests[257410069] = 250 -- Preparations for the Future (Darnassus)
quests[257440054] = 25 -- Negotiations (Gnomeregan)
quests[257451133] = 25 -- Negotiations (Bilgewater Cartel)
quests[257461158] = 250 -- Sethria's Brood (Guardians of Hyjal)
quests[257551135] = 500 -- Visions of the Past: The Slaughter of Biel'aran Ridge (The Earthen Ring)
quests[257560054] = 250 -- Get Zherin! (Gnomeregan)
quests[257571133] = 250 -- Get Koalbeard! (Bilgewater Cartel)
quests[257581158] = 250 -- A Gap in Their Armor (Guardians of Hyjal)
quests[257601135] = 500 -- Visions of the Past: The Invasion of Vashj'ir (The Earthen Ring)
quests[257611158] = 250 -- Disassembly (Guardians of Hyjal)
quests[257631158] = 10 -- The Codex of Shadows (Guardians of Hyjal)
quests[257641158] = 250 -- Egg Hunt (Guardians of Hyjal)
quests[257650072] = 75 -- Tell 'Em Koko Sent You (Stormwind)
quests[257660069] = 500 -- Arcane Legacy (Darnassus)
quests[257670069] = 250 -- Capturing Memories (Darnassus)
quests[257680069] = 25 -- Bombs Away: Mirkfallon Post! (Darnassus)
quests[257690069] = 10 -- Fallowmere Beckons (Darnassus)
quests[257700047] = 75 -- Keg Run (Ironforge)
quests[257740054] = 250 -- Fool's Gold (Gnomeregan)
quests[257751133] = 250 -- Fool's Gold (Bilgewater Cartel)
quests[257761158] = 250 -- Sethria's Demise (Guardians of Hyjal)
quests[257770047] = 75 -- Onwards to Menethil (Ironforge)
quests[257800047] = 250 -- Assault on Menethil Keep (Ironforge)
quests[257900054] = 250 -- Back to Crazzle (Gnomeregan)
quests[257911133] = 250 -- Back to Riznek (Bilgewater Cartel)
quests[257920047] = 250 -- Pushing Forward (Ironforge)
quests[257920054] = 250 -- Pushing Forward (Gnomeregan)
quests[257930072] = 10 -- The Deep Reaches (Stormwind)
quests[257941135] = 150 -- Undersea Sanctuary (The Earthen Ring)
quests[257951158] = 10 -- Return to the Shrine (Guardians of Hyjal)
quests[257960054] = 250 -- Eminent Domain (Gnomeregan)
quests[257971133] = 250 -- Eminent Domain (Bilgewater Cartel)
quests[257980054] = 250 -- Defend the Drill (Gnomeregan)
quests[257991133] = 250 -- Defend the Drill (Bilgewater Cartel)
quests[258000072] = 250 -- When Life Gives You Crabs (Stormwind)
quests[258010072] = 250 -- Claws from the Deep (Stormwind)
quests[258020072] = 75 -- Reclaiming Goods (Stormwind)
quests[258030072] = 25 -- The Search Continues (Stormwind)
quests[258040072] = 75 -- Search More Hovels (Stormwind)
quests[258050072] = 350 -- Return the Statuette (Stormwind)
quests[258060072] = 250 -- They Put the Assass in... Never Mind (Stormwind)
quests[258071158] = 10 -- An Ancient Reborn (Guardians of Hyjal)
quests[258080072] = 250 -- Is This Thing On? (Stormwind)
quests[258090072] = 350 -- Leave No Man Behind!  (Stormwind)
quests[258101158] = 10 -- The Hatchery Must Burn (Guardians of Hyjal)
quests[258110072] = 350 -- Shuttin Her Down (Stormwind)
quests[258130054] = 350 -- Go Blow that Horn (Gnomeregan)
quests[258141133] = 350 -- Go Blow that Horn (Bilgewater Cartel)
quests[258170072] = 250 -- The Cursed Crew (Stormwind)
quests[258180072] = 75 -- Lifting the Curse (Stormwind)
quests[258190072] = 250 -- The Eye of Paleth (Stormwind)
quests[258210072] = 10 -- A Special Kind of Job (Stormwind)
quests[258220072] = 250 -- Armaments for War (Stormwind)
quests[258230072] = 250 -- The Unrelenting Horde (Stormwind)
quests[258250054] = 25 -- Deliver the Goods (Gnomeregan)
quests[258261133] = 25 -- Deliver the Goods (Bilgewater Cartel)
quests[258301158] = 10 -- The Last Living Lorekeeper (Guardians of Hyjal)
quests[258321158] = 10 -- Return to Aviana (Guardians of Hyjal)
quests[258340072] = 10 -- Death by Proxy (Stormwind)
quests[258350069] = 25 -- Free Freewind Post (Darnassus)
quests[258360081] = 25 -- Free Freewind Post (Thunder Bluff)
quests[258370072] = 350 -- A Proper Peace Offerin' (Stormwind)
quests[258380047] = 10 -- Help from Steelgrill's Depot (Ironforge)
quests[258380054] = 10 -- Help from Steelgrill's Depot (Gnomeregan)
quests[258390047] = 150 -- The Ultrasafe Personnel Launcher (Ironforge)
quests[258390054] = 150 -- The Ultrasafe Personnel Launcher (Gnomeregan)
quests[258400047] = 250 -- Eliminate the Resistance (Ironforge)
quests[258400054] = 250 -- Eliminate the Resistance (Gnomeregan)
quests[258410047] = 250 -- Strike From Above (Ironforge)
quests[258410054] = 250 -- Strike From Above (Gnomeregan)
quests[258421158] = 250 -- Firefight (Guardians of Hyjal)
quests[258431158] = 10 -- Tortolla's Revenge (Guardians of Hyjal)
quests[258440072] = 75 -- Sating the Savage Beast (Stormwind)
quests[258450072] = 10 -- Terms of Service (Stormwind)
quests[258460072] = 75 -- Grundig Darkcloud, Chieftain of the Grimtotem (Stormwind)
quests[258490047] = 250 -- When Archaeology Attacks (Ironforge)
quests[258500047] = 250 -- Strike the Earth! (Ironforge)
quests[258510072] = 500 -- Dances with Grimtotem (Stormwind)
quests[258520072] = 10 -- All's Clear (Stormwind)
quests[258530047] = 250 -- Tooling Around (Ironforge)
quests[258540054] = 250 -- I'll Call Him Bitey (Gnomeregan)
quests[258550054] = 250 -- Gizmos and Gadgets (Gnomeregan)
quests[258560072] = 250 -- Crocolisk Hides (Stormwind)
quests[258570072] = 350 -- Hunting Horrorjaw (Stormwind)
quests[258640047] = 250 -- Dinosaur Crisis (Ironforge)
quests[258650047] = 250 -- The Mosshide Job (Ironforge)
quests[258660047] = 250 -- Dark Iron Trappers (Ironforge)
quests[258670047] = 250 -- Gnoll Escape (Ironforge)
quests[258680047] = 350 -- Yorla Darksnare (Ironforge)
quests[258690069] = 250 -- Grimtotem in the Post (Darnassus)
quests[258700081] = 250 -- Grimtotem in the Post (Thunder Bluff)
quests[258710069] = 250 -- Save the Sentinel (Darnassus)
quests[258720081] = 250 -- The Brave and the Bold (Thunder Bluff)
quests[258730069] = 350 -- Horn of the Traitor (Darnassus)
quests[258740081] = 350 -- Horn of the Traitor (Thunder Bluff)
quests[258750069] = 75 -- Gnome on the Inside (Darnassus)
quests[258760069] = 250 -- It's Up There! (Darnassus)
quests[258770069] = 75 -- No Time for Goodbyes! (Darnassus)
quests[258780072] = 250 -- Schemin' That Sabotage (Stormwind)
quests[258780069] = 250 -- Schemin' That Sabotage (Darnassus)
quests[258790072] = 250 -- The Lumbering Oaf Problem (Stormwind)
quests[258790069] = 250 -- The Lumbering Oaf Problem (Darnassus)
quests[258800069] = 10 -- Warn Master Thal'darah (Darnassus)
quests[258811158] = 250 -- Lost Wardens (Guardians of Hyjal)
quests[258820047] = 10 -- A Hand at the Ranch (Ironforge)
quests[258841135] = 250 -- Come Hell or High Water (The Earthen Ring)
quests[258861158] = 350 -- Pressing the Advantage (Guardians of Hyjal)
quests[258871135] = 250 -- Wake of Destruction (The Earthen Ring)
quests[258881135] = 250 -- Decompression (The Earthen Ring)
quests[258890069] = 350 -- Save the Children! (Darnassus)
quests[258901135] = 10 -- Nespirah (The Earthen Ring)
quests[258910069] = 250 -- Last Ditch Effort (Darnassus)
quests[258991158] = 250 -- Breakthrough (Guardians of Hyjal)
quests[259001135] = 75 -- Making Contact (The Earthen Ring)
quests[259011158] = 250 -- Hyjal Recycling Program (Guardians of Hyjal)
quests[259041158] = 250 -- The Hammer and the Key (Guardians of Hyjal)
quests[259050047] = 250 -- Rams on the Lam (Ironforge)
quests[259050054] = 250 -- Rams on the Lam (Gnomeregan)
quests[259061158] = 10 -- The Third Flamegate (Guardians of Hyjal)
quests[259071135] = 250 -- Slave Labor (The Earthen Ring)
quests[259081135] = 250 -- Stick it to Them (The Earthen Ring)
quests[259091135] = 150 -- Capture the Crab (The Earthen Ring)
quests[259101158] = 250 -- The Time for Mercy has Passed (Guardians of Hyjal)
quests[259120069] = 250 -- Seen Better Days (Darnassus)
quests[259130069] = 350 -- Putting Them to Rest (Darnassus)
quests[259140069] = 250 -- Back to the Depths! (Darnassus)
quests[259151158] = 250 -- The Strength of Tortolla (Guardians of Hyjal)
quests[259161135] = 75 -- Breaking Through (The Earthen Ring)
quests[259170469] = 75 -- We Are Not Alone (Alliance)
quests[259180469] = 250 -- Hopelessly Gearless (Alliance)
quests[259190469] = 250 -- Body Blows (Alliance)
quests[259200469] = 250 -- Still Valuable (Alliance)
quests[259210469] = 500 -- Overseer Idra'kess (Alliance)
quests[259211135] = 500 -- Overseer Idra'kess (The Earthen Ring)
quests[259220469] = 150 -- Waking the Beast (Alliance)
quests[259221135] = 150 -- Waking the Beast (The Earthen Ring)
quests[259231158] = 250 -- Finish Nemesis (Guardians of Hyjal)
quests[259241135] = 250 -- Call of Duty (The Earthen Ring)
quests[259250069] = 500 -- Thal'darah's Vengeance (Darnassus)
quests[259260069] = 250 -- Mired in Hatred (Darnassus)
quests[259270069] = 250 -- The Threat of Flame (Darnassus)
quests[259281158] = 10 -- Tortolla's Triumph (Guardians of Hyjal)
quests[259291135] = 250 -- Sea Legs (The Earthen Ring)
quests[259301134] = 250 -- Ascending the Vale (Gilneas)
quests[259311134] = 250 -- Brood of Seldarria (Gilneas)
quests[259320047] = 250 -- It's Raid Night Every Night (Ironforge)
quests[259320054] = 250 -- It's Raid Night Every Night (Gnomeregan)
quests[259330047] = 10 -- Help for the Quarry (Ironforge)
quests[259330054] = 10 -- Help for the Quarry (Gnomeregan)
quests[259341134] = 250 -- World First: Gnomegen (Gilneas)
quests[259340054] = 250 -- World First: Gnomegen (Gnomeregan)
quests[259351134] = 250 -- Hungry Pups (Gilneas)
quests[259361135] = 250 -- Pay It Forward (The Earthen Ring)
quests[259370047] = 250 -- Priceless Treasures (Ironforge)
quests[259370054] = 250 -- Priceless Treasures (Gnomeregan)
quests[259381134] = 25 -- Help for Desolace (Gilneas)
quests[259380072] = 25 -- Help for Desolace (Stormwind)
quests[259390069] = 250 -- For Peat's Sake (Darnassus)
quests[259401158] = 10 -- Last Stand at Whistling Grove (Guardians of Hyjal)
quests[259411135] = 10 -- Rest For the Weary (The Earthen Ring)
quests[259420076] = 250 -- Buy Us Some Time (Orgrimmar)
quests[259430076] = 250 -- Traveling on Our Stomachs (Orgrimmar)
quests[259440076] = 250 -- Girding Our Loins (Orgrimmar)
quests[259450076] = 10 -- We're Here to Do One Thing, Maybe Two... (Orgrimmar)
quests[259460076] = 250 -- Helm's Deep (Orgrimmar)
quests[259470076] = 250 -- Finders, Keepers (Orgrimmar)
quests[259480076] = 250 -- Bring It On! (Orgrimmar)
quests[259571135] = 500 -- Visions of the Past: The Invasion of Vashj'ir (The Earthen Ring)
quests[259661135] = 500 -- Visions of the Past: The Slaughter of Biel'aran Ridge (The Earthen Ring)
quests[259780047] = 250 -- Entombed in Ice (Ironforge)
quests[259780054] = 250 -- Entombed in Ice (Gnomeregan)
quests[259790047] = 250 -- Dealing with the Surge (Ironforge)
quests[259851158] = 10 -- Wings Over Mount Hyjal (Guardians of Hyjal)
quests[259860047] = 250 -- Trouble at the Lake (Ironforge)
quests[259860054] = 250 -- Trouble at the Lake (Gnomeregan)
quests[259891135] = 150 -- Capture the Crab (The Earthen Ring)
quests[259901135] = 75 -- Breaking Through (The Earthen Ring)
quests[259910067] = 75 -- We Are Not Alone (Horde)
quests[259920067] = 250 -- Hopelessly Gearless (Horde)
quests[259930067] = 250 -- Body Blows (Horde)
quests[259940067] = 250 -- Still Valuable (Horde)
quests[259950067] = 500 -- Overseer Idra'kess (Horde)
quests[259951135] = 500 -- Overseer Idra'kess (The Earthen Ring)
quests[259960067] = 150 -- Waking the Beast (Horde)
quests[259961135] = 150 -- Waking the Beast (The Earthen Ring)
quests[259970047] = 250 -- Dark Iron Scheming (Ironforge)
quests[259970054] = 250 -- Dark Iron Scheming (Gnomeregan)
quests[259980047] = 10 -- Get to the Airfield (Ironforge)
quests[259990076] = 250 -- Barrier to Entry (Orgrimmar)
quests[260010076] = 250 -- The Missing Blastgineer (Orgrimmar)
quests[260020076] = 250 -- Alliance Attack Plans (Orgrimmar)
quests[260030076] = 250 -- Lessons from the Lost Isles (Orgrimmar)
quests[260040076] = 350 -- Krom'gar Fortress (Orgrimmar)
quests[260090076] = 350 -- Seek and Destroy (Orgrimmar)
quests[260100076] = 250 -- Ashes to Ashes (Orgrimmar)
quests[260110076] = 250 -- Enemy of the Horde: Marshal Paltrow (Orgrimmar)
quests[260140469] = 10 -- The Brothers Digsong (Alliance)
quests[260150469] = 250 -- Phosphora Hunting (Alliance)
quests[260160076] = 250 -- Orders from Base Camp (Orgrimmar)
quests[260170469] = 250 -- A Lure (Alliance)
quests[260180469] = 250 -- Coldlights Out (Alliance)
quests[260190469] = 75 -- Enormous Eel Egg (Alliance)
quests[260200076] = 10 -- Report to Bombgutz (Orgrimmar)
quests[260210469] = 350 -- The Brothers Digsong 2: Eel-Egg-Trick Boogaloo (Alliance)
quests[260240072] = 350 -- Wanted! Marez Cowl (Stormwind)
quests[260260076] = 350 -- Dream of a Better Tomorrow (Orgrimmar)
quests[260270068] = 250 -- Call to Arms (Undercity)
quests[260280076] = 500 -- Mr. D's Wild Ride (Orgrimmar)
quests[260290068] = 350 -- The Real Threat (Undercity)
quests[260300068] = 250 -- Sigil of Strom (Undercity)
quests[260310068] = 250 -- Sigil of Thoradin (Undercity)
quests[260320068] = 250 -- Sigil of Arathor (Undercity)
quests[260330068] = 75 -- Trol'kalar (Undercity)
quests[260350047] = 350 -- Worth Its Weight in Gold (Ironforge)
quests[260360047] = 250 -- Wand over Fist (Ironforge)
quests[260370072] = 250 -- Trelane's Defenses (Stormwind)
quests[260380047] = 350 -- Attack on the Tower (Ironforge)
quests[260430076] = 500 -- BEWARE OF CRAGJAW! (Orgrimmar)
quests[260440076] = 25 -- Where Are the Parts? (Orgrimmar)
quests[260450076] = 250 -- I Got Your Parts Right Here... (Orgrimmar)
quests[260460076] = 250 -- Between a Rock and a Hard Place (Orgrimmar)
quests[260470076] = 250 -- And That's Why They Call Them Peons... (Orgrimmar)
quests[260480076] = 10 -- Spare Parts Up In Here! (Orgrimmar)
quests[260500021] = 250 -- Goggle Boggle (Booty Bay)
quests[260500087] = -250 -- Goggle Boggle (Bloodsail Buccaneers)
quests[260510021] = 350 -- Sunken Treasure (Booty Bay)
quests[260510087] = -350 -- Sunken Treasure (Bloodsail Buccaneers)
quests[260520021] = 75 -- Speak to Shakes (Booty Bay)
quests[260520087] = -75 -- Speak to Shakes (Bloodsail Buccaneers)
quests[260530068] = 250 -- Clearing the Highlands (Undercity)
quests[260550021] = 250 -- Drowned Sorrows (Booty Bay)
quests[260550087] = -250 -- Drowned Sorrows (Bloodsail Buccaneers)
quests[260580076] = 500 -- In Defense of Krom'gar Fortress (Orgrimmar)
quests[260590076] = 25 -- Eyes and Ears: Malaka'jin (Orgrimmar)
quests[260600076] = 250 -- Da Voodoo: Stormer Heart (Orgrimmar)
quests[260610076] = 250 -- Da Voodoo: Ram Horns (Orgrimmar)
quests[260620076] = 500 -- Da Voodoo: Resonite Crystal (Orgrimmar)
quests[260630076] = 250 -- The Queen and Her Court (Orgrimmar)
quests[260640076] = 250 -- Fight On Their Stomachs (Orgrimmar)
quests[260651135] = 350 -- Free Wil'hai (The Earthen Ring)
quests[260660076] = 250 -- Reinforcements... (Orgrimmar)
quests[260670076] = 500 -- Jin'Zil's Blessing (Orgrimmar)
quests[260680076] = 250 -- Kobold Fury! (Orgrimmar)
quests[260690076] = 10 -- Nura Pathfinder (Orgrimmar)
quests[260721135] = 250 -- Into the Totem (The Earthen Ring)
quests[260730076] = 500 -- All's Quiet on the Southern Front (Orgrimmar)
quests[260740076] = 75 -- Beginning of the End (Orgrimmar)
quests[260750076] = 250 -- The Turd Problem (Orgrimmar)
quests[260760076] = 350 -- Spy Infestation (Orgrimmar)
quests[260770076] = 75 -- Final Delivery (Orgrimmar)
quests[260780047] = 250 -- Extinguish the Fires (Ironforge)
quests[260780054] = 250 -- Extinguish the Fires (Gnomeregan)
quests[260790072] = 350 -- Wanted! Otto and Falconcrest (Stormwind)
quests[260800469] = 10 -- One Last Favor (Alliance)
quests[260810068] = 350 -- Alina's Reward (Undercity)
quests[260820076] = 350 -- To Battlescar! (Orgrimmar)
quests[260840076] = 75 -- Hammerfall (Orgrimmar)
quests[260850047] = 250 -- Rallying the Defenders (Ironforge)
quests[260850054] = 250 -- Rallying the Defenders (Gnomeregan)
quests[260860067] = 10 -- Orako (Horde)
quests[260870067] = 250 -- "Glow-Juice" (Horde)
quests[260880067] = 250 -- Here Fishie Fishie (Horde)
quests[260890067] = 250 -- Die Fishman Die (Horde)
quests[260900067] = 75 -- I Brought You This Egg (Horde)
quests[260910067] = 350 -- Here Fishie Fishie 2: Eel-Egg-Trick Boogaloo (Horde)
quests[260920067] = 10 -- Orako's Report (Horde)
quests[260930072] = 250 -- Northfold Manor (Stormwind)
quests[260940047] = 250 -- Striking Back (Ironforge)
quests[260940054] = 250 -- Striking Back (Gnomeregan)
quests[260950072] = 350 -- Stromgarde Badges (Stormwind)
quests[260961135] = 250 -- Scalding Shrooms (The Earthen Ring)
quests[260970076] = 250 -- Proof of Lies (Orgrimmar)
quests[260980076] = 250 -- Betrayal at the Grove (Orgrimmar)
quests[260990076] = 250 -- Is This Justice? (Orgrimmar)
quests[261000076] = 10 -- The General is Dead (Orgrimmar)
quests[261010076] = 10 -- Might Makes Right (Orgrimmar)
quests[261020047] = 250 -- Grimaxe's Demise (Ironforge)
quests[261020054] = 250 -- Grimaxe's Demise (Gnomeregan)
quests[261030054] = 250 -- Bio-Fuel (Gnomeregan)
quests[261050469] = 250 -- Claim Korthun's End (Alliance)
quests[261060054] = 250 -- Fuel-ology 101 (Gnomeregan)
quests[261070076] = 250 -- Raising Spirits (Orgrimmar)
quests[261080076] = 250 -- Guile of the Raptor (Orgrimmar)
quests[261100072] = 250 -- Just Like Old Times (Stormwind)
quests[261111135] = 350 -- ... It Will Come (The Earthen Ring)
quests[261120047] = 250 -- Demanding Answers (Ironforge)
quests[261120054] = 250 -- Demanding Answers (Gnomeregan)
quests[261130072] = 75 -- Quae and Kinelory (Stormwind)
quests[261150076] = 500 -- To Be Horde... (Orgrimmar)
quests[261160072] = 250 -- Kinelory Strikes (Stormwind)
quests[261170072] = 350 -- For Southshore (Stormwind)
quests[261180047] = 250 -- Seize the Ambassador (Ironforge)
quests[261180054] = 250 -- Seize the Ambassador (Gnomeregan)
quests[261200069] = 250 -- The Crazed Dragonmaw (Darnassus)
quests[261210469] = 250 -- Claim Korthun's End (Alliance)
quests[261221133] = 250 -- Environmental Awareness (Bilgewater Cartel)
quests[261240067] = 250 -- Secure Seabrush (Horde)
quests[261250067] = 250 -- Secure Seabrush (Horde)
quests[261261133] = 250 -- The Perfect Fuel (Bilgewater Cartel)
quests[261270047] = 250 -- The Twilight's Hammer Revealed (Ironforge)
quests[261280047] = 350 -- The Battle of Thandol Span (Ironforge)
quests[261291134] = 10 -- Brothers In Arms (Gilneas)
quests[261301135] = 250 -- Unplug L'ghorek (The Earthen Ring)
quests[261310047] = 10 -- Reinforcements for Loch Modan (Ironforge)
quests[261310054] = 10 -- Reinforcements for Loch Modan (Gnomeregan)
quests[261340076] = 75 -- Nothing Left for You Here (Orgrimmar)
quests[261351135] = 500 -- Visions of the Past: Rise from the Deep (The Earthen Ring)
quests[261370047] = 75 -- Checking on the Boys (Ironforge)
quests[261380069] = 75 -- Ferilon Leafborn (Darnassus)
quests[261390469] = 75 -- Into Arathi (Alliance)
quests[261450047] = 250 -- The Trogg Threat (Ironforge)
quests[261460047] = 250 -- In Defense of the King's Lands (Ironforge)
quests[261470047] = 250 -- Bigger and Uglier (Ironforge)
quests[261480047] = 350 -- A Decisive Strike (Ironforge)
quests[261500072] = 25 -- A Visit With Maybell (Stormwind)
quests[261520072] = 350 -- Wanted: James Clark (Stormwind)
quests[261530072] = 250 -- Penny's Pumpkin Pancakes (Stormwind)
quests[261570072] = 250 -- Kasim Sharim (Stormwind)
quests[261570054] = 250 -- Kasim Sharim (Gnomeregan)
quests[261580054] = 25 -- Attune the Bloodstone (Gnomeregan)
quests[261590054] = 250 -- The First Step (Gnomeregan)
quests[261600054] = 25 -- Blood Ritual (Gnomeregan)
quests[261610054] = 250 -- Not Just Any Body (Gnomeregan)
quests[261620054] = 500 -- The Altar of Storms (Gnomeregan)
quests[261630930] = 25 -- Time is Short (Exodar)
quests[261640930] = 250 -- The Charred Granite of the Dark Portal (Exodar)
quests[261650930] = 250 -- The Vile Blood of Demons (Exodar)
quests[261660930] = 75 -- Enhancing the Stone (Exodar)
quests[261670054] = 250 -- The Amulet of Allistarj (Gnomeregan)
quests[261680054] = 250 -- The Amulet of Sevine (Gnomeregan)
quests[261690054] = 250 -- The Amulet of Grol (Gnomeregan)
quests[261710469] = 1000 -- You Are Rakh'likh, Demon (Alliance)
quests[261720054] = 250 -- A Bloodmage's Gotta Eat Too (Gnomeregan)
quests[261730072] = 350 -- Protecting Our Rear (Stormwind)
quests[261740072] = 250 -- Watching Our Back (Stormwind)
quests[261751134] = 75 -- Surwich (Gilneas)
quests[261760047] = 150 -- Onward to Thelsamar (Ironforge)
quests[261770072] = 250 -- Feeling Crabby? (Stormwind)
quests[261781073] = 250 -- Planning for the Future (The Kalu'ak)
quests[261790081] = 250 -- The Venture Co. (Thunder Bluff)
quests[261800081] = 350 -- Supervisor Fizsprocket (Thunder Bluff)
quests[261811135] = 25 -- Back to Darkbreak Cove (The Earthen Ring)
quests[261821135] = 25 -- Back to the Tenebrous Cavern (The Earthen Ring)
quests[261830072] = 250 -- The King's Cider (Stormwind)
quests[261841134] = 350 -- Wormthorn's Dream (Gilneas)
quests[261850609] = 10 -- Advice from the Cenarion Circle (Cenarion Circle)
quests[261861134] = 250 -- Demoniac Vessel (Gilneas)
quests[261860942] = 250 -- Demoniac Vessel (Cenarion Expedition)
quests[261871134] = 500 -- The Downfall of Marl Wormthorn (Gilneas)
quests[261880081] = 250 -- Mazzranache (Thunder Bluff)
quests[261890047] = 350 -- The Angerfang Menace (Ironforge)
quests[261900072] = 250 -- A Fisherman's Feast (Stormwind)
quests[261920072] = 250 -- Orphans Like Cookies Too! (Stormwind)
quests[261930072] = 250 -- Defending the Rift (Stormwind)
quests[261940076] = 250 -- Defending the Rift (Orgrimmar)
quests[261950069] = 75 -- Who Wards The Greenwarden (Darnassus)
quests[261960047] = 75 -- Longbraid the Grim (Ironforge)
quests[261970054] = 75 -- The Future of Gnomeregan (Gnomeregan)
quests[261980054] = 150 -- The Arts of a Mage (Gnomeregan)
quests[261990054] = 75 -- The Future of Gnomeregan (Gnomeregan)
quests[262000054] = 150 -- The Arts of a Priest (Gnomeregan)
quests[262010054] = 150 -- The Power of a Warlock (Gnomeregan)
quests[262020054] = 75 -- The Future of Gnomeregan (Gnomeregan)
quests[262030054] = 75 -- The Future of Gnomeregan (Gnomeregan)
quests[262040054] = 150 -- The Arts of a Warrior (Gnomeregan)
quests[262050054] = 250 -- A Job for the Multi-Bot (Gnomeregan)
quests[262060054] = 75 -- The Future of Gnomeregan (Gnomeregan)
quests[262070054] = 150 -- The Arts of a Rogue (Gnomeregan)
quests[262080054] = 250 -- The Fight Continues (Gnomeregan)
quests[262090072] = 250 -- Murder Was The Case That They Gave Me (Stormwind)
quests[262130072] = 250 -- Hot On the Trail: The Riverpaw Clan (Stormwind)
quests[262140072] = 250 -- Hot On the Trail: Murlocs (Stormwind)
quests[262200076] = 250 -- Everything Is Better with Bacon (Orgrimmar)
quests[262220054] = 250 -- Scrounging for Parts (Gnomeregan)
quests[262251133] = 250 -- Pupellyverbos Port (Bilgewater Cartel)
quests[262260076] = 250 -- Crawfish Creole (Orgrimmar)
quests[262270076] = 250 -- Careful, This Fruit Bites Back (Orgrimmar)
quests[262320072] = 150 -- Lou's Parting Thoughts (Stormwind)
quests[262330076] = 250 -- Stealing From Our Own (Orgrimmar)
quests[262340076] = 250 -- Stealing From Our Own (Orgrimmar)
quests[262350076] = 250 -- Even Thieves Get Hungry (Orgrimmar)
quests[262431135] = 75 -- The World Pillar (The Earthen Ring)
quests[262441135] = 250 -- The Earth Claims All (The Earthen Ring)
quests[262451135] = 250 -- Gunship Down (The Earthen Ring)
quests[262461135] = 75 -- Captain's Log (The Earthen Ring)
quests[262471135] = 75 -- Diplomacy First (The Earthen Ring)
quests[262481135] = 250 -- All Our Friends Are Dead (The Earthen Ring)
quests[262491135] = 75 -- The Admiral's Cabin (The Earthen Ring)
quests[262511135] = 250 -- Take No Prisoners (The Earthen Ring)
quests[262541135] = 150 -- Some Spraining to Do (The Earthen Ring)
quests[262551135] = 150 -- Return to the Temple of Earth (The Earthen Ring)
quests[262561135] = 150 -- Bleed the Bloodshaper (The Earthen Ring)
quests[262581135] = 150 -- Deathwing's Fall (The Earthen Ring)
quests[262591135] = 250 -- Blood of the Earthwarder (The Earthen Ring)
quests[262601135] = 250 -- The Forgemaster's Log (The Earthen Ring)
quests[262611135] = 250 -- Question the Slaves (The Earthen Ring)
quests[262640054] = 250 -- What's Left Behind (Gnomeregan)
quests[262650054] = 250 -- Dealing with the Fallout (Gnomeregan)
quests[262720530] = 75 -- The Rise of the Darkspear (Darkspear Trolls)
quests[262730530] = 250 -- The Basics: Hitting Things (Darkspear Trolls)
quests[262740530] = 150 -- The Arts of a Warlock (Darkspear Trolls)
quests[262750530] = 250 -- A Rough Start (Darkspear Trolls)
quests[262760530] = 250 -- Proving Pit (Darkspear Trolls)
quests[262770530] = 75 -- More Than Expected (Darkspear Trolls)
quests[262780076] = 250 -- The Defense of Grom'gol: Raptor Risk (Orgrimmar)
quests[262790076] = 250 -- The Defense of Grom'gol: Ogre Oppression (Orgrimmar)
quests[262800076] = 250 -- The Defense of Grom'gol: Murloc Menace (Orgrimmar)
quests[262810067] = 350 -- The Defense of Grom'gol: Trollish Thievery (Horde)
quests[262830068] = 250 -- Prime Slime (Undercity)
quests[262840047] = 250 -- Missing in Action (Ironforge)
quests[262840054] = 250 -- Missing in Action (Gnomeregan)
quests[262850047] = 250 -- Get Me Explosives Back! (Ironforge)
quests[262850054] = 250 -- Get Me Explosives Back! (Gnomeregan)
quests[262860072] = 250 -- In Defense of Westfall (Stormwind)
quests[262870072] = 250 -- The Westfall Brigade (Stormwind)
quests[262880072] = 250 -- Jango Spothide (Stormwind)
quests[262890072] = 75 -- Find Agent Kearnen (Stormwind)
quests[262900072] = 250 -- Secrets of the Tower (Stormwind)
quests[262910072] = 250 -- Big Trouble in Moonbrook (Stormwind)
quests[262920072] = 75 -- To Moonbrook! (Stormwind)
quests[262930076] = 25 -- Machines of War (Orgrimmar)
quests[262940076] = 250 -- Weapons of Mass Dysfunction (Orgrimmar)
quests[262950072] = 250 -- Propaganda (Stormwind)
quests[262960072] = 250 -- Evidence Collection (Stormwind)
quests[262970072] = 150 -- The Dawning of a New Day (Stormwind)
quests[262980530] = 250 -- Hunt for Yenniku (Darkspear Trolls)
quests[262990530] = 250 -- Headhunting (Darkspear Trolls)
quests[263000530] = 250 -- Nezzliok Will Know (Darkspear Trolls)
quests[263010530] = 250 -- Speaking with Nezzliok (Darkspear Trolls)
quests[263020530] = 75 -- Chasing Yenniku's Fate (Darkspear Trolls)
quests[263030530] = 250 -- The Mind's Eye (Darkspear Trolls)
quests[263040530] = 350 -- Nighttime in the Jungle (Darkspear Trolls)
quests[263050530] = 350 -- Saving Yenniku (Darkspear Trolls)
quests[263090068] = 250 -- Dark Vessels (Undercity)
quests[263110076] = 250 -- Unfamiliar Waters (Orgrimmar)
quests[263121135] = 250 -- Crumbling Defenses (The Earthen Ring)
quests[263131135] = 250 -- Core of Our Troubles (The Earthen Ring)
quests[263141135] = 250 -- On Even Ground (The Earthen Ring)
quests[263151135] = 250 -- Imposing Confrontation (The Earthen Ring)
quests[263160054] = 10 -- What's Keeping Jessup? (Gnomeregan)
quests[263180047] = 250 -- Finishin' the Job (Ironforge)
quests[263180054] = 250 -- Finishin' the Job (Gnomeregan)
quests[263190072] = 10 -- Secrets Revealed (Stormwind)
quests[263200072] = 250 -- A Vision of the Past (Stormwind)
quests[263220072] = 500 -- Rise of the Brotherhood (Stormwind)
quests[263240076] = 150 -- Where Is My Warfleet? (Orgrimmar)
quests[263261135] = 10 -- The Very Earth Beneath Our Feet (The Earthen Ring)
quests[263270047] = 25 -- Anvilmar the Hero (Ironforge)
quests[263281171] = 500 -- Rocky Relations (Therazane)
quests[263310054] = 250 -- Crushcog's Minions (Gnomeregan)
quests[263330054] = 250 -- No Tanks! (Gnomeregan)
quests[263350076] = 10 -- Ready the Navy (Orgrimmar)
quests[263370076] = 250 -- Beating the Market (Orgrimmar)
quests[263380530] = 250 -- Population Con-Troll (Darkspear Trolls)
quests[263390054] = 10 -- Staging in Brewnall (Gnomeregan)
quests[263420054] = 250 -- Paint it Black (Gnomeregan)
quests[263460047] = 75 -- Myzrael's Tale (Ironforge)
quests[263500530] = 250 -- Priestess Hu'rala (Darkspear Trolls)
quests[263510530] = 350 -- Mind Vision (Darkspear Trolls)
quests[263520530] = 350 -- Cozzle's Plan (Darkspear Trolls)
quests[263581133] = 10 -- Ready the Air Force (Bilgewater Cartel)
quests[263590530] = 350 -- See Raptor (Darkspear Trolls)
quests[263600530] = 350 -- Mind Control (Darkspear Trolls)
quests[263611133] = 250 -- Smoot's Samophlange (Bilgewater Cartel)
quests[263620530] = 350 -- Be Raptor (Darkspear Trolls)
quests[263640054] = 350 -- Down with Crushcog! (Gnomeregan)
quests[263640047] = 350 -- Down with Crushcog! (Ironforge)
quests[263650072] = 250 -- Hero's Call: Redridge Mountains! (Stormwind)
quests[263680068] = 250 -- Venomous Secrets (Undercity)
quests[263700072] = 10 -- Return to Sentinel Hill (Stormwind)
quests[263721133] = 10 -- Pre-Flight Checklist (Bilgewater Cartel)
quests[263730047] = 10 -- On to Kharanos (Ironforge)
quests[263740076] = 250 -- Ready the Ground Troops (Orgrimmar)
quests[263751171] = 7000 -- Loose Stones (Therazane)
quests[263761171] = 7000 -- Hatred Runs Deep (Therazane)
quests[263771171] = 7000 -- Unsolid Ground (Therazane)
quests[263780072] = 75 -- Hero's Call: Westfall! (Stormwind)
quests[263790069] = 25 -- In Aid of the Refugees (Darnassus)
quests[263800047] = 10 -- Bound for Kharanos (Ironforge)
quests[263810068] = 250 -- The Eye of Shadra (Undercity)
quests[263820068] = 250 -- Can't Make An Omelette Without... (Undercity)
quests[263830069] = 25 -- Breaking Waves of Change (Darnassus)
quests[263850069] = 25 -- Breaking Waves of Change (Darnassus)
quests[263860530] = 500 -- Surkhan (Darkspear Trolls)
quests[263870068] = 250 -- Starvation Diet (Undercity)
quests[263880076] = 250 -- Twilight Skies (Orgrimmar)
quests[263890072] = 250 -- Blackrock Invasion (Stormwind)
quests[263900072] = 350 -- Ending the Invasion! (Stormwind)
quests[263910072] = 250 -- Extinguishing Hope (Stormwind)
quests[263930072] = 25 -- A Swift Message (Stormwind)
quests[263940072] = 75 -- Continue to Stormwind (Stormwind)
quests[263950072] = 25 -- Dungar Longdrink (Stormwind)
quests[263960072] = 350 -- Return to Argus (Stormwind)
quests[263970076] = 250 -- Walk With The Earth Mother (Orgrimmar)
quests[263980076] = 250 -- Walk With The Earth Mother (Orgrimmar)
quests[263991133] = 250 -- The Mosh'Ogg Bounty (Bilgewater Cartel)
quests[264001133] = 250 -- The Universal Key (Bilgewater Cartel)
quests[264010069] = 250 -- Return to Vestia (Darnassus)
quests[264020069] = 75 -- General Shandris Feathermoon (Darnassus)
quests[264031133] = 250 -- Venture Company Mining (Bilgewater Cartel)
quests[264041133] = 250 -- Above My Pay Grade (Bilgewater Cartel)
quests[264050530] = 350 -- Zul'Mamwe Mambo (Darkspear Trolls)
quests[264060068] = 250 -- The Fang of Shadra (Undercity)
quests[264070076] = 150 -- Mosh'Ogg Handiwork (Orgrimmar)
quests[264080069] = 25 -- Ashes in Ashenvale (Darnassus)
quests[264091135] = 10 -- Where's Goldmine? (The Earthen Ring)
quests[264101135] = 250 -- Explosive Bonding Compound (The Earthen Ring)
quests[264111135] = 75 -- Apply and Flash Dry (The Earthen Ring)
quests[264121133] = 250 -- The Mosh'Ogg Bounty (Bilgewater Cartel)
quests[264131135] = 10 -- Take Him to the Earthcaller (The Earthen Ring)
quests[264140072] = 250 -- Hitting a Walleye (Stormwind)
quests[264160076] = 25 -- Well, Come to the Jungle (Orgrimmar)
quests[264170076] = 10 -- Northern Stranglethorn: The Fallen Empire (Orgrimmar)
quests[264180068] = 250 -- The Shell of Shadra (Undercity)
quests[264190068] = 350 -- Shadra the Venom Queen (Undercity)
quests[264200072] = 250 -- Diggin' For Worms (Stormwind)
quests[264210054] = 10 -- Meet the High Tinker (Gnomeregan)
quests[264220054] = 10 -- Meet the High Tinker (Gnomeregan)
quests[264230054] = 10 -- Meet the High Tinker (Gnomeregan)
quests[264240054] = 10 -- Meet the High Tinker (Gnomeregan)
quests[264250054] = 10 -- Meet the High Tinker (Gnomeregan)
quests[264261171] = 7000 -- Violent Gale (Therazane)
quests[264271135] = 150 -- Without a Captain or Crew (The Earthen Ring)
quests[264280068] = 250 -- To Steal From Thieves (Undercity)
quests[264290076] = 250 -- Crush the Witherbark (Orgrimmar)
quests[264331133] = 250 -- The Holy Water of Clarity (Bilgewater Cartel)
quests[264341133] = 250 -- Primal Reagents of Power (Bilgewater Cartel)
quests[264350911] = 25 -- I'm A Huge Liar and a Fraud (Silvermoon City)
quests[264361171] = 250 -- Entrenched (Therazane)
quests[264371171] = 500 -- Making Things Crystal Clear (Therazane)
quests[264381171] = 500 -- Intervention (Therazane)
quests[264391171] = 500 -- Putting the Pieces Together (Therazane)
quests[264401171] = 500 -- Clingy (Therazane)
quests[264411171] = 250 -- So Big, So Round... (Therazane)
quests[264420072] = 250 -- Rock Lobster (Stormwind)
quests[264430930] = 75 -- Diabolical Plans (Exodar)
quests[264440930] = 250 -- Destroy the Legion (Exodar)
quests[264450930] = 350 -- Never Again! (Exodar)
quests[264460069] = 250 -- Reclaiming Felfire Hill (Darnassus)
quests[264470076] = 75 -- Diabolical Plans (Orgrimmar)
quests[264480076] = 250 -- Destroy the Legion (Orgrimmar)
quests[264490076] = 350 -- Never Again! (Orgrimmar)
quests[264501133] = 250 -- If They're Just Going to Leave Them Lying Around... (Bilgewater Cartel)
quests[264530930] = 25 -- A Helping Hand (Exodar)
quests[264540069] = 250 -- A Shameful Waste (Darnassus)
quests[264550069] = 250 -- The Lost Chalice (Darnassus)
quests[264560890] = 75 -- Report from the Northern Front (Silverwing Sentinels)
quests[264570069] = 250 -- Agents of Destruction (Darnassus)
quests[264630069] = 250 -- Finding Teronis (Darnassus)
quests[264640069] = 250 -- The Lost Gem (Darnassus)
quests[264650069] = 250 -- The Ancient Statuettes (Darnassus)
quests[264660069] = 350 -- Ruuzel (Darnassus)
quests[264670069] = 250 -- Vile Satyr! Dryads in Danger! (Darnassus)
quests[264680069] = 250 -- The Branch of Cenarius (Darnassus)
quests[264690069] = 250 -- Satyr Slaying! (Darnassus)
quests[264700069] = 250 -- Retaking Mystral Lake (Darnassus)
quests[264720069] = 350 -- Insane Druids (Darnassus)
quests[264730069] = 250 -- Bathran's Hair (Darnassus)
quests[264740069] = 350 -- Orendil's Cure (Darnassus)
quests[264750069] = 250 -- Elune's Tear (Darnassus)
quests[264760069] = 75 -- Dryad Delivery (Darnassus)
quests[264770069] = 250 -- Search the Bole (Darnassus)
quests[264780069] = 250 -- Playing Possum (Darnassus)
quests[264790069] = 10 -- Return to Raene (Darnassus)
quests[264800069] = 150 -- Dartol's Rod (Darnassus)
quests[264810069] = 250 -- A New Adornment (Darnassus)
quests[264820069] = 500 -- True Power of the Rod (Darnassus)
quests[264871133] = 250 -- Akiris by the Bundle (Bilgewater Cartel)
quests[264880072] = 250 -- Big Gulp (Stormwind)
quests[264891133] = 250 -- It's You!! (Bilgewater Cartel)
quests[264900068] = 250 -- Prime Slime (Undercity)
quests[264931133] = 350 -- There's Somebody Out There Who Wants It (Bilgewater Cartel)
quests[264941133] = 10 -- Mixmaster Jasper (Bilgewater Cartel)
quests[264951133] = 10 -- Chabal (Bilgewater Cartel)
quests[265030072] = 350 -- Still Assessing the Threat (Stormwind)
quests[265040072] = 250 -- Wanted: Redridge Gnolls (Stormwind)
quests[265050072] = 75 -- Parker's Report (Stormwind)
quests[265060072] = 250 -- Franks and Beans (Stormwind)
quests[265071171] = 500 -- Petrified Delicacies (Therazane)
quests[265080072] = 250 -- Nida's Necklace (Stormwind)
quests[265090072] = 250 -- An Unwelcome Guest (Stormwind)
quests[265100072] = 250 -- We Must Prepare! (Stormwind)
quests[265110072] = 250 -- Lake Everstill Clean Up (Stormwind)
quests[265120072] = 150 -- Tuning the Gnomecorder (Stormwind)
quests[265130072] = 350 -- Like a Fart in the Wind (Stormwind)
quests[265140072] = 350 -- Canyon Romp (Stormwind)
quests[265200072] = 500 -- Saving Foreman Oslow (Stormwind)
quests[265331133] = 250 -- Zanzil's Secret (Bilgewater Cartel)
quests[265341133] = 150 -- Let's See What You've Got, Zanzil (Bilgewater Cartel)
quests[265351133] = 350 -- Recipe for Disaster (Bilgewater Cartel)
quests[265360072] = 250 -- Thunder Falls (Stormwind)
quests[265380076] = 150 -- Emergency Aid (Orgrimmar)
quests[265390076] = 75 -- Stalled Negotiations (Orgrimmar)
quests[265400076] = 75 -- Dangerous Compassion (Orgrimmar)
quests[265430076] = 250 -- Clammy Hands (Orgrimmar)
quests[265440072] = 250 -- They've Wised Up... (Stormwind)
quests[265450072] = 500 -- Yowler Must Die! (Stormwind)
quests[265490076] = 250 -- Madness (Orgrimmar)
quests[265560076] = 250 -- No Dumping Allowed (Orgrimmar)
quests[265570076] = 250 -- A Staggering Effort (Orgrimmar)
quests[265610072] = 250 -- Krakauer (Stormwind)
quests[265620072] = 250 -- And Last But Not Least... Danforth (Stormwind)
quests[265630072] = 500 -- Return of the Bravo Company (Stormwind)
quests[265660054] = 10 -- A Triumph of Gnomish Ingenuity (Gnomeregan)
quests[265670072] = 10 -- John J. Keeshan (Stormwind)
quests[265680072] = 10 -- This Ain't My War (Stormwind)
quests[265690072] = 250 -- Surveying Equipment (Stormwind)
quests[265700072] = 250 -- Render's Army (Stormwind)
quests[265710072] = 250 -- Weapons of War (Stormwind)
quests[265720076] = 250 -- A Golden Opportunity (Orgrimmar)
quests[265730072] = 250 -- His Heart Must Be In It (Stormwind)
quests[265751171] = 500 -- Rock Bottom (Therazane)
quests[265761171] = 250 -- Steady Hand (Therazane)
quests[265771171] = 250 -- Rocky Upheaval (Therazane)
quests[265781171] = 500 -- Doomshrooms (Therazane)
quests[265791171] = 500 -- Gone Soft (Therazane)
quests[265801135] = 25 -- Familiar Intruders (The Earthen Ring)
quests[265811135] = 250 -- A Head Full of Wind (The Earthen Ring)
quests[265821171] = 250 -- Unnatural Causes (Therazane)
quests[265831171] = 500 -- Wrath of the Fungalmancer (Therazane)
quests[265841171] = 500 -- Shaken and Stirred (Therazane)
quests[265851171] = 500 -- Corruption Destruction (Therazane)
quests[265860072] = 250 -- In Search of Bravo Company (Stormwind)
quests[265870072] = 250 -- Breaking Out is Hard to Do (Stormwind)
quests[265880076] = 250 -- A Furious Catch (Orgrimmar)
quests[265901133] = 250 -- The Holy Water of Clarity (Bilgewater Cartel)
quests[265921133] = 250 -- Diffractory Chromascope (Bilgewater Cartel)
quests[265930021] = 250 -- Scaring Shaky (Booty Bay)
quests[265930087] = -250 -- Scaring Shaky (Bloodsail Buccaneers)
quests[265940021] = 150 -- Return to MacKinley (Booty Bay)
quests[265940087] = -150 -- Return to MacKinley (Bloodsail Buccaneers)
quests[265950021] = 250 -- Kill-Collect (Booty Bay)
quests[265950087] = -250 -- Kill-Collect (Bloodsail Buccaneers)
quests[265970021] = 500 -- Stranglethorn Fever (Booty Bay)
quests[265970087] = -500 -- Stranglethorn Fever (Bloodsail Buccaneers)
quests[265990021] = 250 -- The Captain's Chest (Booty Bay)
quests[265990087] = -250 -- The Captain's Chest (Bloodsail Buccaneers)
quests[266000021] = 150 -- A Giant's Feast (Booty Bay)
quests[266000087] = -150 -- A Giant's Feast (Bloodsail Buccaneers)
quests[266010021] = 500 -- Mok'rash the Cleaver (Booty Bay)
quests[266010087] = -500 -- Mok'rash the Cleaver (Bloodsail Buccaneers)
quests[266020021] = 350 -- A Dish Best Served Huge (Booty Bay)
quests[266020087] = -350 -- A Dish Best Served Huge (Bloodsail Buccaneers)
quests[266070072] = 75 -- They Drew First Blood (Stormwind)
quests[266090021] = 150 -- The Bloodsail Buccaneers (Booty Bay)
quests[266090087] = -150 -- The Bloodsail Buccaneers (Bloodsail Buccaneers)
quests[266100021] = 150 -- Bloodsail Treachery (Booty Bay)
quests[266100087] = -150 -- Bloodsail Treachery (Bloodsail Buccaneers)
quests[266110021] = 25 -- The Baron Must Be Told (Booty Bay)
quests[266110087] = -25 -- The Baron Must Be Told (Bloodsail Buccaneers)
quests[266120021] = 350 -- Details of the Attack (Booty Bay)
quests[266120087] = -350 -- Details of the Attack (Bloodsail Buccaneers)
quests[266130021] = 250 -- Up to Snuff (Booty Bay)
quests[266130087] = -250 -- Up to Snuff (Bloodsail Buccaneers)
quests[266140021] = 250 -- Keep An Eye Out (Booty Bay)
quests[266140087] = -250 -- Keep An Eye Out (Bloodsail Buccaneers)
quests[266160072] = 25 -- It's Never Over (Stormwind)
quests[266170021] = 150 -- An Old Sea Dog (Booty Bay)
quests[266170087] = -150 -- An Old Sea Dog (Bloodsail Buccaneers)
quests[266180072] = 250 -- Wolves at Our Heels (Stormwind)
quests[266191172] = 250 -- You Say You Want a Revolution (Dragonmaw Clan)
quests[266200072] = 250 -- Seasoned Wolf Kabobs (Stormwind)
quests[266211172] = 500 -- Insurrection (Dragonmaw Clan)
quests[266221172] = 1750 -- Death to Mor'ghor (Dragonmaw Clan)
quests[266230072] = 250 -- Dusky Crab Cakes (Stormwind)
quests[266280021] = 350 -- Death From Below (Booty Bay)
quests[266280087] = -350 -- Death From Below (Bloodsail Buccaneers)
quests[266360072] = 250 -- Bravo Company Field Kit: Camouflage (Stormwind)
quests[266370072] = 250 -- Bravo Company Field Kit: Chloroform (Stormwind)
quests[266380072] = 250 -- Hunting the Hunters (Stormwind)
quests[266390072] = 250 -- Point of Contact: Brubaker (Stormwind)
quests[266400072] = 250 -- Unspeakable Atrocities (Stormwind)
quests[266420076] = 75 -- Preserving the Barrens (Orgrimmar)
quests[266450072] = 250 -- The Night Watch (Stormwind)
quests[266460072] = 350 -- Prisoners of War (Stormwind)
quests[266510072] = 250 -- To Win a War, You Gotta Become War (Stormwind)
quests[266561171] = 500 -- Don't. Stop. Moving. (Therazane)
quests[266571171] = 500 -- Hard Falls (Therazane)
quests[266581171] = 500 -- Fragile Values (Therazane)
quests[266591171] = 500 -- Resonating Blow (Therazane)
quests[266620021] = 250 -- The Brashtide Crew (Booty Bay)
quests[266630021] = 250 -- Sinking From Within (Booty Bay)
quests[266640021] = 250 -- Making Mutiny (Booty Bay)
quests[266650021] = 250 -- Call of Booty (Booty Bay)
quests[266660072] = 25 -- The Legend of Stalvan (Stormwind)
quests[266670072] = 75 -- The Stolen Letters (Stormwind)
quests[266680072] = 500 -- Detonation (Stormwind)
quests[266690072] = 75 -- In A Dark Corner (Stormwind)
quests[266700072] = 250 -- Roland's Doom (Stormwind)
quests[266711134] = 75 -- The Fate of Stalvan Mistmantle (Gilneas)
quests[266741134] = 500 -- Mistmantle's Revenge (Gilneas)
quests[266780021] = 10 -- Doublerum (Booty Bay)
quests[266790021] = 75 -- Return to Revilgaz (Booty Bay)
quests[266820076] = 250 -- A Shambling Threat (Orgrimmar)
quests[266840072] = 250 -- The Insane Ghoul (Stormwind)
quests[266850072] = 250 -- Classy Glass (Stormwind)
quests[266860072] = 250 -- Bones That Walk (Stormwind)
quests[266870072] = 75 -- Reinforcements From Theramore (Stormwind)
quests[266880072] = 250 -- Worgen in the Woods (Stormwind)
quests[266890072] = 250 -- The Rotting Orchard (Stormwind)
quests[266900072] = 150 -- Vile and Tainted (Stormwind)
quests[266910072] = 350 -- Worgen in the Woods (Stormwind)
quests[266920072] = 350 -- Shadowhide Extinction (Stormwind)
quests[266930072] = 350 -- The Dark Tower (Stormwind)
quests[266940072] = 350 -- The Grand Magus Doane (Stormwind)
quests[266950021] = 250 -- Prepare for Takeoff (Booty Bay)
quests[266970021] = 350 -- The Final Voyage of the Brashtide (Booty Bay)
quests[266980021] = 10 -- Seeking Seahorn (Booty Bay)
quests[266990021] = 250 -- Turning the Brashtide (Booty Bay)
quests[267000021] = 250 -- The Damsel's (Bad) Luck (Booty Bay)
quests[267010067] = 10 -- Flight to Brackenwall (Horde)
quests[267020469] = 10 -- Flight to Theramore (Alliance)
quests[267030021] = 500 -- Bloodsail's End (Booty Bay)
quests[267061134] = 500 -- Endgame (Gilneas)
quests[267071134] = 250 -- A Deadly Vine (Gilneas)
quests[267080072] = 500 -- AHHHHHHHHHHHH! AHHHHHHHHH!!! (Stormwind)
quests[267091171] = 150 -- The Stone Throne (Therazane)
quests[267101171] = 250 -- Lost In The Deeps (Therazane)
quests[267111133] = 25 -- Off to the Bank (Bilgewater Cartel)
quests[267121133] = 25 -- Off to the Bank (Bilgewater Cartel)
quests[267130072] = 500 -- Showdown at Stonewatch (Stormwind)
quests[267140072] = 500 -- Darkblaze, Brood of the Worldbreaker (Stormwind)
quests[267171134] = 250 -- The Yorgen Worgen (Gilneas)
quests[267191134] = 75 -- Delivery to Master Harris (Gilneas)
quests[267201134] = 250 -- A Curse We Cannot Lift (Gilneas)
quests[267210072] = 250 -- The Jitters-Bugs (Stormwind)
quests[267231134] = 250 -- The Fate of Morbent Fel (Gilneas)
quests[267260072] = 500 -- Triumphant Return (Stormwind)
quests[267270072] = 500 -- The Embalmer's Revenge (Stormwind)
quests[267280072] = 250 -- Hero's Call: Duskwood! (Stormwind)
quests[267290072] = 250 -- Water Elementals (Stormwind)
quests[267300072] = 250 -- You Can Take the Murloc Out of the Ocean... (Stormwind)
quests[267310072] = 250 -- The Altar of Naias (Stormwind)
quests[267320072] = 250 -- Bad Medicine (Stormwind)
quests[267330072] = 250 -- Control Sample (Stormwind)
quests[267340072] = 250 -- The Source of the Madness (Stormwind)
quests[267350072] = 250 -- The Fate of Kurzen (Stormwind)
quests[267360072] = 250 -- Spared from Madness (Stormwind)
quests[267370072] = 500 -- Stopping Kurzen's Legacy (Stormwind)
quests[267420047] = 250 -- Bloodscalp Insight (Ironforge)
quests[267430072] = 250 -- Sacred to the Bloodscalp (Stormwind)
quests[267440072] = 250 -- Deep Roots (Stormwind)
quests[267450072] = 250 -- Favored Skull (Stormwind)
quests[267490930] = 10 -- Priestess Thaalia (Exodar)
quests[267501171] = 250 -- At the Stonemother's Call (Therazane)
quests[267510072] = 250 -- Population Con-Troll (Stormwind)
quests[267521135] = 250 -- Audience with the Stonemother (The Earthen Ring)
quests[267521171] = 500 -- Audience with the Stonemother (Therazane)
quests[267541134] = 500 -- Morbent's Bane (Gilneas)
quests[267601134] = 75 -- Cry For The Moon (Gilneas)
quests[267610072] = 250 -- Threat to the Kingdom (Stormwind)
quests[267630021] = 250 -- Venture Company Mining (Booty Bay)
quests[267640021] = 250 -- A New Low (Booty Bay)
quests[267650072] = 250 -- Return to Corporal Kaleb (Stormwind)
quests[267661135] = 250 -- Big Game, Big Bait (The Earthen Ring)
quests[267681135] = 150 -- To Catch a Dragon (The Earthen Ring)
quests[267691133] = 75 -- Raging River Ride (Bilgewater Cartel)
quests[267711135] = 250 -- Testing the Trap (The Earthen Ring)
quests[267720930] = 250 -- Mind Vision (Exodar)
quests[267730930] = 250 -- See Raptor (Exodar)
quests[267740930] = 250 -- Mind Control (Exodar)
quests[267750930] = 350 -- Be Raptor (Exodar)
quests[267760930] = 500 -- Ghaliri (Exodar)
quests[267771134] = 250 -- Soothing Spirits (Gilneas)
quests[267781134] = 250 -- The Cries of the Dead (Gilneas)
quests[267790530] = 350 -- Zul'Mamwe Mambo (Darkspear Trolls)
quests[267800054] = 250 -- Nighttime in the Jungle (Gnomeregan)
quests[267821134] = 250 -- The Mosh'Ogg Bounty (Gilneas)
quests[267831134] = 250 -- The Mosh'Ogg Bounty (Gilneas)
quests[267840076] = 250 -- Muddied Waters (Orgrimmar)
quests[267860076] = 10 -- Securing the Beach Head (Orgrimmar)
quests[267870072] = 250 -- Bear In Mind (Stormwind)
quests[267880076] = 250 -- Cementing Our Victory (Orgrimmar)
quests[267930072] = 25 -- The Weathered Grave (Stormwind)
quests[267940072] = 25 -- Morgan Ladimore (Stormwind)
quests[267950072] = 150 -- Mor'Ladim (Stormwind)
quests[267960072] = 25 -- The Daughter Who Lived (Stormwind)
quests[267980076] = 10 -- The Warchief Will be Pleased (Orgrimmar)
quests[267990068] = 350 -- Those That Couldn't Be Saved (Undercity)
quests[268000068] = 350 -- Recruitment (Undercity)
quests[268010068] = 250 -- Scourge on our Perimeter (Undercity)
quests[268020068] = 250 -- The Damned (Undercity)
quests[268030076] = 10 -- Missing Reports (Orgrimmar)
quests[268040076] = 10 -- Flight to Razor Hill (Orgrimmar)
quests[268050047] = 25 -- To the Cape! (Ironforge)
quests[268060076] = 10 -- Reports to Orgrimmar (Orgrimmar)
quests[268070076] = 150 -- Return to Eitrigg (Orgrimmar)
quests[268080047] = 10 -- Maywiki (Ironforge)
quests[268150047] = 250 -- Zanzil's Secret (Ironforge)
quests[268160054] = 350 -- Recipe for Disaster (Gnomeregan)
quests[268170054] = 250 -- Perfectly Pure (Gnomeregan)
quests[268180047] = 250 -- Plush Pelts (Ironforge)
quests[268190047] = 250 -- Akiris by the Bundle (Ironforge)
quests[268200047] = 250 -- If They're Just Going to Leave Them Lying Around... (Ironforge)
quests[268210054] = 25 -- Airwyn Bantamflax (Gnomeregan)
quests[268220047] = 350 -- Good-Fer-Nothin' Slither-Dogs (Ironforge)
quests[268230054] = 250 -- Too Big For His Britches (Gnomeregan)
quests[268271135] = 250 -- Rallying the Earthen Ring (The Earthen Ring)
quests[268281171] = 250 -- Our Part of the Bargain (Therazane)
quests[268291171] = 500 -- The Stone March (Therazane)
quests[268300076] = 250 -- Traitor's Bait (Orgrimmar)
quests[268311171] = 500 -- The Twilight Flight (Therazane)
quests[268321171] = 500 -- Therazane's Mercy (Therazane)
quests[268331135] = 250 -- Word In Stone (The Earthen Ring)
quests[268400076] = 10 -- Return to the Highlands (Orgrimmar)
quests[268410069] = 75 -- Forbidden Sigil (Darnassus)
quests[268420047] = 250 -- Out of Gnoll-where (Ironforge)
quests[268430047] = 250 -- A Tiny, Clever Commander (Ironforge)
quests[268440047] = 250 -- Kobold and Kobolder (Ironforge)
quests[268450047] = 250 -- Who's In Charge Here? (Ironforge)
quests[268460047] = 250 -- A Nasty Exploit (Ironforge)
quests[268540047] = 75 -- The Lost Pilot (Ironforge)
quests[268540054] = 75 -- The Lost Pilot (Gnomeregan)
quests[268550047] = 150 -- A Pilot's Revenge (Ironforge)
quests[268550054] = 150 -- A Pilot's Revenge (Gnomeregan)
quests[268571135] = 250 -- Abyssion's Minions (The Earthen Ring)
quests[268600047] = 250 -- Thelsamar Blood Sausages (Ironforge)
quests[268611135] = 250 -- Block the Gates (The Earthen Ring)
quests[268630047] = 350 -- Filthy Paws (Ironforge)
quests[268640047] = 250 -- The Bearer of Gnoll-edge (Ironforge)
quests[268680047] = 350 -- Axis of Awful (Ironforge)
quests[268691171] = 7000 -- Depth of the Depths (Therazane)
quests[268711171] = 7000 -- A Rock Amongst Many (Therazane)
quests[268751135] = 250 -- Undying Twilight (The Earthen Ring)
quests[268761135] = 350 -- The World Pillar Fragment (The Earthen Ring)
quests[268860054] = 250 -- Going Off-Task (Gnomeregan)
quests[268870054] = 250 -- Filling Our Pockets (Gnomeregan)
quests[268900076] = 250 -- The Essence of Aku'Mai (Orgrimmar)
quests[269010068] = 500 -- Going, Going, Guano! (Undercity)
quests[269040047] = 75 -- Corruption (Ironforge)
quests[269100072] = 75 -- Etched Letter (Stormwind)
quests[269110076] = 75 -- Myzrael's Tale (Orgrimmar)
quests[269130072] = 250 -- Charging into Battle (Stormwind)
quests[269140072] = 250 -- Corruption (Stormwind)
quests[269150072] = 250 -- The Deepest Cut (Stormwind)
quests[269160072] = 250 -- Mastering the Arcane (Stormwind)
quests[269170072] = 250 -- The Hunter's Path (Stormwind)
quests[269180072] = 250 -- The Power of the Light (Stormwind)
quests[269190072] = 250 -- Learning the Word (Stormwind)
quests[269200068] = 25 -- The Battle for Andorhal (Undercity)
quests[269210068] = 250 -- Scourge First... Alliance Later (Undercity)
quests[269220068] = 250 -- The Endless Flow (Undercity)
quests[269230068] = 350 -- War Machines (Undercity)
quests[269240068] = 500 -- Scholomancer (Undercity)
quests[269250068] = 350 -- Araj the Summoner (Undercity)
quests[269260068] = 10 -- Victory, For Now (Undercity)
quests[269280047] = 250 -- Smells Like A Plan (Ironforge)
quests[269290047] = 250 -- A Load of Croc (Ironforge)
quests[269300068] = 150 -- After the Crusade (Undercity)
quests[269310068] = 250 -- Foxes and Hounds (Undercity)
quests[269320047] = 250 -- Buzz Off (Ironforge)
quests[269330068] = 250 -- Foes Before Hoes (Undercity)
quests[269340068] = 250 -- Latent Disease (Undercity)
quests[269351106] = 10 -- Northridge Lumber Mill (Argent Crusade)
quests[269360068] = 75 -- Lower the Boom (Undercity)
quests[269370068] = 250 -- When Death is Not Enough (Undercity)
quests[269380068] = 350 -- Combat Training (Undercity)
quests[269390054] = 500 -- The G-Team (Gnomeregan)
quests[269400069] = 250 -- Frost Nova (Darnassus)
quests[269410054] = 500 -- The G-Team (Gnomeregan)
quests[269420054] = 500 -- The G-Team (Gnomeregan)
quests[269450069] = 250 -- Learning New Techniques (Darnassus)
quests[269460069] = 250 -- A Rogue's Advantage (Darnassus)
quests[269470069] = 250 -- A Woodsman's Training (Darnassus)
quests[269480069] = 250 -- Moonfire (Darnassus)
quests[269490069] = 250 -- Learning the Word (Darnassus)
quests[269520068] = 10 -- The Menders' Stead (Undercity)
quests[269530609] = 350 -- Zen'Kiki, the Druid (Cenarion Circle)
quests[269540942] = 250 -- A Different Approach (Cenarion Expedition)
quests[269550609] = 350 -- Zen'Kiki and the Cultists (Cenarion Circle)
quests[269560609] = 250 -- I Ain't Sayin' You a Gourd-Digger... (Cenarion Circle)
quests[269571106] = 25 -- The Long Trip Home (Argent Crusade)
quests[269580930] = 250 -- Your First Lesson (Exodar)
quests[269600072] = 10 -- My Son, the Prince (Stormwind)
quests[269610047] = 250 -- Gathering Idols (Ironforge)
quests[269630930] = 250 -- Steadying Your Shot (Exodar)
quests[269640068] = 250 -- Warchief's Command: Silverpine Forest! (Undercity)
quests[269650068] = 150 -- The Warchief Cometh (Undercity)
quests[269660930] = 250 -- The Light's Power (Exodar)
quests[269680930] = 250 -- Frost Nova (Exodar)
quests[269690930] = 250 -- Primal Strike (Exodar)
quests[269700930] = 250 -- Learning the Word (Exodar)
quests[269711135] = 500 -- The Binding (The Earthen Ring)
quests[269711171] = 500 -- The Binding (Therazane)
quests[269750072] = 150 -- Rallying the Fleet (Stormwind)
quests[269770072] = 10 -- Twilight Investigation (Stormwind)
quests[269780068] = 250 -- Who Needs Cauldrons? (Undercity)
quests[269790068] = 75 -- Strange New Faces (Undercity)
quests[269800047] = 75 -- Swiftgear Station (Ironforge)
quests[269810047] = 75 -- Whelgar's Retreat (Ironforge)
quests[269890068] = 250 -- The Gilneas Liberation Front (Undercity)
quests[269920068] = 250 -- Agony Abounds (Undercity)
quests[269950068] = 250 -- Guts and Gore (Undercity)
quests[269970072] = 250 -- The Usual Suspects (Stormwind)
quests[269980068] = 350 -- Iterating Upon Success (Undercity)
quests[269991106] = 250 -- A New Era for the Plaguelands (Argent Crusade)
quests[270001106] = 250 -- Learning the Ropes (Argent Crusade)
quests[270011106] = 250 -- This Means WAR (Wild Arachnid Roundup) (Argent Crusade)
quests[270021106] = 350 -- An Audience with the Highlord (Argent Crusade)
quests[270071135] = 75 -- Silvermarsh Rendezvous (The Earthen Ring)
quests[270101135] = 250 -- Quicksilver Submersion (The Earthen Ring)
quests[270140081] = 75 -- Hallowed Note (Thunder Bluff)
quests[270150081] = 75 -- Consecrated Note (Thunder Bluff)
quests[270171106] = 250 -- Memories from a Lost Past (Argent Crusade)
quests[270200081] = 75 -- The First Lesson (Thunder Bluff)
quests[270210081] = 75 -- The Hunter's Path (Thunder Bluff)
quests[270230081] = 75 -- The Way of the Sunwalkers (Thunder Bluff)
quests[270250054] = 250 -- Thistle While You Work (Gnomeregan)
quests[270260054] = 250 -- Defcon: Bobcat (Gnomeregan)
quests[270270081] = 75 -- Primal Strike (Thunder Bluff)
quests[270280047] = 250 -- Hornet Hunting (Ironforge)
quests[270300047] = 250 -- Foxtails By The Handful (Ironforge)
quests[270310054] = 250 -- Wing Nut (Gnomeregan)
quests[270330054] = 350 -- Skystrider's Heart (Gnomeregan)
quests[270340054] = 75 -- He's That Age (Gnomeregan)
quests[270350054] = 250 -- Standing Up (Gnomeregan)
quests[270380072] = 250 -- Missing Parts (Stormwind)
quests[270391136] = 150 -- Dangerous Intentions (Tranquillien Conversion)
quests[270440072] = 75 -- Peasant Problems (Stormwind)
quests[270450068] = 500 -- Waiting to Exsanguinate (Undercity)
quests[270461171] = 250 -- Fear of Boring (Therazane)
quests[270471171] = 250 -- Motes (Therazane)
quests[270481171] = 250 -- Underground Economy (Therazane)
quests[270491171] = 250 -- Soft Rock (Therazane)
quests[270501171] = 250 -- Fungal Fury (Therazane)
quests[270511171] = 250 -- Through Persistence (Therazane)
quests[270531106] = 75 -- Gahrron's Withering Cauldron (Argent Crusade)
quests[270541106] = 250 -- Desperate Acts (Argent Crusade)
quests[270551106] = 500 -- Students of Krastinov (Argent Crusade)
quests[270560068] = 10 -- Belmont's Report (Undercity)
quests[270571106] = 250 -- Return to the Stead (Argent Crusade)
quests[270600072] = 250 -- Unholy Cow (Stormwind)
quests[270611135] = 25 -- The Twilight Overlook (The Earthen Ring)
quests[270640072] = 150 -- He's Holding Out on Us (Stormwind)
quests[270650068] = 10 -- The Warchief's Fleet (Undercity)
quests[270650076] = 10 -- The Warchief's Fleet (Orgrimmar)
quests[270660081] = 75 -- Learning the Word (Thunder Bluff)
quests[270670081] = 75 -- Moonfire (Thunder Bluff)
quests[270690068] = 250 -- Steel Thunder (Undercity)
quests[270690076] = 250 -- Steel Thunder (Orgrimmar)
quests[270700169] = 500 -- A Fool's Errand (Steamwheedle Cartel)
quests[270720072] = 250 -- The Old Barracks (Stormwind)
quests[270730068] = 250 -- Give 'em Hell! (Undercity)
quests[270730076] = 500 -- Give 'em Hell! (Orgrimmar)
quests[270740047] = 75 -- Fight the Hammer (Ironforge)
quests[270750047] = 350 -- Servants of Cho'gall (Ironforge)
quests[270760169] = 500 -- Breaking and Entering (Steamwheedle Cartel)
quests[270770047] = 350 -- Clutching at Chaos (Ironforge)
quests[270780047] = 500 -- Gor'kresh (Ironforge)
quests[270820068] = 250 -- Playing Dirty (Undercity)
quests[270830068] = 10 -- The Battle Resumes! (Undercity)
quests[270840068] = 250 -- The Farmers' Militia (Undercity)
quests[270850068] = 350 -- Supporting the Troops (Undercity)
quests[270860068] = 250 -- Ashes to Ashes (Undercity)
quests[270870068] = 10 -- Lindsay Ravensun, Revealed (Undercity)
quests[270880068] = 350 -- It's Only Poisonous if You Ingest It (Undercity)
quests[270890068] = 500 -- Ace in the Hole (Undercity)
quests[270900068] = 250 -- Andorhal, Once and For All (Undercity)
quests[270910911] = 250 -- Charge! (Silvermoon City)
quests[270920072] = 350 -- Cataclysm in the Catacombs (Stormwind)
quests[270930068] = 350 -- Lost in the Darkness (Undercity)
quests[270930076] = 350 -- Lost in the Darkness (Orgrimmar)
quests[270940068] = 350 -- Deeper into Darkness (Undercity)
quests[270940076] = 350 -- Deeper into Darkness (Orgrimmar)
quests[270950068] = 250 -- Skitterweb Menace (Undercity)
quests[270950076] = 250 -- Skitterweb Menace (Orgrimmar)
quests[270960068] = 500 -- Orcs are in Order (Undercity)
quests[270960076] = 500 -- Orcs are in Order (Orgrimmar)
quests[270970068] = 350 -- Rise, Forsaken (Undercity)
quests[270980068] = 500 -- Lordaeron (Undercity)
quests[270990068] = 500 -- No Escape (Undercity)
quests[271001135] = 250 -- Twilight Research (The Earthen Ring)
quests[271021135] = 350 -- Maziel's Ascendancy (The Earthen Ring)
quests[271060072] = 350 -- A Villain Unmasked (Stormwind)
quests[271160054] = 250 -- The Winds of Loch Modan (Gnomeregan)
quests[271170068] = 500 -- Brute Strength (Undercity)
quests[271231135] = 10 -- Deepholm, Realm of Earth (The Earthen Ring)
quests[271351135] = 250 -- Something that Burns (The Earthen Ring)
quests[271361135] = 250 -- Elemental Energy (The Earthen Ring)
quests[271391133] = 10 -- Hobart Needs You (Bilgewater Cartel)
quests[271440068] = 250 -- The Reckoning (Undercity)
quests[271511106] = 250 -- Taelan Fordring's Legacy (Argent Crusade)
quests[271521106] = 250 -- Unusual Behavior... Even For Gnolls (Argent Crusade)
quests[271531106] = 250 -- The Good People of Hearthglen (Argent Crusade)
quests[271541106] = 250 -- Bagging Bisp (Argent Crusade)
quests[271551106] = 350 -- Turning Yourself In (Argent Crusade)
quests[271561106] = 250 -- It's About Time! (Argent Crusade)
quests[271571106] = 250 -- Drudges... <Sigh> (Argent Crusade)
quests[271580072] = 25 -- The Battle for Andorhal (Stormwind)
quests[271590072] = 250 -- Scourge First... Horde Later (Stormwind)
quests[271600072] = 350 -- War Machines (Stormwind)
quests[271610072] = 250 -- The Endless Flow (Stormwind)
quests[271620072] = 500 -- Scholomancer (Stormwind)
quests[271630072] = 500 -- Brute Strength (Stormwind)
quests[271640068] = 350 -- Araj the Summoner (Undercity)
quests[271650072] = 10 -- Victory, For Now (Stormwind)
quests[271660072] = 250 -- Go Fletch! (Stormwind)
quests[271670072] = 250 -- A Mighty Hunger (Stormwind)
quests[271680072] = 250 -- Those That Couldn't Let Go (Stormwind)
quests[271710072] = 350 -- Ambushed! (Stormwind)
quests[271720469] = 75 -- The Writhing Haunt (Alliance)
quests[271730469] = 250 -- This Is Our Army (Alliance)
quests[271740469] = 350 -- Combat Training (Alliance)
quests[271750068] = 10 -- The Menders' Stead (Undercity)
quests[271770072] = 250 -- Salvage Operation (Stormwind)
quests[271780072] = 250 -- Naga Reinforcements (Stormwind)
quests[271800068] = 350 -- Honor the Dead (Undercity)
quests[271810068] = 250 -- Excising the Taint (Undercity)
quests[271930068] = 250 -- Seek and Destroy (Undercity)
quests[271940068] = 250 -- Cornered and Crushed! (Undercity)
quests[271950068] = 500 -- Nowhere to Run (Undercity)
quests[271970072] = 10 -- The Battle Resumes! (Stormwind)
quests[271980072] = 350 -- Supporting the Troops (Stormwind)
quests[271990072] = 250 -- Ashes to Ashes (Stormwind)
quests[272000072] = 250 -- Siren's Song (Stormwind)
quests[272010072] = 350 -- Val'kyr Incursion (Stormwind)
quests[272020072] = 350 -- Brother Against Brother (Stormwind)
quests[272031135] = 10 -- The Maelstrom (The Earthen Ring)
quests[272040072] = 500 -- Aradne (Stormwind)
quests[272050072] = 10 -- The Depravity of the Forsaken (Stormwind)
quests[272060072] = 250 -- Alas, Andorhal (Stormwind)
quests[272100072] = 250 -- Traitors Among Us (Stormwind)
quests[272110072] = 250 -- Propaganda War (Stormwind)
quests[272120072] = 250 -- Discrediting the Deserters (Stormwind)
quests[272130072] = 350 -- The End of the Deserters (Stormwind)
quests[272220072] = 350 -- Take Down Tethyr! (Stormwind)
quests[272250072] = 10 -- A Summons from Ander Germaine (Stormwind)
quests[272260076] = 250 -- Hair of the Dog (Orgrimmar)
quests[272260068] = 250 -- Hair of the Dog (Undercity)
quests[272290076] = 350 -- SMASH BROODQUEEN (Orgrimmar)
quests[272310076] = 250 -- Reinforcements from Fenris (Orgrimmar)
quests[272310068] = 250 -- Reinforcements from Fenris (Undercity)
quests[272320076] = 500 -- The Waters Run Red... (Orgrimmar)
quests[272320068] = 500 -- The Waters Run Red... (Undercity)
quests[272330068] = 10 -- The Menders' Stead (Undercity)
quests[272440076] = 350 -- The Lost Report (Orgrimmar)
quests[272460072] = 25 -- The Orc Report (Stormwind)
quests[272470072] = 250 -- Captain Vimes (Stormwind)
quests[272650072] = 10 -- Lord Grayson Shadowbreaker (Stormwind)
quests[272660072] = 10 -- Wulf Calls (Stormwind)
quests[272670072] = 10 -- Make Contact with SI:7 (Stormwind)
quests[272680072] = 10 -- Make Haste to the Cathedral (Stormwind)
quests[272690069] = 10 -- The Temple of the Moon (Darnassus)
quests[272700072] = 10 -- An Audience with the Farseer (Stormwind)
quests[272710072] = 10 -- Journey to the Wizard's Sanctum (Stormwind)
quests[272720072] = 10 -- Demisette Sends Word (Stormwind)
quests[272730609] = 10 -- An Invitation from Moonglade (Cenarion Circle)
quests[272740054] = 10 -- The Chief Surgeon (Gnomeregan)
quests[272750069] = 10 -- Together Again (Darnassus)
quests[272760081] = 10 -- Together Again (Thunder Bluff)
quests[272770076] = 10 -- An Audience with Ureda (Orgrimmar)
quests[272780076] = 10 -- Grimshot's Call (Orgrimmar)
quests[272790076] = 10 -- The Shattered Hand (Orgrimmar)
quests[272800076] = 10 -- The Earthbreaker Calls (Orgrimmar)
quests[272810076] = 10 -- Grezz Ragefist (Orgrimmar)
quests[272820076] = 10 -- Kranosh's Behest (Orgrimmar)
quests[272830609] = 10 -- A Journey to Moonglade (Cenarion Circle)
quests[272900068] = 250 -- To Forsaken Forward Command (Undercity)
quests[272970076] = 500 -- Justice Dispensed (Orgrimmar)
quests[272980911] = 10 -- Seek Out Master Pyreanor (Silvermoon City)
quests[272991135] = 250 -- Torn Ground (The Earthen Ring)
quests[273001135] = 250 -- Pushing Back (The Earthen Ring)
quests[273011135] = 250 -- Unbroken (The Earthen Ring)
quests[273021135] = 250 -- Simple Solutions (The Earthen Ring)
quests[273031135] = 250 -- Mercy for the Bound (The Earthen Ring)
quests[273040081] = 10 -- Follow the Sun (Thunder Bluff)
quests[273100069] = 250 -- No Weapons For You! (Darnassus)
quests[273110081] = 250 -- No Weapons For You! (Thunder Bluff)
quests[273120069] = 250 -- Darkcloud Grimtotem (Darnassus)
quests[273130081] = 250 -- Darkcloud Grimtotem (Thunder Bluff)
quests[273140069] = 350 -- Grimtotem Chiefs: Isha Gloomaxe (Darnassus)
quests[273150081] = 350 -- Grimtotem Chiefs: Isha Gloomaxe (Thunder Bluff)
quests[273160069] = 250 -- The Rattle of Bones (Darnassus)
quests[273170081] = 250 -- The Rattle of Bones (Thunder Bluff)
quests[273180069] = 350 -- Grimtotem Chiefs: Elder Stormhoof (Darnassus)
quests[273190081] = 350 -- Grimtotem Chiefs: Elder Stormhoof (Thunder Bluff)
quests[273200069] = 250 -- The Writ of History (Darnassus)
quests[273210081] = 250 -- The Writ of History (Thunder Bluff)
quests[273220068] = 500 -- Korok the Colossus (Undercity)
quests[273230069] = 350 -- Grimtotem Chiefs: Grundig Darkcloud (Darnassus)
quests[273240081] = 350 -- Grimtotem Chiefs: Grundig Darkcloud (Thunder Bluff)
quests[273250069] = 250 -- The Drums of War (Darnassus)
quests[273260081] = 250 -- The Drums of War (Thunder Bluff)
quests[273270069] = 350 -- Grimtotem Chiefs: The Chief of Chiefs (Darnassus)
quests[273280081] = 350 -- Grimtotem Chiefs: The Chief of Chiefs (Thunder Bluff)
quests[273290069] = 350 -- Invoking the Serpent (Darnassus)
quests[273300081] = 350 -- Invoking the Serpent (Thunder Bluff)
quests[273310076] = 10 -- The Seer's Call (Orgrimmar)
quests[273320076] = 10 -- Seek the Shadow-Walker (Orgrimmar)
quests[273330068] = 350 -- Losing Ground (Undercity)
quests[273340068] = 10 -- Dark Cleric Cecille (Undercity)
quests[273350911] = 10 -- Journey to Orgrimmar (Silvermoon City)
quests[273370072] = 500 -- A Fitting Weapon (Stormwind)
quests[273380072] = 75 -- Our Boys on the Shore (Stormwind)
quests[273410072] = 75 -- Scouting the Shore (Stormwind)
quests[273420068] = 250 -- In Time, All Will Be Revealed (Undercity)
quests[273430072] = 500 -- The Hand of the Light (Stormwind)
quests[273440072] = 500 -- A Well-Earned Reward (Stormwind)
quests[273460369] = 10 -- The Zeppelin Crash (Gadgetzan)
quests[273470369] = 250 -- Corrosion Prevention (Gadgetzan)
quests[273480369] = 250 -- Secure the Cargo! (Gadgetzan)
quests[273490068] = 350 -- Break in Communications: Dreadwatch Outpost (Undercity)
quests[273500068] = 250 -- Break in Communications: Rutsak's Guard (Undercity)
quests[273510072] = 500 -- A Royal Reward (Stormwind)
quests[273530072] = 500 -- Blessings of the Elements (Stormwind)
quests[273540072] = 500 -- Mastering the Arcane (Stormwind)
quests[273550072] = 500 -- A Boon for the Powerful (Stormwind)
quests[273560609] = 500 -- The Circle's Future (Cenarion Circle)
quests[273570069] = 250 -- The Captive Bride (Darnassus)
quests[273580081] = 250 -- The Captive Bride (Thunder Bluff)
quests[273600068] = 250 -- Vengeance for Our Soldiers (Undercity)
quests[273610072] = 500 -- Favored of the Light (Stormwind)
quests[273620072] = 500 -- Favored of Elune (Stormwind)
quests[273630072] = 500 -- A Budding Young Surgeon (Stormwind)
quests[273640068] = 350 -- On Whose Orders? (Undercity)
quests[273650076] = 500 -- A Fitting Weapon (Orgrimmar)
quests[273660072] = 350 -- Landgrab (Stormwind)
quests[273741135] = 25 -- The Maw of Madness (The Earthen Ring)
quests[273751135] = 25 -- The Weeping Wound (The Earthen Ring)
quests[273761135] = 250 -- The Maw of Iso'rath (The Earthen Ring)
quests[273771135] = 250 -- Devoured (The Earthen Ring)
quests[273781135] = 250 -- The Worldbreaker (The Earthen Ring)
quests[273791135] = 250 -- The Terrors of Iso'rath (The Earthen Ring)
quests[273801135] = 500 -- Nightmare (The Earthen Ring)
quests[273821106] = 350 -- Rough Roads (Argent Crusade)
quests[273950076] = 500 -- A Marksman's Weapon (Orgrimmar)
quests[273960076] = 500 -- Blade of the Shattered Hand (Orgrimmar)
quests[273970076] = 500 -- Terga's Task (Orgrimmar)
quests[274000076] = 500 -- Mastering the Arcane (Orgrimmar)
quests[274010068] = 350 -- What Tomorrow Brings (Undercity)
quests[274020076] = 500 -- Token of Power (Orgrimmar)
quests[274030076] = 500 -- A True Sunwalker (Orgrimmar)
quests[274040609] = 500 -- The Circle's Future (Cenarion Circle)
quests[274050068] = 250 -- Fall Back! (Undercity)
quests[274060068] = 500 -- A Man Named Godfrey (Undercity)
quests[274120369] = 250 -- WANTED: Goreclaw the Ravenous (Gadgetzan)
quests[274130369] = 250 -- Catch a Dragon by the Tail (Gadgetzan)
quests[274200609] = 250 -- Postponing the Inevitable (Cenarion Circle)
quests[274210609] = 250 -- Amidst Death, Life (Cenarion Circle)
quests[274230068] = 500 -- Resistance is Futile (Undercity)
quests[274321106] = 250 -- Zaeldarr the Outcast (Argent Crusade)
quests[274330072] = 250 -- Shredderectomy (Stormwind)
quests[274340076] = 500 -- The Adept's Path (Orgrimmar)
quests[274350076] = 500 -- A Seer's Staff (Orgrimmar)
quests[274360076] = 500 -- The Shadow-Walker's Task (Orgrimmar)
quests[274370076] = 500 -- The Dark Cleric's Bidding (Orgrimmar)
quests[274380068] = 1000 -- The Great Escape (Undercity)
quests[274390076] = 500 -- Staff of the Light (Orgrimmar)
quests[274410076] = 500 -- A Seer's Staff (Orgrimmar)
quests[274460054] = 10 -- Tanaris is Calling (Gnomeregan)
quests[274471133] = 10 -- Tanaris is Calling (Bilgewater Cartel)
quests[274501106] = 250 -- Frederick's Fish Fancy (Argent Crusade)
quests[274511106] = 250 -- To Kill With Purpose (Argent Crusade)
quests[274521106] = 250 -- Dark Garb (Argent Crusade)
quests[274531106] = 250 -- Catalysm (Argent Crusade)
quests[274541106] = 250 -- Just a Drop in the Bucket (Argent Crusade)
quests[274561106] = 250 -- Gathering Some Grub(s) (Argent Crusade)
quests[274591106] = 10 -- The Brotherhood of Light (Argent Crusade)
quests[274601106] = 250 -- Soft Landing (Argent Crusade)
quests[274611106] = 10 -- To Take the Abbey (Argent Crusade)
quests[274621106] = 10 -- To Take the Barracks (Argent Crusade)
quests[274631106] = 10 -- A Boyhood Dream (Argent Crusade)
quests[274641106] = 500 -- Argent Call: The Trial of the Crypt (Argent Crusade)
quests[274651106] = 350 -- Argent Call: The Noxious Glade (Argent Crusade)
quests[274671106] = 250 -- Buried Blades (Argent Crusade)
quests[274680072] = 250 -- Siege Tank Rescue (Stormwind)
quests[274720068] = 1000 -- Rise, Godfrey (Undercity)
quests[274740068] = 350 -- Breaking the Barrier (Undercity)
quests[274750068] = 250 -- Unyielding Servitors (Undercity)
quests[274760068] = 250 -- Dalar Dawnweaver (Undercity)
quests[274771106] = 250 -- The Corpulent One (Argent Crusade)
quests[274780068] = 250 -- Relios the Relic Keeper (Undercity)
quests[274791106] = 250 -- Righteous Indignation (Argent Crusade)
quests[274800068] = 350 -- Ley Energies (Undercity)
quests[274830068] = 250 -- Practical Vengeance (Undercity)
quests[274840068] = 250 -- Only One May Enter (Undercity)
quests[274900072] = 250 -- SI:7 Drop (Stormwind)
quests[274910076] = 250 -- Kor'kron Drop (Orgrimmar)
quests[274920072] = 250 -- Ogres & Ettins (Stormwind)
quests[274931172] = 250 -- Ogres & Ettins (Dragonmaw Clan)
quests[274940072] = 250 -- Move the Mountain (Stormwind)
quests[274951172] = 250 -- Move the Mountain (Dragonmaw Clan)
quests[274960072] = 250 -- Call in the Artillery (Stormwind)
quests[274970076] = 250 -- Call in the Artillery (Orgrimmar)
quests[274991172] = 250 -- Signal the Attack (Dragonmaw Clan)
quests[275020072] = 250 -- Up to the Citadel (Stormwind)
quests[275030076] = 250 -- Up to the Citadel (Orgrimmar)
quests[275100076] = 250 -- A Wolf in Bear's Clothing (Orgrimmar)
quests[275100068] = 250 -- A Wolf in Bear's Clothing (Undercity)
quests[275120068] = 250 -- Transdimensional Warfare: Chapter I (Undercity)
quests[275130068] = 350 -- Transdimensional Warfare: Chapter II (Undercity)
quests[275141174] = 25 -- Bird Down! Bird Down! (Wildhammer Clan)
quests[275151174] = 250 -- Roots'll Do 'Er (Wildhammer Clan)
quests[275161174] = 25 -- Wings Over Highbank (Wildhammer Clan)
quests[275180068] = 500 -- Transdimensional Warfare: Chapter III (Undercity)
quests[275191173] = 250 -- Under the Choking Sands (Ramkahen)
quests[275201173] = 250 -- Minions of Al'Akir (Ramkahen)
quests[275251106] = 350 -- Guardians of Stratholme (Argent Crusade)
quests[275261106] = 500 -- Gidwin's Fate Revealed (Argent Crusade)
quests[275271106] = 250 -- Journey's End (Argent Crusade)
quests[275281106] = 250 -- Scourged Mass (Argent Crusade)
quests[275291106] = 250 -- Defenders of Darrowshire (Argent Crusade)
quests[275301106] = 350 -- The Corpsebeasts (Argent Crusade)
quests[275311106] = 250 -- Counter-Plague Research (Argent Crusade)
quests[275321106] = 10 -- The Plaguewood Tower (Argent Crusade)
quests[275351106] = 10 -- Just a Little Touched (Argent Crusade)
quests[275370072] = 250 -- Any Portal in a Storm (Stormwind)
quests[275380072] = 250 -- The Perfect Poultice (Stormwind)
quests[275391106] = 350 -- Add 'em to the Pile (Argent Crusade)
quests[275420068] = 10 -- Taking the Battlefront (Undercity)
quests[275440609] = 10 -- Cenarion Tenacity (Cenarion Circle)
quests[275450072] = 75 -- The Way is Open (Stormwind)
quests[275470068] = 250 -- Of No Consequence (Undercity)
quests[275480068] = 250 -- Lessons in Fear (Undercity)
quests[275500068] = 350 -- Pyrewood's Fall (Undercity)
quests[275511106] = 250 -- The Baroness' Missive (Argent Crusade)
quests[275740068] = 250 -- I Never Forget a Face (Undercity)
quests[275750068] = 350 -- From the Belly of the Beast (Undercity)
quests[275761133] = 75 -- Patchwork Command (Bilgewater Cartel)
quests[275770068] = 250 -- 7th Legion Battle Plans (Undercity)
quests[275800068] = 350 -- Sowing Discord (Undercity)
quests[275831172] = 75 -- The Northern Flank (Dragonmaw Clan)
quests[275841172] = 350 -- Blood in the Surf (Dragonmaw Clan)
quests[275861133] = 250 -- Shells on the Sea Shore (Bilgewater Cartel)
quests[275880072] = 250 -- Signal the Attack (Stormwind)
quests[275901172] = 250 -- Signal the Attack (Dragonmaw Clan)
quests[275940068] = 500 -- On Her Majesty's Secret Service (Undercity)
quests[275951173] = 75 -- The Prophet Hadassi (Ramkahen)
quests[276010068] = 1000 -- Cities in Dust (Undercity)
quests[276021173] = 75 -- The Prophet's Dying Words (Ramkahen)
quests[276061133] = 250 -- Blast Him! (Bilgewater Cartel)
quests[276061172] = 250 -- Blast Him! (Dragonmaw Clan)
quests[276070076] = 75 -- The Southern Flank (Orgrimmar)
quests[276080072] = 250 -- Four Heads are Better than None (Stormwind)
quests[276090076] = 250 -- Four Heads are Better than None (Orgrimmar)
quests[276100076] = 250 -- Scouting the Shore (Orgrimmar)
quests[276110076] = 250 -- Blood on the Sand (Orgrimmar)
quests[276121106] = 250 -- Victory From Within (Argent Crusade)
quests[276131106] = 250 -- The Assassin (Argent Crusade)
quests[276141106] = 250 -- Scarlet Salvage (Argent Crusade)
quests[276151106] = 250 -- The Wrathcaster (Argent Crusade)
quests[276161106] = 250 -- The Huntsman (Argent Crusade)
quests[276171106] = 10 -- Befouled No More (Argent Crusade)
quests[276181106] = 10 -- Argent Upheaval (Argent Crusade)
quests[276191106] = 250 -- The Commander (Argent Crusade)
quests[276201106] = 500 -- Like Rats (Argent Crusade)
quests[276211174] = 25 -- Firebeard's Patrol (Wildhammer Clan)
quests[276221133] = 250 -- Mo' Better Shredder (Bilgewater Cartel)
quests[276231173] = 250 -- Colossal Guardians (Ramkahen)
quests[276250930] = 350 -- In Defense of Quel'Danil (Exodar)
quests[276260930] = 350 -- The Highvale Documents (Exodar)
quests[276281173] = 150 -- Send Word to Phaoris (Ramkahen)
quests[276291173] = 75 -- The Vizier's Vote (Ramkahen)
quests[276301173] = 10 -- The High Priest's Vote (Ramkahen)
quests[276311173] = 75 -- The High Commander's Vote (Ramkahen)
quests[276321173] = 150 -- Tanotep's Son (Ramkahen)
quests[276350054] = 150 -- Decontamination (Gnomeregan)
quests[276360072] = 75 -- Just You and Mathias (Stormwind)
quests[276380076] = 75 -- Just You and Garona (Orgrimmar)
quests[276401174] = 25 -- Dunwalds Don't Die (Wildhammer Clan)
quests[276411174] = 250 -- While Meeting The Family (Wildhammer Clan)
quests[276421174] = 250 -- Sifting Through The Wreckage (Wildhammer Clan)
quests[276431174] = 25 -- Donnelly Dunwald (Wildhammer Clan)
quests[276441174] = 25 -- Eoin Dunwald (Wildhammer Clan)
quests[276451174] = 25 -- Cayden Dunwald (Wildhammer Clan)
quests[276461174] = 250 -- Finding Beak (Wildhammer Clan)
quests[276471174] = 250 -- Anything But Water (Wildhammer Clan)
quests[276481174] = 250 -- Once More Into The Fire (Wildhammer Clan)
quests[276491174] = 250 -- A Steady Supply (Wildhammer Clan)
quests[276501174] = 250 -- Home Again (Wildhammer Clan)
quests[276511174] = 250 -- Doing It Like a Dunwald (Wildhammer Clan)
quests[276520072] = 250 -- Dark Assassins (Stormwind)
quests[276530076] = 250 -- Dark Assassins (Orgrimmar)
quests[276540072] = 250 -- Bring the Hammer Down (Stormwind)
quests[276550076] = 250 -- Bring the Hammer Down (Orgrimmar)
quests[276571135] = 75 -- Help from the Earthcaller (The Earthen Ring)
quests[276581135] = 75 -- Help from the Earthcaller (The Earthen Ring)
quests[276591135] = 250 -- Portal Overload (The Earthen Ring)
quests[276601135] = 75 -- Spirit of the Loch (The Earthen Ring)
quests[276611135] = 250 -- Fire the Cannon (The Earthen Ring)
quests[276621135] = 250 -- Unbinding (The Earthen Ring)
quests[276640909] = 350 -- Darkmoon Volcanic Deck (Darkmoon Faire)
quests[276650909] = 350 -- Darkmoon Hurricane Deck (Darkmoon Faire)
quests[276660909] = 350 -- Darkmoon Tsunami Deck (Darkmoon Faire)
quests[276670909] = 350 -- Darkmoon Earthquake Deck (Darkmoon Faire)
quests[276700054] = 250 -- Pinned Down (Gnomeregan)
quests[276710054] = 250 -- See to the Survivors (Gnomeregan)
quests[276880072] = 350 -- Distract Them for Me (Stormwind)
quests[276890076] = 350 -- Distract Them for Me (Orgrimmar)
quests[276901172] = 75 -- Narkrall, the Drake-Tamer (Dragonmaw Clan)
quests[276950072] = 250 -- The Elementium Axe (Stormwind)
quests[276960076] = 250 -- The Elementium Axe (Orgrimmar)
quests[277000072] = 250 -- Dragon, Unchained (Stormwind)
quests[277010076] = 250 -- Dragon, Unchained (Orgrimmar)
quests[277020072] = 250 -- Coup de Grace (Stormwind)
quests[277030076] = 250 -- Coup de Grace (Orgrimmar)
quests[277061173] = 150 -- The Scepter of Orsis (Ramkahen)
quests[277071173] = 250 -- Neferset Prison (Ramkahen)
quests[277081135] = 350 -- The Warden's Time (The Earthen Ring)
quests[277111135] = 10 -- Back to the Elementium Depths (The Earthen Ring)
quests[277121135] = 10 -- Back to the Elementium Depths (The Earthen Ring)
quests[277180076] = 10 -- Warchief's Command: Vashj'ir! (Orgrimmar)
quests[277190072] = 350 -- Water of Life (Stormwind)
quests[277201135] = 250 -- Mr. Goldmine's Wild Ride (The Earthen Ring)
quests[277211158] = 10 -- Warchief's Command: Mount Hyjal! (Guardians of Hyjal)
quests[277220076] = 10 -- Warchief's Command: Deepholm! (Orgrimmar)
quests[277240072] = 10 -- Hero's Call: Vashj'ir! (Stormwind)
quests[277260072] = 10 -- Hero's Call: Mount Hyjal! (Stormwind)
quests[277270072] = 10 -- Hero's Call: Deepholm! (Stormwind)
quests[277330072] = 500 -- The Good Ol' Switcheroo (Stormwind)
quests[277381173] = 350 -- The Pit of Scales (Ramkahen)
quests[277421135] = 250 -- A Little on the Side (The Earthen Ring)
quests[277431135] = 250 -- While We're Here (The Earthen Ring)
quests[277441135] = 250 -- Rune Ruination (The Earthen Ring)
quests[277451135] = 350 -- A Fiery Reunion (The Earthen Ring)
quests[277460068] = 250 -- Empire of Dirt (Undercity)
quests[277471172] = 350 -- Total War (Dragonmaw Clan)
quests[277501172] = 350 -- War Forage (Dragonmaw Clan)
quests[277511172] = 500 -- Crushing the Wildhammer (Dragonmaw Clan)
quests[277521174] = 250 -- Fight Like a Wildhammer (Wildhammer Clan)
quests[277531174] = 250 -- Never Leave a Dinner Behind (Wildhammer Clan)
quests[277541174] = 250 -- Dragonmaw Takedown (Wildhammer Clan)
quests[277551173] = 250 -- The Curse of the Tombs (Ramkahen)
quests[277560072] = 500 -- The Foreman (Stormwind)
quests[277580072] = 500 -- The Carpenter (Stormwind)
quests[277630169] = 10 -- To Fuselight Proper (Steamwheedle Cartel)
quests[277640169] = 75 -- A Strange Request (Steamwheedle Cartel)
quests[277740169] = 250 -- Easily Swayed (Steamwheedle Cartel)
quests[277760169] = 250 -- It's Goat Time, Baby (Steamwheedle Cartel)
quests[277810072] = 500 -- The Machination (Stormwind)
quests[277840072] = 250 -- The Hammer of Twilight (Stormwind)
quests[277850072] = 500 -- The Admiral (Stormwind)
quests[277860076] = 250 -- The Hammer of Twilight (Orgrimmar)
quests[277871174] = 500 -- Skullcrusher the Mountain (Wildhammer Clan)
quests[277870072] = 250 -- Skullcrusher the Mountain (Stormwind)
quests[277881172] = 500 -- Skullcrusher the Mountain (Dragonmaw Clan)
quests[277880076] = 250 -- Skullcrusher the Mountain (Orgrimmar)
quests[277900072] = 500 -- The Defias Kingpin (Stormwind)
quests[277910047] = 10 -- Dustwind Dig (Ironforge)
quests[277920047] = 250 -- All's Fair in Love, War, and Archaeology (Ironforge)
quests[277950072] = 250 -- Tides of Darkness (Stormwind)
quests[277980076] = 350 -- Water of Life (Orgrimmar)
quests[278031174] = 250 -- Welcome Relief (Wildhammer Clan)
quests[278041174] = 250 -- The Only Homes We Have (Wildhammer Clan)
quests[278051174] = 250 -- Small Comforts (Wildhammer Clan)
quests[278061174] = 25 -- Honorable Bearing (Wildhammer Clan)
quests[278071174] = 25 -- Clan Mullan (Wildhammer Clan)
quests[278081174] = 250 -- Stubborn as a Doyle (Wildhammer Clan)
quests[278091174] = 25 -- Firebeard Bellows (Wildhammer Clan)
quests[278101174] = 250 -- The Fighting Spirit (Wildhammer Clan)
quests[278111174] = 250 -- The Scent of Battle (Wildhammer Clan)
quests[278121174] = 150 -- The Loyalty of Clan Mullan (Wildhammer Clan)
quests[278131174] = 250 -- Death Worthy of a Dragonmaw (Wildhammer Clan)
quests[278141174] = 250 -- Anything We Can Get (Wildhammer Clan)
quests[278151174] = 25 -- Somethin' for the Boys (Wildhammer Clan)
quests[278161174] = 25 -- Personal Request (Wildhammer Clan)
quests[278171174] = 250 -- Dropping the Hammer (Wildhammer Clan)
quests[278210072] = 250 -- Orcs and Humans (Stormwind)
quests[278220072] = 250 -- Lumbering Oafs (Stormwind)
quests[278230047] = 250 -- A Dwarf's Got Needs (Ironforge)
quests[278240047] = 250 -- The Good Stuff (Ironforge)
quests[278250072] = 250 -- Survival of the Fattest (Stormwind)
quests[278330072] = 250 -- Half-Ton Holdouts (Stormwind)
quests[278361173] = 250 -- Stopping the Spread (Ramkahen)
quests[278371173] = 250 -- Trespassers in the Water (Ramkahen)
quests[278381173] = 250 -- The Root of the Corruption (Ramkahen)
quests[278391173] = 250 -- Ancient Weapons (Ramkahen)
quests[278400930] = 250 -- Draenethyst Crystals (Exodar)
quests[278420076] = 500 -- Only the Beginning (Orgrimmar)
quests[278430072] = 250 -- Reinforcements Denied (Stormwind)
quests[278440076] = 500 -- Traitors!!! (Orgrimmar)
quests[278450072] = 250 -- Marking the Fallen (Stormwind)
quests[278470076] = 500 -- Not Quite There (Orgrimmar)
quests[278480076] = 500 -- Good Intentions...Poor Execution (Orgrimmar)
quests[278490072] = 250 -- Assault on Stonard (Stormwind)
quests[278500076] = 500 -- The Defias Kingpin (Orgrimmar)
quests[278510072] = 350 -- Cutting Supply (Stormwind)
quests[278520076] = 250 -- Orcs and Humans (Orgrimmar)
quests[278530076] = 250 -- Lumbering Oafs (Orgrimmar)
quests[278540076] = 250 -- Tides of Darkness (Orgrimmar)
quests[278550076] = 250 -- Reinforcements Denied (Orgrimmar)
quests[278560076] = 250 -- Marking the Fallen (Orgrimmar)
quests[278570076] = 350 -- We're Under Attack! (Orgrimmar)
quests[278600930] = 250 -- The Lost Ones (Exodar)
quests[278741158] = 75 -- Aviana's Legacy (Guardians of Hyjal)
quests[278750930] = 250 -- Prayerblossom (Exodar)
quests[278760930] = 250 -- Secrets of the Mire (Exodar)
quests[278770911] = 250 -- The Morons' League (Silvermoon City)
quests[278780911] = 250 -- Forcible Acquisition (Silvermoon City)
quests[278790076] = 250 -- Survival of the Fattest (Orgrimmar)
quests[278800076] = 250 -- Half-Ton Holdouts (Orgrimmar)
quests[278810911] = 10 -- Bloodwatcher Point (Silvermoon City)
quests[278820911] = 250 -- It's Not About History, It's About Power (Silvermoon City)
quests[279020930] = 250 -- The Purespring (Exodar)
quests[279040930] = 250 -- Breath of Mist (Exodar)
quests[279070076] = 250 -- Prayerblossom (Orgrimmar)
quests[279080076] = 250 -- Secrets of the Mire (Orgrimmar)
quests[279090076] = 250 -- The Purespring (Orgrimmar)
quests[279100076] = 250 -- Last Regrets (Orgrimmar)
quests[279171134] = 500 -- Sniffing Them Out (Gilneas)
quests[279201134] = 500 -- Armored to the Teeth (Gilneas)
quests[279211134] = 500 -- Fighting Tooth and Claw (Gilneas)
quests[279221173] = 250 -- Traitors! (Ramkahen)
quests[279291172] = 350 -- Drag 'em Down (Dragonmaw Clan)
quests[279381135] = 350 -- The Middle Fragment (The Earthen Ring)
quests[279441177] = 250 -- Thinning the Brood (Baradin's Wardens)
quests[279451172] = 250 -- Paint it Black (Dragonmaw Clan)
quests[279481177] = 250 -- A Sticky Task (Baradin's Wardens)
quests[279491177] = 250 -- The Forgotten (Baradin's Wardens)
quests[279580059] = 250 -- A Proper Antivenom (Thorium Brotherhood)
quests[279600059] = 250 -- The Fewer, the Better (Thorium Brotherhood)
quests[279610059] = 250 -- Out of Place (Thorium Brotherhood)
quests[279620059] = 250 -- A Lumbering Relic (Thorium Brotherhood)
quests[279630059] = 250 -- A New Master... But Who? (Thorium Brotherhood)
quests[279640059] = 250 -- Dig-Boss Dinwhisker (Thorium Brotherhood)
quests[279650059] = 10 -- Thorium Point: The Seat of the Brotherhood (Thorium Brotherhood)
quests[279661177] = 250 -- Salvaging the Remains (Baradin's Wardens)
quests[279671177] = 250 -- First Lieutenant Connor (Baradin's Wardens)
quests[279681134] = 500 -- Fury of the Pack (Gilneas)
quests[279701177] = 250 -- Captain P. Harris (Baradin's Wardens)
quests[279711177] = 250 -- Rattling Their Cages (Baradin's Wardens)
quests[279721177] = 250 -- Boosting Morale (Baradin's Wardens)
quests[279731177] = 250 -- Watch Out For Splinters! (Baradin's Wardens)
quests[279740068] = 500 -- This Land is Our Land (Undercity)
quests[279751177] = 250 -- WANTED: Foreman Wellson (Baradin's Wardens)
quests[279760059] = 250 -- Curse These Fat Fingers (Thorium Brotherhood)
quests[279770059] = 250 -- Recon Essentials (Thorium Brotherhood)
quests[279781177] = 250 -- Ghostbuster (Baradin's Wardens)
quests[279790059] = 350 -- Dark Ministry (Thorium Brotherhood)
quests[279800059] = 250 -- The Spiders Have to Go (Thorium Brotherhood)
quests[279810059] = 250 -- Heat That Just Don't Quit (Thorium Brotherhood)
quests[279860059] = 10 -- In the Hall of the Mountain-Lord (Thorium Brotherhood)
quests[279871177] = 250 -- Cannonball! (Baradin's Wardens)
quests[279880068] = 500 -- Plague...Plague Everywhere! (Undercity)
quests[279890576] = 1000 -- Ruumbo Demands Honey (Timbermaw Hold)
quests[279911177] = 250 -- Taking the Overlook Back (Baradin's Wardens)
quests[279921177] = 250 -- Magnets, How Do They Work? (Baradin's Wardens)
quests[279940576] = 250 -- Ruumbo Demands Justice (Timbermaw Hold)
quests[279950576] = 1000 -- Dance for Ruumbo! (Timbermaw Hold)
quests[279960068] = 500 -- Orders Are For the Living (Undercity)
quests[279980068] = 500 -- Sweet, Merciless Revenge (Undercity)
quests[280011174] = 25 -- A Coward's Due (Wildhammer Clan)
quests[280350059] = 350 -- The Mountain-Lord's Support (Thorium Brotherhood)
quests[280411172] = 350 -- Bait and Throttle (Dragonmaw Clan)
quests[280431172] = 350 -- How to Maim Your Dragon (Dragonmaw Clan)
quests[280461177] = 250 -- Finish The Job (Baradin's Wardens)
quests[280501177] = 350 -- Shark Tank (Baradin's Wardens)
quests[280520059] = 350 -- Operation: Stir the Cauldron (Thorium Brotherhood)
quests[280540059] = 250 -- Slavery is Bad (Thorium Brotherhood)
quests[280550059] = 250 -- Sweet, Horrible Freedom (Thorium Brotherhood)
quests[280560059] = 250 -- Rise, Obsidion (Thorium Brotherhood)
quests[280570059] = 250 -- Kill 'em With Sleep Deprivation (Thorium Brotherhood)
quests[280580059] = 250 -- Look at the Size of It! (Thorium Brotherhood)
quests[280591177] = 250 -- Claiming The Keep (Baradin's Wardens)
quests[280600059] = 350 -- Twisted Twilight Ties (Thorium Brotherhood)
quests[280610059] = 250 -- Minions of Calcinder (Thorium Brotherhood)
quests[280620059] = 350 -- From Whence He Came (Thorium Brotherhood)
quests[280631177] = 250 -- Leave No Weapon Behind (Baradin's Wardens)
quests[280640059] = 500 -- Welcome to the Brotherhood (Thorium Brotherhood)
quests[280651177] = 250 -- Walk A Mile In Their Shoes (Baradin's Wardens)
quests[280840076] = 10 -- Might of the Krom'gar (Orgrimmar)
quests[280890068] = 250 -- Warchief's Command: Hillsbrad Foothills! (Undercity)
quests[280901133] = 250 -- Precious Goods (Bilgewater Cartel)
quests[280911133] = 250 -- Easy Pickings (Bilgewater Cartel)
quests[280921133] = 250 -- If The Key Fits (Bilgewater Cartel)
quests[280941133] = 250 -- Paving the Way (Bilgewater Cartel)
quests[280960068] = 250 -- Welcome to the Machine (Undercity)
quests[280970076] = 10 -- The Gates of Grim Batol (Orgrimmar)
quests[280990059] = 350 -- Rasha'krak (Thorium Brotherhood)
quests[281000576] = 300 -- A Talking Totem (Timbermaw Hold)
quests[281010072] = 75 -- Mathias' Command (Stormwind)
quests[281030072] = 250 -- Easy Pickings (Stormwind)
quests[281040072] = 250 -- Precious Goods (Stormwind)
quests[281051173] = 250 -- Kavem the Callous (Ramkahen)
quests[281070072] = 250 -- Paving the Way (Stormwind)
quests[281080072] = 250 -- If The Key Fits (Stormwind)
quests[281110068] = 10 -- Trouble at Azurelode (Undercity)
quests[281121173] = 150 -- Escape From the Lost City (Ramkahen)
quests[281140068] = 250 -- Glorious Harvest (Undercity)
quests[281150068] = 350 -- Past Their Prime (Undercity)
quests[281171177] = 350 -- Clearing the Depths (Baradin's Wardens)
quests[281181177] = 350 -- The Imprisoned Archmage (Baradin's Wardens)
quests[281201177] = 350 -- Learning From The Past (Baradin's Wardens)
quests[281221177] = 350 -- A Huge Problem (Baradin's Wardens)
quests[281231172] = 350 -- The Demon Chain (Dragonmaw Clan)
quests[281301177] = 250 -- Not The Friendliest Town (Baradin's Wardens)
quests[281331172] = 1750 -- Fury Unbound (Dragonmaw Clan)
quests[281341173] = 250 -- Impending Retribution (Ramkahen)
quests[281351173] = 150 -- Al'Akir's Vengeance (Ramkahen)
quests[281371177] = 250 -- Teach A Man To Fish.... Or Steal (Baradin's Wardens)
quests[281380068] = 250 -- Human Infestation (Undercity)
quests[281411173] = 250 -- Relics of the Sun King (Ramkahen)
quests[281440068] = 250 -- Thieving Little Monsters! (Undercity)
quests[281451173] = 250 -- Venomblood Antidote (Ramkahen)
quests[281460068] = 250 -- Coastal Delicacies! (Undercity)
quests[281470076] = 250 -- Purple is Your Color (Orgrimmar)
quests[281490076] = 250 -- Whispers in the Wind (Orgrimmar)
quests[281510076] = 250 -- Dressed to Kill (Orgrimmar)
quests[281540068] = 250 -- Muckgill's Flipper or Something... (Undercity)
quests[281560068] = 500 -- Deep Mine Rescue (Undercity)
quests[281600054] = 500 -- Spread the Word (Gnomeregan)
quests[281611133] = 500 -- Spread the Word (Bilgewater Cartel)
quests[281621177] = 350 -- Swamp Bait (Baradin's Wardens)
quests[281631177] = 350 -- The Leftovers (Baradin's Wardens)
quests[281640076] = 10 -- Seek Brother Silverhallow (Orgrimmar)
quests[281651177] = 350 -- D-Block (Baradin's Wardens)
quests[281660076] = 250 -- Thog's Nightlight (Orgrimmar)
quests[281670054] = 10 -- Report to Carvo Blastbolt (Gnomeregan)
quests[281680068] = 10 -- Trouble at the Sludge Fields (Undercity)
quests[281690054] = 10 -- Withdraw to the Loading Room! (Gnomeregan)
quests[281700076] = 250 -- Night Terrors (Orgrimmar)
quests[281851177] = 350 -- Svarnos (Baradin's Wardens)
quests[281861177] = 350 -- Cursed Shackles (Baradin's Wardens)
quests[281881177] = 350 -- Prison Revolt (Baradin's Wardens)
quests[281890068] = 500 -- Do the Right Thing (Undercity)
quests[281920068] = 250 -- No One Here Gets Out Alive (Undercity)
quests[281960068] = 350 -- The Battle for Hillsbrad (Undercity)
quests[281970068] = 250 -- Discretion is Key (Undercity)
quests[281981173] = 10 -- The Weakest Link (Ramkahen)
quests[281990068] = 250 -- For Science! (Undercity)
quests[282001173] = 250 -- The Element of Supplies (Ramkahen)
quests[282011173] = 250 -- Ploughshares to Swords (Ramkahen)
quests[282020059] = 350 -- A Perfect Costume (Thorium Brotherhood)
quests[282030059] = 350 -- A Perfect Costume (Thorium Brotherhood)
quests[282040059] = 350 -- A Perfect Costume (Thorium Brotherhood)
quests[282050059] = 350 -- A Perfect Costume (Thorium Brotherhood)
quests[282060068] = 75 -- Little Girl Lost (Undercity)
quests[282090068] = 250 -- Freedom for Lydon (Undercity)
quests[282101173] = 250 -- Shaping Up (Ramkahen)
quests[282110072] = 250 -- Tempered in Elemental Flame (Stormwind)
quests[282120072] = 250 -- Hot Stuff (Stormwind)
quests[282150072] = 250 -- Potential Energy (Stormwind)
quests[282160072] = 250 -- Magmalord Falthazar (Stormwind)
quests[282231177] = 350 -- The Warden (Baradin's Wardens)
quests[282250059] = 250 -- Prove Yer Allegiance (Thorium Brotherhood)
quests[282260059] = 250 -- Scrapped Golems (Thorium Brotherhood)
quests[282270059] = 350 -- Golem Training (Thorium Brotherhood)
quests[282300068] = 350 -- Protocol (Undercity)
quests[282310068] = 350 -- Do it for Twinkles (Undercity)
quests[282310911] = 350 -- Do it for Twinkles (Silvermoon City)
quests[282321177] = 350 -- Food From Below (Baradin's Wardens)
quests[282350068] = 350 -- Burnside Must Fall (Undercity)
quests[282370068] = 500 -- A Blight Upon the Land (Undercity)
quests[282380072] = 75 -- Twilight Shores (Stormwind)
quests[282441174] = 75 -- Eye Spy (Wildhammer Clan)
quests[282481174] = 25 -- Victors' Point (Wildhammer Clan)
quests[282491172] = 25 -- Crushblow (Dragonmaw Clan)
quests[282501173] = 150 -- Thieving Little Pluckers (Ramkahen)
quests[282510068] = 25 -- Trouble at Southshore (Undercity)
quests[282540059] = 10 -- A Needle in a Hellhole (Thorium Brotherhood)
quests[282580072] = 10 -- Meet with Ander Germaine (Stormwind)
quests[282590072] = 10 -- Meet with Demisette Cloyce (Stormwind)
quests[282600072] = 10 -- Meet with Wulf Hansreim (Stormwind)
quests[282620072] = 10 -- Meet with Lord Tony Romano (Stormwind)
quests[282630072] = 10 -- Meet with Maginor Dumas (Stormwind)
quests[282680072] = 10 -- Meet with Lord Grayson Shadowbreaker (Stormwind)
quests[282751177] = 250 -- Bombs Away! (Baradin's Wardens)
quests[282761173] = 75 -- Salhet's Secret (Ramkahen)
quests[282771173] = 250 -- Salhet the Tactician (Ramkahen)
quests[282801174] = 250 -- Tear Them From the Sky! (Wildhammer Clan)
quests[282811174] = 250 -- Last Stand at Thundermar (Wildhammer Clan)
quests[282821174] = 250 -- Narkrall, The Drake-Tamer (Wildhammer Clan)
quests[282850072] = 10 -- Meet with High Priestess Laurena (Stormwind)
quests[282870072] = 10 -- Meet with Farseer Umbrua (Stormwind)
quests[282890609] = 10 -- Moonglade Calls (Cenarion Circle)
quests[282900076] = 10 -- Meet with Grezz Ragefist (Orgrimmar)
quests[282911173] = 150 -- Return to Camp (Ramkahen)
quests[282941174] = 75 -- The Kirthaven Summit (Wildhammer Clan)
quests[282970076] = 10 -- Meet with Ormak Grimshot (Orgrimmar)
quests[282980076] = 10 -- Meet with Gordul (Orgrimmar)
quests[282990076] = 10 -- Meet with Kranosh (Orgrimmar)
quests[283000076] = 10 -- Meet with Ureda (Orgrimmar)
quests[283010076] = 10 -- Meet with Terga Earthbreaker (Orgrimmar)
quests[283020076] = 10 -- Meet with Sunwalker Atohmo (Orgrimmar)
quests[283030076] = 10 -- Meet with Master Pyreanor (Orgrimmar)
quests[283040068] = 10 -- Meet with Dark Cleric Cecille (Undercity)
quests[283070911] = 10 -- Meet with Tyelis (Silvermoon City)
quests[283080081] = 10 -- Meet with Seer Liwatha (Thunder Bluff)
quests[283090530] = 10 -- Meet with Shadow-Walker Zuru (Darkspear Trolls)
quests[283231133] = 10 -- Meet with Brother Silverhallow (Bilgewater Cartel)
quests[283240068] = 250 -- Studies in Lethality (Undercity)
quests[283250068] = 250 -- Green Living (Undercity)
quests[283280072] = 350 -- Twilight Scheming (Stormwind)
quests[283290068] = 250 -- Angry Scrubbing Bubbles (Undercity)
quests[283300068] = 350 -- Trail of Filth (Undercity)
quests[283310068] = 500 -- Helcular's Rod Giveth... (Undercity)
quests[283320068] = 250 -- Helcular's Command (Undercity)
quests[283331133] = 250 -- It's Time to Oil Up (Bilgewater Cartel)
quests[283341133] = 250 -- A Flare Fight (Bilgewater Cartel)
quests[283351133] = 250 -- Turn It Off! Turn It Off! (Bilgewater Cartel)
quests[283361133] = 250 -- Slap and Cap (Bilgewater Cartel)
quests[283371134] = 250 -- The Shredders of Irontree (Gilneas)
quests[283380576] = 1400 -- Deadwood of the North (Timbermaw Hold)
quests[283391133] = 250 -- Is Your Oil Running? (Bilgewater Cartel)
quests[283401133] = 250 -- A Bomb Deal (Bilgewater Cartel)
quests[283430609] = 350 -- The Breath of Cenarius (Cenarion Circle)
quests[283440068] = 10 -- Can You Smell What the Lok'tar is Cooking? (Undercity)
quests[283450068] = 75 -- *Gurgle* HELP! *Gurgle* (Undercity)
quests[283450076] = 75 -- *Gurgle* HELP! *Gurgle* (Orgrimmar)
quests[283461174] = 25 -- The Bachelor (Wildhammer Clan)
quests[283480068] = 250 -- Stormpike Rendezvous (Undercity)
quests[283480076] = 250 -- Stormpike Rendezvous (Orgrimmar)
quests[283491133] = 10 -- Megs in Marketing (Bilgewater Cartel)
quests[283540068] = 250 -- Kasha Will Fly Again (Undercity)
quests[283540076] = 250 -- Kasha Will Fly Again (Orgrimmar)
quests[283550068] = 350 -- Terrible Little Creatures (Undercity)
quests[283560068] = 350 -- Preemptive Strike (Undercity)
quests[283571133] = 250 -- Take it to the Tree (Bilgewater Cartel)
quests[283620576] = 1000 -- Stupid Drizle! (Timbermaw Hold)
quests[283640576] = 1400 -- The Chieftain's Key (Timbermaw Hold)
quests[283660576] = 1000 -- Disarming Bears (Timbermaw Hold)
quests[283681133] = 350 -- Fire in the Hole! (Bilgewater Cartel)
quests[283691174] = 10 -- My Sister, Fanny (Wildhammer Clan)
quests[283701133] = 250 -- Wisp-napping (Bilgewater Cartel)
quests[283730576] = 25 -- Timbermaw Hold (Timbermaw Hold)
quests[283750068] = 250 -- The Road to Purgation (Undercity)
quests[283750076] = 250 -- The Road to Purgation (Orgrimmar)
quests[283771174] = 25 -- Rescue at Glopgut's Hollow (Wildhammer Clan)
quests[283781174] = 250 -- Find Fanny (Wildhammer Clan)
quests[283791174] = 250 -- Ogre Bashin' (Wildhammer Clan)
quests[283821134] = 250 -- Enemy at our Roots (Gilneas)
quests[283830069] = 250 -- Wisps of the Woods (Darnassus)
quests[283841134] = 75 -- Into the Clearing (Gilneas)
quests[283851134] = 250 -- Oil and Irony (Gilneas)
quests[283861134] = 250 -- Borrowed Bombs (Gilneas)
quests[283871134] = 250 -- Burnout (Gilneas)
quests[283881134] = 250 -- Bloody Wages (Gilneas)
quests[283891134] = 350 -- Report to the Denmother (Gilneas)
quests[283901171] = 350 -- Glop, Son of Glop (Therazane)
quests[283911171] = 350 -- The Restless Brood (Therazane)
quests[283920576] = 25 -- The Timbermaw Tribe (Timbermaw Hold)
quests[283930072] = 350 -- A Dangerous Alliance (Stormwind)
quests[283940072] = 350 -- The Golem Lord's Creations (Stormwind)
quests[283950576] = 2000 -- Feathers for Nafien (Timbermaw Hold)
quests[283960576] = 2000 -- Feathers for Grazle (Timbermaw Hold)
quests[283970068] = 250 -- They Will Never Expect This... (Undercity)
quests[283970076] = 250 -- They Will Never Expect This... (Orgrimmar)
quests[283980072] = 350 -- The Pyromancer's Grimoire (Stormwind)
quests[283990072] = 350 -- Stones of Binding (Stormwind)
quests[284000068] = 500 -- Heroes of the Horde! (Undercity)
quests[284000076] = 500 -- Heroes of the Horde! (Orgrimmar)
quests[284010072] = 350 -- Slaves of the Firelord (Stormwind)
quests[284050072] = 350 -- Weapons of Darkness (Stormwind)
quests[284060072] = 350 -- The Dark Iron Army (Stormwind)
quests[284071174] = 25 -- The Bachelorette (Wildhammer Clan)
quests[284081174] = 250 -- Something Bold (Wildhammer Clan)
quests[284091174] = 250 -- Something Brewed (Wildhammer Clan)
quests[284101174] = 250 -- Something Stolen (Wildhammer Clan)
quests[284111174] = 250 -- Something Stewed (Wildhammer Clan)
quests[284131174] = 75 -- Words and Music By... (Wildhammer Clan)
quests[284141133] = 250 -- Fourth and Goal (Bilgewater Cartel)
quests[284270059] = 10 -- A Needle in a Hellhole (Thorium Brotherhood)
quests[284280059] = 350 -- A Perfect Costume (Thorium Brotherhood)
quests[284290059] = 350 -- A Perfect Costume (Thorium Brotherhood)
quests[284300059] = 350 -- A Perfect Costume (Thorium Brotherhood)
quests[284310059] = 350 -- A Perfect Costume (Thorium Brotherhood)
quests[284570076] = 350 -- A Dangerous Alliance (Orgrimmar)
quests[284580076] = 350 -- The Pyromancer's Grimoire (Orgrimmar)
quests[284590076] = 350 -- Stones of Binding (Orgrimmar)
quests[284600576] = 1400 -- Threat of the Winterfall (Timbermaw Hold)
quests[284610076] = 350 -- The Golem Lord's Creations (Orgrimmar)
quests[284620576] = 350 -- Winterfall Firewater (Timbermaw Hold)
quests[284630076] = 350 -- The Dark Iron Army (Orgrimmar)
quests[284640576] = 1000 -- Falling to Corruption (Timbermaw Hold)
quests[284650076] = 350 -- Slaves of the Firelord (Orgrimmar)
quests[284660076] = 350 -- Weapons of Darkness (Orgrimmar)
quests[284670576] = 1000 -- Mystery Goo (Timbermaw Hold)
quests[284690576] = 1000 -- Winterfall Runners (Timbermaw Hold)
quests[284700576] = 2000 -- High Chief Winterfall (Timbermaw Hold)
quests[284710576] = 1000 -- The Final Piece (Timbermaw Hold)
quests[284720577] = 250 -- Words of the High Chief (Everlook)
quests[284730911] = 350 -- Weapons of Darkness (Silvermoon City)
quests[284740068] = 350 -- Twilight Scheming (Undercity)
quests[284750911] = 350 -- Twilight Scheming (Silvermoon City)
quests[284760081] = 350 -- Twilight Scheming (Thunder Bluff)
quests[284770530] = 350 -- Twilight Scheming (Darkspear Trolls)
quests[284781133] = 350 -- Twilight Scheming (Bilgewater Cartel)
quests[284801173] = 250 -- Lieutenants of Darkness (Ramkahen)
quests[284831173] = 250 -- Bleeding the Enemy (Ramkahen)
quests[284840068] = 250 -- The Heart of the Matter (Undercity)
quests[284850068] = 500 -- Yetimus the Yeti Lord (Undercity)
quests[284861173] = 250 -- Salhet's Gambit (Ramkahen)
quests[284870068] = 250 -- Humbert's Personal Problems (Undercity)
quests[284881171] = 250 -- Beneath the Surface (Therazane)
quests[284890068] = 250 -- Stagwiches (Undercity)
quests[284900069] = 10 -- Hero's Call: Darkshore! (Darnassus)
quests[284920069] = 10 -- Hero's Call: Ashenvale! (Darnassus)
quests[284930076] = 10 -- Warchief's Command: Ashenvale! (Orgrimmar)
quests[284940076] = 10 -- Warchief's Command: Northern Barrens! (Orgrimmar)
quests[284950068] = 250 -- A Fighting Chance (Undercity)
quests[284950911] = 250 -- A Fighting Chance (Silvermoon City)
quests[284961133] = 10 -- Warchief's Command: Azshara! (Bilgewater Cartel)
quests[284981173] = 250 -- The Secret of Nahom (Ramkahen)
quests[284991173] = 250 -- Punish the Trespassers (Ramkahen)
quests[285001173] = 250 -- The Cypher of Keset (Ramkahen)
quests[285011173] = 350 -- The Defense of Nahom (Ramkahen)
quests[285021173] = 250 -- The Bandit Warlord (Ramkahen)
quests[285030069] = 10 -- Hero's Call: Thousand Needles! (Darnassus)
quests[285040076] = 10 -- Warchief's Command: Thousand Needles! (Orgrimmar)
quests[285050072] = 25 -- The Battle for Andorhal (Stormwind)
quests[285060068] = 250 -- March of the Stormpike (Undercity)
quests[285060076] = 250 -- March of the Stormpike (Orgrimmar)
quests[285070054] = 10 -- Hero's Call: Tanaris! (Gnomeregan)
quests[285080068] = 25 -- The Battle for Andorhal (Undercity)
quests[285091133] = 10 -- Warchief's Command: Tanaris! (Bilgewater Cartel)
quests[285100081] = 10 -- Warchief's Command: Feralas! (Thunder Bluff)
quests[285110069] = 10 -- Hero's Call: Feralas! (Darnassus)
quests[285120059] = 10 -- To the Aid of the Thorium Brotherhood (Thorium Brotherhood)
quests[285140059] = 10 -- Mouton Flamestar (Thorium Brotherhood)
quests[285150059] = 10 -- Mouton Flamestar (Thorium Brotherhood)
quests[285201173] = 500 -- The Fall of Neferset City (Ramkahen)
quests[285210576] = 100 -- Speak to Salfa (Timbermaw Hold)
quests[285220576] = 1400 -- Winterfall Activity (Timbermaw Hold)
quests[285230576] = 2000 -- More Beads for Salfa (Timbermaw Hold)
quests[285240576] = 75 -- Delivery for Donova (Timbermaw Hold)
quests[285270609] = 10 -- Warchief's Command: Silithus! (Cenarion Circle)
quests[285280609] = 10 -- Hero's Call: Silithus! (Cenarion Circle)
quests[285290069] = 500 -- Writings of the Void (Darnassus)
quests[285300576] = 1400 -- Scalding Signs (Timbermaw Hold)
quests[285310072] = 10 -- Hero's Call: Desolace! (Stormwind)
quests[285320076] = 10 -- Warchief's Command: Stonetalon Mountains! (Orgrimmar)
quests[285331173] = 250 -- The High Council's Decision (Ramkahen)
quests[285380068] = 250 -- Cry of the Banshee (Undercity)
quests[285390069] = 10 -- Hero's Call: Stonetalon Mountains! (Darnassus)
quests[285420609] = 10 -- Warchief's Command: Felwood! (Cenarion Circle)
quests[285430609] = 10 -- Hero's Call: Felwood! (Cenarion Circle)
quests[285480911] = 10 -- Warchief's Command: Desolace! (Silvermoon City)
quests[285490081] = 10 -- Warchief's Command: Southern Barrens! (Thunder Bluff)
quests[285500072] = 10 -- Hero's Call: Southern Barrens! (Stormwind)
quests[285510072] = 10 -- Hero's Call: Southern Barrens! (Stormwind)
quests[285520072] = 10 -- Hero's Call: Dustwallow Marsh! (Stormwind)
quests[285540076] = 10 -- Warchief's Command: Dustwallow Marsh! (Orgrimmar)
quests[285551173] = 250 -- Expel the Envoys (Ramkahen)
quests[285560068] = 350 -- Breaking the Hand (Undercity)
quests[285560076] = 350 -- Breaking the Hand (Orgrimmar)
quests[285590930] = 10 -- Hero's Call: Bloodmyst Isle! (Exodar)
quests[285600911] = 10 -- Warchief's Command: Ghostlands! (Silvermoon City)
quests[285611173] = 250 -- Nahom Must Hold (Ramkahen)
quests[285620072] = 75 -- Hero's Call: Westfall! (Stormwind)
quests[285630072] = 250 -- Hero's Call: Redridge Mountains! (Stormwind)
quests[285640072] = 10 -- Hero's Call: Duskwood! (Stormwind)
quests[285650047] = 10 -- Hero's Call: Wetlands! (Ironforge)
quests[285660068] = 250 -- Decimation (Undercity)
quests[285670047] = 10 -- Hero's Call: Loch Modan! (Ironforge)
quests[285680068] = 250 -- Warchief's Command: Silverpine Forest! (Undercity)
quests[285710068] = 10 -- Warchief's Command: Hillsbrad Foothills! (Undercity)
quests[285720068] = 10 -- Warchief's Command: Arathi Highlands! (Undercity)
quests[285730072] = 10 -- Hero's Call: Arathi Highlands! (Stormwind)
quests[285750068] = 10 -- Warchief's Command: Western Plaguelands! (Undercity)
quests[285760072] = 10 -- Hero's Call: Western Plaguelands! (Stormwind)
quests[285810059] = 10 -- Warchief's Command: Searing Gorge! (Thorium Brotherhood)
quests[285820059] = 10 -- Hero's Call: Searing Gorge! (Thorium Brotherhood)
quests[285831133] = 75 -- Krazzworks (Bilgewater Cartel)
quests[285841133] = 250 -- Quality Construction (Bilgewater Cartel)
quests[285861133] = 250 -- Pool Pony Rescue (Bilgewater Cartel)
quests[285870068] = 10 -- Aid of the Frostwolf (Undercity)
quests[285870076] = 10 -- Aid of the Frostwolf (Orgrimmar)
quests[285881133] = 250 -- Wildhammer Infestation (Bilgewater Cartel)
quests[285891133] = 350 -- Everything But the Kitchen Sink (Bilgewater Cartel)
quests[285901133] = 75 -- Reprisal (Bilgewater Cartel)
quests[285911133] = 250 -- Off The Wall (Bilgewater Cartel)
quests[285921133] = 250 -- Parting Packages (Bilgewater Cartel)
quests[285931133] = 250 -- Of Utmost Importance (Bilgewater Cartel)
quests[285941133] = 250 -- Highbank, Crybank (Bilgewater Cartel)
quests[285951133] = 350 -- Krazz Works! (Bilgewater Cartel)
quests[285960072] = 250 -- Welcome to Highbank! (Stormwind)
quests[285980072] = 350 -- Aiming High (Stormwind)
quests[285991174] = 25 -- Kurdran's Waitin' (Wildhammer Clan)
quests[286000068] = 250 -- Matters of Loyalty (Undercity)
quests[286000076] = 250 -- Matters of Loyalty (Orgrimmar)
quests[286030068] = 250 -- Infiltration (Undercity)
quests[286040068] = 250 -- Deception and Trickery (Undercity)
quests[286050068] = 250 -- Domination (Undercity)
quests[286080068] = 250 -- The Shadow Grave (Undercity)
quests[286100577] = 250 -- Rubble Trouble (Everlook)
quests[286111173] = 250 -- The Defilers' Ritual (Ramkahen)
quests[286140576] = 1400 -- Bearzerker (Timbermaw Hold)
quests[286150576] = 1000 -- Turning the Earth (Timbermaw Hold)
quests[286160068] = 500 -- Stormpike Apocalypse (Undercity)
quests[286160076] = 500 -- Stormpike Apocalypse (Orgrimmar)
quests[286190068] = 10 -- Warchief's Command: Arathi Highlands! (Undercity)
quests[286200068] = 25 -- Eastpoint Tower (Undercity)
quests[286231173] = 75 -- The Push Westward (Ramkahen)
quests[286290577] = 250 -- Are We There, Yeti? (Everlook)
quests[286310577] = 150 -- The Perfect Horns (Everlook)
quests[286340068] = 250 -- Extinction (Undercity)
quests[286350068] = 250 -- A Haunting in Hillsbrad (Undercity)
quests[286360068] = 250 -- Silence of the Dwarves (Undercity)
quests[286430068] = 350 -- The Durnholde Challenge: Zephyrus (Undercity)
quests[286440068] = 350 -- The Durnholde Challenge: Teracula (Undercity)
quests[286450068] = 350 -- The Durnholde Challenge: Bloodvenom (Undercity)
quests[286460068] = 350 -- The Durnholde Challenge: Infernus (Undercity)
quests[286470068] = 500 -- The Durnholde Challenge: D-1000 (Undercity)
quests[286551174] = 350 -- Wild, Wild, Wildhammer Wedding (Wildhammer Clan)
quests[286571178] = 350 -- A Huge Problem (Hellscream's Reach)
quests[286581178] = 350 -- Swamp Bait (Hellscream's Reach)
quests[286591178] = 350 -- The Leftovers (Hellscream's Reach)
quests[286601178] = 350 -- Clearing the Depths (Hellscream's Reach)
quests[286611178] = 350 -- The Imprisoned Archmage (Hellscream's Reach)
quests[286621178] = 350 -- Learning From The Past (Hellscream's Reach)
quests[286631178] = 350 -- D-Block (Hellscream's Reach)
quests[286641178] = 350 -- Svarnos (Hellscream's Reach)
quests[286651178] = 350 -- Cursed Shackles (Hellscream's Reach)
quests[286660072] = 10 -- Hero's Call: Burning Steppes! (Stormwind)
quests[286670076] = 10 -- Warchief's Command: Burning Steppes! (Orgrimmar)
quests[286681178] = 350 -- Prison Revolt (Hellscream's Reach)
quests[286691178] = 350 -- The Warden (Hellscream's Reach)
quests[286701178] = 350 -- Food From Below (Hellscream's Reach)
quests[286710076] = 10 -- Warchief's Command: Blasted Lands! (Orgrimmar)
quests[286720068] = 75 -- The Executor In the Field (Undercity)
quests[286730072] = 10 -- Hero's Call: Blasted Lands! (Stormwind)
quests[286781178] = 250 -- Captain P. Harris (Hellscream's Reach)
quests[286791178] = 250 -- Rattling Their Cages (Hellscream's Reach)
quests[286801178] = 250 -- Boosting Morale (Hellscream's Reach)
quests[286811178] = 350 -- Shark Tank (Hellscream's Reach)
quests[286821178] = 250 -- Claiming The Keep (Hellscream's Reach)
quests[286831178] = 250 -- Thinning the Brood (Hellscream's Reach)
quests[286841178] = 250 -- A Sticky Task (Hellscream's Reach)
quests[286851178] = 250 -- Leave No Weapon Behind (Hellscream's Reach)
quests[286861178] = 250 -- Not The Friendliest Town (Hellscream's Reach)
quests[286871178] = 250 -- Teach A Man To Fish.... Or Steal (Hellscream's Reach)
quests[286880076] = 10 -- Warchief's Command: Northern Stranglethorn Vale! (Orgrimmar)
quests[286891178] = 250 -- The Forgotten (Hellscream's Reach)
quests[286901178] = 250 -- Salvaging the Remains (Hellscream's Reach)
quests[286911178] = 250 -- First Lieutenant Connor (Hellscream's Reach)
quests[286921178] = 250 -- Magnets, How Do They Work? (Hellscream's Reach)
quests[286931178] = 250 -- Finish The Job (Hellscream's Reach)
quests[286941178] = 250 -- Watch Out For Splinters! (Hellscream's Reach)
quests[286951178] = 250 -- WANTED: Foreman Wellson (Hellscream's Reach)
quests[286961178] = 250 -- Bombs Away! (Hellscream's Reach)
quests[286971178] = 250 -- Ghostbuster (Hellscream's Reach)
quests[286981178] = 250 -- Cannonball! (Hellscream's Reach)
quests[286990072] = 10 -- Hero's Call: Northern Stranglethorn Vale! (Stormwind)
quests[287001178] = 250 -- Taking the Overlook Back (Hellscream's Reach)
quests[287020047] = 10 -- Hero's Call: The Cape of Stranglethorn! (Ironforge)
quests[287041133] = 10 -- Warchief's Command: The Cape of Stranglethorn! (Bilgewater Cartel)
quests[287050076] = 10 -- Warchief's Command: Outland! (Orgrimmar)
quests[287080072] = 10 -- Hero's Call: Outland! (Stormwind)
quests[287091050] = 10 -- Hero's Call: Northrend! (Valiance Expedition)
quests[287111085] = 10 -- Warchief's Command: Northrend! (Warsong Offensive)
quests[287130069] = 250 -- The Balance of Nature (Darnassus)
quests[287140069] = 250 -- Fel Moss Corruption (Darnassus)
quests[287150069] = 250 -- Demonic Thieves (Darnassus)
quests[287160072] = 10 -- Hero's Call: Twilight Highlands! (Stormwind)
quests[287170076] = 10 -- Warchief's Command: Twilight Highlands! (Orgrimmar)
quests[287211178] = 250 -- Walk A Mile In Their Shoes (Hellscream's Reach)
quests[287220577] = 350 -- Yetiphobia (Everlook)
quests[287230069] = 75 -- Priestess of the Moon (Darnassus)
quests[287240069] = 250 -- Iverron's Antidote (Darnassus)
quests[287250069] = 150 -- The Woodland Protector (Darnassus)
quests[287260069] = 250 -- Webwood Corruption (Darnassus)
quests[287270069] = 250 -- Vile Touch (Darnassus)
quests[287280069] = 75 -- Signs of Things to Come (Darnassus)
quests[287290069] = 150 -- Teldrassil: Crown of Azeroth (Darnassus)
quests[287300069] = 250 -- Precious Waters (Darnassus)
quests[287310069] = 75 -- Teldrassil: Passing Awareness (Darnassus)
quests[287340069] = 10 -- A Favor for Melithar (Darnassus)
quests[287361173] = 250 -- Fire From the Sky (Ramkahen)
quests[287490072] = 25 -- The Battle for Andorhal (Stormwind)
quests[287500068] = 25 -- The Battle for Andorhal (Undercity)
quests[287550529] = 2000 -- Annals of the Silver Hand (Argent Dawn)
quests[287560529] = 1000 -- Aberrations of Bone (Argent Dawn)
quests[287570072] = 250 -- Beating Them Back! (Stormwind)
quests[287590072] = 250 -- Lions for Lambs (Stormwind)
quests[287601173] = 500 -- Vengeance for Orsis (Ramkahen)
quests[287620072] = 250 -- Beating Them Back! (Stormwind)
quests[287630072] = 250 -- Beating Them Back! (Stormwind)
quests[287640072] = 250 -- Beating Them Back! (Stormwind)
quests[287650072] = 250 -- Beating Them Back! (Stormwind)
quests[287660072] = 250 -- Beating Them Back! (Stormwind)
quests[287670072] = 250 -- Beating Them Back! (Stormwind)
quests[287690072] = 250 -- Lions for Lambs (Stormwind)
quests[287700072] = 250 -- Lions for Lambs (Stormwind)
quests[287710072] = 250 -- Lions for Lambs (Stormwind)
quests[287720072] = 250 -- Lions for Lambs (Stormwind)
quests[287730072] = 250 -- Lions for Lambs (Stormwind)
quests[287740072] = 250 -- Lions for Lambs (Stormwind)
quests[287791173] = 500 -- A Long Way from Home (Ramkahen)
quests[287800072] = 25 -- Join the Battle! (Stormwind)
quests[287811173] = 500 -- Targets of Opportunity (Ramkahen)
quests[287831173] = 500 -- The Source of Their Power (Ramkahen)
quests[287840072] = 25 -- Join the Battle! (Stormwind)
quests[287850072] = 25 -- Join the Battle! (Stormwind)
quests[287860072] = 25 -- Join the Battle! (Stormwind)
quests[287870072] = 25 -- Join the Battle! (Stormwind)
quests[287880072] = 25 -- Join the Battle! (Stormwind)
quests[287890072] = 25 -- Join the Battle! (Stormwind)
quests[287910072] = 250 -- They Sent Assassins (Stormwind)
quests[287920072] = 250 -- They Sent Assassins (Stormwind)
quests[287930072] = 250 -- They Sent Assassins (Stormwind)
quests[287940072] = 250 -- They Sent Assassins (Stormwind)
quests[287950072] = 250 -- They Sent Assassins (Stormwind)
quests[287960072] = 250 -- They Sent Assassins (Stormwind)
quests[287970072] = 250 -- They Sent Assassins (Stormwind)
quests[288060072] = 250 -- Fear No Evil (Stormwind)
quests[288070072] = 75 -- Expert Opinion (Stormwind)
quests[288080072] = 250 -- Fear No Evil (Stormwind)
quests[288090072] = 250 -- Fear No Evil (Stormwind)
quests[288100072] = 250 -- Fear No Evil (Stormwind)
quests[288110072] = 250 -- Fear No Evil (Stormwind)
quests[288120072] = 250 -- Fear No Evil (Stormwind)
quests[288130072] = 250 -- Fear No Evil (Stormwind)
quests[288141135] = 500 -- Followers and Leaders (The Earthen Ring)
quests[288151135] = 500 -- Twilight Documents (The Earthen Ring)
quests[288160076] = 10 -- To the Depths (Orgrimmar)
quests[288170072] = 10 -- The Rear is Clear (Stormwind)
quests[288180072] = 10 -- The Rear is Clear (Stormwind)
quests[288190072] = 10 -- The Rear is Clear (Stormwind)
quests[288200072] = 10 -- The Rear is Clear (Stormwind)
quests[288210072] = 10 -- The Rear is Clear (Stormwind)
quests[288220072] = 10 -- The Rear is Clear (Stormwind)
quests[288230072] = 10 -- The Rear is Clear (Stormwind)
quests[288241171] = 500 -- Wayward Child (Therazane)
quests[288270072] = 10 -- To the Depths (Stormwind)
quests[288320072] = 75 -- Twilight Shores (Stormwind)
quests[288451173] = 10 -- The Vortex Pinnacle (Ramkahen)
quests[288490076] = 75 -- Twilight Skies (Orgrimmar)
quests[288501134] = 10 -- The Prison Rooftop (Gilneas)
quests[288531135] = 350 -- Kill the Courier (The Earthen Ring)
quests[288560609] = 75 -- The Sands of Silithus (Cenarion Circle)
quests[288570072] = 75 -- Blasted Lands: The Other Side of the World (Stormwind)
quests[288580076] = 10 -- Blasted Lands: The Other Side of the World (Orgrimmar)
quests[288590609] = 75 -- The Dunes of Silithus (Cenarion Circle)
quests[288601174] = 250 -- Keeping the Dragonmaw at Bay (Wildhammer Clan)
quests[288611174] = 250 -- Fight Like a Wildhammer (Wildhammer Clan)
quests[288621174] = 250 -- Never Leave a Dinner Behind (Wildhammer Clan)
quests[288631174] = 350 -- Warlord Halthar is Back (Wildhammer Clan)
quests[288641174] = 250 -- Beer Run (Wildhammer Clan)
quests[288650076] = 25 -- Call of the Warmatron (Orgrimmar)
quests[288661135] = 10 -- Into the Stonecore (The Earthen Ring)
quests[288670072] = 25 -- Nethergarde Needs You! (Stormwind)
quests[288680054] = 250 -- The View from Down Here (Gnomeregan)
quests[288701173] = 10 -- Return to the Lost City (Ramkahen)
quests[288711172] = 250 -- Crushing the Wildhammer (Dragonmaw Clan)
quests[288721172] = 250 -- Total War (Dragonmaw Clan)
quests[288731172] = 250 -- Another Maw to Feed (Dragonmaw Clan)
quests[288741172] = 250 -- Hook 'em High (Dragonmaw Clan)
quests[288751172] = 350 -- Bring Down the High Shaman (Dragonmaw Clan)
quests[288760076] = 10 -- To the Mor'shan Rampart (Orgrimmar)
quests[288770076] = 75 -- They Call Him Swiftdagger. He Kills Harpies. (Orgrimmar)
quests[288851135] = 250 -- Mr. Goldmine's Wild Ride (The Earthen Ring)
quests[289090076] = 10 -- Sauranok Will Point the Way (Orgrimmar)
quests[290211133] = 150 -- Samophlange (Bilgewater Cartel)
quests[290221133] = 75 -- Samophlange (Bilgewater Cartel)
quests[290231133] = 150 -- Samophlange (Bilgewater Cartel)
quests[290241133] = 250 -- Samophlange (Bilgewater Cartel)
quests[290261133] = 75 -- Wenikee Boltbucket (Bilgewater Cartel)
quests[290271133] = 250 -- Nugget Slugs (Bilgewater Cartel)
quests[290320589] = 3000 -- Get Them While They're Young (Wintersaber Trainers)
quests[290340589] = 42999 -- They Grow Up So Fast (Wintersaber Trainers)
quests[290350589] = 1500 -- A Cub's Cravings (Wintersaber Trainers)
quests[290370589] = 1500 -- 'Borrowing' From the Winterfall (Wintersaber Trainers)
quests[290380589] = 1500 -- Hunting Practice (Wintersaber Trainers)
quests[290400589] = 1500 -- Cub's First Toy (Wintersaber Trainers)
quests[290661158] = 10 -- Good News... and Bad News (Guardians of Hyjal)
quests[290780072] = 250 -- Beating Them Back! (Stormwind)
quests[290790072] = 250 -- Lions for Lambs (Stormwind)
quests[290800072] = 25 -- Join the Battle! (Stormwind)
quests[290810072] = 250 -- They Sent Assassins (Stormwind)
quests[290820072] = 250 -- Fear No Evil (Stormwind)
quests[290830072] = 10 -- The Rear is Clear (Stormwind)
quests[291011158] = 150 -- Punting Season (Guardians of Hyjal)
quests[291221158] = 250 -- Echoes of Nemesis (Guardians of Hyjal)
quests[291231158] = 250 -- Rage Against the Flames (Guardians of Hyjal)
quests[291251158] = 150 -- Between the Trees (Guardians of Hyjal)
quests[291261158] = 250 -- The Power of Malorne (Guardians of Hyjal)
quests[291271158] = 250 -- Rage Against the Flames (Guardians of Hyjal)
quests[291281158] = 350 -- The Protectors of Hyjal (Guardians of Hyjal)
quests[291371158] = 250 -- Breach in the Defenses (Guardians of Hyjal)
quests[291381158] = 250 -- Burn Victims (Guardians of Hyjal)
quests[291391158] = 250 -- Aggressive Growth (Guardians of Hyjal)
quests[291411158] = 250 -- The Harder They Fall (Guardians of Hyjal)
quests[291421158] = 250 -- Traitors Return (Guardians of Hyjal)
quests[291431158] = 250 -- Wisp Away (Guardians of Hyjal)
quests[291451158] = 10 -- Opening the Door (Guardians of Hyjal)
quests[291460076] = 250 -- Ridin' the Rocketway (Orgrimmar)
quests[291471158] = 150 -- Call the Flock (Guardians of Hyjal)
quests[291481158] = 250 -- Wings Aflame (Guardians of Hyjal)
quests[291491158] = 250 -- Rage Against the Flames (Guardians of Hyjal)
quests[291591158] = 250 -- Pyrorachnophobia (Guardians of Hyjal)
quests[291601158] = 250 -- Egg-stinction (Guardians of Hyjal)
quests[291611158] = 150 -- Those Bears Up There (Guardians of Hyjal)
quests[291621158] = 250 -- Nature's Blessing (Guardians of Hyjal)
quests[291631158] = 250 -- Rage Against the Flames (Guardians of Hyjal)
quests[291641158] = 150 -- Perfecting Your Howl (Guardians of Hyjal)
quests[291651158] = 250 -- The Call of the Pack (Guardians of Hyjal)
quests[291661158] = 250 -- Supplies for the Other Side (Guardians of Hyjal)
quests[291670076] = 250 -- The Banshee Queen (Orgrimmar)
quests[291760076] = 250 -- The Fallen Chieftain (Orgrimmar)
quests[291771158] = 250 -- Vigilance on Wings (Guardians of Hyjal)
quests[291791158] = 250 -- Hostile Elements (Guardians of Hyjal)
quests[291811158] = 350 -- Druids of the Talon (Guardians of Hyjal)
quests[291821158] = 25 -- Flight of the Storm Crows (Guardians of Hyjal)
quests[291891158] = 250 -- Wicked Webs (Guardians of Hyjal)
quests[291910076] = 250 -- You Scream, I Scream... (Orgrimmar)
quests[291921158] = 250 -- The Wardens are Watching (Guardians of Hyjal)
quests[291951158] = 250 -- A Ritual of Flame (Guardians of Hyjal)
quests[291961158] = 10 -- To the Sanctuary! (Guardians of Hyjal)
quests[291971158] = 250 -- Caught Unawares (Guardians of Hyjal)
quests[291981158] = 250 -- The Sanctuary Must Not Fall (Guardians of Hyjal)
quests[291991158] = 250 -- Calling for Reinforcements (Guardians of Hyjal)
quests[292001158] = 250 -- Leyara (Guardians of Hyjal)
quests[292011158] = 350 -- Through the Gates of Hell (Guardians of Hyjal)
quests[292021158] = 10 -- The Fate of Runetotem (Guardians of Hyjal)
quests[292031158] = 350 -- Into the Depths (Guardians of Hyjal)
quests[292041158] = 25 -- The Warden's Charge (Guardians of Hyjal)
quests[292051158] = 250 -- The Forlorn Spire (Guardians of Hyjal)
quests[292061158] = 250 -- Into the Fire (Guardians of Hyjal)
quests[292091158] = 25 -- Into the Fiery Depths (Guardians of Hyjal)
quests[292101158] = 250 -- Enduring the Heat (Guardians of Hyjal)
quests[292111158] = 250 -- Solar Core Destruction (Guardians of Hyjal)
quests[292141158] = 350 -- The Shadow Wardens (Guardians of Hyjal)
quests[292151158] = 25 -- The Hunt Begins (Guardians of Hyjal)
quests[292431158] = 350 -- Strike at the Heart (Guardians of Hyjal)
quests[292441158] = 25 -- A Lieutenant of Flame (Guardians of Hyjal)
quests[292451158] = 75 -- The Mysterious Seed (Guardians of Hyjal)
quests[292461158] = 250 -- Relieving the Pain (Guardians of Hyjal)
quests[292471158] = 250 -- Treating the Wounds (Guardians of Hyjal)
quests[292481158] = 250 -- Releasing the Pressure (Guardians of Hyjal)
quests[292491158] = 75 -- Planting Season (Guardians of Hyjal)
quests[292541158] = 250 -- Little Lasher (Guardians of Hyjal)
quests[292551158] = 250 -- Embergris (Guardians of Hyjal)
quests[292571158] = 250 -- Steal Magmolias (Guardians of Hyjal)
quests[292631158] = 250 -- A Bitter Pill (Guardians of Hyjal)
quests[292641158] = 250 -- Flamewakers of the Molten Flow (Guardians of Hyjal)
quests[292651158] = 250 -- Fire Flowers (Guardians of Hyjal)
quests[292721158] = 250 -- Need... Water... Badly... (Guardians of Hyjal)
quests[292731158] = 250 -- How Hot (Guardians of Hyjal)
quests[292741158] = 250 -- Hounds of Shannox (Guardians of Hyjal)
quests[292751158] = 250 -- Fandral's Methods (Guardians of Hyjal)
quests[292761158] = 250 -- The Flame Spider Queen (Guardians of Hyjal)
quests[292781158] = 250 -- Living Obsidium (Guardians of Hyjal)
quests[292791158] = 350 -- Filling the Moonwell (Guardians of Hyjal)
quests[292801158] = 25 -- Nourishing Waters (Guardians of Hyjal)
quests[292811158] = 350 -- Additional Armaments (Guardians of Hyjal)
quests[292821158] = 25 -- Well Armed (Guardians of Hyjal)
quests[292831158] = 350 -- Calling the Ancients (Guardians of Hyjal)
quests[292841158] = 25 -- Aid of the Ancients (Guardians of Hyjal)
quests[292871158] = 250 -- Peaked Interest (Guardians of Hyjal)
quests[292881158] = 250 -- Starting Young (Guardians of Hyjal)
quests[292891158] = 250 -- Mother's Malice (Guardians of Hyjal)
quests[292901158] = 250 -- Fire in the Skies (Guardians of Hyjal)
quests[292931158] = 250 -- Singed Wings (Guardians of Hyjal)
quests[292951158] = 250 -- The Bigger They Are (Guardians of Hyjal)
quests[292961158] = 250 -- Territorial Birds (Guardians of Hyjal)
quests[292971158] = 250 -- Bye Bye Burdy (Guardians of Hyjal)
quests[292981158] = 75 -- A Smoke-Stained Locket (Guardians of Hyjal)
quests[292991158] = 250 -- Some Like It Hot (Guardians of Hyjal)
quests[293021158] = 75 -- Unlocking the Secrets Within (Guardians of Hyjal)
quests[293041158] = 250 -- The Dogs of War (Guardians of Hyjal)
quests[293051158] = 350 -- Strike at the Heart (Guardians of Hyjal)
quests[293101158] = 75 -- The Tipping Point (Guardians of Hyjal)
quests[293111158] = 500 -- The Rest is History (Guardians of Hyjal)
quests[293130069] = 250 -- The Secret to Perfect Kimchi (Darnassus)
quests[293140069] = 250 -- Remembering the Ancestors (Darnassus)
quests[293150068] = 250 -- Fungus Among Us (Undercity)
quests[293160069] = 250 -- Back to Basics (Darnassus)
quests[293170068] = 250 -- Fish Head (Undercity)
quests[293180069] = 250 -- Ribs for the Sentinels (Darnassus)
quests[293190068] = 250 -- Tadpole Terror (Undercity)
quests[293200068] = 250 -- Like Pike? (Undercity)
quests[293210069] = 250 -- Happy as a Clam Digger (Darnassus)
quests[293220068] = 250 -- Time for Slime (Undercity)
quests[293230069] = 250 -- Stocking Up (Darnassus)
quests[293240069] = 250 -- The Sister's Pendant (Darnassus)
quests[293250069] = 250 -- A Slippery Snack (Darnassus)
quests[293261135] = 75 -- The Nordrassil Summit (The Earthen Ring)
quests[293271135] = 350 -- Elemental Bonds: Doubt (The Earthen Ring)
quests[293281135] = 350 -- Elemental Bonds: Desire (The Earthen Ring)
quests[293291135] = 350 -- Elemental Bonds: Patience (The Earthen Ring)
quests[293301135] = 500 -- Elemental Bonds: Fury (The Earthen Ring)
quests[293311135] = 1000 -- Elemental Bonds: The Vow (The Earthen Ring)
quests[293320068] = 250 -- Lily, Oh Lily (Undercity)
quests[293330068] = 250 -- Escargot A Go-Go (Undercity)
quests[293340068] = 250 -- Roach Coach (Undercity)
quests[293351135] = 75 -- Into Slashing Winds (The Earthen Ring)
quests[293361135] = 75 -- Into Coaxing Tides (The Earthen Ring)
quests[293371135] = 75 -- Into Constant Earth (The Earthen Ring)
quests[293381135] = 75 -- Into Unrelenting Flame (The Earthen Ring)
quests[293420047] = 250 -- Cold Water Fishing (Ironforge)
quests[293430047] = 250 -- One fer the Ages (Ironforge)
quests[293440047] = 250 -- Fish fer Squirky (Ironforge)
quests[293450081] = 250 -- Pond Predators (Thunder Bluff)
quests[293460081] = 250 -- The Ring's the Thing (Thunder Bluff)
quests[293470047] = 250 -- Live Bait (Ironforge)
quests[293480081] = 250 -- The Race to Restock (Thunder Bluff)
quests[293490081] = 250 -- Craving Crayfish (Thunder Bluff)
quests[293500047] = 250 -- The Gnomish Bait-o-Matic (Ironforge)
quests[293510047] = 250 -- A Round for the Guards (Ironforge)
quests[293520047] = 250 -- A Fowl Shortage (Ironforge)
quests[293530047] = 250 -- Keepin' the Haggis Flowin' (Ironforge)
quests[293540081] = 250 -- Shiny Baubles (Thunder Bluff)
quests[293550047] = 250 -- Can't Get Enough Spice Bread (Ironforge)
quests[293560047] = 250 -- I Need to Cask a Favor (Ironforge)
quests[293570069] = 250 -- Spice Bread Aplenty (Darnassus)
quests[293580081] = 250 -- Pining for Nuts (Thunder Bluff)
quests[293590069] = 250 -- An Old Favorite (Darnassus)
quests[293600068] = 250 -- Would You Like Some Flies With That? (Undercity)
quests[293610068] = 250 -- Moat Monster! (Undercity)
quests[293620081] = 250 -- "Magic" Mushrooms (Thunder Bluff)
quests[293630081] = 250 -- Mulgore Spice Bread (Thunder Bluff)
quests[293640081] = 250 -- Corn Mash (Thunder Bluff)
quests[293650081] = 250 -- Perfectly Picked Portions (Thunder Bluff)
quests[293920072] = 150 -- Missing Heirlooms (Stormwind)
quests[293980072] = 150 -- Fencing the Goods (Stormwind)
quests[293990072] = 150 -- Shopping Around (Stormwind)
quests[294020072] = 250 -- Taking Precautions (Stormwind)
quests[294061216] = 250 -- The Lesson of the Sandy Fist (Shang Xi's Academy)
quests[294081216] = 250 -- The Lesson of the Burning Scroll (Shang Xi's Academy)
quests[294091216] = 250 -- The Disciple's Challenge (Shang Xi's Academy)
quests[294101216] = 25 -- Aysa of the Tushui (Shang Xi's Academy)
quests[294110072] = 250 -- What Now? (Stormwind)
quests[294141216] = 250 -- The Way of the Tushui (Shang Xi's Academy)
quests[294150076] = 150 -- Missing Heirlooms (Orgrimmar)
quests[294160076] = 150 -- Fencing the Goods (Orgrimmar)
quests[294171216] = 250 -- The Way of the Huojin (Shang Xi's Academy)
quests[294181216] = 250 -- Kindling the Fire (Shang Xi's Academy)
quests[294191216] = 250 -- The Missing Driver (Shang Xi's Academy)
quests[294201216] = 25 -- The Spirit's Guardian (Shang Xi's Academy)
quests[294211216] = 250 -- Only the Worthy Shall Pass (Shang Xi's Academy)
quests[294221216] = 250 -- Huo, the Spirit of Fire (Shang Xi's Academy)
quests[294231216] = 75 -- The Passion of Shen-zin Su (Shang Xi's Academy)
quests[294241216] = 250 -- Items of Utmost Importance (Shang Xi's Academy)
quests[294250076] = 150 -- Shopping Around (Orgrimmar)
quests[294260076] = 250 -- Taking Precautions (Orgrimmar)
quests[294280076] = 250 -- What Now? (Orgrimmar)
quests[294300072] = 10 -- A Friend in Need (Stormwind)
quests[294310068] = 10 -- A Friend in Need (Undercity)
quests[294330909] = 250 -- Test Your Strength (Darkmoon Faire)
quests[294340909] = 250 -- Tonk Commander (Darkmoon Faire)
quests[294360909] = 250 -- The Humanoid Cannonball (Darkmoon Faire)
quests[294371158] = 250 -- The Fallen Guardian (Guardians of Hyjal)
quests[294380909] = 250 -- He Shoots, He Scores! (Darkmoon Faire)
quests[294391135] = 75 -- The Call of the World-Shaman (The Earthen Ring)
quests[294401135] = 75 -- The Call of the World-Shaman (The Earthen Ring)
quests[294430909] = 250 -- A Curious Crystal (Darkmoon Faire)
quests[294440909] = 250 -- An Exotic Egg (Darkmoon Faire)
quests[294450909] = 250 -- An Intriguing Grimoire (Darkmoon Faire)
quests[294460909] = 250 -- A Wondrous Weapon (Darkmoon Faire)
quests[294510909] = 250 -- The Master Strategist (Darkmoon Faire)
quests[294550909] = 250 -- Target: Turtle (Darkmoon Faire)
quests[294560909] = 250 -- A Captured Banner (Darkmoon Faire)
quests[294570909] = 250 -- The Enemy's Insignia (Darkmoon Faire)
quests[294580909] = 250 -- The Captured Journal (Darkmoon Faire)
quests[294630909] = 250 -- It's Hammer Time (Darkmoon Faire)
quests[294640909] = 250 -- Tools of Divination (Darkmoon Faire)
quests[295060909] = 250 -- A Fizzy Fusion (Darkmoon Faire)
quests[295070909] = 250 -- Fun for the Little Ones (Darkmoon Faire)
quests[295080909] = 250 -- Baby Needs Two Pair of Shoes (Darkmoon Faire)
quests[295090909] = 250 -- Putting the Crunch in the Frog (Darkmoon Faire)
quests[295100909] = 250 -- Putting Trash to Good Use (Darkmoon Faire)
quests[295110909] = 250 -- Talkin' Tonks (Darkmoon Faire)
quests[295120909] = 250 -- Putting the Carnies Back Together Again (Darkmoon Faire)
quests[295130909] = 250 -- Spoilin' for Salty Sea Dogs (Darkmoon Faire)
quests[295140909] = 250 -- Herbs for Healing (Darkmoon Faire)
quests[295150909] = 250 -- Writing the Future (Darkmoon Faire)
quests[295160909] = 250 -- Keeping the Faire Sparkling (Darkmoon Faire)
quests[295170909] = 250 -- Eyes on the Prizes (Darkmoon Faire)
quests[295180909] = 250 -- Rearm, Reuse, Recycle (Darkmoon Faire)
quests[295190909] = 250 -- Tan My Hide (Darkmoon Faire)
quests[295200909] = 250 -- Banners, Banners Everywhere! (Darkmoon Faire)
quests[295211216] = 25 -- The Singing Pools (Shang Xi's Academy)
quests[295221216] = 25 -- Ji of the Huojin (Shang Xi's Academy)
quests[295231216] = 250 -- Fanning the Flames (Shang Xi's Academy)
quests[295241216] = 250 -- The Lesson of Stifled Pride (Shang Xi's Academy)
quests[295270947] = 350 -- War on the Ramparts (Thrallmar)
quests[295280946] = 350 -- War on the Ramparts (Honor Hold)
quests[295290946] = 350 -- Demons in the Citadel (Honor Hold)
quests[295300947] = 350 -- Demons in the Citadel (Thrallmar)
quests[295350947] = 350 -- Make Them Bleed (Thrallmar)
quests[295360947] = 350 -- Heart of Rage (Thrallmar)
quests[295370947] = 350 -- Mind the Gap (Thrallmar)
quests[295380946] = 350 -- Make Them Bleed (Honor Hold)
quests[295390946] = 350 -- Heart of Rage (Honor Hold)
quests[295400947] = 350 -- The Breaker (Thrallmar)
quests[295581242] = 2000 -- The Path of War (Pearlfin Jinyu)
quests[295591242] = 2000 -- Freeing Our Brothers (Pearlfin Jinyu)
quests[295601242] = 2000 -- Ancient Power (Pearlfin Jinyu)
quests[295621242] = 1000 -- Jailbreak (Pearlfin Jinyu)
quests[295630942] = 500 -- Lost in Action (Cenarion Expedition)
quests[295670970] = 350 -- Stalk the Stalker (Sporeggar)
quests[295700942] = 500 -- Rescuing the Expedition (Cenarion Expedition)
quests[295730933] = 250 -- Safety is Job One (The Consortium)
quests[295740933] = 150 -- Intriguing Specimens (The Consortium)
quests[295750933] = 250 -- Undercutting the Competition (The Consortium)
quests[295960935] = 500 -- The End of the Exarch (The Sha'tar)
quests[295980989] = 1000 -- Taretha's Diversion (Keepers of Time)
quests[295990989] = 1000 -- Escape from Durnholde (Keepers of Time)
quests[296051011] = 350 -- Brother Against Brother (Lower City)
quests[296081094] = 10 -- Allies in Dalaran (The Silver Covenant)
quests[296091124] = 10 -- Friends in Dalaran (The Sunreavers)
quests[296130942] = 500 -- A Proper Fate (Cenarion Expedition)
quests[296401011] = 250 -- Find Spy To'gun (Lower City)
quests[296411011] = 350 -- The Soul Devices (Lower City)
quests[296421011] = 150 -- Trouble at Auchindoun (Lower City)
quests[296431011] = 350 -- The Codex of Blood (Lower City)
quests[296441011] = 500 -- Into the Heart of the Labyrinth (Lower City)
quests[296490946] = 250 -- How to Save a Life (Honor Hold)
quests[296500946] = 250 -- Fel Ember (Honor Hold)
quests[296510947] = 250 -- Honor the Dying (Thrallmar)
quests[296520946] = 250 -- One Last Favor (Honor Hold)
quests[296530947] = 500 -- The Will of the Warchief (Thrallmar)
quests[296540946] = 500 -- Turning the Tide (Honor Hold)
quests[296550947] = 350 -- A Very Special Cloth (Thrallmar)
quests[296560946] = 350 -- O'mrogg's Warcloth (Honor Hold)
quests[296611216] = 250 -- The Lesson of Dry Fur (Shang Xi's Academy)
quests[296621216] = 250 -- Stronger Than Reeds (Shang Xi's Academy)
quests[296631216] = 250 -- The Lesson of the Balanced Rock (Shang Xi's Academy)
quests[296641216] = 250 -- The Challenger's Fires (Shang Xi's Academy)
quests[296651216] = 250 -- From Bad to Worse (Shang Xi's Academy)
quests[296661216] = 250 -- The Sting of Learning (Shang Xi's Academy)
quests[296761216] = 25 -- Finding an Old Friend (Shang Xi's Academy)
quests[296771216] = 250 -- The Sun Pearl (Shang Xi's Academy)
quests[296781216] = 25 -- Shu, the Spirit of Water (Shang Xi's Academy)
quests[296791216] = 250 -- A New Friend (Shang Xi's Academy)
quests[296801216] = 25 -- The Source of Our Livelihood (Shang Xi's Academy)
quests[296841077] = 10 -- Severed Communications (Shattered Sun Offensive)
quests[296851077] = 250 -- Hard to Kill (Shattered Sun Offensive)
quests[296861077] = 250 -- A Radical Notion (Shattered Sun Offensive)
quests[296871077] = 250 -- Twisted Associations (Shattered Sun Offensive)
quests[296910970] = 250 -- Bring Me A Shrubbery! (Sporeggar)
quests[296920970] = 250 -- Bring Me Another Shrubbery! (Sporeggar)
quests[297340067] = 75 -- Deepforge the Elder (Horde)
quests[297340469] = 75 -- Deepforge the Elder (Alliance)
quests[297350067] = 75 -- Stonebrand the Elder (Horde)
quests[297350469] = 75 -- Stonebrand the Elder (Alliance)
quests[297360067] = 75 -- Darkfeather the Elder (Horde)
quests[297360469] = 75 -- Darkfeather the Elder (Alliance)
quests[297370067] = 75 -- Firebeard the Elder (Horde)
quests[297370469] = 75 -- Firebeard the Elder (Alliance)
quests[297380067] = 75 -- Moonlance the Elder (Horde)
quests[297380469] = 75 -- Moonlance the Elder (Alliance)
quests[297390067] = 75 -- Windsong the Elder (Horde)
quests[297390469] = 75 -- Windsong the Elder (Alliance)
quests[297400067] = 75 -- Evershade the Elder (Horde)
quests[297400469] = 75 -- Evershade the Elder (Alliance)
quests[297410067] = 75 -- Sekhemi the Elder (Horde)
quests[297410469] = 75 -- Sekhemi the Elder (Alliance)
quests[297420067] = 75 -- Menkhaf the Elder (Horde)
quests[297420469] = 75 -- Menkhaf the Elder (Alliance)
quests[297591242] = 4000 -- Kung Din (Pearlfin Jinyu)
quests[297621242] = 2500 -- Family Heirlooms (Pearlfin Jinyu)
quests[297681216] = 250 -- Missing Mallet (Shang Xi's Academy)
quests[297691216] = 250 -- Rascals (Shang Xi's Academy)
quests[297701216] = 250 -- Still Good! (Shang Xi's Academy)
quests[297711216] = 250 -- Stronger Than Wood (Shang Xi's Academy)
quests[297721216] = 250 -- Raucous Rousing (Shang Xi's Academy)
quests[297741216] = 250 -- Not In the Face! (Shang Xi's Academy)
quests[297751216] = 250 -- The Spirit and Body of Shen-zin Su (Shang Xi's Academy)
quests[297761216] = 250 -- Morning Breeze Village (Shang Xi's Academy)
quests[297771216] = 250 -- Tools of the Enemy (Shang Xi's Academy)
quests[297781216] = 250 -- Rewritten Wisdoms (Shang Xi's Academy)
quests[297791216] = 250 -- The Direct Solution (Shang Xi's Academy)
quests[297801216] = 250 -- Do No Evil (Shang Xi's Academy)
quests[297811216] = 250 -- Monkey Advisory Warning (Shang Xi's Academy)
quests[297821216] = 250 -- Stronger Than Bone (Shang Xi's Academy)
quests[297831216] = 250 -- Stronger Than Stone (Shang Xi's Academy)
quests[297841216] = 250 -- Balanced Perspective (Shang Xi's Academy)
quests[297851216] = 250 -- Dafeng, the Spirit of Air (Shang Xi's Academy)
quests[297861216] = 250 -- Battle for the Skies (Shang Xi's Academy)
quests[297871216] = 250 -- Worthy of Passing (Shang Xi's Academy)
quests[297881216] = 250 -- Unwelcome Nature (Shang Xi's Academy)
quests[297891216] = 250 -- Small, But Significant (Shang Xi's Academy)
quests[297901216] = 250 -- Passing Wisdom (Shang Xi's Academy)
quests[297911216] = 250 -- The Suffering of Shen-zin Su (Shang Xi's Academy)
quests[297921216] = 250 -- Bidden to Greatness (Shang Xi's Academy)
quests[297931216] = 250 -- Evil from the Seas (Shang Xi's Academy)
quests[297941216] = 250 -- None Left Behind (Shang Xi's Academy)
quests[297951216] = 250 -- Stocking Stalks (Shang Xi's Academy)
quests[297961216] = 250 -- Urgent News (Shang Xi's Academy)
quests[297971216] = 250 -- Medical Supplies (Shang Xi's Academy)
quests[297981216] = 250 -- An Ancient Evil (Shang Xi's Academy)
quests[297991216] = 250 -- The Healing of Shen-zin Su (Shang Xi's Academy)
quests[298001216] = 250 -- New Allies (Shang Xi's Academy)
quests[298031052] = 500 -- Ears of the Lich King (Horde Expedition)
quests[298301090] = 500 -- Containment (Kirin Tor)
quests[298470349] = 500 -- To Catch a Thief (Ravenholdt)
quests[298601119] = 250 -- Diametrically Opposed (The Sons of Hodir)
quests[298611119] = 500 -- Whatever it Takes! (The Sons of Hodir)
quests[298621119] = 250 -- Clearing the Way (The Sons of Hodir)
quests[298791228] = 200 -- Swallowed Whole (Forest Hozen)
quests[298831242] = 2500 -- The Pearlfin Situation (Pearlfin Jinyu)
quests[298871242] = 2500 -- The Elder's Instruments (Pearlfin Jinyu)
quests[298941242] = 5000 -- Spirits of the Water (Pearlfin Jinyu)
quests[299031242] = 4000 -- A Perfect Match (Pearlfin Jinyu)
quests[299041242] = 4500 -- Bigger Fish to Fry (Pearlfin Jinyu)
quests[299051242] = 5000 -- Let Them Burn (Pearlfin Jinyu)
quests[299061242] = 5000 -- Carp Diem (Pearlfin Jinyu)
quests[299241228] = 300 -- Kill Kher Shan (Forest Hozen)
quests[299331228] = 500 -- The Bees' Knees (Forest Hozen)
quests[299371228] = 3000 -- Furious Fowl (Forest Hozen)
quests[299391228] = 500 -- Boom Bait (Forest Hozen)
quests[299671228] = 41000 -- Boom Goes the Doonamite! (Forest Hozen)
quests[300001341] = 250 -- The Jade Serpent (The August Celestials)
quests[300061341] = 250 -- The Darkness Around Us (The August Celestials)
quests[300271216] = 250 -- The Lesson of the Iron Bough (Shang Xi's Academy)
quests[300331216] = 250 -- The Lesson of the Iron Bough (Shang Xi's Academy)
quests[300341216] = 250 -- The Lesson of the Iron Bough (Shang Xi's Academy)
quests[300351216] = 250 -- The Lesson of the Iron Bough (Shang Xi's Academy)
quests[300361216] = 250 -- The Lesson of the Iron Bough (Shang Xi's Academy)
quests[300371216] = 250 -- The Lesson of the Iron Bough (Shang Xi's Academy)
quests[300381216] = 250 -- The Lesson of the Iron Bough (Shang Xi's Academy)
quests[300391216] = 10 -- Much to Learn (Shang Xi's Academy)
quests[300401216] = 10 -- Much to Learn (Shang Xi's Academy)
quests[300411216] = 10 -- Much to Learn (Shang Xi's Academy)
quests[300421216] = 10 -- Much to Learn (Shang Xi's Academy)
quests[300431216] = 10 -- Much to Learn (Shang Xi's Academy)
quests[300441216] = 10 -- Much to Learn (Shang Xi's Academy)
quests[300451216] = 10 -- Much to Learn (Shang Xi's Academy)
quests[300631341] = 250 -- Behind the Masks (The August Celestials)
quests[300641341] = 250 -- Saving the Sutras (The August Celestials)
quests[300651341] = 250 -- Arrows of Fortune (The August Celestials)
quests[300661341] = 250 -- Hidden Power (The August Celestials)
quests[300671341] = 350 -- The Shadow of Doubt (The August Celestials)
quests[300681341] = 350 -- Flames of the Void (The August Celestials)
quests[300930349] = 500 -- Assassinate Creed (Ravenholdt)
quests[301070349] = 500 -- Cluster Clutch (Ravenholdt)
quests[301080349] = 75 -- Our Man in Karazhan (Ravenholdt)
quests[301090349] = 500 -- Blood of the Betrayer (Ravenholdt)
quests[301121052] = 500 -- A Score to Settle (Horde Expedition)
quests[301130349] = 75 -- Victory in the Depths (Ravenholdt)
quests[301160349] = 500 -- Sharpening Your Fangs (Ravenholdt)
quests[301180349] = 500 -- Patricide (Ravenholdt)
quests[301341271] = 500 -- Wild Things (Order of the Cloud Serpent)
quests[301351271] = 1000 -- Beating the Odds (Order of the Cloud Serpent)
quests[301361271] = 1000 -- Empty Nests (Order of the Cloud Serpent)
quests[301371271] = 1000 -- Egg Collection (Order of the Cloud Serpent)
quests[301381271] = 500 -- Choosing the One (Order of the Cloud Serpent)
quests[301391271] = 500 -- The Rider's Journey (Order of the Cloud Serpent)
quests[301401271] = 500 -- The Rider's Journey (Order of the Cloud Serpent)
quests[301411271] = 500 -- The Rider's Journey (Order of the Cloud Serpent)
quests[301421271] = 1500 -- It's A... (Order of the Cloud Serpent)
quests[301431271] = 200 -- They Grow Like Weeds (Order of the Cloud Serpent)
quests[301441271] = 400 -- Flight Training: Ring Round-Up (Order of the Cloud Serpent)
quests[301451271] = 600 -- Flight Training: Full Speed Ahead (Order of the Cloud Serpent)
quests[301461271] = 125 -- Snack Time (Order of the Cloud Serpent)
quests[301471271] = 125 -- Fragments of the Past (Order of the Cloud Serpent)
quests[301481271] = 125 -- Just a Flesh Wound (Order of the Cloud Serpent)
quests[301491271] = 125 -- A Feast for the Senses (Order of the Cloud Serpent)
quests[301501271] = 450 -- Sweet as Honey (Order of the Cloud Serpent)
quests[301511271] = 450 -- Catch! (Order of the Cloud Serpent)
quests[301521271] = 600 -- The Sky Race (Order of the Cloud Serpent)
quests[301541271] = 450 -- The Easiest Way To A Serpent's Heart (Order of the Cloud Serpent)
quests[301551271] = 450 -- Restoring the Balance (Order of the Cloud Serpent)
quests[301561271] = 450 -- Feeding Time (Order of the Cloud Serpent)
quests[301571271] = 450 -- Emptier Nests (Order of the Cloud Serpent)
quests[301581271] = 450 -- Disarming the Enemy (Order of the Cloud Serpent)
quests[301591271] = 450 -- Preservation (Order of the Cloud Serpent)
quests[301601275] = 540 -- A Ruby Shard for Ella (Ella)
quests[301871271] = 1200 -- Flight Training: In Due Course (Order of the Cloud Serpent)
quests[301881271] = 1000 -- Riding the Skies (Order of the Cloud Serpent)
quests[301891275] = 540 -- A Lovely Apple for Ella (Ella)
quests[302401269] = 300 -- Survival Ring: Flame (Golden Lotus)
quests[302421269] = 300 -- Survival Ring: Blades (Golden Lotus)
quests[302431269] = 300 -- Mantid Under Fire (Golden Lotus)
quests[302521272] = 900 -- A Helping Hand (The Tillers)
quests[302541272] = 500 -- Learn and Grow II: Tilling and Planting (The Tillers)
quests[302551272] = 500 -- Learn and Grow III: Tending Crops (The Tillers)
quests[302561272] = 700 -- Learn and Grow IV: Harvesting (The Tillers)
quests[302571272] = 999 -- Learn and Grow V: Halfhill Market (The Tillers)
quests[302571281] = 3000 -- Learn and Grow V: Halfhill Market (Gina Mudclaw)
quests[302581272] = 400 -- Mung-Mung's Vote I: A Hozen's Problem (The Tillers)
quests[302591272] = 400 -- Mung-Mung's Vote II: Rotten to the Core (The Tillers)
quests[302601272] = 500 -- Growing the Farm I: The Weeds (The Tillers)
quests[302611269] = 300 -- Roll Club: Serpent's Spine (Golden Lotus)
quests[302661269] = 300 -- Bloodied Skies (Golden Lotus)
quests[302731341] = 250 -- In the House of the Red Crane (The August Celestials)
quests[302771269] = 500 -- The Crumbling Hall (Golden Lotus)
quests[302801269] = 500 -- The Thunder Below (Golden Lotus)
quests[303061269] = 300 -- The Battle Ring (Golden Lotus)
quests[303171283] = 2000 -- Water, Water Everywhere (Farmer Fung)
quests[303171272] = 150 -- Water, Water Everywhere (The Tillers)
quests[303181276] = 3000 -- Chasing the Chicken (Old Hillpaw)
quests[303181272] = 150 -- Chasing the Chicken (The Tillers)
quests[303191279] = 3300 -- Pest Problems (Haohan Mudclaw)
quests[303191272] = 150 -- Pest Problems (The Tillers)
quests[303211273] = 2000 -- Weed War II (Jogu the Drunk)
quests[303211272] = 150 -- Weed War II (The Tillers)
quests[303221281] = 2600 -- Money Matters (Gina Mudclaw)
quests[303221272] = 150 -- Money Matters (The Tillers)
quests[303231280] = 2000 -- They Don't Even Wear Them (Tina Mudclaw)
quests[303231272] = 150 -- They Don't Even Wear Them (The Tillers)
quests[303241277] = 2000 -- Not in Chee-Chee's Backyard (Chee Chee)
quests[303241272] = 150 -- Not in Chee-Chee's Backyard (The Tillers)
quests[303251278] = 2000 -- Where It Counts (Sho)
quests[303251272] = 150 -- Where It Counts (The Tillers)
quests[303261282] = 2000 -- The Kunzen Legend-Chief (Fish Fellreed)
quests[303261272] = 150 -- The Kunzen Legend-Chief (The Tillers)
quests[303271275] = 2000 -- You Have to Burn the Ropes (Ella)
quests[303271272] = 150 -- You Have to Burn the Ropes (The Tillers)
quests[303331272] = 275 -- The Lesser of Two Evils (The Tillers)
quests[303341272] = 275 -- Stealing is Bad... Re-Stealing is OK (The Tillers)
quests[303351272] = 275 -- Stalling the Ravage (The Tillers)
quests[303361272] = 275 -- The Kunzen Hunter-Chief (The Tillers)
quests[303371272] = 275 -- Simian Sabotage (The Tillers)
quests[303791281] = 540 -- A Ruby Shard for Gina (Gina Mudclaw)
quests[303801281] = 540 -- A Lovely Apple for Gina (Gina Mudclaw)
quests[303811275] = 900 -- A Jade Cat for Ella (Ella)
quests[303821275] = 540 -- A Blue Feather for Ella (Ella)
quests[303831275] = 540 -- A Marsh Lily for Ella (Ella)
quests[303861275] = 1800 -- A Dish for Ella (Ella)
quests[303871281] = 540 -- A Jade Cat for Gina (Gina Mudclaw)
quests[303881281] = 540 -- A Blue Feather for Gina (Gina Mudclaw)
quests[303891281] = 900 -- A Marsh Lily for Gina (Gina Mudclaw)
quests[303901281] = 1800 -- A Dish for Gina (Gina Mudclaw)
quests[303911276] = 540 -- A Ruby Shard for Old Hillpaw (Old Hillpaw)
quests[303921276] = 540 -- A Lovely Apple for Old Hillpaw (Old Hillpaw)
quests[303931276] = 540 -- A Jade Cat for Old Hillpaw (Old Hillpaw)
quests[303941276] = 900 -- A Blue Feather for Old Hillpaw (Old Hillpaw)
quests[303951276] = 540 -- A Marsh Lily for Old Hillpaw (Old Hillpaw)
quests[303961276] = 1800 -- A Dish for Old Hillpaw (Old Hillpaw)
quests[303971277] = 540 -- A Ruby Shard for Chee Chee (Chee Chee)
quests[303981277] = 540 -- A Lovely Apple for Chee Chee (Chee Chee)
quests[303991277] = 540 -- A Jade Cat for Chee Chee (Chee Chee)
quests[304001277] = 900 -- A Blue Feather for Chee Chee (Chee Chee)
quests[304011277] = 540 -- A Marsh Lily for Chee Chee (Chee Chee)
quests[304021277] = 1800 -- A Dish for Chee Chee (Chee Chee)
quests[304031278] = 540 -- A Ruby Shard for Sho (Sho)
quests[304041278] = 900 -- A Lovely Apple for Sho (Sho)
quests[304051278] = 540 -- A Jade Cat for Sho (Sho)
quests[304061278] = 540 -- A Blue Feather for Sho (Sho)
quests[304071278] = 540 -- A Marsh Lily for Sho (Sho)
quests[304081278] = 1800 -- A Dish for Sho (Sho)
quests[304091279] = 900 -- A Ruby Shard for Haohan (Haohan Mudclaw)
quests[304101279] = 540 -- A Lovely Apple for Haohan (Haohan Mudclaw)
quests[304111279] = 540 -- A Jade Cat for Haohan (Haohan Mudclaw)
quests[304121279] = 540 -- A Blue Feather for Haohan (Haohan Mudclaw)
quests[304131279] = 540 -- A Marsh Lily for Haohan (Haohan Mudclaw)
quests[304141279] = 1800 -- A Dish for Haohan (Haohan Mudclaw)
quests[304151277] = 800 -- Tracking Quest - Successful (Chee Chee)
quests[304161283] = 540 -- A Ruby Shard for Farmer Fung (Farmer Fung)
quests[304171283] = 540 -- A Lovely Apple for Farmer Fung (Farmer Fung)
quests[304181283] = 540 -- A Jade Cat for Farmer Fung (Farmer Fung)
quests[304191283] = 540 -- A Blue Feather for Farmer Fung (Farmer Fung)
quests[304201283] = 900 -- A Marsh Lily for Farmer Fung (Farmer Fung)
quests[304211283] = 1800 -- A Dish for Farmer Fung (Farmer Fung)
quests[304221282] = 540 -- A Ruby Shard for Fish (Fish Fellreed)
quests[304231282] = 540 -- A Lovely Apple for Fish (Fish Fellreed)
quests[304241282] = 900 -- A Jade Cat for Fish (Fish Fellreed)
quests[304251282] = 540 -- A Blue Feather for Fish (Fish Fellreed)
quests[304261282] = 540 -- A Marsh Lily for Fish (Fish Fellreed)
quests[304271282] = 1800 -- A Dish for Fish (Fish Fellreed)
quests[304281280] = 900 -- A Ruby Shard for Tina (Tina Mudclaw)
quests[304291280] = 540 -- A Lovely Apple for Tina (Tina Mudclaw)
quests[304301280] = 540 -- A Jade Cat for Tina (Tina Mudclaw)
quests[304311280] = 540 -- A Blue Feather for Tina (Tina Mudclaw)
quests[304321280] = 540 -- A Marsh Lily for Tina (Tina Mudclaw)
quests[304331280] = 1800 -- A Dish for Tina (Tina Mudclaw)
quests[304341273] = 540 -- A Ruby Shard for Jogu (Jogu the Drunk)
quests[304351273] = 900 -- A Lovely Apple for Jogu (Jogu the Drunk)
quests[304361273] = 540 -- A Jade Cat for Jogu (Jogu the Drunk)
quests[304371273] = 540 -- A Blue Feather for Jogu (Jogu the Drunk)
quests[304381273] = 540 -- A Marsh Lily for Jogu (Jogu the Drunk)
quests[304391273] = 1800 -- A Dish for Jogu (Jogu the Drunk)
quests[304490909] = 350 -- Darkmoon Crane Deck (Darkmoon Faire)
quests[304500909] = 350 -- Darkmoon Ox Deck (Darkmoon Faire)
quests[304510909] = 350 -- Darkmoon Serpent Deck (Darkmoon Faire)
quests[304520909] = 350 -- Darkmoon Tiger Deck (Darkmoon Faire)
quests[304701280] = 1400 -- A Gift For Tina (Tina Mudclaw)
quests[304701272] = 150 -- A Gift For Tina (The Tillers)
quests[304711277] = 1400 -- A Gift For Chee Chee (Chee Chee)
quests[304711272] = 150 -- A Gift For Chee Chee (The Tillers)
quests[304721278] = 1400 -- A Gift For Sho (Sho)
quests[304721272] = 150 -- A Gift For Sho (The Tillers)
quests[304731282] = 1400 -- A Gift For Fish (Fish Fellreed)
quests[304731272] = 150 -- A Gift For Fish (The Tillers)
quests[304741275] = 1400 -- A Gift For Ella (Ella)
quests[304741272] = 150 -- A Gift For Ella (The Tillers)
quests[304751283] = 1400 -- A Gift For Fung (Farmer Fung)
quests[304751272] = 150 -- A Gift For Fung (The Tillers)
quests[304761276] = 1400 -- A Gift For Old Hillpaw (Old Hillpaw)
quests[304761272] = 150 -- A Gift For Old Hillpaw (The Tillers)
quests[304771279] = 1400 -- A Gift For Haohan (Haohan Mudclaw)
quests[304771272] = 150 -- A Gift For Haohan (The Tillers)
quests[304781273] = 1400 -- A Gift For Jogu (Jogu the Drunk)
quests[304781272] = 150 -- A Gift For Jogu (The Tillers)
quests[304791281] = 1400 -- A Gift For Gina (Gina Mudclaw)
quests[304791272] = 150 -- A Gift For Gina (The Tillers)
quests[305161272] = 500 -- Growing the Farm I: A Little Problem (The Tillers)
quests[305171272] = 250 -- Farmer Fung's Vote I: Yak Attack (The Tillers)
quests[305171283] = 1000 -- Farmer Fung's Vote I: Yak Attack (Farmer Fung)
quests[305181272] = 250 -- Farmer Fung's Vote II: On the Loose (The Tillers)
quests[305181283] = 7000 -- Farmer Fung's Vote II: On the Loose (Farmer Fung)
quests[305191272] = 400 -- Nana's Vote I: Nana's Secret Recipe (The Tillers)
quests[305211272] = 300 -- Haohan's Vote I: Bungalow Break-In (The Tillers)
quests[305211279] = 2000 -- Haohan's Vote I: Bungalow Break-In (Haohan Mudclaw)
quests[305221272] = 300 -- Haohan's Vote II: The Real Culprits (The Tillers)
quests[305231272] = 500 -- Growing the Farm II: The Broken Wagon (The Tillers)
quests[305241272] = 500 -- Growing the Farm II: Knock on Wood (The Tillers)
quests[305251272] = 300 -- Haohan's Vote III: Pure Poison (The Tillers)
quests[305261272] = 500 -- Lost and Lonely (The Tillers)
quests[305271272] = 300 -- Haohan's Vote IV: Melons For Felons (The Tillers)
quests[305281272] = 1000 -- Haohan's Vote V: Chief Yip-Yip (The Tillers)
quests[305351272] = 500 -- Learn and Grow I: Seeds (The Tillers)
quests[305841302] = 500 -- Shocking! (The Anglers)
quests[305851302] = 500 -- What Lurks Below (The Anglers)
quests[305861302] = 500 -- Jagged Abalone (The Anglers)
quests[305881302] = 500 -- Fishing for a Bruising (The Anglers)
quests[305891216] = 250 -- Wrecking the Wreck (Shang Xi's Academy)
quests[305901216] = 250 -- Handle With Care (Shang Xi's Academy)
quests[305911216] = 250 -- Preying on the Predators (Shang Xi's Academy)
quests[305981302] = 500 -- Who Knew Fish Liked Eggs? (The Anglers)
quests[306131302] = 500 -- Armored Carp (The Anglers)
quests[306581302] = 500 -- Huff & Puff (The Anglers)
quests[306651270] = 75 -- The Defense of Shado-Pan Fallback (Shado-Pan)
quests[306701270] = 75 -- Turnabout (Shado-Pan)
quests[306781302] = 500 -- Like Bombing Fish In A Barrel (The Anglers)
quests[306901270] = 100 -- Unmasking the Yaungol (Shado-Pan)
quests[306981302] = 500 -- Scavenger Hunt (The Anglers)
quests[306991270] = 25 -- To Winter's Blossom (Shado-Pan)
quests[307001302] = 500 -- Snapclaw (The Anglers)
quests[307011302] = 500 -- Viseclaw Soup (The Anglers)
quests[307151270] = 75 -- A Line Unbroken (Shado-Pan)
quests[307161341] = 250 -- Chasing Hope (The August Celestials)
quests[307171341] = 250 -- Gifts of the Great Crane (The August Celestials)
quests[307181341] = 250 -- Students of Chi-Ji (The August Celestials)
quests[307231270] = 75 -- Honor, Even in Death (Shado-Pan)
quests[307241270] = 25 -- To the Wall! (Shado-Pan)
quests[307251341] = 350 -- Ellia Ravenmane (The August Celestials)
quests[307261341] = 350 -- Minh Do-Tan (The August Celestials)
quests[307271341] = 350 -- Ellia Ravenmane: Rematch (The August Celestials)
quests[307281341] = 350 -- Fat Long-Fat (The August Celestials)
quests[307291341] = 350 -- Julia Bates (The August Celestials)
quests[307301341] = 350 -- Dextrous Izissha (The August Celestials)
quests[307311341] = 350 -- Kuo-Na Quillpaw (The August Celestials)
quests[307321341] = 350 -- Ellia Ravenmane: Revenge (The August Celestials)
quests[307331341] = 350 -- Tukka-Tuk (The August Celestials)
quests[307341341] = 350 -- Huck Wheelbarrow (The August Celestials)
quests[307351341] = 350 -- Mindel Sunspeaker (The August Celestials)
quests[307361341] = 350 -- Yan Quillpaw (The August Celestials)
quests[307371341] = 350 -- Fat Long-Fat: Rematch (The August Celestials)
quests[307381341] = 350 -- Thelonius (The August Celestials)
quests[307391341] = 350 -- Ellia Ravenmane: Redemption (The August Celestials)
quests[307401341] = 350 -- Champion of Chi-Ji (The August Celestials)
quests[307501270] = 75 -- Off the Wall! (Shado-Pan)
quests[307511270] = 75 -- A Terrible Sacrifice (Shado-Pan)
quests[307521270] = 100 -- Unbelievable! (Shado-Pan)
quests[307531302] = 500 -- Jumping the Shark (The Anglers)
quests[307541302] = 500 -- Bright Bait (The Anglers)
quests[307571270] = 500 -- Lord of the Shado-Pan (Shado-Pan)
quests[307631302] = 500 -- Piranha! (The Anglers)
quests[307671216] = 250 -- Risking It All (Shang Xi's Academy)
quests[307681270] = 25 -- My Husband... (Shado-Pan)
quests[307691270] = 100 -- First Assault (Shado-Pan)
quests[307701270] = 75 -- Running Rampant (Shado-Pan)
quests[307711270] = 75 -- Perfect Pitch (Shado-Pan)
quests[307721270] = 100 -- Seeing Red (Shado-Pan)
quests[307731270] = 75 -- Pitching In (Shado-Pan)
quests[307741270] = 75 -- Ranger Rescue (Shado-Pan)
quests[307751270] = 75 -- The Exile (Shado-Pan)
quests[307761270] = 100 -- Jung Duk (Shado-Pan)
quests[307771270] = 75 -- In Search of Suna (Shado-Pan)
quests[307781270] = 75 -- Dust to Dust (Shado-Pan)
quests[307791270] = 75 -- Slaying the Scavengers (Shado-Pan)
quests[307801270] = 75 -- Totemic Research (Shado-Pan)
quests[307811270] = 100 -- Last Toll of the Yaungol (Shado-Pan)
quests[307821270] = 75 -- Spiteful Spirits (Shado-Pan)
quests[307831270] = 75 -- Hatred Becomes Us (Shado-Pan)
quests[307841270] = 100 -- The Point of No Return (Shado-Pan)
quests[307851270] = 25 -- Gao-Ran Battlefront (Shado-Pan)
quests[308141270] = 75 -- A Foot in the Door (Shado-Pan)
quests[308271270] = 75 -- What Lies Beneath (Shado-Pan)
quests[308791341] = 250 -- Round 1: Brewmaster Chani (The August Celestials)
quests[308801341] = 250 -- Round 1: The Streetfighter (The August Celestials)
quests[308811341] = 250 -- Round 2: Clever Ashyo & Ken-Ken (The August Celestials)
quests[308821341] = 250 -- Round 2: Kang Bramblestaff (The August Celestials)
quests[308831341] = 250 -- Round 3: The Wrestler (The August Celestials)
quests[308841270] = 25 -- Behind the Battlefront (Shado-Pan)
quests[308851341] = 250 -- Round 3: Master Boom Boom (The August Celestials)
quests[308871270] = 75 -- Unwelcome Intruders (Shado-Pan)
quests[308881270] = 75 -- Breach in the Defenses (Shado-Pan)
quests[308891270] = 75 -- Trap Setting (Shado-Pan)
quests[308901270] = 100 -- The Restless Watch (Shado-Pan)
quests[308911270] = 75 -- Treatment for the Troops (Shado-Pan)
quests[308921270] = 75 -- Back on Their Feet (Shado-Pan)
quests[308931270] = 100 -- The Endless Swarm (Shado-Pan)
quests[308941270] = 75 -- Rummaging Through the Remains (Shado-Pan)
quests[308951270] = 100 -- Improvised Ammunition (Shado-Pan)
quests[308961270] = 75 -- Thieves and Troublemakers (Shado-Pan)
quests[308971270] = 75 -- In the Wrong Hands (Shado-Pan)
quests[308981270] = 75 -- Cutting the Swarm (Shado-Pan)
quests[308991270] = 75 -- In Skilled Hands (Shado-Pan)
quests[309001270] = 100 -- Terror of the Dread Wastes (Shado-Pan)
quests[309011270] = 25 -- Along the Southern Front (Shado-Pan)
quests[309021341] = 250 -- Round 4: Master Windfur (The August Celestials)
quests[309071341] = 250 -- Round 4: The P.U.G. (The August Celestials)
quests[309211270] = 100 -- The Motives of the Mantid (Shado-Pan)
quests[309221270] = 75 -- Natural Antiseptic (Shado-Pan)
quests[309231270] = 75 -- Set the Mantid Back (Shado-Pan)
quests[309241270] = 75 -- The Wisdom of Niuzao (Shado-Pan)
quests[309251270] = 100 -- Niuzao's Price (Shado-Pan)
quests[309251341] = 250 -- Niuzao's Price (The August Celestials)
quests[309261270] = 100 -- The Terrible Truth (Shado-Pan)
quests[309271270] = 75 -- Give Them Peace (Shado-Pan)
quests[309281270] = 75 -- A Trail of Fear (Shado-Pan)
quests[309521341] = 350 -- The Unending Siege (The August Celestials)
quests[309531341] = 250 -- Fallen Sentinels (The August Celestials)
quests[309541341] = 250 -- A Blade is a Blade (The August Celestials)
quests[309551341] = 250 -- Paying Tribute (The August Celestials)
quests[309561341] = 350 -- The Siege Swells (The August Celestials)
quests[309571341] = 250 -- The Overwhelming Swarm (The August Celestials)
quests[309581341] = 250 -- In Battle's Shadow (The August Celestials)
quests[309591341] = 250 -- The Big Guns (The August Celestials)
quests[309601270] = 25 -- Returning from the Pass (Shado-Pan)
quests[309631270] = 100 -- The Field Armorer (Shado-Pan)
quests[309641270] = 75 -- A Proper Poultice (Shado-Pan)
quests[309681270] = 500 -- The Sha of Hatred (Shado-Pan)
quests[309690076] = 350 -- A New Enemy (Orgrimmar)
quests[309701270] = 100 -- Enraged By Hatred (Shado-Pan)
quests[309711270] = 75 -- Taking Stock (Shado-Pan)
quests[309721270] = 25 -- Joining the Fight (Shado-Pan)
quests[309731270] = 100 -- Up In Flames (Shado-Pan)
quests[309741270] = 250 -- Raising Spirits (Shado-Pan)
quests[309751270] = 100 -- The Taking of Dusklight Bridge (Shado-Pan)
quests[309761270] = 25 -- Joining the Hunt (Shado-Pan)
quests[309771270] = 100 -- Grounded Welcome (Shado-Pan)
quests[309781270] = 75 -- Hostile Skies (Shado-Pan)
quests[309791270] = 100 -- Devastation Below (Shado-Pan)
quests[309801270] = 25 -- Heroes of the Shado-Pan (Shado-Pan)
quests[309811270] = 75 -- Taoshi and Korvexxis (Shado-Pan)
quests[309820076] = 350 -- Animal Control (Orgrimmar)
quests[309830076] = 350 -- The Dark Shaman (Orgrimmar)
quests[309840076] = 350 -- No Orc Left Behind (Orgrimmar)
quests[309870469] = 250 -- Joining the Alliance (Alliance)
quests[309880469] = 250 -- The Alliance Way (Alliance)
quests[309890469] = 500 -- An Old Pit Fighter (Alliance)
quests[309911270] = 75 -- Do a Barrel Roll! (Shado-Pan)
quests[309921270] = 100 -- Finish This! (Shado-Pan)
quests[309931270] = 25 -- Where are My Reinforcements? (Shado-Pan)
quests[309941270] = 25 -- Lao-Chin's Gambit (Shado-Pan)
quests[309950072] = 350 -- No Man Left Behind (Stormwind)
quests[309960072] = 350 -- The Dark Shaman (Stormwind)
quests[309970072] = 350 -- Animal Control (Stormwind)
quests[309980072] = 350 -- A New Enemy (Stormwind)
quests[310031337] = 250 -- Psycho Mantid (The Klaxxi)
quests[310041337] = 500 -- Preserved in Amber (The Klaxxi)
quests[310051337] = 250 -- Wakening Sickness (The Klaxxi)
quests[310061337] = 250 -- The Klaxxi Council (The Klaxxi)
quests[310071337] = 250 -- The Dread Clutches (The Klaxxi)
quests[310081337] = 250 -- Amber Arms (The Klaxxi)
quests[310091337] = 250 -- Dead Zone (The Klaxxi)
quests[310101337] = 500 -- In Her Clutch (The Klaxxi)
quests[310120067] = 250 -- Joining the Horde (Horde)
quests[310130067] = 250 -- The Horde Way (Horde)
quests[310140067] = 500 -- Hellscream's Gift (Horde)
quests[310151345] = 50 -- Your Private Collection (The Lorewalkers)
quests[310181337] = 250 -- Beneath the Heart of Fear (The Klaxxi)
quests[310191337] = 250 -- Amber Is Life (The Klaxxi)
quests[310201337] = 250 -- Feeding the Beast (The Klaxxi)
quests[310211337] = 250 -- Living Amber (The Klaxxi)
quests[310221337] = 250 -- Kypari Zar (The Klaxxi)
quests[310231337] = 250 -- Relics of the Swarm (The Klaxxi)
quests[310241337] = 130 -- Kunchong Treats (The Klaxxi)
quests[310261337] = 500 -- The Root of the Problem (The Klaxxi)
quests[310301270] = 25 -- Into the Monastery (Shado-Pan)
quests[310311270] = 25 -- Into the Monastery (Shado-Pan)
quests[310321270] = 75 -- Choking the Skies (Shado-Pan)
quests[310381270] = 350 -- The Challenger's Ring: Snow Blossom (Shado-Pan)
quests[310391270] = 250 -- The Mogu Menace (Shado-Pan)
quests[310401270] = 250 -- Spiteful Sprites (Shado-Pan)
quests[310411270] = 250 -- Egg Rescue! (Shado-Pan)
quests[310421270] = 250 -- Onyx Hearts (Shado-Pan)
quests[310431270] = 250 -- Dark Arts (Shado-Pan)
quests[310441270] = 250 -- Bronze Claws (Shado-Pan)
quests[310451270] = 250 -- Illusions Of The Past (Shado-Pan)
quests[310461270] = 250 -- Little Hatchlings (Shado-Pan)
quests[310471270] = 250 -- Born Free (Shado-Pan)
quests[310481270] = 250 -- Grave Consequences (Shado-Pan)
quests[310491270] = 250 -- In Sprite Of Everything (Shado-Pan)
quests[310551345] = 2850 -- Between a Saurok and a Hard Place (The Lorewalkers)
quests[310611270] = 250 -- Riding the Storm (Shado-Pan)
quests[310621270] = 250 -- When The Dead Speak (Shado-Pan)
quests[310631270] = 75 -- Lao-Chin and Serevex (Shado-Pan)
quests[310641270] = 75 -- Nurong and Rothek (Shado-Pan)
quests[310651270] = 25 -- Buried Beneath (Shado-Pan)
quests[310661337] = 500 -- A Cry From Darkness (The Klaxxi)
quests[310851337] = 500 -- Fires and Fears of Old (The Klaxxi)
quests[310871337] = 250 -- Extending Our Coverage (The Klaxxi)
quests[310881337] = 250 -- Crime and Punishment (The Klaxxi)
quests[310891337] = 500 -- By the Sea, Nevermore (The Klaxxi)
quests[310901337] = 250 -- Better With Age (The Klaxxi)
quests[310911337] = 250 -- Reunited (The Klaxxi)
quests[310921337] = 250 -- Feed or Be Eaten (The Klaxxi)
quests[310931345] = 2850 -- Hozen in the Mist (The Lorewalkers)
quests[310941345] = 2850 -- Fish Tales (The Lorewalkers)
quests[310951345] = 2850 -- The Dark Heart of the Mogu (The Lorewalkers)
quests[310961345] = 2850 -- What is Worth Fighting For (The Lorewalkers)
quests[310971345] = 2850 -- Heart of the Mantid Swarm (The Lorewalkers)
quests[311001345] = 2850 -- The Song of the Yaungol (The Lorewalkers)
quests[311021345] = 2850 -- The Seven Burdens of Shaohao (The Lorewalkers)
quests[311031345] = 2850 -- The Ballad of Liu Lang (The Lorewalkers)
quests[311041270] = 350 -- The Challenger's Ring: Yalia Sagewhisper (Shado-Pan)
quests[311051270] = 250 -- The Mogu Menace (Shado-Pan)
quests[311061270] = 250 -- The Mogu Menace (Shado-Pan)
quests[311071337] = 250 -- Citizens of a New Empire (The Klaxxi)
quests[311081337] = 250 -- Concentrated Fear (The Klaxxi)
quests[311091337] = 130 -- Culling the Swarm (The Klaxxi)
quests[311101270] = 250 -- Through the Portal (Shado-Pan)
quests[311111337] = 130 -- Eradicating the Zan'thik (The Klaxxi)
quests[311131270] = 250 -- Assault Fire Camp Gai-Cho (Shado-Pan)
quests[311141270] = 250 -- Assault Deadtalker's Plateau (Shado-Pan)
quests[311161270] = 250 -- Spirit Dust (Shado-Pan)
quests[311171270] = 250 -- Uruk! (Shado-Pan)
quests[311181270] = 250 -- The Deadtalker Cipher (Shado-Pan)
quests[311191270] = 250 -- The Enemy of My Enemy... Is Still My Enemy! (Shado-Pan)
quests[311201270] = 250 -- Cheng Bo! (Shado-Pan)
quests[311271270] = 350 -- The Challenger's Ring: Chao the Voice (Shado-Pan)
quests[311281270] = 350 -- The Challenger's Ring: Lao-Chin the Iron Belly (Shado-Pan)
quests[311350054] = 75 -- The Future of Gnomeregan (Gnomeregan)
quests[311370054] = 10 -- Meet the High Tinker (Gnomeregan)
quests[311380054] = 150 -- The Arts of a Monk (Gnomeregan)
quests[311390072] = 250 -- Beating Them Back! (Stormwind)
quests[311400072] = 250 -- Lions for Lambs (Stormwind)
quests[311410072] = 75 -- Calligraphed Letter (Stormwind)
quests[311420072] = 250 -- Palm of the Tiger (Stormwind)
quests[311430072] = 25 -- Join the Battle! (Stormwind)
quests[311440072] = 250 -- They Sent Assassins (Stormwind)
quests[311450072] = 10 -- The Rear is Clear (Stormwind)
quests[311460068] = 75 -- Scribbled Scroll (Undercity)
quests[311470068] = 75 -- Tiger Palm (Undercity)
quests[311500047] = 75 -- Elegant Rune (Ironforge)
quests[311510047] = 75 -- Kick, Punch, It's All in the Mind (Ironforge)
quests[311560076] = 75 -- Calligraphed Parchment (Orgrimmar)
quests[311570076] = 75 -- Tiger Palm (Orgrimmar)
quests[311580530] = 250 -- The Basics: Hitting Things (Darkspear Trolls)
quests[311590530] = 75 -- The Rise of the Darkspear (Darkspear Trolls)
quests[311600530] = 250 -- A Rough Start (Darkspear Trolls)
quests[311610530] = 250 -- Proving Pit (Darkspear Trolls)
quests[311620530] = 150 -- The Arts of a Monk (Darkspear Trolls)
quests[311630530] = 75 -- More Than Expected (Darkspear Trolls)
quests[311650081] = 75 -- Calligraphed Note (Thunder Bluff)
quests[311660081] = 75 -- Tiger Palm (Thunder Bluff)
quests[311680069] = 75 -- Calligraphed Sigil (Darnassus)
quests[311690069] = 250 -- The Art of the Monk (Darnassus)
quests[311700911] = 25 -- Monk Training (Silvermoon City)
quests[311710911] = 250 -- Tiger Palm (Silvermoon City)
quests[311720930] = 75 -- Monk Training (Exodar)
quests[311730930] = 250 -- The Tiger Palm (Exodar)
quests[311751337] = 250 -- Skeer the Bloodseeker (The Klaxxi)
quests[311761337] = 250 -- A Strange Appetite (The Klaxxi)
quests[311771337] = 250 -- Fine Dining (The Klaxxi)
quests[311781337] = 250 -- A Bloody Delight (The Klaxxi)
quests[311791337] = 250 -- The Scent of Blood (The Klaxxi)
quests[311811302] = 250 -- Fresh Pots (The Anglers)
quests[311821302] = 250 -- You Otter Know (The Anglers)
quests[311831302] = 10 -- Meet the Cap'n (The Anglers)
quests[311841302] = 250 -- Old Age and Treachery (The Anglers)
quests[311851302] = 250 -- Walking Dog (The Anglers)
quests[311861302] = 250 -- Dog Food (The Anglers)
quests[311871302] = 250 -- On the Crab (The Anglers)
quests[311881302] = 250 -- Shark Week (The Anglers)
quests[311891302] = 250 -- Reeltime Strategy (The Anglers)
quests[311901302] = 500 -- The Mariner's Revenge (The Anglers)
quests[311941271] = 600 -- Slitherscale Suppression (Order of the Cloud Serpent)
quests[311961270] = 250 -- Sra'vess Wetwork (Shado-Pan)
quests[311971270] = 250 -- The Bigger They Come... (Shado-Pan)
quests[311981270] = 250 -- A Morale Victory (Shado-Pan)
quests[311991270] = 250 -- Destroy the Siege Weapons! (Shado-Pan)
quests[312001270] = 250 -- Fumigation (Shado-Pan)
quests[312011270] = 250 -- Friends, Not Food! (Shado-Pan)
quests[312031270] = 250 -- Target of Opportunity: Sra'thik Swarmlord (Shado-Pan)
quests[312041270] = 250 -- Target of Opportunity: Sra'thik Hivelord (Shado-Pan)
quests[312081337] = 250 -- Venomous Intent (The Klaxxi)
quests[312091337] = 250 -- Dark Wings, Dark Things (The Klaxxi)
quests[312101337] = 250 -- A Shade of Dread (The Klaxxi)
quests[312111337] = 250 -- The Poisoned Mind (The Klaxxi)
quests[312161337] = 130 -- Dark Skies (The Klaxxi)
quests[312201270] = 350 -- The Challenger's Ring: Hawkmaster Nurong (Shado-Pan)
quests[312211270] = 350 -- The Challenger's Ring: Tenwu of the Red Smoke (Shado-Pan)
quests[312311337] = 130 -- Dreadspinner Extermination (The Klaxxi)
quests[312321337] = 130 -- An Ancient Empire (The Klaxxi)
quests[312331337] = 130 -- Sap Tapping (The Klaxxi)
quests[312341337] = 130 -- Putting An Eye Out (The Klaxxi)
quests[312351337] = 130 -- Nope Nope Nope (The Klaxxi)
quests[312371337] = 130 -- Debugging the Terrace (The Klaxxi)
quests[312381337] = 130 -- Brain Food (The Klaxxi)
quests[312391228] = 350 -- What's in a Name Name? (Forest Hozen)
quests[312611228] = 350 -- Captain Jack's Dead (Forest Hozen)
quests[312651302] = 10 -- Mazu's Breath (The Anglers)
quests[312671337] = 130 -- Mistblade Destruction (The Klaxxi)
quests[312681337] = 130 -- A Little Brain Work (The Klaxxi)
quests[312691337] = 130 -- The Scale-Lord (The Klaxxi)
quests[312701337] = 130 -- The Fight Against Fear (The Klaxxi)
quests[312711337] = 130 -- Bad Genes (The Klaxxi)
quests[312721337] = 130 -- Infection (The Klaxxi)
quests[312771270] = 1000 -- Surprise Attack! (Shado-Pan)
quests[313201273] = 1000 -- Buy A Fish A Drink? (Jogu the Drunk)
quests[313211273] = 3000 -- Buy A Fish A Round? (Jogu the Drunk)
quests[313221273] = 5000 -- Buy A Fish A Keg? (Jogu the Drunk)
quests[313231273] = 6000 -- Buy A Fish A Brewery? (Jogu the Drunk)
quests[313251280] = 3750 -- A Very Nice Necklace (Tina Mudclaw)
quests[313261280] = 3750 -- Tina's Tasteful Tiara (Tina Mudclaw)
quests[313281280] = 3750 -- An Exquisite Earring (Tina Mudclaw)
quests[313291280] = 3750 -- A Beautiful Brooch (Tina Mudclaw)
quests[313321357] = 1000 -- Lesson 1: Sliced Peaches (Nomi)
quests[313331357] = 1000 -- Lesson 2: Instant Noodles (Nomi)
quests[313341357] = 1000 -- Lesson 3: Toasted Fish Jerky (Nomi)
quests[313351357] = 1000 -- Lesson 4: Dried Needle Mushrooms (Nomi)
quests[313361357] = 1000 -- Lesson 5: Pounded Rice Cake (Nomi)
quests[313371357] = 1000 -- A Token of Appreciation (Nomi)
quests[313381277] = 3750 -- Lost Sheepie (Chee Chee)
quests[313391277] = 3750 -- Lost Sheepie... Again (Chee Chee)
quests[313401277] = 3750 -- Oh Sheepie... (Chee Chee)
quests[313411277] = 3750 -- A Wolf In Sheep's Clothing (Chee Chee)
quests[313421270] = 500 -- The Path to Respect Lies in Violence (Shado-Pan)
quests[313541337] = 500 -- Mazu's Bounty (The Klaxxi)
quests[313591337] = 250 -- The Kunchong Whisperer (The Klaxxi)
quests[313671345] = 20 -- The Lorewalkers (The Lorewalkers)
quests[313681345] = 20 -- The Lorewalkers (The Lorewalkers)
quests[313941341] = 250 -- A Celestial Experience (The August Celestials)
quests[313951341] = 250 -- A Celestial Experience (The August Celestials)
quests[313981337] = 250 -- Falling to Pieces (The Klaxxi)
quests[314391337] = 10 -- Dropping Our Signal (The Klaxxi)
quests[314411337] = 250 -- Corruption Runs Deep (The Klaxxi)
quests[314431302] = 350 -- Flying Tiger Gourami (The Anglers)
quests[314431358] = 600 -- Flying Tiger Gourami (Nat Pagle)
quests[314441302] = 350 -- Spinefish Alpha (The Anglers)
quests[314441358] = 600 -- Spinefish Alpha (Nat Pagle)
quests[314461302] = 350 -- Mimic Octopus (The Anglers)
quests[314461358] = 600 -- Mimic Octopus (Nat Pagle)
quests[314581337] = 250 -- Damage Control (The Klaxxi)
quests[314651337] = 250 -- Extracting Answers (The Klaxxi)
quests[314871337] = 130 -- Sonic Disruption (The Klaxxi)
quests[314921341] = 250 -- The Torch of Strength (The August Celestials)
quests[314941337] = 130 -- Free From Her Clutches (The Klaxxi)
quests[314961337] = 130 -- Sampling the Empire's Finest (The Klaxxi)
quests[315021337] = 130 -- Wing Clip (The Klaxxi)
quests[315031337] = 130 -- Shortcut to Ruin (The Klaxxi)
quests[315041337] = 130 -- Ordnance Disposal (The Klaxxi)
quests[315051337] = 130 -- Vess-Guard Duty (The Klaxxi)
quests[315061337] = 130 -- Shackles of Manipulation (The Klaxxi)
quests[315071337] = 130 -- Meltdown (The Klaxxi)
quests[315081337] = 130 -- Specimen Request (The Klaxxi)
quests[315091337] = 130 -- Fear Takes Root (The Klaxxi)
quests[315101337] = 130 -- Quiet Now (The Klaxxi)
quests[315111269] = 75 -- A Witness to History (Golden Lotus)
quests[315121269] = 75 -- A Witness to History (Golden Lotus)
quests[315171341] = 250 -- Contending With Bullies (The August Celestials)
quests[315291278] = 3750 -- Mission: Culling The Vermin (Sho)
quests[315301278] = 3750 -- Mission: The Hozen Dozen (Sho)
quests[315311278] = 3750 -- Mission: Aerial Threat (Sho)
quests[315321278] = 3750 -- Mission: Predator of the Cliffs (Sho)
quests[315341275] = 5000 -- The Beginner's Brew (Ella)
quests[315371275] = 5000 -- Ella's Taste Test (Ella)
quests[315381275] = 5000 -- A Worthy Brew (Ella)
quests[315981337] = 130 -- Kypa'rak's Core (The Klaxxi)
quests[315991337] = 130 -- The Matriarch's Maw (The Klaxxi)
quests[316031337] = 250 -- Seeds of Fear (The Klaxxi)
quests[316051337] = 10 -- The Zan'thik Dig (The Klaxxi)
quests[316061337] = 250 -- The Dissector Wakens (The Klaxxi)
quests[316091337] = 350 -- The Wrath of Shek'zeer (The Klaxxi)
quests[316121337] = 1000 -- Shadow of the Empire (The Klaxxi)
quests[316601337] = 250 -- Not Fit to Swarm (The Klaxxi)
quests[316611337] = 250 -- A Source of Terrifying Power (The Klaxxi)
quests[316691272] = 350 -- The Cabbage Test (The Tillers)
quests[316701272] = 350 -- That Dangling Carrot (The Tillers)
quests[316711272] = 350 -- Why Not Scallions? (The Tillers)
quests[316721272] = 350 -- A Pumpkin-y Perfume (The Tillers)
quests[316731272] = 350 -- Red Blossom Leeks, You Make the Croc-in' World Go Down (The Tillers)
quests[316741272] = 350 -- The Pink Turnip Challenge (The Tillers)
quests[316751272] = 350 -- The White Turnip Treatment (The Tillers)
quests[316761337] = 250 -- Ancient Vengeance (The Klaxxi)
quests[316771337] = 130 -- The Warlord's Ashes (The Klaxxi)
quests[316791337] = 250 -- Extending Our Coverage (The Klaxxi)
quests[316801337] = 250 -- Crime and Punishment (The Klaxxi)
quests[316811337] = 250 -- Better With Age (The Klaxxi)
quests[316821337] = 500 -- By the Sea, Nevermore (The Klaxxi)
quests[316871270] = 100 -- Thinning the Sik'thik (Shado-Pan)
quests[316881270] = 75 -- The Search for Restless Leng (Shado-Pan)
quests[316891337] = 250 -- The Dreadsworn (The Klaxxi)
quests[316981271] = 450 -- Thinning The Pack (Order of the Cloud Serpent)
quests[316991271] = 450 -- Sprite Fright (Order of the Cloud Serpent)
quests[317001271] = 450 -- The Shoe Is On The Other Foot (Order of the Cloud Serpent)
quests[317011271] = 600 -- Dark Huntress (Order of the Cloud Serpent)
quests[317021271] = 600 -- On The Prowl (Order of the Cloud Serpent)
quests[317031271] = 600 -- Madcap Mayhem (Order of the Cloud Serpent)
quests[317041271] = 450 -- Pooped (Order of the Cloud Serpent)
quests[317051271] = 600 -- Needle Me Not (Order of the Cloud Serpent)
quests[317061271] = 450 -- Weeping Widows (Order of the Cloud Serpent)
quests[317071271] = 450 -- A Tangled Web (Order of the Cloud Serpent)
quests[317081271] = 450 -- Serpent's Scale (Order of the Cloud Serpent)
quests[317091271] = 450 -- Lingering Doubt (Order of the Cloud Serpent)
quests[317101271] = 450 -- Tiny Treats (Order of the Cloud Serpent)
quests[317111271] = 600 -- The Seed of Doubt (Order of the Cloud Serpent)
quests[317121271] = 450 -- Monkey Mischief (Order of the Cloud Serpent)
quests[317131271] = 450 -- The Big Brew-haha (Order of the Cloud Serpent)
quests[317141271] = 450 -- Saving the Serpents (Order of the Cloud Serpent)
quests[317151271] = 600 -- The Big Kah-Oona (Order of the Cloud Serpent)
quests[317161271] = 450 -- Pooped (Order of the Cloud Serpent)
quests[317171271] = 450 -- The Trainer's Challenge: Ace Longpaw (Order of the Cloud Serpent)
quests[317181271] = 450 -- The Trainer's Challenge: Big Bao (Order of the Cloud Serpent)
quests[317191271] = 450 -- The Trainer's Challenge: Ningna Darkwheel (Order of the Cloud Serpent)
quests[317201271] = 450 -- The Trainer's Challenge: Suchi the Sweet (Order of the Cloud Serpent)
quests[317211271] = 450 -- The Trainer's Challenge: Qua-Ro Whitebrow (Order of the Cloud Serpent)
quests[317371270] = 5 -- The Cost of War (Shado-Pan)
quests[317711270] = 5 -- Face to Face With Consequence (Shado-Pan)
quests[317841271] = 500 -- Onyx To Goodness (Order of the Cloud Serpent)
quests[318081337] = 130 -- Rampage Against the Machine (The Klaxxi)
quests[318091270] = 500 -- Remnants of Anger (Shado-Pan)
quests[318101271] = 1000 -- Riding the Skies (Order of the Cloud Serpent)
quests[318111271] = 1000 -- Riding the Skies (Order of the Cloud Serpent)
quests[318201357] = 1000 -- A Present for Teacher (Nomi)
quests[319361272] = 250 -- The "Jinyu Princess" Irrigation System (The Tillers)
quests[319371272] = 250 -- "Thunder King" Pest Repellers (The Tillers)
quests[319381272] = 250 -- The "Earth-Slasher" Master Plow (The Tillers)
quests[319411272] = 350 -- Squash Those Foul Odors (The Tillers)
quests[319421272] = 350 -- It's Melon Time (The Tillers)
quests[319431272] = 350 -- Which Berries? Witchberries. (The Tillers)
quests[319451272] = 1400 -- Learn and Grow VI: Gina's Vote (The Tillers)
quests[319451281] = 6000 -- Learn and Grow VI: Gina's Vote (Gina Mudclaw)
quests[319461272] = 400 -- Mung-Mung's Vote III: The Great Carrot Caper (The Tillers)
quests[319471272] = 250 -- Farmer Fung's Vote III: Crazy For Cabbage (The Tillers)
quests[319471283] = 7000 -- Farmer Fung's Vote III: Crazy For Cabbage (Farmer Fung)
quests[319481272] = 400 -- Nana's Vote II: The Sacred Springs (The Tillers)
quests[319491272] = 400 -- Nana's Vote III: Witchberry Julep (The Tillers)
quests[319591337] = 250 -- The Empress' Gambit (The Klaxxi)
quests[320301337] = 10 -- Once in a Hundred Lifetimes (The Klaxxi)
quests[321071302] = 350 -- Flying Tiger Gourami (The Anglers)
quests[321071358] = 500 -- Flying Tiger Gourami (Nat Pagle)
quests[321081375] = 500 -- Domination Point (Dominance Offensive)
quests[321080067] = 250 -- Domination Point (Horde)
quests[321091376] = 500 -- Lion's Landing (Operation: Shieldwall)
quests[321090469] = 250 -- Lion's Landing (Alliance)
quests[321151376] = 150 -- Shackles of the Past (Operation: Shieldwall)
quests[321191376] = 150 -- It Is A Mystery (Operation: Shieldwall)
quests[321201375] = 150 -- Legacy of Ogudei (Dominance Offensive)
quests[321211376] = 150 -- The Spirit Trap (Operation: Shieldwall)
quests[321221376] = 300 -- Ogudei's Lieutenants (Operation: Shieldwall)
quests[321231375] = 150 -- Death on Two Legs (Dominance Offensive)
quests[321261375] = 150 -- Tear It Up (Dominance Offensive)
quests[321271375] = 300 -- All Dead, All Dead (Dominance Offensive)
quests[321281375] = 150 -- Another One Bites the Dust (Dominance Offensive)
quests[321301375] = 150 -- Good Luck, Have Fun (Dominance Offensive)
quests[321311375] = 150 -- We Require More Minerals! (Dominance Offensive)
quests[321321375] = 150 -- Worker Harassment (Dominance Offensive)
quests[321331375] = 150 -- Sentry Wards (Dominance Offensive)
quests[321341375] = 150 -- Hard Counter (Dominance Offensive)
quests[321351375] = 300 -- Hero Killer (Dominance Offensive)
quests[321371375] = 150 -- Runnin' On Empty (Dominance Offensive)
quests[321391375] = 150 -- Stacked! (Dominance Offensive)
quests[321411375] = 150 -- Power Metal (Dominance Offensive)
quests[321421376] = 150 -- We Will Rock You (Operation: Shieldwall)
quests[321431376] = 150 -- A Kind of Magic (Operation: Shieldwall)
quests[321441376] = 150 -- Under Pressure (Operation: Shieldwall)
quests[321451376] = 300 -- Don't Lose Your Head (Operation: Shieldwall)
quests[321461376] = 150 -- Hammer to Fall (Operation: Shieldwall)
quests[321471376] = 250 -- Attack Daily 06 (Operation: Shieldwall)
quests[321481376] = 150 -- Attack! Move! (Operation: Shieldwall)
quests[321491376] = 150 -- Resource Gathering (Operation: Shieldwall)
quests[321501376] = 150 -- Supply Block (Operation: Shieldwall)
quests[321511376] = 150 -- Tower Defense (Operation: Shieldwall)
quests[321521376] = 150 -- Siege Damage (Operation: Shieldwall)
quests[321531376] = 300 -- Hero Killer (Operation: Shieldwall)
quests[321541376] = 150 -- Burn Out! (Operation: Shieldwall)
quests[321551376] = 150 -- Necessary Breaks (Operation: Shieldwall)
quests[321561376] = 150 -- It's Only Right (Operation: Shieldwall)
quests[321571376] = 150 -- The Only Good Goblin... (Operation: Shieldwall)
quests[321581376] = 300 -- Two Step Program (Operation: Shieldwall)
quests[321591376] = 150 -- Circle of Life (Operation: Shieldwall)
quests[321671375] = 200 -- Ancient's Fall (Dominance Offensive)
quests[321681375] = 200 -- End of an Elder (Dominance Offensive)
quests[321691375] = 200 -- A Colossal Victory (Dominance Offensive)
quests[321701376] = 200 -- Ancient's Fall (Operation: Shieldwall)
quests[321711376] = 200 -- End of an Elder (Operation: Shieldwall)
quests[321721376] = 200 -- A Colossal Victory (Operation: Shieldwall)
quests[321891272] = 10 -- A Shabby New Face (The Tillers)
quests[321901375] = 150 -- To Mogujia (Dominance Offensive)
quests[321911375] = 250 -- Ancient Guardians (Dominance Offensive)
quests[321921375] = 250 -- Bloodlines (Dominance Offensive)
quests[321931376] = 150 -- To Mogujia (Operation: Shieldwall)
quests[321941376] = 400 -- Bad Blood (Operation: Shieldwall)
quests[321971375] = 150 -- Mystery Meatloaf (Dominance Offensive)
quests[321981272] = 75 -- One Magical, Flying Kingdom's Trash... (The Tillers)
quests[321991375] = 150 -- Krasarang Steampot (Dominance Offensive)
quests[322001388] = 150 -- Dangers of Za'Tual (Sunreaver Onslaught)
quests[322011388] = 150 -- Grave Circumstances (Sunreaver Onslaught)
quests[322041388] = 150 -- The Skumblade Threat (Sunreaver Onslaught)
quests[322051388] = 150 -- Gather Scout Reports (Sunreaver Onslaught)
quests[322061388] = 150 -- This Just Won't Do (Sunreaver Onslaught)
quests[322071388] = 150 -- Saur Loser (Sunreaver Onslaught)
quests[322081388] = 150 -- Maximum Capacitor (Sunreaver Onslaught)
quests[322091388] = 150 -- Save Our Scouts! (Sunreaver Onslaught)
quests[322121388] = 300 -- The Assault on Zeb'tula (Sunreaver Onslaught)
quests[322131388] = 150 -- Old Enemies (Sunreaver Onslaught)
quests[322141375] = 150 -- Bilgewater Infiltrators (Dominance Offensive)
quests[322151388] = 150 -- Heinous Sacrifice (Sunreaver Onslaught)
quests[322161388] = 150 -- Pterrible Ptorment (Sunreaver Onslaught)
quests[322171388] = 150 -- Dark Offerings (Sunreaver Onslaught)
quests[322181388] = 150 -- Ashes of the Enemy (Sunreaver Onslaught)
quests[322191388] = 150 -- Stone Cold (Sunreaver Onslaught)
quests[322201388] = 150 -- Soul Surrender (Sunreaver Onslaught)
quests[322211375] = 150 -- Storming the Beach (Dominance Offensive)
quests[322221375] = 300 -- Wanted: Chief Engineer Cogwrench (Dominance Offensive)
quests[322231375] = 300 -- Wanted: Lieutenant Ethan Jacobson (Dominance Offensive)
quests[322241388] = 150 -- Rise No More! (Sunreaver Onslaught)
quests[322251388] = 150 -- The Call of Thunder (Sunreaver Onslaught)
quests[322261388] = 150 -- Into the Crypts (Sunreaver Onslaught)
quests[322271388] = 150 -- Preventing a Future Threat (Sunreaver Onslaught)
quests[322281388] = 150 -- The Shuddering Moor (Sunreaver Onslaught)
quests[322301388] = 150 -- The Zandalari Colossus (Sunreaver Onslaught)
quests[322321388] = 150 -- What's Inside Counts (Sunreaver Onslaught)
quests[322331388] = 150 -- Very Disarming (Sunreaver Onslaught)
quests[322341388] = 150 -- Knowledge Is Power (Sunreaver Onslaught)
quests[322351375] = 150 -- Flash! Aaaaaahhhh! (Dominance Offensive)
quests[322361375] = 150 -- Bug Off! (Dominance Offensive)
quests[322371375] = 150 -- Precious Resource (Dominance Offensive)
quests[322381375] = 150 -- Universal Remote-Explode (Dominance Offensive)
quests[322421375] = 150 -- Buried Secrets (Dominance Offensive)
quests[322431376] = 400 -- The Source of Korune Power (Operation: Shieldwall)
quests[322441375] = 250 -- The Korune (Dominance Offensive)
quests[322461376] = 150 -- Meet the Scout (Operation: Shieldwall)
quests[322460469] = 75 -- Meet the Scout (Alliance)
quests[322471376] = 300 -- A King Among Men (Operation: Shieldwall)
quests[322470469] = 250 -- A King Among Men (Alliance)
quests[322481376] = 150 -- A Little Patience (Operation: Shieldwall)
quests[322491375] = 150 -- Meet the Scout (Dominance Offensive)
quests[322490067] = 10 -- Meet the Scout (Horde)
quests[322501375] = 300 -- The Might of the Warchief (Dominance Offensive)
quests[322500067] = 250 -- The Might of the Warchief (Horde)
quests[322511375] = 150 -- Dagger in the Dark (Dominance Offensive)
quests[322521388] = 150 -- Harbingers of the Loa (Sunreaver Onslaught)
quests[322541388] = 150 -- Manipulating the Saurok (Sunreaver Onslaught)
quests[322551388] = 150 -- De-Constructed (Sunreaver Onslaught)
quests[322561375] = 150 -- Rise Of An Empire (Dominance Offensive)
quests[322571375] = 400 -- Voice of the Gods (Dominance Offensive)
quests[322581388] = 700 -- Horde Quest Choice: PvP (Sunreaver Onslaught)
quests[322591388] = 1100 -- Horde Quest Choice: PvE (Sunreaver Onslaught)
quests[322601387] = 1100 -- Alliance Quest Choice: PvE (Kirin Tor Offensive)
quests[322611387] = 700 -- Alliance Quest Choice: PvP (Kirin Tor Offensive)
quests[322621388] = 150 -- Captive Audience (Sunreaver Onslaught)
quests[322641388] = 150 -- Spellbound (Sunreaver Onslaught)
quests[322651388] = 150 -- Charged Moganite (Sunreaver Onslaught)
quests[322661388] = 150 -- Mana Manifestations (Sunreaver Onslaught)
quests[322681388] = 150 -- Tactical Mana Bombs (Sunreaver Onslaught)
quests[322691388] = 150 -- Breaking Down the Defenses (Sunreaver Onslaught)
quests[322741388] = 150 -- The Residents of Ihgaluk (Sunreaver Onslaught)
quests[322751388] = 150 -- Surgical Death (Sunreaver Onslaught)
quests[322761388] = 300 -- Tear Down This Wall! (Sunreaver Onslaught)
quests[322771388] = 300 -- To the Skies! (Sunreaver Onslaught)
quests[322781388] = 300 -- Decisive Action (Sunreaver Onslaught)
quests[322791388] = 300 -- The Fall of Shan Bu (Sunreaver Onslaught)
quests[322821388] = 150 -- Compy Stomp (Sunreaver Onslaught)
quests[322831388] = 150 -- Loa-saur (Sunreaver Onslaught)
quests[322840530] = 75 -- Someone You Should See (Darkspear Trolls)
quests[322851388] = 150 -- The Sleepless Legion (Sunreaver Onslaught)
quests[322871388] = 150 -- Enemies Beneath the Tower (Sunreaver Onslaught)
quests[322881388] = 150 -- Bolstering the Defenses (Sunreaver Onslaught)
quests[322921388] = 200 -- Forge Ahead! (Sunreaver Onslaught)
quests[322931388] = 200 -- Among the Bones (Sunreaver Onslaught)
quests[322941388] = 200 -- Raining Bones (Sunreaver Onslaught)
quests[322961435] = 500 -- Treasures of the Thunder King (Shado-Pan Assault)
quests[322971388] = 150 -- Direhorn or Devilsaur (Sunreaver Onslaught)
quests[322981388] = 150 -- Dino Might (Sunreaver Onslaught)
quests[322991388] = 150 -- Just Some Light Clean-Up Work (Sunreaver Onslaught)
quests[323001388] = 150 -- Disarming Irony (Sunreaver Onslaught)
quests[323011387] = 150 -- Old Enemies (Kirin Tor Offensive)
quests[323021388] = 150 -- Desconstruction (Sunreaver Onslaught)
quests[323031388] = 150 -- Made for War (Sunreaver Onslaught)
quests[323041388] = 200 -- High Recognition (Sunreaver Onslaught)
quests[323051388] = 150 -- Overpowered (Sunreaver Onslaught)
quests[323151376] = 200 -- Anduin's Plea (Operation: Shieldwall)
quests[323150469] = 75 -- Anduin's Plea (Alliance)
quests[323161376] = 500 -- Heart Of The Alliance (Operation: Shieldwall)
quests[323160469] = 500 -- Heart Of The Alliance (Alliance)
quests[323180530] = 250 -- Regeneration Takes Time (Darkspear Trolls)
quests[323191375] = 200 -- Find Thrall! (Dominance Offensive)
quests[323190530] = 250 -- Find Thrall! (Darkspear Trolls)
quests[323201375] = 350 -- The Horde Is Family (Dominance Offensive)
quests[323200530] = 250 -- The Horde Is Family (Darkspear Trolls)
quests[323211376] = 150 -- The Monkey King (Operation: Shieldwall)
quests[323261375] = 150 -- Insertion (Dominance Offensive)
quests[323271375] = 400 -- The Darnassus Mission (Dominance Offensive)
quests[323270067] = 350 -- The Darnassus Mission (Horde)
quests[323281375] = 400 -- Victorious Return (Dominance Offensive)
quests[323291375] = 200 -- Get My Results! (Dominance Offensive)
quests[323290911] = 75 -- Get My Results! (Silvermoon City)
quests[323301375] = 500 -- What's in the Box? (Dominance Offensive)
quests[323300911] = 500 -- What's in the Box? (Silvermoon City)
quests[323311376] = 200 -- The Kirin Tor (Operation: Shieldwall)
quests[323321376] = 150 -- The First Riddle: Mercy (Operation: Shieldwall)
quests[323331376] = 150 -- The Second Riddle: Fellowship (Operation: Shieldwall)
quests[323341376] = 150 -- The Third Riddle: Strength (Operation: Shieldwall)
quests[323351376] = 250 -- The Greatest Prank (Operation: Shieldwall)
quests[323361376] = 250 -- The Handle (Operation: Shieldwall)
quests[323371376] = 250 -- The Head (Operation: Shieldwall)
quests[323381376] = 250 -- The Harmonic Ointment (Operation: Shieldwall)
quests[323391376] = 250 -- Rebuilding the Mallet (Tracker) (Operation: Shieldwall)
quests[323421375] = 150 -- The Spirit Trap (Dominance Offensive)
quests[323431375] = 300 -- Ogudei's Lieutenants (Dominance Offensive)
quests[323441375] = 150 -- It Is A Mystery (Dominance Offensive)
quests[323451375] = 150 -- We're Not Monsters! (Dominance Offensive)
quests[323461376] = 150 -- Oi Ain't Afraid o' No Ghosts! (Operation: Shieldwall)
quests[323471376] = 150 -- Eviction Notice (Operation: Shieldwall)
quests[323481375] = 150 -- Kick 'em While They're Down (Dominance Offensive)
quests[323511375] = 200 -- Echoes of Thunder (Dominance Offensive)
quests[323521375] = 500 -- A Gathering Storm (Dominance Offensive)
quests[323520067] = 500 -- A Gathering Storm (Horde)
quests[323551376] = 100 -- The Harmonic Mallet (Operation: Shieldwall)
quests[323621376] = 500 -- The Fate of Dalaran (Operation: Shieldwall)
quests[323620469] = 350 -- The Fate of Dalaran (Alliance)
quests[323631375] = 150 -- The Kun-Lai Expedition (Dominance Offensive)
quests[323681375] = 500 -- Memory Wine (Dominance Offensive)
quests[323701376] = 150 -- The Kun-Lai Expedition (Operation: Shieldwall)
quests[323711376] = 400 -- Memory Wine (Operation: Shieldwall)
quests[323721375] = 450 -- De-Subjugation (Dominance Offensive)
quests[323720530] = 500 -- De-Subjugation (Darkspear Trolls)
quests[323751375] = 125 -- A Dash of That (Dominance Offensive)
quests[323761375] = 150 -- To the Valley! (Dominance Offensive)
quests[323771376] = 400 -- A Kor'kron In Our Midst (Operation: Shieldwall)
quests[323781375] = 200 -- Clearing a Path (Dominance Offensive)
quests[323791375] = 200 -- Legacy of the Korune (Dominance Offensive)
quests[323801376] = 150 -- The Best Around (Operation: Shieldwall)
quests[323811376] = 400 -- To Catch A Spy (Operation: Shieldwall)
quests[323821376] = 150 -- He's In Deep (Operation: Shieldwall)
quests[323831376] = 400 -- Bugging Out (Operation: Shieldwall)
quests[323841375] = 300 -- Trapping the Leader (Dominance Offensive)
quests[323911375] = 200 -- The Ruins of Korune (Dominance Offensive)
quests[323910067] = 10 -- The Ruins of Korune (Horde)
quests[323921375] = 500 -- The Divine Bell (Dominance Offensive)
quests[323920067] = 350 -- The Divine Bell (Horde)
quests[323931376] = 200 -- The Ruins of Korune (Operation: Shieldwall)
quests[323930469] = 10 -- The Ruins of Korune (Alliance)
quests[323941376] = 500 -- The Divine Bell (Operation: Shieldwall)
quests[323940469] = 350 -- The Divine Bell (Alliance)
quests[323971376] = 400 -- He Won't Even Miss It (Operation: Shieldwall)
quests[323980067] = 10 -- The Bell Speaks (Horde)
quests[323990067] = 500 -- Breath of Darkest Shadow (Horde)
quests[324021375] = 100 -- The Situation In Dalaran (Dominance Offensive)
quests[324031375] = 200 -- It Starts in the Sewers (Dominance Offensive)
quests[324030067] = 250 -- It Starts in the Sewers (Horde)
quests[324041375] = 200 -- Violence in the Arena (Dominance Offensive)
quests[324040067] = 250 -- Violence in the Arena (Horde)
quests[324051375] = 200 -- Hand of the Silver Covenant (Dominance Offensive)
quests[324050067] = 250 -- Hand of the Silver Covenant (Horde)
quests[324061375] = 200 -- A Tactical Assault (Dominance Offensive)
quests[324060067] = 250 -- A Tactical Assault (Horde)
quests[324081375] = 200 -- The Silver Covenant's Stronghold (Dominance Offensive)
quests[324080067] = 250 -- The Silver Covenant's Stronghold (Horde)
quests[324091375] = 200 -- The Kirin Tor's True Colors (Dominance Offensive)
quests[324090067] = 250 -- The Kirin Tor's True Colors (Horde)
quests[324101375] = 200 -- Krasus' Landing (Dominance Offensive)
quests[324100067] = 250 -- Krasus' Landing (Horde)
quests[324111375] = 250 -- The Remaining Sunreavers (Dominance Offensive)
quests[324110067] = 250 -- The Remaining Sunreavers (Horde)
quests[324121375] = 350 -- One Last Grasp (Dominance Offensive)
quests[324120067] = 500 -- One Last Grasp (Horde)
quests[324141376] = 100 -- Darnassus Attacked? (Operation: Shieldwall)
quests[324161376] = 100 -- Jaina's Resolution (Operation: Shieldwall)
quests[324171376] = 200 -- Sewer Cleaning (Operation: Shieldwall)
quests[324170469] = 250 -- Sewer Cleaning (Alliance)
quests[324181376] = 200 -- Unfair Trade (Operation: Shieldwall)
quests[324180469] = 250 -- Unfair Trade (Alliance)
quests[324191376] = 200 -- Nowhere to Hide (Operation: Shieldwall)
quests[324190469] = 250 -- Nowhere to Hide (Alliance)
quests[324201376] = 300 -- Cashing Out (Operation: Shieldwall)
quests[324200469] = 250 -- Cashing Out (Alliance)
quests[324211376] = 300 -- Nowhere to Run (Operation: Shieldwall)
quests[324210469] = 250 -- Nowhere to Run (Alliance)
quests[324231376] = 350 -- What Had To Be Done (Operation: Shieldwall)
quests[324230469] = 350 -- What Had To Be Done (Alliance)
quests[324261376] = 400 -- Stirred, Not Shaken (Operation: Shieldwall)
quests[324331376] = 300 -- Undermining the Under Miner (Operation: Shieldwall)
quests[324461376] = 150 -- Dis-Assembly Required (Operation: Shieldwall)
quests[324481375] = 300 -- Ties with the Past (Dominance Offensive)
quests[324550469] = 500 -- The Silence (Alliance)
quests[324601376] = 200 -- Tracking the Thieves (Operation: Shieldwall)
quests[324851387] = 150 -- Bolstering the Defenses (Kirin Tor Offensive)
quests[324891388] = 150 -- The Creeping Carpet of Ihgaluk (Sunreaver Onslaught)
quests[324911388] = 150 -- Left To Rot (Sunreaver Onslaught)
quests[324921388] = 250 -- Destroy Cauldrons (Sunreaver Onslaught)
quests[324931388] = 150 -- They All Fall Down (Sunreaver Onslaught)
quests[324941388] = 150 -- Power Play (Sunreaver Onslaught)
quests[324951388] = 150 -- The Bloodletter (Sunreaver Onslaught)
quests[325061388] = 150 -- A Wing to Fly On (Sunreaver Onslaught)
quests[325171388] = 150 -- The Conquest of Stone (Sunreaver Onslaught)
quests[325201388] = 200 -- Competing Magic (Sunreaver Onslaught)
quests[325211388] = 200 -- Encroaching Force (Sunreaver Onslaught)
quests[325221388] = 200 -- Imposing Threat (Sunreaver Onslaught)
quests[325251387] = 150 -- Ashes of the Enemy (Kirin Tor Offensive)
quests[325261387] = 150 -- Soul Surrender (Kirin Tor Offensive)
quests[325271387] = 150 -- Grave Circumstances (Kirin Tor Offensive)
quests[325281387] = 150 -- Into the Crypts (Kirin Tor Offensive)
quests[325291387] = 150 -- The Call of Thunder (Kirin Tor Offensive)
quests[325301387] = 150 -- The Bloodletter (Kirin Tor Offensive)
quests[325311387] = 150 -- The Conquest of Stone (Kirin Tor Offensive)
quests[325321387] = 150 -- Rise No More! (Kirin Tor Offensive)
quests[325331387] = 150 -- Stone Cold (Kirin Tor Offensive)
quests[325351387] = 150 -- The Skumblade Threat (Kirin Tor Offensive)
quests[325361387] = 150 -- Manipulating the Saurok (Kirin Tor Offensive)
quests[325371387] = 150 -- De-Constructed (Kirin Tor Offensive)
quests[325381387] = 150 -- Heinous Sacrifice (Kirin Tor Offensive)
quests[325391387] = 150 -- Dark Offerings (Kirin Tor Offensive)
quests[325401387] = 150 -- Harbingers of the Loa (Kirin Tor Offensive)
quests[325411387] = 150 -- Preventing a Future Threat (Kirin Tor Offensive)
quests[325421387] = 150 -- Surgical Strike (Kirin Tor Offensive)
quests[325431387] = 150 -- Dangers of Za'Tual (Kirin Tor Offensive)
quests[325441387] = 150 -- Pterrible Ptorment (Kirin Tor Offensive)
quests[325451387] = 150 -- The Residents of Ihgaluk (Kirin Tor Offensive)
quests[325461387] = 150 -- Just Some Light Clean-Up Work (Kirin Tor Offensive)
quests[325471387] = 150 -- The Creeping Carpet of Ihgaluk (Kirin Tor Offensive)
quests[325481387] = 150 -- Left To Rot (Kirin Tor Offensive)
quests[325501387] = 150 -- Saur Loser (Kirin Tor Offensive)
quests[325511387] = 150 -- Compy Stomp (Kirin Tor Offensive)
quests[325521387] = 150 -- Loa-saur (Kirin Tor Offensive)
quests[325531387] = 150 -- Direhorn vs Devilsaur (Kirin Tor Offensive)
quests[325541387] = 150 -- Dino Might (Kirin Tor Offensive)
quests[325551387] = 150 -- Centuries in Sentries (Kirin Tor Offensive)
quests[325561387] = 150 -- The More You Know (Kirin Tor Offensive)
quests[325571387] = 150 -- Out of Enemy Hands (Kirin Tor Offensive)
quests[325581387] = 150 -- All In the Family (Kirin Tor Offensive)
quests[325591387] = 150 -- Even Giants Fall (Kirin Tor Offensive)
quests[325601387] = 150 -- Keep It Secret (Kirin Tor Offensive)
quests[325611388] = 200 -- Competing Magic (Sunreaver Onslaught)
quests[325621388] = 200 -- Imposing Threat (Sunreaver Onslaught)
quests[325631388] = 200 -- Encroaching Force (Sunreaver Onslaught)
quests[325641388] = 200 -- Zandalari on the Rise (Sunreaver Onslaught)
quests[325651388] = 200 -- Encroaching Force (Sunreaver Onslaught)
quests[325701387] = 150 -- Gather Scout Reports (Kirin Tor Offensive)
quests[325711387] = 150 -- A Wing to Fly On (Kirin Tor Offensive)
quests[325721387] = 150 -- The Sleepless Legion (Kirin Tor Offensive)
quests[325731387] = 150 -- Enemies Beneath the Tower (Kirin Tor Offensive)
quests[325741387] = 150 -- The Shuddering Moor (Kirin Tor Offensive)
quests[325751387] = 150 -- The Zandalari Colossus (Kirin Tor Offensive)
quests[325761387] = 200 -- Competing Magic (Kirin Tor Offensive)
quests[325771387] = 200 -- Imposing Threat (Kirin Tor Offensive)
quests[325781387] = 200 -- Among the Bones (Kirin Tor Offensive)
quests[325791387] = 200 -- Competing Magic (Kirin Tor Offensive)
quests[325801387] = 200 -- Encroaching Force (Kirin Tor Offensive)
quests[325811387] = 200 -- Imposing Threat (Kirin Tor Offensive)
quests[325821387] = 200 -- Raining Bones (Kirin Tor Offensive)
quests[325831387] = 200 -- Zandalari on the Rise (Kirin Tor Offensive)
quests[325841387] = 200 -- Encroaching Force (Kirin Tor Offensive)
quests[325851387] = 200 -- Encroaching Force (Kirin Tor Offensive)
quests[325861387] = 150 -- Maximum Capacitor (Kirin Tor Offensive)
quests[325871387] = 200 -- Forge Ahead! (Kirin Tor Offensive)
quests[325881387] = 150 -- On Her Magic-ey Secret Service (Kirin Tor Offensive)
quests[325891388] = 1000 -- Life Blood (Sunreaver Onslaught)
quests[325991387] = 1000 -- Securing A Future (Kirin Tor Offensive)
quests[326001345] = 2850 -- The Zandalari Prophecy (The Lorewalkers)
quests[326011345] = 2850 -- Rumbles of Thunder (The Lorewalkers)
quests[326021345] = 2850 -- Gods and Monsters (The Lorewalkers)
quests[326051388] = 150 -- Subtle Encouragement (Sunreaver Onslaught)
quests[326061387] = 150 -- Subtle Encouragement (Kirin Tor Offensive)
quests[326071387] = 400 -- Extended Shore Leave (Kirin Tor Offensive)
quests[326081387] = 400 -- Raiding the Vault (Kirin Tor Offensive)
quests[326221388] = 250 -- Intro Flight (Sunreaver Onslaught)
quests[326271387] = 150 -- Charged Moganite (Kirin Tor Offensive)
quests[326281387] = 150 -- Tactical Mana Bombs (Kirin Tor Offensive)
quests[326311387] = 200 -- High Recognition (Kirin Tor Offensive)
quests[326321387] = 150 -- Made for War (Kirin Tor Offensive)
quests[326331387] = 150 -- Spellbound (Kirin Tor Offensive)
quests[326341387] = 150 -- Breaking Down the Defenses (Kirin Tor Offensive)
quests[326351387] = 150 -- Enough with the Bombs! (Kirin Tor Offensive)
quests[326361387] = 150 -- Captive Audience (Kirin Tor Offensive)
quests[326371387] = 150 -- Overpowered (Kirin Tor Offensive)
quests[326381387] = 150 -- Mana Scavengers (Kirin Tor Offensive)
quests[326391387] = 150 -- Deconstruction (Kirin Tor Offensive)
quests[326401435] = 300 -- Champions of the Thunder King (Shado-Pan Assault)
quests[326411435] = 300 -- Champions of the Thunder King (Shado-Pan Assault)
quests[326431375] = 400 -- Work Order: Dominance Offensive II (Dominance Offensive)
quests[326441387] = 300 -- The Assault on Shaol'mara (Kirin Tor Offensive)
quests[326461376] = 400 -- Work Order: Operation: Shieldwall II (Operation: Shieldwall)
quests[326481269] = 400 -- Work Order: Golden Lotus II (Golden Lotus)
quests[326501270] = 400 -- Work Order: Shado-Pan II (Shado-Pan)
quests[326521387] = 300 -- To the Skies! (Kirin Tor Offensive)
quests[326541387] = 300 -- Tear Down This Wall! (Kirin Tor Offensive)
quests[326551387] = 300 -- A Bold Idea (Kirin Tor Offensive)
quests[326561387] = 300 -- The Fall of Shan Bu (Kirin Tor Offensive)
quests[326571341] = 400 -- Work Order: The August Celestials II (The August Celestials)
quests[326591337] = 400 -- Work Order: The Klaxxi II (The Klaxxi)
quests[326761388] = 400 -- Extended Shore Leave (Sunreaver Onslaught)
quests[326771388] = 200 -- Raiding the Vault (Sunreaver Onslaught)
quests[326801388] = 150 -- The Storm Gathers (Sunreaver Onslaught)
quests[326811387] = 150 -- The Storm Gathers (Kirin Tor Offensive)
quests[326941388] = 1250 -- Quest Reward Summary: PvP Stage 1-5 (Sunreaver Onslaught)
quests[326951388] = 1850 -- Quest Reward Summary: PvE Stage 0-2 (Sunreaver Onslaught)
quests[326971388] = 1900 -- Quest Reward Summary: PvE Stage 3 (Sunreaver Onslaught)
quests[326981388] = 2150 -- Quest Reward Summary: PvE Stage 4-5 (Sunreaver Onslaught)
quests[327001387] = 1250 -- Quest Reward Summary: PvP Stage 1-5 (Kirin Tor Offensive)
quests[327011387] = 1850 -- Quest Reward Summary: PvE Stage 0-2 (Kirin Tor Offensive)
quests[327031387] = 1900 -- Quest Reward Summary: PvE Stage 3 (Kirin Tor Offensive)
quests[327041387] = 2150 -- Quest Reward Summary: PvE Stage 4-5 (Kirin Tor Offensive)
quests[327061387] = 150 -- Allies in the Shadows (Kirin Tor Offensive)
quests[327071435] = 200 -- Secrets in the Isle of Thunder (Shado-Pan Assault)
quests[327081435] = 300 -- Setting the Trap (Shado-Pan Assault)
quests[327091388] = 150 -- Allies in the Shadows (Sunreaver Onslaught)
quests[327221388] = 200 -- Forge Ahead! (Sunreaver Onslaught)
quests[327241387] = 200 -- Forge Ahead! (Kirin Tor Offensive)
quests[327831445] = 10 -- Ga'nar's Vengeance (Frostwolf Orcs)
quests[327911445] = 250 -- Let the Hunt Begin! (Frostwolf Orcs)
quests[327921445] = 250 -- Leave Nothing Behind! (Frostwolf Orcs)
quests[327941445] = 250 -- Wrath of Gronn (Frostwolf Orcs)
quests[327951445] = 350 -- The Eldest (Frostwolf Orcs)
quests[327961445] = 10 -- To the Garrison (Frostwolf Orcs)
quests[328031445] = 350 -- Treasure: Thunderlord Cache 01 (Frostwolf Orcs)
quests[328041445] = 250 -- Articles of the Fallen (Frostwolf Orcs)
quests[329181445] = 250 -- Vignette: Giant-Slayer Kul (Frostwolf Orcs)
quests[329291445] = 250 -- They Rely on Numbers (Frostwolf Orcs)
quests[329411445] = 250 -- Vignette: Oasis Icemother (Frostwolf Orcs)
quests[329431388] = 400 -- Work Order: Sunreaver Onslaught II (Sunreaver Onslaught)
quests[329451387] = 400 -- Work Order: Kirin Tor Offensive II (Kirin Tor Offensive)
quests[329791445] = 10 -- Garrison Campaign: Farseer's Rock (Frostwolf Orcs)
quests[329801445] = 250 -- Eaters of the Elements (Frostwolf Orcs)
quests[329831445] = 250 -- In the Shadow of Giants (Frostwolf Orcs)
quests[329841445] = 250 -- The Ascent (Frostwolf Orcs)
quests[329851445] = 350 -- Fury of Frostfire (Frostwolf Orcs)
quests[329891445] = 10 -- Securing the South (Frostwolf Orcs)
quests[329901445] = 250 -- They Who Held Fast (Frostwolf Orcs)
quests[329911445] = 250 -- Into the Boneslag (Frostwolf Orcs)
quests[329921445] = 250 -- Moving Target (Frostwolf Orcs)
quests[329931445] = 250 -- The Strength of Our Bonds (Frostwolf Orcs)
quests[329941445] = 250 -- Enfilade (Frostwolf Orcs)
quests[330101445] = 350 -- The Iron Wolf (Frostwolf Orcs)
quests[330131445] = 250 -- Of Fire and Thunder (Frostwolf Orcs)
quests[330591731] = 250 -- The Fate of Karabor (Council of Exarchs)
quests[330701731] = 250 -- Think of the Children! (Council of Exarchs)
quests[330721731] = 75 -- Into Twilight (Council of Exarchs)
quests[330751731] = 250 -- A Hero's Welcome (Council of Exarchs)
quests[330761731] = 250 -- The Clarity Elixir (Council of Exarchs)
quests[330771731] = 250 -- A Grandmother's Remedy (Council of Exarchs)
quests[330781731] = 250 -- The Dark that Blinds Us (Council of Exarchs)
quests[330801731] = 250 -- Going Undercover (Council of Exarchs)
quests[330811731] = 500 -- Escape From Shaz'gul (Council of Exarchs)
quests[330831731] = 250 -- On the Offensive (Council of Exarchs)
quests[331191445] = 250 -- Slavery and Strife (Frostwolf Orcs)
quests[331251445] = 250 -- A Proper Parting (Frostwolf Orcs)
quests[331321445] = 250 -- Gut Guttra (Frostwolf Orcs)
quests[331451445] = 500 -- Bonus Objective: Grimfrost Hill (Frostwolf Orcs)
quests[332501271] = 1000 -- A Time-Lost Treasure (Order of the Cloud Serpent)
quests[332551731] = 250 -- The Righteous March (Council of Exarchs)
quests[332561731] = 250 -- The Defense of Karabor (Council of Exarchs)
quests[333351492] = 250 -- The Last Emperor (Emperor Shaohao)
quests[333401492] = 250 -- Timeless Nutriment (Emperor Shaohao)
quests[333411492] = 250 -- Wayshrines Of The Celestials (Emperor Shaohao)
quests[333421492] = 250 -- Drive Back The Flame (Emperor Shaohao)
quests[333431492] = 250 -- The Archiereus Of Flame (Emperor Shaohao)
quests[333441445] = 250 -- Armed and Dangerous (Frostwolf Orcs)
quests[333540909] = 250 -- Den Mother's Demise (Darkmoon Faire)
quests[333741492] = 250 -- Path of the Mistwalker (Emperor Shaohao)
quests[334081445] = 250 -- Great Balls of Fire! (Frostwolf Orcs)
quests[334101445] = 350 -- The Butcher of Bladespire (Frostwolf Orcs)
quests[334121445] = 10 -- Mopping Up (Frostwolf Orcs)
quests[334191731] = 250 -- A Matter of Life and Death (Council of Exarchs)
quests[334271445] = 250 -- Ours is the Fury (Frostwolf Orcs)
quests[334441445] = 250 -- Eatercology (Frostwolf Orcs)
quests[334501445] = 250 -- The Warlord's Guard (Frostwolf Orcs)
quests[334541445] = 250 -- The Cure (Frostwolf Orcs)
quests[334621445] = 250 -- Thunderlord for a Day (Frostwolf Orcs)
quests[334671445] = 250 -- The Fall of the Warlord (Frostwolf Orcs)
quests[334681445] = 250 -- Save Wolf Home (Frostwolf Orcs)
quests[334691445] = 250 -- The Farseer (Frostwolf Orcs)
quests[334701445] = 250 -- Pool of Visions (Frostwolf Orcs)
quests[334731445] = 250 -- Back to Bladespire Citadel (Frostwolf Orcs)
quests[334831445] = 250 -- Mulverick's Plight (Frostwolf Orcs)
quests[334841445] = 10 -- The Slavemaster's Demise (Frostwolf Orcs)
quests[334931445] = 250 -- Return to the Pack (Frostwolf Orcs)
quests[335261445] = 250 -- These Colors Don't Run (Frostwolf Orcs)
quests[335271445] = 500 -- Last Steps (Frostwolf Orcs)
quests[335301731] = 250 -- Nightmare in the Tomb (Council of Exarchs)
quests[335461445] = 250 -- Deeds Left Undone (Frostwolf Orcs)
quests[335791515] = 250 -- What the Draenei Found (Arakkoa Outcasts)
quests[335801515] = 250 -- The Purge of Veil Shadar (Arakkoa Outcasts)
quests[335811515] = 250 -- Forbidden Knowledge (Arakkoa Outcasts)
quests[335821515] = 250 -- Kura's Vengeance (Arakkoa Outcasts)
quests[336221445] = 250 -- To the Slaughter (Frostwolf Orcs)
quests[336451515] = 250 -- Tracking Quest - Iskar's Hired Hand (Arakkoa Outcasts)
quests[336571445] = 10 -- Moving In (Frostwolf Orcs)
quests[337201445] = 250 -- Decommissioned Mission (Frostwolf Orcs)
quests[337211445] = 250 -- Gazlowe's Solution (Frostwolf Orcs)
quests[337221445] = 250 -- Too Many Irons in the Fire (Frostwolf Orcs)
quests[337241445] = 250 -- Vol. X Pages ? (Frostwolf Orcs)
quests[337281445] = 250 -- An Eye for a Spy (Frostwolf Orcs)
quests[337291445] = 250 -- Born to Shred (Frostwolf Orcs)
quests[337301445] = 350 -- Engineering Her Demise (Frostwolf Orcs)
quests[337311445] = 500 -- The Battle for Shattrath (Frostwolf Orcs)
quests[337341731] = 250 -- Pieces of Us (Council of Exarchs)
quests[337351445] = 250 -- In Short Supply (Frostwolf Orcs)
quests[337361445] = 250 -- Iron Them Out (Frostwolf Orcs)
quests[337401731] = 250 -- Burning Sky (Council of Exarchs)
quests[337611731] = 250 -- Barum's Notes (Council of Exarchs)
quests[337651731] = 250 -- Dark Enemies (Council of Exarchs)
quests[337841445] = 250 -- Gormaul Tower (Frostwolf Orcs)
quests[337851445] = 250 -- Karg Unchained (Frostwolf Orcs)
quests[337931731] = 250 -- Harbingers of the Void (Council of Exarchs)
quests[337941731] = 250 -- The Great Salvation (Council of Exarchs)
quests[337951731] = 250 -- Across the Stars (Council of Exarchs)
quests[338071445] = 250 -- Free Our Brothers and Sisters (Frostwolf Orcs)
quests[338131731] = 250 -- In Need of a Hero (Council of Exarchs)
quests[338151445] = 10 -- A Song of Frost and Fire (Frostwolf Orcs)
quests[338161445] = 250 -- Honor Has Its Rewards (Frostwolf Orcs)
quests[338261445] = 250 -- Where's My Wolf?! (Frostwolf Orcs)
quests[338281445] = 250 -- The Master Siegesmith (Frostwolf Orcs)
quests[338331731] = 500 -- Wanted: Kuu'rat's Tusks (Council of Exarchs)
quests[338341731] = 500 -- Wanted: Kliaa's Stinger (Council of Exarchs)
quests[338361731] = 350 -- Wanted: Maa'run's Hoof (Council of Exarchs)
quests[338371731] = 500 -- Darkness Falls (Council of Exarchs)
quests[338681445] = 75 -- The Home of the Frostwolves (Frostwolf Orcs)
quests[338981445] = 250 -- Slaying Slavers (Frostwolf Orcs)
quests[339051731] = 250 -- Closing the Door (Council of Exarchs)
quests[339151445] = 250 -- Young Hearts (Frostwolf Orcs)
quests[339171731] = 250 -- Disrupting the Flow (Council of Exarchs)
quests[339191445] = 500 -- Wanted: Gutsmash the Destroyer (Frostwolf Orcs)
quests[339201731] = 250 -- The Heart of Auchindoun (Council of Exarchs)
quests[339251445] = 350 -- Treasure: Frostwolf Cache 01 (Frostwolf Orcs)
quests[339261445] = 350 -- Treasure: Lagoon Pool (Frostwolf Orcs)
quests[339301445] = 350 -- Treasure: Saberon Stash (Frostwolf Orcs)
quests[339331445] = 350 -- Treasure: Deceptia's Smoldering Boots (Frostwolf Orcs)
quests[339401445] = 350 -- Treasure: Crag-Leaper's Cache (Frostwolf Orcs)
quests[339421445] = 350 -- Treasure: Frostwolf Cache 02 (Frostwolf Orcs)
quests[339451445] = 350 -- Treasure: Frostwolf Cache 03 (Frostwolf Orcs)
quests[339461445] = 350 -- Treasure: Frostwolf Cache 04 (Frostwolf Orcs)
quests[339471445] = 350 -- Treasure: Grimfrost Cache (Frostwolf Orcs)
quests[339481445] = 350 -- Treasure: Goren Leftovers (Frostwolf Orcs)
quests[339491445] = 350 -- Treasure: Frostwolf Cache 07 (Frostwolf Orcs)
quests[339581731] = 250 -- Into the Hollow (Council of Exarchs)
quests[339671731] = 250 -- Antivenin (Council of Exarchs)
quests[339691731] = 250 -- Vile Defilers (Council of Exarchs)
quests[339751445] = 350 -- Treasure: Blasting Charges - Rubble A (Frostwolf Orcs)
quests[339761731] = 250 -- The Final Piece (Council of Exarchs)
quests[339881731] = 250 -- Invasion of the Soul Eaters (Council of Exarchs)
quests[339901445] = 350 -- Treasure: Blasting Charges - Rubble B (Frostwolf Orcs)
quests[339951445] = 350 -- Treasure: Blasting Charges - Rubble C (Frostwolf Orcs)
quests[339971445] = 350 -- Treasure: Blasting Charges - Rubble D (Frostwolf Orcs)
quests[340131731] = 250 -- Never Forget (Council of Exarchs)
quests[340191731] = 250 -- Shadows Awaken (Council of Exarchs)
quests[340431731] = 250 -- Darkest Night (Council of Exarchs)
quests[340541731] = 500 -- The Dark Side of the Moon (Council of Exarchs)
quests[340761731] = 500 -- Bonus Objective: The Burial Fields (Council of Exarchs)
quests[340881731] = 250 -- Too Many Irons in the Fire (Council of Exarchs)
quests[340891731] = 250 -- In Short Supply (Council of Exarchs)
quests[340901731] = 250 -- Thaelin's Quick Fix (Council of Exarchs)
quests[340911731] = 250 -- Decommissioned Mission (Council of Exarchs)
quests[340921731] = 250 -- Desperate Measures (Council of Exarchs)
quests[340941731] = 250 -- Vol. X Pages ? (Council of Exarchs)
quests[340951731] = 250 -- Iron Them Out (Council of Exarchs)
quests[340961731] = 250 -- An Eye for a Spy (Council of Exarchs)
quests[340971731] = 250 -- Born to Shred (Council of Exarchs)
quests[340981731] = 250 -- Engineering Her Demise (Council of Exarchs)
quests[340991731] = 500 -- The Battle for Shattrath (Council of Exarchs)
quests[341231445] = 10 -- To Thunder Pass (Frostwolf Orcs)
quests[341241445] = 500 -- The Battle of Thunder Pass (Frostwolf Orcs)
quests[341461445] = 350 -- Treasure: Gorgrond Treasure 05 (Frostwolf Orcs)
quests[341471445] = 350 -- Treasure: Gorgrond Treasure 01 (Frostwolf Orcs)
quests[341491445] = 350 -- Treasure: Gorgrond Treasure 02 (Frostwolf Orcs)
quests[341511445] = 350 -- Treasure: Gorgrond Treasure 03 (Frostwolf Orcs)
quests[341521445] = 350 -- Treasure: Gorgrond Treasure 07 (Frostwolf Orcs)
quests[341531445] = 350 -- Treasure: Gorgrond Treasure 04 (Frostwolf Orcs)
quests[341541731] = 500 -- Destination: Unknown (Council of Exarchs)
quests[341551445] = 350 -- Treasure: Gorgrond Treasure 06 (Frostwolf Orcs)
quests[341571731] = 75 -- Into the Heart of Madness (Council of Exarchs)
quests[341631731] = 250 -- Hiding in the Shadows (Council of Exarchs)
quests[341641731] = 250 -- Book Burning (Council of Exarchs)
quests[341721445] = 350 -- Treasure: Gorgrond Treasure 08 (Frostwolf Orcs)
quests[341731445] = 350 -- Treasure: Gorgrond Treasure 09 (Frostwolf Orcs)
quests[341741445] = 350 -- Treasure: Shadowmoon Treasure 03 (Frostwolf Orcs)
quests[341761445] = 350 -- Treasure: Gorgrond Treasure 11 (Frostwolf Orcs)
quests[341771445] = 350 -- Treasure: Gorgrond Treasure 14 (Frostwolf Orcs)
quests[341791445] = 350 -- Treasure: Gorgrond Treasure 18 (Frostwolf Orcs)
quests[341951445] = 350 -- Treasure: Gorgrond Treasure 16 (Frostwolf Orcs)
quests[342011445] = 350 -- Treasure: Gorgrond Treasure 13 (Frostwolf Orcs)
quests[342061445] = 350 -- Treasure: Gorgrond Treasure 17 (Frostwolf Orcs)
quests[342101445] = 350 -- Treasure: Gorgrond Treasure 19 (Frostwolf Orcs)
quests[342111445] = 350 -- Treasure: Gorgrond Treasure 20 (Frostwolf Orcs)
quests[342141445] = 350 -- Treasure: Gorgrond Treasure 21 (Frostwolf Orcs)
quests[342151445] = 350 -- Treasure: Gorgrond Treasure 22 (Frostwolf Orcs)
quests[342181445] = 350 -- Treasure: Gorgrond Treasure 23 (Frostwolf Orcs)
quests[342191445] = 350 -- Treasure: Gorgrond Treasure 15 (Frostwolf Orcs)
quests[342341731] = 250 -- Payback (Council of Exarchs)
quests[342401731] = 75 -- Scheduled Pickup (Council of Exarchs)
quests[342411445] = 350 -- Treasure: Gorgrond Treasure 12 (Frostwolf Orcs)
quests[343261731] = 10 -- Changing the Tide (Council of Exarchs)
quests[343441445] = 250 -- Lurkers (Frostwolf Orcs)
quests[343451445] = 250 -- Poulticide (Frostwolf Orcs)
quests[343461445] = 250 -- Frosted Fury (Frostwolf Orcs)
quests[343481445] = 250 -- The Real Prey (Frostwolf Orcs)
quests[343511731] = 250 -- We Must Construct Additional Pylons (Council of Exarchs)
quests[343771445] = 75 -- The Portal Home (Frostwolf Orcs)
quests[343791445] = 75 -- Den of Wolves (Frostwolf Orcs)
quests[343801445] = 250 -- Rally the Frostwolves (Frostwolf Orcs)
quests[344021445] = 75 -- Of Wolves and Warriors (Frostwolf Orcs)
quests[344071731] = 250 -- Holding the Line (Council of Exarchs)
quests[344521731] = 75 -- Light's Rest (Council of Exarchs)
quests[344581731] = 250 -- Powering the Defenses (Council of Exarchs)
quests[345011445] = 500 -- Bonus Objective: Frostbite Hollow (Frostwolf Orcs)
quests[345041731] = 500 -- Bonus Objective: Bloodthorn Cave (Council of Exarchs)
quests[345051445] = 500 -- Bonus Objective: Forbidden Glacier (Frostwolf Orcs)
quests[345081731] = 250 -- Restalaan, Captain of the Guard (Council of Exarchs)
quests[345121732] = 250 -- Declawing The Competition (Steamwheedle Draenor Expedition)
quests[345131732] = 350 -- WANTED: Razorpaw! (Steamwheedle Draenor Expedition)
quests[345141732] = 250 -- They've Got The Goods! (Steamwheedle Draenor Expedition)
quests[345151732] = 250 -- Gobnapped (Steamwheedle Draenor Expedition)
quests[345161732] = 250 -- My Precious! (Steamwheedle Draenor Expedition)
quests[345851731] = 250 -- Quakefist (Council of Exarchs)
quests[345981732] = 10 -- The Missing Caravan (Steamwheedle Draenor Expedition)
quests[346531515] = 10 -- Arakkoa Exodus (Arakkoa Outcasts)
quests[346551515] = 250 -- The Shadows of Skettis (Arakkoa Outcasts)
quests[346561515] = 250 -- Hidden in Plain Sight (Arakkoa Outcasts)
quests[346571515] = 250 -- Adherents of the Sun God (Arakkoa Outcasts)
quests[346581515] = 250 -- Orders From On High (Arakkoa Outcasts)
quests[346591515] = 250 -- The Crone (Arakkoa Outcasts)
quests[346621732] = 350 -- Ring of Trials: Crushmaul (Steamwheedle Draenor Expedition)
quests[346631732] = 350 -- Ring of Trials: Raketalon (Steamwheedle Draenor Expedition)
quests[346641732] = 350 -- Ring of Trials: Hol'yelaa (Steamwheedle Draenor Expedition)
quests[346651732] = 350 -- Ring of Trials: Captain Boomspark (Steamwheedle Draenor Expedition)
quests[346661732] = 350 -- Ring of Trials: Roakk the Zealot (Steamwheedle Draenor Expedition)
quests[346671731] = 500 -- Bonus Objective: Court of Souls (Council of Exarchs)
quests[347011731] = 10 -- Speaker for the Dead (Council of Exarchs)
quests[347131445] = 350 -- Treasure: Gronnstalker's Cache (Frostwolf Orcs)
quests[347141445] = 350 -- Frostfire Ridge: (RRP) - Treasure - Pale Leather Cache (Frostwolf Orcs)
quests[347281731] = 500 -- Bonus Objective: The Shimmer Moor (Council of Exarchs)
quests[347321445] = 250 -- Mulverick's Offer of Service (Frostwolf Orcs)
quests[347351445] = 350 -- Thunderlord Invasion! (Frostwolf Orcs)
quests[347561515] = 250 -- A Charming Deception (Arakkoa Outcasts)
quests[347801731] = 250 -- Invisible Ramparts (Council of Exarchs)
quests[347811731] = 250 -- Defenstrations (Council of Exarchs)
quests[347821731] = 10 -- The Exarch Council (Council of Exarchs)
quests[347831731] = 250 -- Naielle, The Rangari (Council of Exarchs)
quests[347841731] = 250 -- Fun with Fungus (Council of Exarchs)
quests[347851731] = 250 -- Hataaru, the Artificer (Council of Exarchs)
quests[347861731] = 250 -- Engorged Goren (Council of Exarchs)
quests[347871731] = 250 -- Exarch Maladaar (Council of Exarchs)
quests[347881731] = 250 -- Friend of the Exarchs (Council of Exarchs)
quests[347891731] = 250 -- Speaker for the Dead (Council of Exarchs)
quests[347901731] = 250 -- The Sting (Council of Exarchs)
quests[347911731] = 250 -- Warning the Exarchs (Council of Exarchs)
quests[347921731] = 250 -- The Traitor's True Name (Council of Exarchs)
quests[348051515] = 250 -- Echo Hunters (Arakkoa Outcasts)
quests[348091711] = 250 -- A Lesson in Mineralogy (Steamwheedle Preservation Society)
quests[348111711] = 250 -- Good Help is Hard to Find (Steamwheedle Preservation Society)
quests[348251445] = 250 -- Vignette: Gruuk (Frostwolf Orcs)
quests[348271515] = 250 -- Last of the Talonpriests (Arakkoa Outcasts)
quests[348281515] = 250 -- Ishaal's Orb (Arakkoa Outcasts)
quests[348291515] = 250 -- New Neighbors (Arakkoa Outcasts)
quests[348301515] = 250 -- Behind the Veil (Arakkoa Outcasts)
quests[348381515] = 250 -- Ikky's Egg (Arakkoa Outcasts)
quests[348391445] = 250 -- Vignette: Skog (Frostwolf Orcs)
quests[348411445] = 350 -- Treasure: Ice-Covered Supplies (Frostwolf Orcs)
quests[348431445] = 250 -- Vignette: Chillfang (Frostwolf Orcs)
quests[348651445] = 250 -- Vignette: Skog (Frostwolf Orcs)
quests[348761731] = 250 -- Loola's Lost Love (Council of Exarchs)
quests[348821515] = 250 -- Blades in the Dark (Arakkoa Outcasts)
quests[348831515] = 250 -- A Feast of Shadows (Arakkoa Outcasts)
quests[348841515] = 250 -- The Kaliri Whisperer (Arakkoa Outcasts)
quests[348851515] = 250 -- Mother of Thorns (Arakkoa Outcasts)
quests[348861515] = 250 -- Baby Bird (Arakkoa Outcasts)
quests[348971731] = 250 -- Rangari Roundup (Council of Exarchs)
quests[349211515] = 250 -- Cult of the Ravenspeakers (Arakkoa Outcasts)
quests[349221515] = 250 -- Words of the Raven Mother (Arakkoa Outcasts)
quests[349231515] = 250 -- The Bloodmane (Arakkoa Outcasts)
quests[349241515] = 250 -- The Egg Thieves (Arakkoa Outcasts)
quests[349311445] = 350 -- Treasure: Frostwolf Cache 04 (Frostwolf Orcs)
quests[349381515] = 350 -- Ralshiara's Demise (Arakkoa Outcasts)
quests[349391515] = 250 -- Declawing The Bloodmane (Arakkoa Outcasts)
quests[349421515] = 10 -- Back from Beyond (Arakkoa Outcasts)
quests[349501445] = 250 -- Dreadpiston (Frostwolf Orcs)
quests[349591731] = 250 -- Dreadpiston (Council of Exarchs)
quests[349631731] = 250 -- Khadgar's Plan (Council of Exarchs)
quests[349671445] = 350 -- Treasure: Raided Loot (Frostwolf Orcs)
quests[349681445] = 350 -- Treasure: Ice-Covered Supplies (Frostwolf Orcs)
quests[349911515] = 250 -- To the... Rescue? (Arakkoa Outcasts)
quests[349961731] = 250 -- Rotting Riverbeasts (Council of Exarchs)
quests[349981515] = 10 -- Talon Watch (Arakkoa Outcasts)
quests[349991515] = 250 -- Hatred Undying (Arakkoa Outcasts)
quests[350001515] = 250 -- Banished From the Sky (Arakkoa Outcasts)
quests[350011515] = 250 -- Gaze of the Raven God (Arakkoa Outcasts)
quests[350021515] = 250 -- Sons of Sethe (Arakkoa Outcasts)
quests[350031515] = 250 -- Ritual Severance (Arakkoa Outcasts)
quests[350041515] = 250 -- Servants of a Dead God (Arakkoa Outcasts)
quests[350071515] = 10 -- Rendezvous with the Ritualists (Arakkoa Outcasts)
quests[350091515] = 250 -- Call of the Raven Mother (Arakkoa Outcasts)
quests[350101515] = 10 -- The High Ravenspeaker (Arakkoa Outcasts)
quests[350111515] = 250 -- The False Talon King (Arakkoa Outcasts)
quests[350121515] = 250 -- Sethe, the Dead God (Arakkoa Outcasts)
quests[350131515] = 250 -- Lithic's Gift (Arakkoa Outcasts)
quests[350141731] = 250 -- Blademoon Bloom (Council of Exarchs)
quests[350151731] = 250 -- Gestating Genesaur (Council of Exarchs)
quests[350321731] = 250 -- Into Anguish (Council of Exarchs)
quests[350621731] = 250 -- Lok-rath is Secured (Council of Exarchs)
quests[350701731] = 250 -- Shut 'er Down (Council of Exarchs)
quests[350931731] = 500 -- The Dark Side of the Moon (Council of Exarchs)
quests[351471711] = 250 -- Fragments of the Past (Steamwheedle Preservation Society)
quests[351691731] = 500 -- And Justice for Thrall (Council of Exarchs)
quests[351711445] = 500 -- And Justice for Thrall (Frostwolf Orcs)
quests[352261445] = 250 -- Old Friends, New Enemies (Frostwolf Orcs)
quests[352271731] = 250 -- Ogre Diplomacy (Council of Exarchs)
quests[352361731] = 500 -- Bonus Objective: Orunai Coast (Council of Exarchs)
quests[352371445] = 500 -- Bonus Objective: Mor'gran Logworks (Frostwolf Orcs)
quests[352451515] = 25 -- A Sentimental Relic (Arakkoa Outcasts)
quests[352521445] = 350 -- Chest Reward Tracking Quest (Frostwolf Orcs)
quests[352541731] = 250 -- Retribution for the Light (Council of Exarchs)
quests[352571515] = 10 -- Power Unearthed (Arakkoa Outcasts)
quests[352581515] = 250 -- Legacy of the Apexis (Arakkoa Outcasts)
quests[352591515] = 250 -- Sol Sisters (Arakkoa Outcasts)
quests[352601515] = 250 -- Hardly Working (Arakkoa Outcasts)
quests[352611515] = 250 -- Shot-Caller (Arakkoa Outcasts)
quests[352641682] = 125 -- A Bunch of Artifact Fragments (Wrynn's Vanguard)
quests[352731515] = 250 -- Hot Seat (Arakkoa Outcasts)
quests[352791445] = 350 -- Treasure: Shadowmoon Treasure 01 (Frostwolf Orcs)
quests[352801445] = 350 -- Treasure: Shadowmoon Treasure 02 (Frostwolf Orcs)
quests[352821445] = 350 -- Treasure - Glowing Blue Mushroom (Frostwolf Orcs)
quests[352891445] = 350 -- Treasure: Shadowmoon Treasure 04 (Frostwolf Orcs)
quests[353471445] = 350 -- Treasure: Ogre Booty 000 (Frostwolf Orcs)
quests[353671445] = 350 -- Treasure: Ogre Booty 001 (Frostwolf Orcs)
quests[353681445] = 350 -- Treasure: Ogre Booty 002 (Frostwolf Orcs)
quests[353691445] = 350 -- Treasure: Ogre Booty 003 (Frostwolf Orcs)
quests[353701445] = 350 -- Treasure: Ogre Booty 004 (Frostwolf Orcs)
quests[353711445] = 350 -- Treasure: Ogre Booty 005 (Frostwolf Orcs)
quests[353731445] = 350 -- Treasure: Ogre Booty 006 (Frostwolf Orcs)
quests[353811445] = 350 -- Treasure: Shadowmoon Treasure 05 (Frostwolf Orcs)
quests[353821445] = 350 -- Treasure: Shadowmoon Treasure 06 (Frostwolf Orcs)
quests[353831445] = 350 -- Treasure: Shadowmoon Treasure 07 (Frostwolf Orcs)
quests[353841445] = 350 -- Treasure: Shadowmoon Treasure 08 (Frostwolf Orcs)
quests[355040909] = 250 -- Iron Joker (Darkmoon Faire)
quests[355301445] = 350 -- Treasure: Shadowmoon Treasure 38 (Frostwolf Orcs)
quests[355521731] = 250 -- Trust No One (Council of Exarchs)
quests[355671445] = 350 -- Treasure: Ogre Booty 007 (Frostwolf Orcs)
quests[355681445] = 350 -- Treasure: Ogre Booty 008 (Frostwolf Orcs)
quests[355691445] = 350 -- Treasure: Ogre Booty 009 (Frostwolf Orcs)
quests[355701445] = 350 -- Treasure: Ogre Booty 010 (Frostwolf Orcs)
quests[355991445] = 250 -- Vignette: Blade-Dancer Aeryx (Frostwolf Orcs)
quests[356321711] = 250 -- A Lesson in Archaeology (Steamwheedle Preservation Society)
quests[356341515] = 250 -- Control is King (Arakkoa Outcasts)
quests[356361515] = 250 -- All Due Respect (Arakkoa Outcasts)
quests[356381682] = 25 -- A Few Artifact Fragments (Wrynn's Vanguard)
quests[356391682] = 500 -- A TON of Artifact Fragments (Wrynn's Vanguard)
quests[356491515] = 500 -- Bonus Objective: The Writhing Mire (Arakkoa Outcasts)
quests[356631711] = 250 -- A Lesson in Teamwork (Steamwheedle Preservation Society)
quests[356681515] = 250 -- Syth's Secret (Arakkoa Outcasts)
quests[356711515] = 75 -- A Gathering of Shadows (Arakkoa Outcasts)
quests[356891682] = 25 -- A Few Artifact Fragments (Wrynn's Vanguard)
quests[356901682] = 125 -- A Bunch of Artifact Fragments (Wrynn's Vanguard)
quests[356911682] = 500 -- A TON of Artifact Fragments (Wrynn's Vanguard)
quests[357041515] = 250 -- When All Is Aligned (Arakkoa Outcasts)
quests[357091445] = 350 -- Treasure: Laughing Skull Cache I (Frostwolf Orcs)
quests[357331515] = 10 -- Rites of the Talonpriests (Arakkoa Outcasts)
quests[357341515] = 250 -- The Talon King (Arakkoa Outcasts)
quests[357981445] = 350 -- Treasure: Glowing Cave Blossom (Frostwolf Orcs)
quests[358671682] = 25 -- A Few Artifact Fragments (Wrynn's Vanguard)
quests[358681682] = 125 -- A Bunch of Artifact Fragments (Wrynn's Vanguard)
quests[358691682] = 500 -- A TON of Artifact Fragments (Wrynn's Vanguard)
quests[358711682] = 25 -- A Few Artifact Fragments (Wrynn's Vanguard)
quests[358721682] = 125 -- A Bunch of Artifact Fragments (Wrynn's Vanguard)
quests[358731682] = 500 -- A TON of Artifact Fragments (Wrynn's Vanguard)
quests[358951515] = 250 -- Terokk's Fall (Arakkoa Outcasts)
quests[358961515] = 500 -- The Avatar of Terokk (Arakkoa Outcasts)
quests[358971515] = 250 -- The Missing Piece (Arakkoa Outcasts)
quests[359141445] = 350 -- Internal: Default Invasion Reward (Frostwolf Orcs)
quests[359271682] = 2500 -- Forsaken Brains (Wrynn's Vanguard)
quests[359371682] = 2500 -- Blood Elf Ear (Wrynn's Vanguard)
quests[359381682] = 2500 -- Goblin Nose (Wrynn's Vanguard)
quests[359391682] = 2500 -- Tauren Hoof (Wrynn's Vanguard)
quests[359401682] = 2500 -- Orc Tooth (Wrynn's Vanguard)
quests[359411682] = 2500 -- Troll Feet (Wrynn's Vanguard)
quests[359421682] = 2500 -- Pandaren Hide (Wrynn's Vanguard)
quests[359451711] = 10 -- Greblin Fastfizzle (Steamwheedle Preservation Society)
quests[359701711] = 250 -- Preservation Capitalization (Steamwheedle Preservation Society)
quests[359721711] = 350 -- Ogre Onslaught (Steamwheedle Preservation Society)
quests[359731711] = 350 -- Brokyo Beatdown (Steamwheedle Preservation Society)
quests[359851711] = 350 -- Garrison Campaign: The Broken Precipice (Steamwheedle Preservation Society)
quests[360241681] = 25 -- A Few Artifact Fragments (Vol'jin's Spear)
quests[360251681] = 500 -- A TON of Artifact Fragments (Vol'jin's Spear)
quests[360261681] = 125 -- A Bunch of Artifact Fragments (Vol'jin's Spear)
quests[360301681] = 25 -- A Few Artifact Fragments (Vol'jin's Spear)
quests[360311681] = 125 -- A Bunch of Artifact Fragments (Vol'jin's Spear)
quests[360321681] = 500 -- A TON of Artifact Fragments (Vol'jin's Spear)
quests[360331681] = 2500 -- Severed Night Elf Head (Vol'jin's Spear)
quests[360341681] = 2500 -- Dwarf Spine (Vol'jin's Spear)
quests[360381681] = 2500 -- Pandaren Hide (Vol'jin's Spear)
quests[360401681] = 2500 -- Tuft of Gnome Hair (Vol'jin's Spear)
quests[360411681] = 2500 -- Draenei Tail (Vol'jin's Spear)
quests[360421681] = 2500 -- Human Bone Chip (Vol'jin's Spear)
quests[360591515] = 250 -- A Worthy Vessel (Arakkoa Outcasts)
quests[360631681] = 25 -- A Few Artifact Fragments (Vol'jin's Spear)
quests[360641681] = 125 -- A Bunch of Artifact Fragments (Vol'jin's Spear)
quests[360651681] = 500 -- A TON of Artifact Fragments (Vol'jin's Spear)
quests[360661681] = 25 -- A Few Artifact Fragments (Vol'jin's Spear)
quests[360671681] = 125 -- A Bunch of Artifact Fragments (Vol'jin's Spear)
quests[360681681] = 500 -- A TON of Artifact Fragments (Vol'jin's Spear)
quests[360851515] = 250 -- On Ebon Wings (Arakkoa Outcasts)
quests[360951711] = 350 -- Azuka Bladefury (Steamwheedle Preservation Society)
quests[361171711] = 350 -- Garrison Campaign: The Broken Precipice (Steamwheedle Preservation Society)
quests[361181445] = 350 -- Treasure: Pile of Rubble (Frostwolf Orcs)
quests[361191682] = 75 -- Welcome to Ashran (Wrynn's Vanguard)
quests[361231682] = 250 -- Into The Wilderness (Wrynn's Vanguard)
quests[361251682] = 250 -- The Road of Glory (Wrynn's Vanguard)
quests[361271682] = 250 -- Survey Ashran (Wrynn's Vanguard)
quests[361301682] = 250 -- Reporting For Duty (Wrynn's Vanguard)
quests[361331682] = 250 -- Uncovering the Artifact Fragments (Wrynn's Vanguard)
quests[361631731] = 250 -- Garrison Campaign: The Exarch's Call (Council of Exarchs)
quests[361641731] = 250 -- The Trial of Courage (Council of Exarchs)
quests[361671731] = 250 -- The Trial of Heart (Council of Exarchs)
quests[361681731] = 250 -- The Trial of Faith (Council of Exarchs)
quests[361691731] = 500 -- The Trial of Champions (Council of Exarchs)
quests[361701445] = 350 -- Treasure: Femur of Improbability (Frostwolf Orcs)
quests[361761731] = 250 -- A Call for Huntsman (Council of Exarchs)
quests[361771731] = 250 -- Friendly Competition (Council of Exarchs)
quests[361851731] = 250 -- Fair Trade (Council of Exarchs)
quests[361931731] = 250 -- An Act of Kindness (Council of Exarchs)
quests[361961681] = 75 -- Welcome to Ashran (Vol'jin's Spear)
quests[361971681] = 150 -- Reporting For Duty (Vol'jin's Spear)
quests[361981681] = 250 -- Uncovering the Artifact Fragments (Vol'jin's Spear)
quests[362051445] = 250 -- The Fractured Hammer (Frostwolf Orcs)
quests[362071445] = 25 -- Waruk the Frostforger (Frostwolf Orcs)
quests[362261681] = 250 -- Survey Ashran (Vol'jin's Spear)
quests[362271681] = 250 -- The Road of Glory (Vol'jin's Spear)
quests[362281681] = 250 -- Into The Wilderness (Vol'jin's Spear)
quests[362301445] = 250 -- The Restless Spirit (Frostwolf Orcs)
quests[362361731] = 250 -- The Cryptic Tome of Tailoring (Council of Exarchs)
quests[362381445] = 250 -- Mending A Broken Heart (Frostwolf Orcs)
quests[362391731] = 250 -- A Mysterious Satchel (Council of Exarchs)
quests[362401731] = 250 -- Slow and Steady (Council of Exarchs)
quests[362411731] = 250 -- The Power of Preservation (Council of Exarchs)
quests[362621731] = 250 -- Ameeka, Master Tailor (Council of Exarchs)
quests[362661731] = 250 -- The Clothes on Their Backs (Council of Exarchs)
quests[362691731] = 250 -- Hexcloth (Council of Exarchs)
quests[363011445] = 250 -- Trega's Tailoring Kit (Frostwolf Orcs)
quests[363091731] = 250 -- The Strength of Iron (Council of Exarchs)
quests[363111731] = 250 -- Father and Son (Council of Exarchs)
quests[363141445] = 250 -- The Intricate Pendant (Frostwolf Orcs)
quests[363161515] = 500 -- Victory is Within Reach (Arakkoa Outcasts)
quests[363521445] = 25 -- Locating the Lapidarist (Frostwolf Orcs)
quests[363781445] = 250 -- No Pressure, No Diamonds (Frostwolf Orcs)
quests[363801445] = 250 -- Diamonds Are Forever (Frostwolf Orcs)
quests[364081731] = 250 -- A Power Lost (Council of Exarchs)
quests[364091731] = 250 -- Restoration (Council of Exarchs)
quests[364171445] = 250 -- From Their Cold Dead Hands (Frostwolf Orcs)
quests[364191445] = 250 -- Dyed in the Fur (Frostwolf Orcs)
quests[364251515] = 250 -- Egg Punt (Arakkoa Outcasts)
quests[364810909] = 250 -- Firebird's Challenge (Darkmoon Faire)
quests[364921445] = 350 -- Garrison Pet Battle (Frostwolf Orcs)
quests[365051445] = 250 -- A Warrior's Shroud (Frostwolf Orcs)
quests[365121731] = 250 -- Together We Are Strong (Council of Exarchs)
quests[365161445] = 250 -- Cut 'Em Out! (Frostwolf Orcs)
quests[365901515] = 500 -- Bonus Objective: The Howling Crag (Arakkoa Outcasts)
quests[366091302] = 250 -- The Great Angler Returns (The Anglers)
quests[366111302] = 250 -- A True Draenor Angler (The Anglers)
quests[366161302] = 250 -- An Angler on Our Team (The Anglers)
quests[366601515] = 500 -- Bonus Objective: Bloodmane Pridelands (Arakkoa Outcasts)
quests[367411682] = 25 -- Vintage Free Action Potion (Wrynn's Vanguard)
quests[367421681] = 25 -- Vintage Free Action Potion (Vol'jin's Spear)
quests[367901515] = 250 -- The Initiate's Revenge (Arakkoa Outcasts)
quests[367911682] = 25 -- Phantom Potion (Wrynn's Vanguard)
quests[367921515] = 500 -- Bonus Objective: Bladefist Hold (Arakkoa Outcasts)
quests[368001358] = 350 -- Fire Ammonite Lunker (Nat Pagle)
quests[368021358] = 350 -- Abyssal Gulper Lunker (Nat Pagle)
quests[368031358] = 350 -- Blackwater Whiptail Lunker (Nat Pagle)
quests[368041358] = 350 -- Blind Lake Lunker (Nat Pagle)
quests[368051358] = 350 -- Fat Sleeper Lunker (Nat Pagle)
quests[368061358] = 350 -- Jawless Skulker Lunker (Nat Pagle)
quests[368821302] = 250 -- The Great Angler Returns (The Anglers)
quests[368841681] = 2500 -- Worgen Snout (Vol'jin's Spear)
quests[369351682] = 25 -- Phantom Potion (Wrynn's Vanguard)
quests[369511515] = 10 -- Arakkoa Exodus (Arakkoa Outcasts)
quests[371120072] = 25 -- Rest and Relaxation (Stormwind)
quests[371251711] = 350 -- A Rare Find (Steamwheedle Preservation Society)
quests[372101711] = 500 -- Aogexon's Fang (Steamwheedle Preservation Society)
quests[372111711] = 500 -- Bergruu's Horn (Steamwheedle Preservation Society)
quests[372211711] = 500 -- Dekorhan's Tusk (Steamwheedle Preservation Society)
quests[372221711] = 500 -- Direhoof's Hide (Steamwheedle Preservation Society)
quests[372231711] = 500 -- Gagrog's Skull (Steamwheedle Preservation Society)
quests[372241711] = 500 -- Mu'gra's Head (Steamwheedle Preservation Society)
quests[372251711] = 500 -- Thek'talon's Talon (Steamwheedle Preservation Society)
quests[372261711] = 500 -- Xelganak's Stinger (Steamwheedle Preservation Society)
quests[372951681] = 75 -- Bonus Objective: Kill Rylai Crestfall (Vol'jin's Spear)
quests[372971682] = 75 -- Bonus Objective: Kill Jeron Emberfall (Wrynn's Vanguard)
quests[374211731] = 500 -- Bonus Objective: Kuuro's Claim (Council of Exarchs)
quests[374221731] = 500 -- Bonus Objective: Zangarra (Council of Exarchs)
quests[374390911] = 250 -- Thirst Unending (Silvermoon City)
quests[374400911] = 250 -- A Fistful of Slivers (Silvermoon City)
quests[374420911] = 250 -- The Shrine of Dath'Remar (Silvermoon City)
quests[374430911] = 250 -- Solanian's Belongings (Silvermoon City)
quests[374440930] = 350 -- Inoculation (Exodar)
quests[374450930] = 250 -- Spare Parts (Exodar)
quests[374460076] = 350 -- Lazy Peons (Orgrimmar)
quests[374831682] = 100 -- Captian's Whistle (Wrynn's Vanguard)
quests[374841681] = 100 -- Captian's Whistle (Vol'jin's Spear)
quests[375111711] = 350 -- Dizzy Sparkshift (Steamwheedle Preservation Society)
quests[375161711] = 500 -- News for Nixxie (Steamwheedle Preservation Society)
quests[375171711] = 500 -- News for Nixxie (Steamwheedle Preservation Society)
quests[375201711] = 500 -- Vileclaw's Claw (Steamwheedle Preservation Society)
quests[378190909] = 250 -- Welcome to the Darkmoon Races (Darkmoon Faire)
quests[378911847] = 500 -- Assault on Ironhold Harbor (Hand of the Prophet)
quests[379100909] = 250 -- The Real Race (Darkmoon Faire)
quests[379110909] = 250 -- The Real Big Race (Darkmoon Faire)
quests[379351848] = 1500 -- The Assault Base (Vol'jin's Headhunters)
quests[379401848] = 500 -- Assault on Ironhold Harbor (Vol'jin's Headhunters)
quests[379681847] = 500 -- Assault on the Temple of Sha'naar (Hand of the Prophet)
quests[380011848] = 350 -- Commander in the Field! (Vol'jin's Headhunters)
quests[380081682] = 250 -- The Road of Glory (Wrynn's Vanguard)
quests[380441848] = 500 -- Bleeding the Bleeding Hollow (Vol'jin's Headhunters)
quests[380451847] = 500 -- Bleeding the Bleeding Hollow (Hand of the Prophet)
quests[380461847] = 500 -- Battle At The Iron Front (Hand of the Prophet)
quests[380471848] = 500 -- Battle At The Iron Front (Vol'jin's Headhunters)
quests[382231847] = 1000 -- Dark Ascension (Hand of the Prophet)
quests[382231848] = 1000 -- Dark Ascension (Vol'jin's Headhunters)
quests[382501847] = 500 -- Assault on the Ruins of Kra'nak (Hand of the Prophet)
quests[382521848] = 500 -- Assault on the Ruins of Kra'nak (Vol'jin's Headhunters)
quests[382741847] = 1000 -- The Eye of Kilrogg (Hand of the Prophet)
quests[382741848] = 1000 -- The Eye of Kilrogg (Vol'jin's Headhunters)
quests[383990021] = 75 -- Fencing the Rose (Booty Bay)
quests[384061358] = 350 -- Sea Scorpion Lunker (Nat Pagle)
quests[384161848] = 1000 -- Secrets of the Sargerei (Vol'jin's Headhunters)
quests[384401847] = 500 -- Assault on The Fel Forge (Hand of the Prophet)
quests[384411848] = 500 -- Assault on The Fel Forge (Vol'jin's Headhunters)
quests[384451847] = 1500 -- The Assault Base (Hand of the Prophet)
quests[384461847] = 350 -- Commander in the Field! (Hand of the Prophet)
quests[384491848] = 500 -- Assault on the Temple of Sha'naar (Vol'jin's Headhunters)
quests[384631848] = 2500 -- The Cipher of Damnation (Vol'jin's Headhunters)
quests[385621847] = 1000 -- Secrets of the Sargerei (Hand of the Prophet)
quests[385771848] = 350 -- The Battle for the West (Vol'jin's Headhunters)
quests[385811847] = 350 -- The Battle for the West (Hand of the Prophet)
quests[385851847] = 500 -- Assault on the Throne of Kil'jaeden (Hand of the Prophet)
quests[385861848] = 500 -- Assault on the Throne of Kil'jaeden (Vol'jin's Headhunters)
quests[392031090] = 10 -- Warchief's Command: Dragonblight (Kirin Tor)
quests[392041090] = 10 -- Hero's Call: Dragonblight (Kirin Tor)
quests[392061085] = 10 -- Warchief's Command: Grizzly Hills (Warsong Offensive)
quests[392071050] = 25 -- Hero's Call: Grizzly Hills (Valiance Expedition)
quests[392081098] = 25 -- Zul'Drak (Knights of the Ebon Blade)
quests[392101090] = 10 -- The Magical Kingdom of Dalaran (Kirin Tor)
quests[392831358] = 350 -- Felmouth Frenzy Lunker (Nat Pagle)
quests[393941847] = 2500 -- The Cipher of Damnation (Hand of the Prophet)
quests[394321849] = 1500 -- Unknowable Power (Order of the Awakened)
quests[394331849] = 1500 -- Unseen Influence (Order of the Awakened)
quests[395091848] = 250 -- Bloodied Blades of Zeth'Gol (Vol'jin's Headhunters)
quests[395101848] = 250 -- A Little Wetwork (Vol'jin's Headhunters)
quests[395111848] = 250 -- Blood of Fallen Brothers (Vol'jin's Headhunters)
quests[395121848] = 250 -- Secrets of the Shadow Council (Vol'jin's Headhunters)
quests[395131848] = 250 -- Disarming Sha'naar (Vol'jin's Headhunters)
quests[395141848] = 250 -- Harnessing Their Power (Vol'jin's Headhunters)
quests[395191848] = 250 -- Demonslayer of Tanaan (Vol'jin's Headhunters)
quests[395261848] = 250 -- Pressing the Attack (Vol'jin's Headhunters)
quests[395291850] = 1500 -- Tooth and Claw (The Saberstalkers)
quests[395321848] = 250 -- Bringing Down the Iron Horde (Vol'jin's Headhunters)
quests[395651850] = 3500 -- Rumble in the Jungle (The Saberstalkers)
quests[395671847] = 250 -- Bloodied Blades of Zeth'Gol (Hand of the Prophet)
quests[395681847] = 250 -- A Little Wetwork (Hand of the Prophet)
quests[395691847] = 250 -- Breaking Their Fel Bonds (Hand of the Prophet)
quests[395701847] = 250 -- Secrets of the Shadow Council (Hand of the Prophet)
quests[395711847] = 250 -- Disarming Sha'naar (Hand of the Prophet)
quests[395731847] = 250 -- Harnessing Their Power (Hand of the Prophet)
quests[395741847] = 250 -- Pressing the Attack (Hand of the Prophet)
quests[395811847] = 250 -- Demonslayer of Tanaan (Hand of the Prophet)
quests[395821850] = 1500 -- Tooth and Claw (The Saberstalkers)
quests[395861847] = 250 -- Bringing Down the Iron Horde (Hand of the Prophet)
end
